#! /bin/sh
# \
exec tclsh${TCL} "$0" ${1+"$@"}
#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------
namespace eval ::thatch {
	if {[package provide thatch] ne {}} { return }

	proc thatchBootstrap {{fn ""}} {
		if {$fn eq ""} { set fn [info script] }
		if {[catch {
			set f [open $fn r]
			chan configure $f -eofchar \x1a
			read $f
			seek $f 1 current
			set d [read $f]
			close $f
		} err]} {
			return 0
		}
		if {[string length $d] == 0} { return 0 }
		uplevel #0 $d
		package require thatch
	}
	package ifneeded thatch 0.5 [namespace current]::thatchBootstrap
}

package require thatch
::thatch::thatchSetup [info script] 

source /herbie/tennessee.tcl
#-------------------------------------------------------------------

# THATCH /thatch/thatch.tcl 
#
# Thatch - Tcl Attach
#
#
# Stuart Cassoff
#
# May 2009
# December 2009
# January 2010
# October 2010
# March 2012
# Summer 2012
# Spring 2018
# Version 0.1
#
# Spring 2018
# Version 0.2
#
# Summer 2018
# Version 0.3
#
# Summer 2022
# Version 0.4
#
# Summer 2022
# Version 0.5
#


[namespace eval ::thatch {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	thatch
		title	Thatch
		descr	"App Builder"
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.5
	}]
	dict set pkgconfig importfrom [namespace current]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2009,2018,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}; # pkgconfig


#-------------------------------------------------------------------
# Thatch preamble and bootstrap code
#-------------------------------------------------------------------


proc preamble {{wish {0}}} {
	return "#! /bin/sh\n# \\\nexec [expr {$wish ? "wi" : "tcl"}]sh\${TCL} \"\$0\" \${1+\"\$@\"}"
}


variable bootstrap [string range {
#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------
namespace eval ::thatch {
	if {[package provide thatch] ne {}} { return }

	proc thatchBootstrap {{fn ""}} {
		if {$fn eq ""} { set fn [info script] }
		if {[catch {
			set f [open $fn r]
			chan configure $f -eofchar \x1a
			read $f
			seek $f 1 current
			set d [read $f]
			close $f
		} err]} {
			return 0
		}
		if {[string length $d] == 0} { return 0 }
		uplevel #0 $d
		package require thatch
	}
	package ifneeded thatch @VER@ [namespace current]::thatchBootstrap
}

package require thatch
::thatch::thatchSetup [info script] @THATCHOPTS@

source @MAIN@
#-------------------------------------------------------------------
} 1 end-1]


#-------------------------------------------------------------------
# Other
#-------------------------------------------------------------------

proc linterleave {l1 l2} {
	set l {}
	foreach e1 $l1 e2 $l2 { lappend l $e1 $e2 }
	return $l
}

proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

# In My Namespace
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

proc aliasconst {name val args} {
	if {[llength $args] % 2 != 0} { error "must be %2!" }
	foreach {n v} [linsert $args 0 $name $val] {
		interp alias {} [imn $n] {} return -level 0 $v
	}
}

proc unknownOption {opt opts} {
	set msg "unknown option \"$opt\""
	if {[llength $opts] == 0} { return $msg }
	append msg ", must be: "
	append msg [lindex $opts 0]
	if {[llength $opts] == 2} {
		append msg " or [lindex $opts 1]"
	} elseif {[llength $opts] > 2} {
		append msg ", " [join [lrange $opts 1 end-1] ", "] " or " [lindex $opts end]
	}
	return $msg
}

#-------------------------------------------------------------------
# File
#-------------------------------------------------------------------

proc openItIn {fn {mode r} args} {
	if {$fn in {- stdin}} {
		set f stdin
	} else {
		if {[string index $fn 0] eq "@"} {
			set f [string range $fn 1 end]
		} else {
			set f [open $fn $mode]
		}
	}
	if {[llength $args] > 0} {
		chan configure $f {*}$args
	}
#puts [chan configure $f]
	return $f
}

proc openItOut {fn {mode w} args} {
	if {$fn in {- stdout}} {
		set f stdout
	} else {
		set f [open $fn $mode]
	}
	if {[llength $args] > 0} {
		chan configure $f {*}$args
	}
	return $f
}

proc closeIt {f} {
	if {$f in {stdin stdout}} {
		return {}
	}
	return [chan close $f]
}

proc inhaleIt {fn args} {
	return [read [set f [openItIn $fn r {*}$args]]][closeIt $f]
}

interp alias {} [imn openIt] {} [imn openItIn]

#-------------------------------------------------------------------
# Transforms
#-------------------------------------------------------------------

proc thatchWhatTransforms {{pat *}} {
	return [lsort -dictionary [lsearch -inline -all -glob [list copy base64 zip] $pat]]
}

if {[namespace ensemble exists binary] && "encode" in [dict keys [namespace ensemble configure binary -map]] && "decode" in [dict keys [namespace ensemble configure binary -map]]} {
	namespace eval ::base64 {
		proc encode {args} { binary encode base64 {*}$args }
		proc decode {args} { binary decode base64 {*}$args }
		package provide base64 2000
	}
}; # if binary encode/decode

namespace eval ::zip {
	proc encode {args} { zlib compress   {*}$args }
	proc decode {args} { zlib decompress {*}$args }
	package provide zip 2000
}

proc thatchTransform {data process args} {
	foreach alg $args {
		if {$alg eq "copy"} { continue }
		package require $alg
		set data [::${alg}::${process} $data]
	}
	return $data
}

#-------------------------------------------------------------------
# Checksums
#-------------------------------------------------------------------

proc thatchWhatChecksums {{pat *}} {
	return [lsort -dictionary [lsearch -inline -all -glob [list md5 sha256] $pat]]
}

proc thatchChecksum {data alg} {
	package require $alg
	set ns [string map [list md5 md5 sha256 sha2] $alg]
	return [::${ns}::${alg} -hex $data]
}

proc thatchMakeChecksums {data args} {
	set cksums {}
	foreach alg [lsort -nocase -dictionary $args] {
		set alg [string tolower $alg]
		lappend cksums $alg:[thatchChecksum $data $alg]
	}
	return [join $cksums ,]
}

proc thatchCheckChecksums {data args} {
	set errs {}
	foreach cksum [lsort -nocase -dictionary [split $args ,]] {
		lassign [split $cksum :] alg sum
		set alg [string tolower $alg]
		set csum [thatchChecksum $data $alg]
		if {$csum ne $sum} {
			lappend errs [list $cksum $csum]
		}
	}
	return $errs
}

#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#
#	i == file info
#	N == \n
#	Z == ^Z
#
#	ZN# THATCH iiiiiiiiiiiiiiiN
#	||__sig___||__file info__||
#	|          |file info line|
#	|_____file info chunk_____|
#
#
#	file info       fis
#	file info line  fil
#	file info chunk fic
#	file info dict  fid
#	file spec       fs
#
#-------------------------------------------------------------------


#
# File Spec (fs) fields
#
proc fsFields {} {
	return [list name size mtime base64 zip md5 sha256 remove rehome modbase]
}

proc fid->fis {fid} {
	set name [dict get $fid name]
	if {[dict exists $fid rehome]} {
		if {[dict get $fid rehome] ne ""} {
			set name [dict get $fid rehome]
		}
		dict unset fid name
		dict unset fid rehome
	}
	if {[dict exists $fid modbase] && [dict get $fid modbase] eq {}} {
		dict unset fid modbase
	}
	return "$name $fid"
}

proc fis->fid {fis} {
	set q [dict create name [lindex $fis 0] {*}[lrange $fis 1 end]]
}

proc fidFields {args} {
	return "name transforms checksums size stored-size mtime modbase"
}

proc newFid {args} {
	return [dict create ${*}[fidFields] {*}$args]
}

#
# File Spec (fs) -> File Info Dict (fid)
#
proc fs->fid {fs {dv {}}} {
	if {$dv ne ""} { upvar $dv data }
	lassign $fs {*}[fsFields]
	set transforms [list]
	set chanconf [dict create]

	# If zip & base64 do zip first
	if {$zip == 1} {
		lappend transforms zip
		dict set chanconf -translation binary
	}
	if {$base64 == 1} {
		lappend transforms base64
		dict set chanconf -translation binary
	}

	set fid [dict create name $name]

	if {[llength $transforms] > 0} { dict set fid transforms $transforms }

	if {$dv eq ""} {
		return $fid
	}

	set filename $name
	file stat $filename s
	set data [inhaleIt $filename {*}$chanconf]

	set algs [list]
	if {$md5 == 1}  { lappend algs md5 }
	if {$sha256 == 1} { lappend algs sha256 }

	set checksums [thatchMakeChecksums $data {*}$algs]
	set data [thatchTransform $data encode {*}$transforms]
	set s(stored-size) [string bytelength $data]

	if {$checksums ne ""} {
		dict set fid checksums $checksums
	}

	if {$size == 1} {
		dict set fid size $s(size)
		if {$s(size) != $s(stored-size)} {
			dict set fid stored-size $s(stored-size)
		}
	}

	if {$mtime == 1} {
		dict set fid mtime $s(mtime)
	}

	dict set fid rehome $rehome
	dict set fid modbase $modbase

	return $fid
}

#
# Writes a Thatch header
#
proc fid->fic {of fid} {
	puts -nonewline $of [^Z][sig][fid->fis $fid]\n
}

#
# Writes a File Info Chunk (fic): the Thatch header and the file
# from a File Spec (fs)
#
proc fs->fic {of fs} {
	set fid [fs->fid $fs data]
	fid->fic $of $fid
	puts -nonewline $of $data
}

#
# Arguments (ags) -> File Spec List (fsl)
#
proc ags->fsl {ags} {
	set fsl [list]
	#         name      var     val default
	set al [list {*}{
		{ -         doArgs    0  1 }
		{ allsize   allSize   1 -1 }
		{ allmtime  allMtime  1 -1 }
		{ allbase64 allBase64 1 -1 }
		{ allzip    allZip    1 -1 }
		{ allmd5    allMd5    1 -1 }
		{ allsha256 allSha256 1 -1 }
		{ noallsize   allSize   0 -1 }
		{ noallmtime  allMtime  0 -1 }
		{ noallbase64 allBase64 0 -1 }
		{ noallzip    allZip    0 -1 }
		{ noallmd5    allMd5    0 -1 }
		{ noallsha256 allSha256 0 -1 }
		{ size      size      1 -1 }
		{ mtime     mtime     1 -1 }
		{ base64    base64    1 -1 }
		{ zip       zip       1 -1 }
		{ md5       md5       1 -1 }
		{ sha256    sha256    1 -1 }
		{ nosize    size      0 -1 }
		{ nomtime   mtime     0 -1 }
		{ nobase64  base64    0 -1 }
		{ nozip     zip       0 -1 }
		{ nomd5     md5       0 -1 }
		{ nosha256  sha256    0 -1 }
		{ allremove allRemove 1  0 }
		{ remove    remove    1 -1 }
		{ file      argIsFs   1  0 }
		{ home      argIsHm   1  0 }
		{ rehome    argIsRHm  1  0 }
		{ modbase   argIsMb   1  0 }
		{ rename    argIsRen  1  0 }
	}]
	foreach a $al { set [lindex $a 1] [lindex $a 3] }

	set    homedir ""
	set  rehomedir ""
	set modbasedir ""
	set resetOpts 1
	set gopts {}
	# Arg options override FileSpec options
	foreach arg $ags {
		if {$resetOpts} {
			set resetOpts 0
			set rename ""
			foreach a $al {
				# Only reset specific options
				if {[lindex $a 0] ni [fsFields]} { continue }
				set [lindex $a 1] [lindex $a 3]
			}
		}
		set home    ""
		set rehome  ""
		set modbase ""
		if {$argIsHm} {
			set argIsHm 0
			set homedir $arg
			continue
		}
		if {$argIsRHm} {
			set argIsRHm 0
			set rehomedir $arg
			continue
		}
		if {$argIsMb} {
			set argIsMb 0
			set modbasedir $arg
			continue
		}
		if {$argIsRen} {
			set argIsRen 0
			set rename $arg
			continue
		}
		# Does arg look like an option or a FileSpec?
		if {$doArgs && !$argIsFs && [string index $arg 0] eq "-"} {
			# Valid options are assigned to vars and added to gopts
			if {[set i [lsearch -index 0 $al [string range $arg 1 end]]] != -1} {
				set [lindex $al $i 1] [lindex $al $i 2]
				if {[string range $arg 1 3] eq "all" || [string range $arg 1 5] eq "noall"} {
					lappend gopts [string range $arg 1 end]
				}
				continue
			}
		}
		set argIsFs 0
		# Check FileSpec for options
		set opts [lassign [split $arg %] name]
		foreach opt $opts {
			if {$argIsHm} {
				set argIsHm 0
				set home $opt
				continue
			}
			if {($opt ni $gopts || $opt eq "home") && [set i [lsearch -index 0 $al $opt]] != -1} {
				set [lindex $al $i 1] [lindex $al $i 2]
			}
		}

		# Empty name can be used to set global options
		if {$name eq ""} {
			set resetOpts 1
			continue
		}

		# Use defaults if options not supplied
		#foreach n {size mtime base64 zip md5 sha256 remove rehome modbase} {}
		foreach n {size mtime base64 zip md5 sha256 remove} {
			if {[set $n] == -1} { set $n [set all[string totitle $n]] }
		}

		if {$home eq ""} {
			set home [expr {[string index $name 0] eq "/" ? "" : $homedir}]
		}

		set file [file join $home $name]

		if {$rename ne ""} {
			set name $rename
		}
		if {$rehomedir ne ""} {
			set rehome [file join [expr {[string index $name 0] eq "/" ? "" : $rehomedir}] $name]
		}
		if {$modbasedir ne "" && [regexp -- {^([_[:alpha:]][:_[:alnum:]]*)-([[:digit:]].*)[.]tm$} [file tail $file] -> pn pv]} {
			set modbase $modbasedir
		}

		lappend fsl [list $file $size $mtime $base64 $zip $md5 $sha256 $remove $rehome $modbase]

		set resetOpts 1
	}

	return $fsl
}

proc thatchOpen {fn {dir {in}}} {
	if {$dir eq "out"} {
		set mode w
		set eofc ""
	} else {
		set mode r
		set eofc [^Z]
	}
	return [openIt[string totitle $dir] $fn $mode -eofchar $eofc -encoding utf-8 -translation lf]
}

#
# Create or modify a Thatch file
#
proc thatchOp {op ifn ofn args} {
	set if [openItIn $ifn]
	if {$op eq "create"} {
		chan configure $if -eofchar [^Z]
	} else {
		chan configure $if -eofchar ""
	}
	set of [thatchOpen $ofn out]
	chan copy $if $of
	closeIt $if
	foreach fs [ags->fsl $args] {
		fs->fic $of $fs
	}
	closeIt $of
	return 1
}

proc thatchCopyCallbackMain {t fn f fid args} {
	lassign $args fsl of
	puts -nonewline $of [read $f]
	return skip
}

proc thatchCopyCallbackFile {t fn f fid args} {
	lassign $args fsl of

	set dosize 0
	set dobase64 0
	set undobase64 0
	set dozip 0
	set undozip 0
	set ofid $fid
	if {[set i [lsearch -index 0 $fsl [dict get $fid name]]] != -1} {
		if {[lindex $fsl $i [lsearch [fsFields] remove]]} {
			return ""
		}
		foreach fld [list size mtime] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0 && [dict exists $fid $fld]} {
				set fid [dict remove $fid $fld]
			} elseif {$v == 1} {
				if {$fld eq "size"} {
					set dosize 1
				}
			}
		}
		foreach fld [list base64 zip] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0} {
				if {![dict exists $fid transforms]} { continue }
				set transforms [dict get $fid transforms]
				if {[set i [lsearch $transforms $fld]] != -1} {
					set transforms [lreplace $transforms $i $i]
				}
				if {$fld eq "base64"} {set undobase64 1}
				if {$fld eq "zip"}    {set undozip 1}
			} elseif {$v == 1} {
				if {[dict exists $fid transforms]} { continue }
				lappend transforms $fld
				if {$fld eq "base64"} {set dobase64 1}
				if {$fld eq "zip"}    {set dozip 1}
			}
			if {[llength $transforms] == 0} {
				set fid [dict remove $fid transforms]
			} else {
				dict set fid transforms $transforms
			}
		}
		set checksums [dgwd {} $fid checksums]
		set sums1 [split $checksums :,]
		foreach fld [list md5 sha256] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0} {
				if {[set i [lsearch $sums1 $fld]] != -1} {
					set sums1 [lreplace $sums1 $i $i+1]
				}
			} elseif {$v == 1} {
			}
		}
		if {[llength $sums1] == 0} {
			if {[dict exists $fid checksums]} {
				set fid [dict remove $fid checksums]
			}
		} else {
			#dict set fid checksums something
		}
	}

	set d [read $f]

	if {$undobase64} {
		set d [thatchTransform $d decode {*}[dgwd {} $ofid transforms]]
		if {[dict exists $fid stored-size]} {
			set fid [dict remove $fid stored-size]
		}
	}
	if {$undozip} {
		set d [thatchTransform $d decode {*}[dgwd {} $ofid transforms]]
		if {[dict exists $fid stored-size]} {
			set fid [dict remove $fid stored-size]
		}
	}
	if {$dosize} {
		dict set fid size [string length $d]
	}
	if {$dobase64} {
		set d [thatchTransform $d encode {*}[dict get $fid transforms]]
		dict set fid stored-size [string length $d]
	}
	if {$dozip} {
		set d [thatchTransform $d encode {*}[dict get $fid transforms]]
		dict set fid stored-size [string length $d]
	}
	fid->fic $of [dict remove $fid offset]
	puts -nonewline $of $d

	return skip
}

proc thatchCopy {ifn ofn args} {
	set of [thatchOpen $ofn out]
	thatchTraverse $ifn [list [imn thatchCopyCallbackFile] [imn thatchCopyCallbackMain]] [ags->fsl $args] $of
	closeIt $of
	return 1
}

proc thatchNameMatch {fid g} {
	return [string match $g [dict get $fid name]]
}

proc zthatchFullFileGlobCallback {t fn f fid args} {
	variable res
	if {$t == 2} {
		dict set res main [read $f]
		return skip
	}
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res thIndex [dict get $fid name] $fid
		}
	}
	return ""
}

proc zthatchFullFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFullFileGlobCallback]] {*}$args
	return $res
}

proc thatchFullFileGlobCallback {t fn f fid args} {
	variable res
	if {$t == 2} {
		set size [string length [read $f]]
		dict set res $fn [dict set fid size $size]
		return skip
	}
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res [dict get $fid name] $fid
		}
	}
	return ""
}

proc thatchFullFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFullFileGlobCallback]] {*}$args
	return $res
}

proc thatchFileGlobCallback {t fn f fid args} {
	if {$t == 2} { return "" }
	variable res
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res [dict get $fid name] $fid
		}
	}
	return ""
}

proc thatchFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFileGlobCallback]] {*}$args
	return $res
}

proc thatchTraverse {fn cb args} {
	# This is supposed to be a Thatch file so it should conform
	set f [thatchOpen $fn]

	# Setup callbacks
	lassign $cb cb mcb
	if {$mcb eq ""} {
		set mcb $cb
	}

	# Get the 'main'
	set fid [dict create name $fn offset 0]
	set r [$mcb 2 $fn $f $fid {*}$args]
	if {$r ne "skip"} {
		read $f
	}

	# Note its largesse
	set offset [tell $f]

	# Now let's see if there's anything else
	while {1} {
		# We have hit eof so check if it's a ^Z eof or a real eof.
		# Do this by turning off -eofchar detection and reading one char:
		# it should be a ^Z. Set -eofchar back to ^Z after.
		chan configure $f -eofchar {}
		set cz [read $f 1]
		# Real eof? If so, $cz should also be "".
		if {[eof $f] || $cz eq ""} {
			# Normal case
			if {[eof $f] && $cz eq ""} {
				break
			}
			# todo, strange case
			break
		}
		# Gotta be control-zee
		if {$cz ne [^Z]} { binary scan $cz c c; error "Expected ^Z at $offset, got \"[format %#x $c]\"::[gets $f]" }
		chan configure $f -eofchar [^Z]
		# After the ^Z should be the sig
		set something [read $f [string length [sig]]]
		if {$something ne [sig]} {
			if {[eof $f]} {
				if {$something eq ""} {
					# null block
					continue
				}
			} else {
				append something [read $f]
			}
			# anon block
			continue
		}
		# Now we should be at the start of the file info line
		set fis [gets $f]
		# Do some checks ...
		if {[eof $f]} { explode }

		# Recalc offset:
		#    ^Z           \n                                 fis                               \n
		incr offset; incr offset [string length [sig]]; incr offset [string length $fis]; incr offset

		set fid [fis->fid $fis]
		dict set fid offset $offset
		set r [$cb 1 $fn $f $fid {*}$args]
		if {$r eq "stop"} { break }
		set size  [dgwd -1 $fid size]
		set ssize [dgwd -1 $fid stored-size]
		if {$ssize > 0} { set size $ssize }
		if {$size == -1} {
			if {$r ne "skip"} {
				read $f
			}
			set offset [tell $f]
		} else {
			if {$r ne "skip"} {
				seek $f $size current
			}
			incr offset $size
		}
	}
	closeIt $f
}

proc thatchExtract {fn fid} {
	set f [thatchOpen $fn]

	set name [lindex [dict keys $fid] 0]
	seek $f [dict get $fid $name offset] start

	set size [dgwd -1 $fid $name size]
	set ssize [dgwd -1 $fid $name stored-size]

	set transforms [dgwd {} $fid $name transforms]
	set checksums [dgwd "" $fid $name checksums]

	if {$ssize < 1} {
		set ssize $size
	}

	if {$ssize > 0} {
		set data [read $f $ssize]
	} else {
		set data [read $f]
	}

	closeIt $f

	set data [thatchTransform $data decode {*}[lreverse $transforms]]

	set checksum_errs [thatchCheckChecksums $data {*}$checksums]
	if {[llength $checksum_errs] > 0} {
		error "checksum mismatches ($fid):\n![join $checksum_errs \n]!"
	}

	return $data
}

proc thatchExtractMain {fn} {
	set f [thatchOpen $fn]
	set data [read $f]
	closeIt $f
	return $data
}

proc thatchIndexGlob {fn args} {
	variable thIndex
	if {[llength $args] == 0} { set args [list *] }
	set res [dict create]
	foreach a $args {
		set res [dict merge $res [dict filter $thIndex key $a]]
	}
	return $res
}

proc thatchCreateIndex {fn} {
	thatchDestroyIndex
	variable thIndex [thatchFileGlob $fn]
	interp alias {} [imn thatchGlob] {} [imn thatchIndexGlob]
}

proc thatchDestroyIndex {} {
	variable thIndex
	unset -nocomplain thIndex
	interp alias {} [imn thatchGlob] {} [imn thatchFileGlob]
}

#-------------------------------------------------------------------
# Overrides
#-------------------------------------------------------------------

proc thatchSourceOverride {fn args} {
	if {[dict size [set fid [thatchGlob $fn {*}$args]]] > 0} {
		set  script [info script]
		info script [lindex [dict keys $fid]]
		set  retval [uplevel 1 [thatchExtract $fn $fid]]
		info script $script
	} else {
		set retval [uplevel 1 [list ::tcl::tcl_source {*}$args]]
	}
	return $retval
}

proc thatchImageOverride {fn args} {
	if {
		   [llength $args] > 3
		&& [lindex $args 0] eq "create"
		&& [set ind [lsearch $args -file]] != -1
		&& [dict size [set fid [thatchGlob $fn [lindex $args $ind+1]]]] > 0
	} {
		lset args $ind -data
		lset args $ind+1 [thatchExtract $fn $fid]
	}
	return [uplevel 1 [list ::tcl::tcl_image {*}$args]]
}

proc thatchLoadOverride {fn args} {
	set lfi -1
	foreach lfn $args {
		incr lfi
		if {[string index $lfn 0] ne "-"} {
			break
		}
	}
	if {$lfi > -1 && [dict size [set fid [thatchGlob $fn $lfn]]] > 0} {
		set c [file tempfile t thatch]
		chan configure $c -translation binary
		puts -nonewline $c [thatchExtract $fn $fid]
		close $c
		lset args $lfi $t
		if {[llength $args] > ($lfi - 1)} {
			lappend args [string totitle [lindex [regexp -inline {^lib([a-zA-z_]+).*} [file tail $lfn]] end]]
		}
		set code [catch {uplevel 1 [list ::tcl::tcl_load {*}$args]} res]
		file delete -force $t
	} else {
		set code [catch {uplevel 1 [list ::tcl::tcl_load {*}$args]} res]
	}
	return -code $code $res
}

proc thatchFileOverride {fn args} {
	set ok 0
	if {[llength $args] == 2} {
		set fcmd [lindex $args 0]
		if {$fcmd in {exists isfile owned readable executable isdirectory writable atime mtime attributes size type}} {
			set ok 1
		}
	}
	if {[llength $args] == 3} {
		set fcmd [lindex $args 0]
		if {$fcmd in {lstat stat}} {
			set ok 1
		}
	}

	if {[lindex $args 0] eq "type" && [lindex $args 1] eq ""} { return "directory" }
	if {$ok && [dict size [set fid [thatchGlob $fn [lindex $args 1]]]] == 1} {
		switch -exact -- [lindex $args 0] {
		exists -
		isfile -
		owned -
		readable {
			return 1
		}
		executable -
		isdirectory -
		writable {
			return 0
		}
		type {
			return "file"
		}
		atime -
		mtime {
			return [dict get $fid mtime]
		}
		attributes {
			return 0o444
		}
		size {
			return $size
		}
		lstat -
		stat {
			upvar 1 [lindex $args 2] a
			unset -nocomplain a
			array set a [list atime [dict get $fid mtime] ctime [dict get $fid mtime] dev -1 gid -1 ino -1 mode 0o444 mtime [dict get $fid mtime] nlink 1 size [dict get $fid size] type "file" uid -1]
			return ""
		}
		default { explode! }
		}
	}
	return [uplevel 1 [list ::tcl::tcl_file {*}$args]]
}

proc thatchGlobOverride {fn args} {
#puts $fn:$args
#if {[catch {info level -1} e]} { set e :: }
#puts $e:$args
	variable thIndex
	array set g {-join 0 -tails 0 -types {} -directory "" -path "" patterns {}}
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -- $arg {
			-join - -tails { set g($arg) 1 }
			-dir - -directory - -path - -types {
				if {$arg eq "-dir"} { set arg "-directory" }
				set state setelem
				set elem $arg
			}
			-nocomplain {}
			-- { set state endopts }
			default { lappend g(patterns) $arg }
			}
		}
		endopts { lappend g(patterns) $arg }
		setelem { set state opt; set g($elem) $arg }
		}
	}
	if {$g(-join)} { set g(patterns) [file join {*}$g(patterns)] }
	set fid [dict create]
	foreach p $g(patterns) {
		set fid [dict merge $fid [thatchGlob $fn [file join $g(-directory) $g(-path) $p]]]
	}
	if {[dict size $fid] > 0} {
		return [dict keys $fid]
	}
	tailcall ::tcl::tcl_glob {*}$args
}

#-------------------------------------------------------------------
# Setup
#-------------------------------------------------------------------

proc overrideAProc {fn proc} {
	if {[llength [info commands ::$proc]]} {
		rename ::$proc ::tcl::tcl_$proc
		interp alias {} ::$proc {} [imn thatch[string totitle $proc]Override] $fn
	} else {
		
	}
}

proc thatchSetup {fn args} {
	set oo {}; foreach o [set op {file glob image load source index packages modules}] { lappend oo -no$o }
	foreach a $args {
		if {[lsearch -exact $oo $a] == -1} {
			error [unknownOption $a $oo]
		}
		set i [lsearch -exact $op [string range $a 3 end]]
		set op [lreplace $op $i $i]
	}
	set createIndex 0
	if {[set i [lsearch -exact $op index]] != -1} {
		set op [lreplace $op $i $i]
		set createIndex 1
	}
	set setupPackages 0
	if {[set i [lsearch -exact $op packages]] != -1} {
		set op [lreplace $op $i $i]
		set setupPackages 1
	}
	set setupModules 0
	if {[set i [lsearch -exact $op modules]] != -1} {
		set op [lreplace $op $i $i]
		set setupModules 1
	}
	foreach p $op { overrideAProc $fn $p }
	if {$createIndex} { thatchCreateIndex $fn }
	if {$setupPackages} {
		dict for {k v} [thatchGlob $fn *pkgIndex.tcl *tclIndex] {
			set ::dir [file dirname $k]
			uplevel #0 [thatchExtract $fn [dict create $k $v]]
		}
		unset -nocomplain ::dir
	}
	if {$setupModules} {
		dict for {k v} [thatchGlob $fn {*-[0-9]*.tm}] {
			if {![regexp -- {^([_[:alpha:]][:_[:alnum:]]*)-([[:digit:]].*)[.]tm$} [file tail $k] -> pn pv]} { continue }
			if {[dict exists $v modbase]} {
				set pn [string map {/ ::} [string trimleft [string range $k [string length [dict get $v modbase]] end-[string length [file tail $k]]] /]]$pn
			}
			package ifneeded $pn $pv "[::list package provide $pn $pv];[::list source -encoding utf-8 $k]"
		}
	}
}

#-------------------------------------------------------------------
# Package Setup
#-------------------------------------------------------------------

proc setup {} {
	package require Tcl 8.5-
	aliasconst ^Z \x1a sig "\n# THATCH "
	interp alias {} [imn thatchCreate] {} [imn thatchOp] create
	interp alias {} [imn thatchAdd]    {} [imn thatchOp] add
	thatchDestroyIndex
	namespace export thatch*
	package provide thatch 0.5
}

namespace current}]::setup

# EOF

# THATCH /herbie/tennessee.tcl 
#! /bin/sh
# \
exec tclsh${TCL} "$0" "$@"

#
# Tennessee
#
#
#
# Stuart Cassoff
#
# Summer 2018
# Version 0.1
#
# Fall 2022
# Version 0.2
#
# Spring 2023
# Version 0.3
#

package require Tcl 8.6-

[namespace eval ::tennessee {

proc setup {} {
	package require Tk
	package require herbie
	namespace import ::herbie::*

	::herbie::ignition
	::herbie::scaleFonts 2

	setupTws
}

proc setupTws {} { oo::class create TennesseeWorkspace {
	superclass ::herbie::Workspace
	variable histFile
	variable saveHistOnExit
constructor {args} {
	next {*}$args
	set saveHistOnExit no
	set histFile ""
}
method done {} {
	if {$saveHistOnExit} { my saveHist }
	next
}
eval [::tu::tcl::makeRWAccessors histFile saveHistOnExit]
method saveHist {} {
	if {$histFile eq ""} { return }
	catch {
	set f [open $histFile w]
	puts $f [join [[my wCombo] cget -values] \n]
	close $f
	}
}
method loadHist {} {
	if {$histFile eq ""} { return }
	if {![file readable $histFile]} { return }
	catch {
	set f [open $histFile]
	set d [read -nonewline $f]
	close $f
	[my wCombo] configure -values [split $d \n]
	}
}
}; # oo::class create TennesseeWorkspace
}; # proc setupTws

proc openTennesseeWorkspace {{on ::} {histFile {}} args} {
	set textFns [list]
	set sourceFns [list]
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-l      { set state $arg }
			default { lappend sourceFns $arg}
			}; # hctiws
		}
		-l { set state opt; lappend textFns $arg }
		}; # hctiws
	}
	set ws [TennesseeWorkspace new $on {} {} -focus combo]
	bind [$ws wW] <Alt-q> [bind [$ws wW] <Control-q>]
	bind [$ws wW] <Alt-q> +[list destroy .]
	bind [$ws wW] <Control-q> {}
	[$ws wM] entryconfigure end -command {::event generate [::tu::epm::epmW] <Alt-q>}
	::tu::toplevels::tk::onToplevelDelete [$ws wW] "[::tu::toplevels::tk::onToplevelDelete [$ws wW]];destroy ."
	$ws textAppend \n {} [$ws banner] {} \n
	if {$histFile ne ""} {
		$ws histFile $histFile
	}
	$ws saveHistOnExit yes
	$ws loadHist
	foreach textFn $textFns { $ws loadFileIntoText $textFn }
	foreach sourceFn $sourceFns { $ws doBlob "source $sourceFn" }
}

proc go {args} {
	set histFile ~/.config/tennessee/history
	setup
	tk appname Tennessee
	wm withdraw .
	openTennesseeWorkspace {::} $histFile {*}$args
}

namespace current}]::go {*}$::argv


# EOF

# THATCH /herbie/packages/struct/pkgIndex.tcl 
if {![package vsatisfies [package provide Tcl] 8.2]} {return}
package ifneeded struct            2.1   [list source [file join $dir struct.tcl]]
package ifneeded struct            1.4   [list source [file join $dir struct1.tcl]]

package ifneeded struct::queue     1.4.5 [list source [file join $dir queue.tcl]]
package ifneeded struct::stack     1.5.3 [list source [file join $dir stack.tcl]]
package ifneeded struct::tree      2.1.2 [list source [file join $dir tree.tcl]]
package ifneeded struct::matrix    2.0.4 [list source [file join $dir matrix.tcl]]
package ifneeded struct::pool      1.2.3 [list source [file join $dir pool.tcl]]
package ifneeded struct::record    1.2.2 [list source [file join $dir record.tcl]]
package ifneeded struct::set       2.2.3 [list source [file join $dir sets.tcl]]
package ifneeded struct::prioqueue 1.4   [list source [file join $dir prioqueue.tcl]]
package ifneeded struct::skiplist  1.3   [list source [file join $dir skiplist.tcl]]

package ifneeded struct::graph     1.2.1 [list source [file join $dir graph1.tcl]]
package ifneeded struct::tree      1.2.2 [list source [file join $dir tree1.tcl]]
package ifneeded struct::matrix    1.2.2 [list source [file join $dir matrix1.tcl]]

if {![package vsatisfies [package provide Tcl] 8.4]} {return}
package ifneeded struct::list        1.8.5  [list source [file join $dir list.tcl]]
package ifneeded struct::list::test  1.8.4  [list source [file join $dir list.test.tcl]]
package ifneeded struct::graph     2.4.3  [list source [file join $dir graph.tcl]]
package ifneeded struct::map       1      [list source [file join $dir map.tcl]]

if {![package vsatisfies [package provide Tcl] 8.5]} {return}

if {![package vsatisfies [package provide Tcl] 8.6]} {return}
package ifneeded struct::disjointset 1.1 [list source [file join $dir disjointset.tcl]]
package ifneeded struct::graph::op 0.11.3 [list source [file join $dir graphops.tcl]]

# THATCH /herbie/packages/struct/list.tcl 
#----------------------------------------------------------------------
#
# list.tcl --
#
#	Definitions for extended processing of Tcl lists.
#
# Copyright (c) 2003 by Kevin B. Kenny.  All rights reserved.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: list.tcl,v 1.27 2011/09/17 14:35:36 mic42 Exp $
#
#----------------------------------------------------------------------

package require Tcl 8.4
package require cmdline

namespace eval ::struct { namespace eval list {} }

namespace eval ::struct::list {
    namespace export list

    if {0} {
	# Possibly in the future.
	namespace export Lassign
	namespace export LdbJoin
	namespace export LdbJoinOuter
	namespace export Ldelete
	namespace export Lequal
	namespace export Lfilter
	namespace export Lfilterfor
	namespace export Lfirstperm
	namespace export Lflatten
	namespace export Lfold
	namespace export Lforeachperm
	namespace export Liota
	namespace export LlcsInvert
	namespace export LlcsInvert2
	namespace export LlcsInvertMerge
	namespace export LlcsInvertMerge2
	namespace export LlongestCommonSubsequence
	namespace export LlongestCommonSubsequence2
	namespace export Lmap
	namespace export Lmapfor
	namespace export Lnextperm
	namespace export Lpermutations
	namespace export Lrepeat
	namespace export Lrepeatn
	namespace export Lreverse
	namespace export Lshift
	namespace export Lswap
	namespace export Lshuffle
    }
}

##########################
# Public functions

# ::struct::list::list --
#
#	Command that access all list commands.
#
# Arguments:
#	cmd	Name of the subcommand to dispatch to.
#	args	Arguments for the subcommand.
#
# Results:
#	Whatever the result of the subcommand is.

proc ::struct::list::list {cmd args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 1 } {
	return -code error "wrong # args: should be \"$cmd ?arg arg ...?\""
    }
    set sub L$cmd
    if { [llength [info commands ::struct::list::$sub]] == 0 } {
	set optlist [info commands ::struct::list::L*]
	set xlist {}
	foreach p $optlist {
	    lappend xlist [string range $p 1 end]
	}
	return -code error \
		"bad option \"$cmd\": must be [linsert [join $xlist ", "] "end-1" "or"]"
    }
    return [uplevel 1 [linsert $args 0 ::struct::list::$sub]]
}

##########################
# Private functions follow

proc ::struct::list::K { x y } { set x }

##########################
# Implementations of the functionality.
#

# ::struct::list::LlongestCommonSubsequence --
#
#       Computes the longest common subsequence of two lists.
#
# Parameters:
#       sequence1, sequence2 -- Two lists to compare.
#	maxOccurs -- If provided, causes the procedure to ignore
#		     lines that appear more than $maxOccurs times
#		     in the second sequence.  See below for a discussion.
# Results:
#       Returns a list of two lists of equal length.
#       The first sublist is of indices into sequence1, and the
#       second sublist is of indices into sequence2.  Each corresponding
#       pair of indices corresponds to equal elements in the sequences;
#       the sequence returned is the longest possible.
#
# Side effects:
#       None.
#
# Notes:
#
#	While this procedure is quite rapid for many tasks of file
# comparison, its performance degrades severely if the second list
# contains many equal elements (as, for instance, when using this
# procedure to compare two files, a quarter of whose lines are blank.
# This drawback is intrinsic to the algorithm used (see the References
# for details).  One approach to dealing with this problem that is
# sometimes effective in practice is arbitrarily to exclude elements
# that appear more than a certain number of times.  This number is
# provided as the 'maxOccurs' parameter.  If frequent lines are
# excluded in this manner, they will not appear in the common subsequence
# that is computed; the result will be the longest common subsequence
# of infrequent elements.
#
#	The procedure struct::list::LongestCommonSubsequence2
# functions as a wrapper around this procedure; it computes the longest
# common subsequence of infrequent elements, and then subdivides the
# subsequences that lie between the matches to approximate the true
# longest common subsequence.
#
# References:
#	J. W. Hunt and M. D. McIlroy, "An algorithm for differential
#	file comparison," Comp. Sci. Tech. Rep. #41, Bell Telephone
#	Laboratories (1976). Available on the Web at the second
#	author's personal site: http://www.cs.dartmouth.edu/~doug/

proc ::struct::list::LlongestCommonSubsequence {
    sequence1
    sequence2
    {maxOccurs 0x7fffffff}
} {
    # Construct a set of equivalence classes of lines in file 2

    set index 0
    foreach string $sequence2 {
	lappend eqv($string) $index
	incr index
    }

    # K holds descriptions of the common subsequences.
    # Initially, there is one common subsequence of length 0,
    # with a fence saying that it includes line -1 of both files.
    # The maximum subsequence length is 0; position 0 of
    # K holds a fence carrying the line following the end
    # of both files.

    lappend K [::list -1 -1 {}]
    lappend K [::list [llength $sequence1] [llength $sequence2] {}]
    set k 0

    # Walk through the first file, letting i be the index of the line and
    # string be the line itself.

    set i 0
    foreach string $sequence1 {
	# Consider each possible corresponding index j in the second file.

	if { [info exists eqv($string)]
	     && [llength $eqv($string)] <= $maxOccurs } {

	    # c is the candidate match most recently found, and r is the
	    # length of the corresponding subsequence.

	    set r 0
	    set c [lindex $K 0]

	    foreach j $eqv($string) {
		# Perform a binary search to find a candidate common
		# subsequence to which may be appended this match.

		set max $k
		set min $r
		set s [expr { $k + 1 }]
		while { $max >= $min } {
		    set mid [expr { ( $max + $min ) / 2 }]
		    set bmid [lindex [lindex $K $mid] 1]
		    if { $j == $bmid } {
			break
		    } elseif { $j < $bmid } {
			set max [expr {$mid - 1}]
		    } else {
			set s $mid
			set min [expr { $mid + 1 }]
		    }
		}

		# Go to the next match point if there is no suitable
		# candidate.

		if { $j == [lindex [lindex $K $mid] 1] || $s > $k} {
		    continue
		}

		# s is the sequence length of the longest sequence
		# to which this match point may be appended. Make
		# a new candidate match and store the old one in K
		# Set r to the length of the new candidate match.

		set newc [::list $i $j [lindex $K $s]]
		if { $r >= 0 } {
		    lset K $r $c
		}
		set c $newc
		set r [expr { $s + 1 }]

		# If we've extended the length of the longest match,
		# we're done; move the fence.

		if { $s >= $k } {
		    lappend K [lindex $K end]
		    incr k
		    break
		}
	    }

	    # Put the last candidate into the array

	    lset K $r $c
	}

	incr i
    }

    # Package the common subsequence in a convenient form

    set seta {}
    set setb {}
    set q [lindex $K $k]

    for { set i 0 } { $i < $k } {incr i } {
	lappend seta {}
	lappend setb {}
    }
    while { [lindex $q 0] >= 0 } {
	incr k -1
	lset seta $k [lindex $q 0]
	lset setb $k [lindex $q 1]
	set q [lindex $q 2]
    }

    return [::list $seta $setb]
}

# ::struct::list::LlongestCommonSubsequence2 --
#
#	Derives an approximation to the longest common subsequence
#	of two lists.
#
# Parameters:
#	sequence1, sequence2 - Lists to be compared
#	maxOccurs - Parameter for imprecise matching - see below.
#
# Results:
#       Returns a list of two lists of equal length.
#       The first sublist is of indices into sequence1, and the
#       second sublist is of indices into sequence2.  Each corresponding
#       pair of indices corresponds to equal elements in the sequences;
#       the sequence returned is an approximation to the longest possible.
#
# Side effects:
#       None.
#
# Notes:
#	This procedure acts as a wrapper around the companion procedure
#	struct::list::LongestCommonSubsequence and accepts the same
#	parameters.  It first computes the longest common subsequence of
#	elements that occur no more than $maxOccurs times in the
#	second list.  Using that subsequence to align the two lists,
#	it then tries to augment the subsequence by computing the true
#	longest common subsequences of the sublists between matched pairs.

proc ::struct::list::LlongestCommonSubsequence2 {
    sequence1
    sequence2
    {maxOccurs 0x7fffffff}
} {
    # Derive the longest common subsequence of elements that occur at
    # most $maxOccurs times

    foreach { l1 l2 } \
	[LlongestCommonSubsequence $sequence1 $sequence2 $maxOccurs] {
	    break
	}

    # Walk through the match points in the sequence just derived.

    set result1 {}
    set result2 {}
    set n1 0
    set n2 0
    foreach i1 $l1 i2 $l2 {
	if { $i1 != $n1 && $i2 != $n2 } {
	    # The match points indicate that there are unmatched
	    # elements lying between them in both input sequences.
	    # Extract the unmatched elements and perform precise
	    # longest-common-subsequence analysis on them.

	    set subl1 [lrange $sequence1 $n1 [expr { $i1 - 1 }]]
	    set subl2 [lrange $sequence2 $n2 [expr { $i2 - 1 }]]
	    foreach { m1 m2 } [LlongestCommonSubsequence $subl1 $subl2] break
	    foreach j1 $m1 j2 $m2 {
		lappend result1 [expr { $j1 + $n1 }]
		lappend result2 [expr { $j2 + $n2 }]
	    }
	}

	# Add the current match point to the result

	lappend result1 $i1
	lappend result2 $i2
	set n1 [expr { $i1 + 1 }]
	set n2 [expr { $i2 + 1 }]
    }

    # If there are unmatched elements after the last match in both files,
    # perform precise longest-common-subsequence matching on them and
    # add the result to our return.

    if { $n1 < [llength $sequence1] && $n2 < [llength $sequence2] } {
	set subl1 [lrange $sequence1 $n1 end]
	set subl2 [lrange $sequence2 $n2 end]
	foreach { m1 m2 } [LlongestCommonSubsequence $subl1 $subl2] break
	foreach j1 $m1 j2 $m2 {
	    lappend result1 [expr { $j1 + $n1 }]
	    lappend result2 [expr { $j2 + $n2 }]
	}
    }

    return [::list $result1 $result2]
}

# ::struct::list::LlcsInvert --
#
#	Takes the data describing a longest common subsequence of two
#	lists and inverts the information in the sense that the result
#	of this command will describe the differences between the two
#	sequences instead of the identical parts.
#
# Parameters:
#	lcsData		longest common subsequence of two lists as
#			returned by longestCommonSubsequence(2).
# Results:
#	Returns a single list whose elements describe the differences
#	between the original two sequences. Each element describes
#	one difference through three pieces, the type of the change,
#	a pair of indices in the first sequence and a pair of indices
#	into the second sequence, in this order.
#
# Side effects:
#       None.

proc ::struct::list::LlcsInvert {lcsData len1 len2} {
    return [LlcsInvert2 [::lindex $lcsData 0] [::lindex $lcsData 1] $len1 $len2]
}

proc ::struct::list::LlcsInvert2 {idx1 idx2 len1 len2} {
    set result {}
    set last1 -1
    set last2 -1

    foreach a $idx1 b $idx2 {
	# Four possible cases.
	# a) last1 ... a and last2 ... b are not empty.
	#    This is a 'change'.
	# b) last1 ... a is empty, last2 ... b is not.
	#    This is an 'addition'.
	# c) last1 ... a is not empty, last2 ... b is empty.
	#    This is a deletion.
	# d) If both ranges are empty we can ignore the
	#    two current indices.

	set empty1 [expr {($a - $last1) <= 1}]
	set empty2 [expr {($b - $last2) <= 1}]

	if {$empty1 && $empty2} {
	    # Case (d), ignore the indices
	} elseif {$empty1} {
	    # Case (b), 'addition'.
	    incr last2 ; incr b -1
	    lappend result [::list added [::list $last1 $a] [::list $last2 $b]]
	    incr b
	} elseif {$empty2} {
	    # Case (c), 'deletion'
	    incr last1 ; incr a -1
	    lappend result [::list deleted [::list $last1 $a] [::list $last2 $b]]
	    incr a
	} else {
	    # Case (q), 'change'.
	    incr last1 ; incr a -1
	    incr last2 ; incr b -1
	    lappend result [::list changed [::list $last1 $a] [::list $last2 $b]]
	    incr a
	    incr b
	}

	set last1 $a
	set last2 $b
    }

    # Handle the last chunk, using the information about the length of
    # the original sequences.

    set empty1 [expr {($len1 - $last1) <= 1}]
    set empty2 [expr {($len2 - $last2) <= 1}]

    if {$empty1 && $empty2} {
	# Case (d), ignore the indices
    } elseif {$empty1} {
	# Case (b), 'addition'.
	incr last2 ; incr len2 -1
	lappend result [::list added [::list $last1 $len1] [::list $last2 $len2]]
    } elseif {$empty2} {
	# Case (c), 'deletion'
	incr last1 ; incr len1 -1
	lappend result [::list deleted [::list $last1 $len1] [::list $last2 $len2]]
    } else {
	# Case (q), 'change'.
	incr last1 ; incr len1 -1
	incr last2 ; incr len2 -1
	lappend result [::list changed [::list $last1 $len1] [::list $last2 $len2]]
    }

    return $result
}

proc ::struct::list::LlcsInvertMerge {lcsData len1 len2} {
    return [LlcsInvertMerge2 [::lindex $lcsData 0] [::lindex $lcsData 1] $len1 $len2]
}

proc ::struct::list::LlcsInvertMerge2 {idx1 idx2 len1 len2} {
    set result {}
    set last1 -1
    set last2 -1

    foreach a $idx1 b $idx2 {
	# Four possible cases.
	# a) last1 ... a and last2 ... b are not empty.
	#    This is a 'change'.
	# b) last1 ... a is empty, last2 ... b is not.
	#    This is an 'addition'.
	# c) last1 ... a is not empty, last2 ... b is empty.
	#    This is a deletion.
	# d) If both ranges are empty we can ignore the
	#    two current indices. For merging we simply
	#    take the information from the input.

	set empty1 [expr {($a - $last1) <= 1}]
	set empty2 [expr {($b - $last2) <= 1}]

	if {$empty1 && $empty2} {
	    # Case (d), add 'unchanged' chunk.
	    set type --
	    foreach {type left right} [lindex $result end] break
	    if {[string match unchanged $type]} {
		# There is an existing result to extend
		lset left end $a
		lset right end $b
		lset result end [::list unchanged $left $right]
	    } else {
		# There is an unchanged result at the start of the list;
		# it may be extended.
		lappend result [::list unchanged [::list $a $a] [::list $b $b]]
	    }
	} else {
	    if {$empty1} {
		# Case (b), 'addition'.
		incr last2 ; incr b -1
		lappend result [::list added [::list $last1 $a] [::list $last2 $b]]
		incr b
	    } elseif {$empty2} {
		# Case (c), 'deletion'
		incr last1 ; incr a -1
		lappend result [::list deleted [::list $last1 $a] [::list $last2 $b]]
		incr a
	    } else {
		# Case (a), 'change'.
		incr last1 ; incr a -1
		incr last2 ; incr b -1
		lappend result [::list changed [::list $last1 $a] [::list $last2 $b]]
		incr a
		incr b
	    }
	    # Finally, the two matching lines are a new unchanged region
	    lappend result [::list unchanged [::list $a $a] [::list $b $b]]
	}
	set last1 $a
	set last2 $b
    }

    # Handle the last chunk, using the information about the length of
    # the original sequences.

    set empty1 [expr {($len1 - $last1) <= 1}]
    set empty2 [expr {($len2 - $last2) <= 1}]

    if {$empty1 && $empty2} {
	# Case (d), ignore the indices
    } elseif {$empty1} {
	# Case (b), 'addition'.
	incr last2 ; incr len2 -1
	lappend result [::list added [::list $last1 $len1] [::list $last2 $len2]]
    } elseif {$empty2} {
	# Case (c), 'deletion'
	incr last1 ; incr len1 -1
	lappend result [::list deleted [::list $last1 $len1] [::list $last2 $len2]]
    } else {
	# Case (q), 'change'.
	incr last1 ; incr len1 -1
	incr last2 ; incr len2 -1
	lappend result [::list changed [::list $last1 $len1] [::list $last2 $len2]]
    }

    return $result
}

# ::struct::list::Lreverse --
#
#	Reverses the contents of the list and returns the reversed
#	list as the result of the command.
#
# Parameters:
#	sequence	List to be reversed.
#
# Results:
#	The sequence in reverse.
#
# Side effects:
#       None.

proc ::struct::list::Lreverse {sequence} {
    set l [::llength $sequence]

    # Shortcut for lists where reversing yields the list itself
    if {$l < 2} {return $sequence}

    # Perform true reversal
    set res [::list]
    while {$l} {
	::lappend res [::lindex $sequence [incr l -1]]
    }
    return $res
}


# ::struct::list::Lassign --
#
#	Assign list elements to variables.
#
# Parameters:
#	sequence	List to assign
#	args		Names of the variables to assign to.
#
# Results:
#	The unassigned part of the sequence. Can be empty.
#
# Side effects:
#       None.

# Do a compatibility version of [assign] for pre-8.5 versions of Tcl.

if { [package vcompare [package provide Tcl] 8.5] < 0 } {
    # 8.4
    proc ::struct::list::Lassign {sequence v args} {
	set args [linsert $args 0 $v]
	set a [::llength $args]

	# Nothing to assign.
	#if {$a == 0} {return $sequence}

	# Perform assignments
	set i 0
	foreach v $args {
	    upvar 1 $v var
	    set      var [::lindex $sequence $i]
	    incr i
	}

	# Return remainder, if there is any.
	return [::lrange $sequence $a end]
}

} else {
    # For 8.5+ simply redirect the method to the core command.

    interp alias {} ::struct::list::Lassign {} lassign
}


# ::struct::list::Lshift --
#
#	Shift a list in a variable one element down, and return first element
#
# Parameters:
#	listvar		Name of variable containing the list to shift.
#
# Results:
#	The first element of the list.
#
# Side effects:
#       After the call the list variable will contain
#	the second to last elements of the list.

proc ::struct::list::Lshift {listvar} {
    upvar 1 $listvar list
    set list [Lassign [K $list [set list {}]] v]
    return $v
}


# ::struct::list::Lflatten --
#
#	Remove nesting from the input
#
# Parameters:
#	sequence	List to flatten
#
# Results:
#	The input list with one or all levels of nesting removed.
#
# Side effects:
#       None.

proc ::struct::list::Lflatten {args} {
    if {[::llength $args] < 1} {
	return -code error \
		"wrong#args: should be \"::struct::list::Lflatten ?-full? ?--? sequence\""
    }

    set full 0
    while {[string match -* [set opt [::lindex $args 0]]]} {
	switch -glob -- $opt {
	    -full   {set full 1}
	    --      {
                set args [::lrange $args 1 end]
                break ; # fix ticket 6e778502b8 -- break exits while loop
            }
	    default {
		return -code error "Unknown option \"$opt\", should be either -full, or --"
	    }
	}
	set args [::lrange $args 1 end]
    }

    if {[::llength $args] != 1} {
	return -code error \
		"wrong#args: should be \"::struct::list::Lflatten ?-full? ?--? sequence\""
    }

    set sequence [::lindex $args 0]
    set cont 1
    while {$cont} {
	set cont 0
	set result [::list]
	foreach item $sequence {
	    # catch/llength detects if the item is following the list
	    # syntax.

	    if {[catch {llength $item} len]} {
		# Element is not a list in itself, no flatten, add it
		# as is.
		lappend result $item
	    } else {
		# Element is parseable as list, add all sub-elements
		# to the result.
		foreach e $item {
		    lappend result $e
		}
	    }
	}
	if {$full && [string compare $sequence $result]} {set cont 1}
	set sequence $result
    }
    return $result
}


# ::struct::list::Lmap --
#
#	Apply command to each element of a list and return concatenated results.
#
# Parameters:
#	sequence	List to operate on
#	cmdprefix	Operation to perform on the elements.
#
# Results:
#	List containing the result of applying cmdprefix to the elements of the
#	sequence.
#
# Side effects:
#       None of its own, but the command prefix can perform arbitry actions.

proc ::struct::list::Lmap {sequence cmdprefix} {
    # Shortcut when nothing is to be done.
    if {[::llength $sequence] == 0} {return $sequence}

    set res [::list]
    foreach item $sequence {
	lappend res [uplevel 1 [linsert $cmdprefix end $item]]
    }
    return $res
}

# ::struct::list::Lmapfor --
#
#	Apply a script to each element of a list and return concatenated results.
#
# Parameters:
#	sequence	List to operate on
#	script		The script to run on the elements.
#
# Results:
#	List containing the result of running script on the elements of the
#	sequence.
#
# Side effects:
#       None of its own, but the script can perform arbitry actions.

proc ::struct::list::Lmapfor {var sequence script} {
    # Shortcut when nothing is to be done.
    if {[::llength $sequence] == 0} {return $sequence}
    upvar 1 $var item

    set res [::list]
    foreach item $sequence {
	lappend res [uplevel 1 $script]
    }
    return $res
}

# ::struct::list::Lfilter --
#
#	Apply command to each element of a list and return elements passing the test.
#
# Parameters:
#	sequence	List to operate on
#	cmdprefix	Test to perform on the elements.
#
# Results:
#	List containing the elements of the input passing the test command.
#
# Side effects:
#       None of its own, but the command prefix can perform arbitrary actions.

proc ::struct::list::Lfilter {sequence cmdprefix} {
    # Shortcut when nothing is to be done.
    if {[::llength $sequence] == 0} {return $sequence}
    return [uplevel 1 [::list ::struct::list::Lfold $sequence {} [::list ::struct::list::FTest $cmdprefix]]]
}

proc ::struct::list::FTest {cmdprefix result item} {
    set pass [uplevel 1 [::linsert $cmdprefix end $item]]
    if {$pass} {::lappend result $item}
    return $result
}

# ::struct::list::Lfilterfor --
#
#	Apply expr condition to each element of a list and return elements passing the test.
#
# Parameters:
#	sequence	List to operate on
#	expr		Test to perform on the elements.
#
# Results:
#	List containing the elements of the input passing the test expression.
#
# Side effects:
#       None of its own, but the command prefix can perform arbitrary actions.

proc ::struct::list::Lfilterfor {var sequence expr} {
    # Shortcut when nothing is to be done.
    if {[::llength $sequence] == 0} {return $sequence}

    upvar 1 $var item
    set result {}
    foreach item $sequence {
	if {[uplevel 1 [::list ::expr $expr]]} {
	    lappend result $item
	}
    }
    return $result
}

# ::struct::list::Lsplit --
#
#	Apply command to each element of a list and return elements passing
#	and failing the test. Basic idea by Salvatore Sanfilippo
#	(http://wiki.tcl.tk/lsplit). The implementation here is mine (AK),
#	and the interface is slightly different (Command prefix with the
#	list element given to it as argument vs. variable + script).
#
# Parameters:
#	sequence	List to operate on
#	cmdprefix	Test to perform on the elements.
#	args = empty | (varPass varFail)
#
# Results:
#	If the variables are specified then a list containing the
#	numbers of passing and failing elements, in this
#	order. Otherwise a list having two elements, the lists of
#	passing and failing elements, in this order.
#
# Side effects:
#       None of its own, but the command prefix can perform arbitrary actions.

proc ::struct::list::Lsplit {sequence cmdprefix args} {
    set largs [::llength $args]
    if {$largs == 0} {
	# Shortcut when nothing is to be done.
	if {[::llength $sequence] == 0} {return {{} {}}}
	return [uplevel 1 [::list [namespace which Lfold] $sequence {} [
		::list ::struct::list::PFTest $cmdprefix]]]
    } elseif {$largs == 2} {
	# Shortcut when nothing is to be done.
	foreach {pv fv} $args break
	upvar 1 $pv pass $fv fail
	if {[::llength $sequence] == 0} {
	    set pass {}
	    set fail {}
	    return {0 0}
	}
	foreach {pass fail} [uplevel 1 [
		::list ::struct::list::Lfold $sequence {} [
			::list ::struct::list::PFTest $cmdprefix]]] break
	return [::list [llength $pass] [llength $fail]]
    } else {
	return -code error \
		"wrong#args: should be \"::struct::list::Lsplit sequence cmdprefix ?passVar failVar?"
    }
}

proc ::struct::list::PFTest {cmdprefix result item} {
    set passing [uplevel 1 [::linsert $cmdprefix end $item]]
    set pass {} ; set fail {}
    foreach {pass fail} $result break
    if {$passing} {
	::lappend pass $item
    } else {
	::lappend fail $item
    }
    return [::list $pass $fail]
}

# ::struct::list::Lfold --
#
#	Fold list into one value.
#
# Parameters:
#	sequence	List to operate on
#	cmdprefix	Operation to perform on the elements.
#
# Results:
#	Result of applying cmdprefix to the elements of the
#	sequence.
#
# Side effects:
#       None of its own, but the command prefix can perform arbitry actions.

proc ::struct::list::Lfold {sequence initialvalue cmdprefix} {
    # Shortcut when nothing is to be done.
    if {[::llength $sequence] == 0} {return $initialvalue}

    set res $initialvalue
    foreach item $sequence {
	set res [uplevel 1 [linsert $cmdprefix end $res $item]]
    }
    return $res
}

# ::struct::list::Liota --
#
#	Return a list containing the integer numbers 0 ... n-1
#
# Parameters:
#	n	First number not in the generated list.
#
# Results:
#	A list containing integer numbers.
#
# Side effects:
#       None

proc ::struct::list::Liota {n} {
    set retval [::list]
    for {set i 0} {$i < $n} {incr i} {
	::lappend retval $i
    }
    return $retval
}

# ::struct::list::Ldelete --
#
#	Delete an element from a list by name.
#	Similar to 'struct::set exclude', however
#	this here preserves order and list intrep.
#
# Parameters:
#	a	First list to compare.
#	b	Second list to compare.
#
# Results:
#	A boolean. True if the lists are delete.
#
# Side effects:
#       None

proc ::struct::list::Ldelete {var item} {
    upvar 1 $var list
    set pos [lsearch -exact $list $item]
    if {$pos < 0} return
    set list [lreplace [K $list [set list {}]] $pos $pos]
    return
}

# ::struct::list::Lequal --
#
#	Compares two lists for equality
#	(Same length, Same elements in same order).
#
# Parameters:
#	a	First list to compare.
#	b	Second list to compare.
#
# Results:
#	A boolean. True if the lists are equal.
#
# Side effects:
#       None

proc ::struct::list::Lequal {a b} {
    # Author of this command is "Richard Suchenwirth"

    if {[::llength $a] != [::llength $b]} {return 0}
    if {[::lindex $a 0] == $a && [::lindex $b 0] == $b} {return [string equal $a $b]}
    foreach i $a j $b {if {![Lequal $i $j]} {return 0}}
    return 1
}

# ::struct::list::Lrepeatn --
#
#	Create a list repeating the same value over again.
#
# Parameters:
#	value	value to use in the created list.
#	args	Dimension(s) of the (nested) list to create.
#
# Results:
#	A list
#
# Side effects:
#       None

proc ::struct::list::Lrepeatn {value args} {
    if {[::llength $args] == 1} {set args [::lindex $args 0]}
    set buf {}
    foreach number $args {
	incr number 0 ;# force integer (1)
	set buf {}
	for {set i 0} {$i<$number} {incr i} {
	    ::lappend buf $value
	}
	set value $buf
    }
    return $buf
    # (1): See 'Stress testing' (wiki) for why this makes the code safer.
}

# ::struct::list::Lrepeat --
#
#	Create a list repeating the same value over again.
#	[Identical to the Tcl 8.5 lrepeat command]
#
# Parameters:
#	n	Number of replications.
#	args	values to use in the created list.
#
# Results:
#	A list
#
# Side effects:
#       None

# Do a compatibility version of [repeat] for pre-8.5 versions of Tcl.

if { [package vcompare [package provide Tcl] 8.5] < 0 } {

    proc ::struct::list::Lrepeat {positiveCount value args} {
	if {![string is integer -strict $positiveCount]} {
	    return -code error "expected integer but got \"$positiveCount\""
	} elseif {$positiveCount < 1} {
	    return -code error {must have a count of at least 1}
	}

	set args   [linsert $args 0 $value]

	if {$positiveCount == 1} {
	    # Tcl itself has already listified the incoming parameters
	    # via 'args'.
	    return $args
	}

	set result [::list]
	while {$positiveCount > 0} {
	    if {($positiveCount % 2) == 0} {
		set args [concat $args $args]
		set positiveCount [expr {$positiveCount/2}]
	    } else {
		set result [concat $result $args]
		incr positiveCount -1
	    }
	}
	return $result
    }

} else {
    # For 8.5 simply redirect the method to the core command.

    interp alias {} ::struct::list::Lrepeat {} lrepeat
}

# ::struct::list::LdbJoin(Keyed) --
#
#	Relational table joins.
#
# Parameters:
#	args	key specs and tables to join
#
# Results:
#	A table/matrix as nested list. See
#	struct/matrix set/get rect for structure.
#
# Side effects:
#       None

proc ::struct::list::LdbJoin {args} {
    # --------------------------------
    # Process options ...

    set mode   inner
    set keyvar {}

    while {[llength $args]} {
        set err [::cmdline::getopt args {inner left right full keys.arg} opt arg]
	if {$err == 1} {
	    if {[string equal $opt keys]} {
		set keyvar $arg
	    } else {
		set mode $opt
	    }
	} elseif {$err < 0} {
	    return -code error "wrong#args: dbJoin ?-inner|-left|-right|-full? ?-keys varname? \{key table\}..."
	} else {
	    # Non-option argument found, stop processing.
	    break
	}
    }

    set inner       [string equal $mode inner]
    set innerorleft [expr {$inner || [string equal $mode left]}]

    # --------------------------------
    # Process tables ...

    if {([llength $args] % 2) != 0} {
	return -code error "wrong#args: dbJoin ?-inner|-left|-right|-full? \{key table\}..."
    }

    # One table only, join is identity
    if {[llength $args] == 2} {return [lindex $args 1]}

    # Use first table for setup.

    foreach {key table} $args break

    # Check for possible early abort
    if {$innerorleft && ([llength $table] == 0)} {return {}}

    set width 0
    array set state {}

    set keylist [InitMap state width $key $table]

    # Extend state with the remaining tables.

    foreach {key table} [lrange $args 2 end] {
	# Check for possible early abort
	if {$inner && ([llength $table] == 0)} {return {}}

	switch -exact -- $mode {
	    inner {set keylist [MapExtendInner      state       $key $table]}
	    left  {set keylist [MapExtendLeftOuter  state width $key $table]}
	    right {set keylist [MapExtendRightOuter state width $key $table]}
	    full  {set keylist [MapExtendFullOuter  state width $key $table]}
	}

	# Check for possible early abort
	if {$inner && ([llength $keylist] == 0)} {return {}}
    }

    if {[string length $keyvar]} {
	upvar 1 $keyvar keys
	set             keys $keylist
    }

    return [MapToTable state $keylist]
}

proc ::struct::list::LdbJoinKeyed {args} {
    # --------------------------------
    # Process options ...

    set mode   inner
    set keyvar {}

    while {[llength $args]} {
        set err [::cmdline::getopt args {inner left right full keys.arg} opt arg]
	if {$err == 1} {
	    if {[string equal $opt keys]} {
		set keyvar $arg
	    } else {
		set mode $opt
	    }
	} elseif {$err < 0} {
	    return -code error "wrong#args: dbJoin ?-inner|-left|-right|-full? table..."
	} else {
	    # Non-option argument found, stop processing.
	    break
	}
    }

    set inner       [string equal $mode inner]
    set innerorleft [expr {$inner || [string equal $mode left]}]

    # --------------------------------
    # Process tables ...

    # One table only, join is identity
    if {[llength $args] == 1} {
	return [Dekey [lindex $args 0]]
    }

    # Use first table for setup.

    set table [lindex $args 0]

    # Check for possible early abort
    if {$innerorleft && ([llength $table] == 0)} {return {}}

    set width 0
    array set state {}

    set keylist [InitKeyedMap state width $table]

    # Extend state with the remaining tables.

    foreach table [lrange $args 1 end] {
	# Check for possible early abort
	if {$inner && ([llength $table] == 0)} {return {}}

	switch -exact -- $mode {
	    inner {set keylist [MapKeyedExtendInner      state       $table]}
	    left  {set keylist [MapKeyedExtendLeftOuter  state width $table]}
	    right {set keylist [MapKeyedExtendRightOuter state width $table]}
	    full  {set keylist [MapKeyedExtendFullOuter  state width $table]}
	}

	# Check for possible early abort
	if {$inner && ([llength $keylist] == 0)} {return {}}
    }

    if {[string length $keyvar]} {
	upvar 1 $keyvar keys
	set             keys $keylist
    }

    return [MapToTable state $keylist]
}

## Helpers for the relational joins.
## Map is an array mapping from keys to a list
## of rows with that key

proc ::struct::list::Cartesian {leftmap rightmap key} {
    upvar $leftmap left $rightmap right
    set joined [::list]
    foreach lrow $left($key) {
	foreach row $right($key) {
	    lappend joined [concat $lrow $row]
	}
    }
    set left($key) $joined
    return
}

proc ::struct::list::SingleRightCartesian {mapvar key rightrow} {
    upvar $mapvar map
    set joined [::list]
    foreach lrow $map($key) {
	lappend joined [concat $lrow $rightrow]
    }
    set map($key) $joined
    return
}

proc ::struct::list::MapToTable {mapvar keys} {
    # Note: keys must not appear multiple times in the list.

    upvar $mapvar map
    set table [::list]
    foreach k $keys {
	foreach row $map($k) {lappend table $row}
    }
    return $table
}

## More helpers, core join operations: Init, Extend.

proc ::struct::list::InitMap {mapvar wvar key table} {
    upvar $mapvar map $wvar width
    set width [llength [lindex $table 0]]
    foreach row $table {
	set keyval [lindex $row $key]
	if {[info exists map($keyval)]} {
	    lappend map($keyval) $row
	} else {
	    set map($keyval) [::list $row]
	}
    }
    return [array names map]
}

proc ::struct::list::MapExtendInner {mapvar key table} {
    upvar $mapvar map
    array set used {}

    # Phase I - Find all keys in the second table matching keys in the
    # first. Remember all their rows.
    foreach row $table {
	set keyval [lindex $row $key]
	if {[info exists map($keyval)]} {
	    if {[info exists used($keyval)]} {
		lappend used($keyval) $row
	    } else {
		set used($keyval) [::list $row]
	    }
	} ; # else: Nothing to do for missing keys.
    }

    # Phase II - Merge the collected rows of the second (right) table
    # into the map, and eliminate all entries which have no keys in
    # the second table.
    foreach k [array names map] {
	if {[info exists  used($k)]} {
	    Cartesian map used $k
	} else {
	    unset map($k)
	}
    }
    return [array names map]
}

proc ::struct::list::MapExtendRightOuter {mapvar wvar key table} {
    upvar $mapvar map $wvar width
    array set used {}

    # Phase I - We keep all keys of the right table, even if they are
    # missing in the left one <=> Definition of right outer join.

    set w [llength [lindex $table 0]]
    foreach row $table {
	set keyval [lindex $row $key]
	if {[info exists used($keyval)]} {
	    lappend used($keyval) $row
	} else {
	    set used($keyval) [::list $row]
	}
    }

    # Phase II - Merge the collected rows of the second (right) table
    # into the map, and eliminate all entries which have no keys in
    # the second table. If there is nothing in the left table we
    # create an appropriate empty row for the cartesian => definition
    # of right outer join.

    # We go through used, because map can be empty for outer

    foreach k [array names map] {
	if {![info exists used($k)]} {
	    unset map($k)
	}
    }
    foreach k [array names used] {
	if {![info exists map($k)]} {
	    set map($k) [::list [Lrepeatn {} $width]]
	}
	Cartesian map used $k
    }

    incr width $w
    return [array names map]
}

proc ::struct::list::MapExtendLeftOuter {mapvar wvar key table} {
    upvar $mapvar map $wvar width
    array set used {}

    ## Keys: All in inner join + additional left keys 
    ##       == All left keys = array names map after
    ##          all is said and done with it.

    # Phase I - Find all keys in the second table matching keys in the
    # first. Remember all their rows.
    set w [llength [lindex $table 0]]
    foreach row $table {
	set keyval [lindex $row $key]
	if {[info exists map($keyval)]} {
	    if {[info exists used($keyval)]} {
		lappend used($keyval) $row
	    } else {
		set used($keyval) [::list $row]
	    }
	} ; # else: Nothing to do for missing keys.
    }

    # Phase II - Merge the collected rows of the second (right) table
    # into the map. We keep entries which have no keys in the second
    # table, we actually extend them <=> Left outer join.

    foreach k [array names map] {
	if {[info exists  used($k)]} {
	    Cartesian map used $k
	} else {
	    SingleRightCartesian map $k [Lrepeatn {} $w]
	}
    }
    incr width $w
    return [array names map]
}

proc ::struct::list::MapExtendFullOuter {mapvar wvar key table} {
    upvar $mapvar map $wvar width
    array set used {}

    # Phase I - We keep all keys of the right table, even if they are
    # missing in the left one <=> Definition of right outer join.

    set w [llength [lindex $table 0]]
    foreach row $table {
	set keyval [lindex $row $key]
	if {[info exists used($keyval)]} {
	    lappend used($keyval) $row
	} else {
	    lappend keylist $keyval
	    set used($keyval) [::list $row]
	}
    }

    # Phase II - Merge the collected rows of the second (right) table
    # into the map. We keep entries which have no keys in the second
    # table, we actually extend them <=> Left outer join.
    # If there is nothing in the left table we create an appropriate
    # empty row for the cartesian => definition of right outer join.

    # We go through used, because map can be empty for outer

    foreach k [array names map] {
	if {![info exists used($k)]} {
	    SingleRightCartesian map $k [Lrepeatn {} $w]
	}
    }
    foreach k [array names used] {
	if {![info exists map($k)]} {
	    set map($k) [::list [Lrepeatn {} $width]]
	}
	Cartesian map used $k
    }

    incr width $w
    return [array names map]
}

## Keyed helpers

proc ::struct::list::InitKeyedMap {mapvar wvar table} {
    upvar $mapvar map $wvar width
    set width [llength [lindex [lindex $table 0] 1]]
    foreach row $table {
	foreach {keyval rowdata} $row break
	if {[info exists map($keyval)]} {
	    lappend map($keyval) $rowdata
	} else {
	    set map($keyval) [::list $rowdata]
	}
    }
    return [array names map]
}

proc ::struct::list::MapKeyedExtendInner {mapvar table} {
    upvar $mapvar map
    array set used {}

    # Phase I - Find all keys in the second table matching keys in the
    # first. Remember all their rows.
    foreach row $table {
	foreach {keyval rowdata} $row break
	if {[info exists map($keyval)]} {
	    if {[info exists used($keyval)]} {
		lappend used($keyval) $rowdata
	    } else {
		set used($keyval) [::list $rowdata]
	    }
	} ; # else: Nothing to do for missing keys.
    }

    # Phase II - Merge the collected rows of the second (right) table
    # into the map, and eliminate all entries which have no keys in
    # the second table.
    foreach k [array names map] {
	if {[info exists  used($k)]} {
	    Cartesian map used $k
	} else {
	    unset map($k)
	}
    }

    return [array names map]
}

proc ::struct::list::MapKeyedExtendRightOuter {mapvar wvar table} {
    upvar $mapvar map $wvar width
    array set used {}

    # Phase I - We keep all keys of the right table, even if they are
    # missing in the left one <=> Definition of right outer join.

    set w [llength [lindex $table 0]]
    foreach row $table {
	foreach {keyval rowdata} $row break
	if {[info exists used($keyval)]} {
	    lappend used($keyval) $rowdata
	} else {
	    set used($keyval) [::list $rowdata]
	}
    }

    # Phase II - Merge the collected rows of the second (right) table
    # into the map, and eliminate all entries which have no keys in
    # the second table. If there is nothing in the left table we
    # create an appropriate empty row for the cartesian => definition
    # of right outer join.

    # We go through used, because map can be empty for outer

    foreach k [array names map] {
	if {![info exists used($k)]} {
	    unset map($k)
	}
    }
    foreach k [array names used] {
	if {![info exists map($k)]} {
	    set map($k) [::list [Lrepeatn {} $width]]
	}
	Cartesian map used $k
    }

    incr width $w
    return [array names map]
}

proc ::struct::list::MapKeyedExtendLeftOuter {mapvar wvar table} {
    upvar $mapvar map $wvar width
    array set used {}

    ## Keys: All in inner join + additional left keys 
    ##       == All left keys = array names map after
    ##          all is said and done with it.

    # Phase I - Find all keys in the second table matching keys in the
    # first. Remember all their rows.
    set w [llength [lindex $table 0]]
    foreach row $table {
	foreach {keyval rowdata} $row break
	if {[info exists map($keyval)]} {
	    if {[info exists used($keyval)]} {
		lappend used($keyval) $rowdata
	    } else {
		set used($keyval) [::list $rowdata]
	    }
	} ; # else: Nothing to do for missing keys.
    }

    # Phase II - Merge the collected rows of the second (right) table
    # into the map. We keep entries which have no keys in the second
    # table, we actually extend them <=> Left outer join.

    foreach k [array names map] {
	if {[info exists  used($k)]} {
	    Cartesian map used $k
	} else {
	    SingleRightCartesian map $k [Lrepeatn {} $w]
	}
    }
    incr width $w
    return [array names map]
}

proc ::struct::list::MapKeyedExtendFullOuter {mapvar wvar table} {
    upvar $mapvar map $wvar width
    array set used {}

    # Phase I - We keep all keys of the right table, even if they are
    # missing in the left one <=> Definition of right outer join.

    set w [llength [lindex $table 0]]
    foreach row $table {
	foreach {keyval rowdata} $row break
	if {[info exists used($keyval)]} {
	    lappend used($keyval) $rowdata
	} else {
	    lappend keylist $keyval
	    set used($keyval) [::list $rowdata]
	}
    }

    # Phase II - Merge the collected rows of the second (right) table
    # into the map. We keep entries which have no keys in the second
    # table, we actually extend them <=> Left outer join.
    # If there is nothing in the left table we create an appropriate
    # empty row for the cartesian => definition of right outer join.

    # We go through used, because map can be empty for outer

    foreach k [array names map] {
	if {![info exists used($k)]} {
	    SingleRightCartesian map $k [Lrepeatn {} $w]
	}
    }
    foreach k [array names used] {
	if {![info exists map($k)]} {
	    set map($k) [::list [Lrepeatn {} $width]]
	}
	Cartesian map used $k
    }

    incr width $w
    return [array names map]
}

proc ::struct::list::Dekey {keyedtable} {
    set table [::list]
    foreach row $keyedtable {lappend table [lindex $row 1]}
    return $table
}

# ::struct::list::Lswap --
#
#	Exchange two elements of a list.
#
# Parameters:
#	listvar	Name of the variable containing the list to manipulate.
#	i, j	Indices of the list elements to exchange.
#
# Results:
#	The modified list
#
# Side effects:
#       None

proc ::struct::list::Lswap {listvar i j} {
    upvar $listvar list

    if {($i < 0) || ($j < 0)} {
	return -code error {list index out of range}
    }
    set len [llength $list]
    if {($i >= $len) || ($j >= $len)} {
	return -code error {list index out of range}
    }

    if {$i != $j} {
	set tmp      [lindex $list $i]
	lset list $i [lindex $list $j]
	lset list $j $tmp
    }
    return $list
}

# ::struct::list::Lfirstperm --
#
#	Returns the lexicographically first permutation of the
#	specified list.
#
# Parameters:
#	list	The list whose first permutation is sought.
#
# Results:
#	A modified list containing the lexicographically first
#	permutation of the input.
#
# Side effects:
#       None

proc ::struct::list::Lfirstperm {list} {
    return [lsort $list]
}

# ::struct::list::Lnextperm --
#
#	Accepts a permutation of a set of elements and returns the
#	next permutatation in lexicographic sequence.
#
# Parameters:
#	list	The list containing the current permutation.
#
# Results:
#	A modified list containing the lexicographically next
#	permutation after the input permutation.
#
# Side effects:
#       None

proc ::struct::list::Lnextperm {perm} {
    # Find the smallest subscript j such that we have already visited
    # all permutations beginning with the first j elements.

    set len [expr {[llength $perm] - 1}]

    set j $len
    set ajp1 [lindex $perm $j]
    while { $j > 0 } {
	incr j -1
	set aj [lindex $perm $j]
	if { [string compare $ajp1 $aj] > 0 } {
	    set foundj {}
	    break
	}
	set ajp1 $aj
    }
    if { ![info exists foundj] } return

    # Find the smallest element greater than the j'th among the elements
    # following aj. Let its index be l, and interchange aj and al.

    set l $len
    while { [string compare $aj [set al [lindex $perm $l]]] >= 0 } {
	incr l -1
    }
    lset perm $j $al
    lset perm $l $aj

    # Reverse a_j+1 ... an

    set k [expr {$j + 1}]
    set l $len
    while { $k < $l } {
	set al [lindex $perm $l]
	lset perm $l [lindex $perm $k]
	lset perm $k $al
	incr k
	incr l -1
    }

    return $perm
}

# ::struct::list::Lpermutations --
#
#	Returns a list containing all the permutations of the
#	specified list, in lexicographic order.
#
# Parameters:
#	list	The list whose permutations are sought.
#
# Results:
#	A list of lists, containing all	permutations of the
#	input.
#
# Side effects:
#       None

proc ::struct::list::Lpermutations {list} {

    if {[llength $list] < 2} {
	return [::list $list]
    }

    set res {}
    set p [Lfirstperm $list]
    while {[llength $p]} {
	lappend res $p
	set p [Lnextperm $p]
    }
    return $res
}

# ::struct::list::Lforeachperm --
#
#	Executes a script for all the permutations of the
#	specified list, in lexicographic order.
#
# Parameters:
#	var	Name of the loop variable.
#	list	The list whose permutations are sought.
#	body	The tcl script to run per permutation of
#		the input.
#
# Results:
#	The empty string.
#
# Side effects:
#       None

proc ::struct::list::Lforeachperm {var list body} {
    upvar $var loopvar

    if {[llength $list] < 2} {
	set loopvar $list
	# TODO run body.

	# The first invocation of the body, also the last, as only one
	# permutation is possible. That makes handling of the result
	# codes easier.

	set code [catch {uplevel 1 $body} result]

	# decide what to do upon the return code:
	#
	#               0 - the body executed successfully
	#               1 - the body raised an error
	#               2 - the body invoked [return]
	#               3 - the body invoked [break]
	#               4 - the body invoked [continue]
	# everything else - return and pass on the results
	#
	switch -exact -- $code {
	    0 {}
	    1 {
		return -errorinfo [ErrorInfoAsCaller uplevel foreachperm]  \
		    -errorcode $::errorCode -code error $result
	    }
	    3 {}
	    4 {}
	    default {
		# Includes code 2
		return -code $code $result
	    }
	}
	return
    }

    set p [Lfirstperm $list]
    while {[llength $p]} {
	set loopvar $p

	set code [catch {uplevel 1 $body} result]

	# decide what to do upon the return code:
	#
	#               0 - the body executed successfully
	#               1 - the body raised an error
	#               2 - the body invoked [return]
	#               3 - the body invoked [break]
	#               4 - the body invoked [continue]
	# everything else - return and pass on the results
	#
	switch -exact -- $code {
	    0 {}
	    1 {
		return -errorinfo [ErrorInfoAsCaller uplevel foreachperm]  \
		    -errorcode $::errorCode -code error $result
	    }
	    3 {
		# FRINK: nocheck
		return
	    }
	    4 {}
	    default {
		return -code $code $result
	    }
	}
	set p [Lnextperm $p]
    }
    return
}

proc ::struct::list::Lshuffle {list} {
    for {set i [llength $list]} {$i > 1} {lset list $j $t} {
	set j [expr {int(rand() * $i)}]
	set t [lindex $list [incr i -1]]
	lset list $i [lindex $list $j]
    }
    return $list
}

# ### ### ### ######### ######### #########

proc ::struct::list::ErrorInfoAsCaller {find replace} {
    set info $::errorInfo
    set i [string last "\n    (\"$find" $info]
    if {$i == -1} {return $info}
    set result [string range $info 0 [incr i 6]]	;# keep "\n    (\""
    append result $replace			;# $find -> $replace
    incr i [string length $find]
    set j [string first ) $info [incr i]]	;# keep rest of parenthetical
    append result [string range $info $i $j]
    return $result
}

# ### ### ### ######### ######### #########
## Ready

namespace eval ::struct {
    # Get 'list::list' into the general structure namespace.
    namespace import -force list::list
    namespace export list
}
package provide struct::list 1.8.5

# THATCH /herbie/packages/struct/tree.tcl 
# tree.tcl --
#
#	Implementation of a tree data structure for Tcl.
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: tree.tcl,v 1.45 2009/06/22 18:21:59 andreas_kupries Exp $

# @mdgen EXCLUDE: tree_c.tcl

package require Tcl 8.2
package require struct::list

namespace eval ::struct::tree {}

# ### ### ### ######### ######### #########
## Management of tree implementations.

# ::struct::tree::LoadAccelerator --
#
#	Loads a named implementation, if possible.
#
# Arguments:
#	key	Name of the implementation to load.
#
# Results:
#	A boolean flag. True if the implementation
#	was successfully loaded; and False otherwise.

proc ::struct::tree::LoadAccelerator {key} {
    variable accel
    set r 0
    switch -exact -- $key {
	critcl {
	    # Critcl implementation of tree requires Tcl 8.4.
	    if {![package vsatisfies [package provide Tcl] 8.4]} {return 0}
	    if {[catch {package require tcllibc}]} {return 0}
	    set r [llength [info commands ::struct::tree_critcl]]
	}
	tcl {
	    variable selfdir
	    source [file join $selfdir tree_tcl.tcl]
	    set r 1
	}
        default {
            return -code error "invalid accelerator/impl. package $key:\
                must be one of [join [KnownImplementations] {, }]"
        }
    }
    set accel($key) $r
    return $r
}

# ::struct::tree::SwitchTo --
#
#	Activates a loaded named implementation.
#
# Arguments:
#	key	Name of the implementation to activate.
#
# Results:
#	None.

proc ::struct::tree::SwitchTo {key} {
    variable accel
    variable loaded

    if {[string equal $key $loaded]} {
	# No change, nothing to do.
	return
    } elseif {![string equal $key ""]} {
	# Validate the target implementation of the switch.

	if {![info exists accel($key)]} {
	    return -code error "Unable to activate unknown implementation \"$key\""
	} elseif {![info exists accel($key)] || !$accel($key)} {
	    return -code error "Unable to activate missing implementation \"$key\""
	}
    }

    # Deactivate the previous implementation, if there was any.

    if {![string equal $loaded ""]} {
	rename ::struct::tree        ::struct::tree_$loaded
	rename ::struct::tree::prune ::struct::tree::prune_$loaded
    }

    # Activate the new implementation, if there is any.

    if {![string equal $key ""]} {
	rename ::struct::tree_$key        ::struct::tree
	rename ::struct::tree::prune_$key ::struct::tree::prune
    }

    # Remember the active implementation, for deactivation by future
    # switches.

    set loaded $key
    return
}

# ::struct::tree::Implementations --
#
#	Determines which implementations are
#	present, i.e. loaded.
#
# Arguments:
#	None.
#
# Results:
#	A list of implementation keys.

proc ::struct::tree::Implementations {} {
    variable accel
    set res {}
    foreach n [array names accel] {
	if {!$accel($n)} continue
	lappend res $n
    }
    return $res
}

# ::struct::tree::KnownImplementations --
#
#	Determines which implementations are known
#	as possible implementations.
#
# Arguments:
#	None.
#
# Results:
#	A list of implementation keys. In the order
#	of preference, most prefered first.

proc ::struct::tree::KnownImplementations {} {
    return {critcl tcl}
}

proc ::struct::tree::Names {} {
    return {
	critcl {tcllibc based}
	tcl    {pure Tcl}
    }
}

# ### ### ### ######### ######### #########
## Initialization: Data structures.

namespace eval ::struct::tree {
    variable  selfdir [file dirname [info script]]
    variable  accel
    array set accel   {tcl 0 critcl 0}
    variable  loaded  {}
}

# ### ### ### ######### ######### #########
## Initialization: Choose an implementation,
## most prefered first. Loads only one of the
## possible implementations. And activates it.

namespace eval ::struct::tree {
    variable e
    foreach e [KnownImplementations] {
	if {[LoadAccelerator $e]} {
	    SwitchTo $e
	    break
	}
    }
    unset e
}

# ### ### ### ######### ######### #########
## Ready

namespace eval ::struct {
    # Export the constructor command.
    namespace export tree
}

package provide struct::tree 2.1.2

# THATCH /herbie/packages/struct/tree_tcl.tcl 
# tree.tcl --
#
#	Implementation of a tree data structure for Tcl.
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: tree_tcl.tcl,v 1.5 2009/06/22 18:21:59 andreas_kupries Exp $

package require Tcl 8.2
package require struct::list

namespace eval ::struct::tree {
    # Data storage in the tree module
    # -------------------------------
    #
    # There's a lot of bits to keep track of for each tree:
    #	nodes
    #	node values
    #	node relationships
    #
    # It would quickly become unwieldy to try to keep these in arrays or lists
    # within the tree namespace itself.  Instead, each tree structure will get
    # its own namespace.  Each namespace contains:
    #	children	array mapping nodes to their children list
    #	parent		array mapping nodes to their parent node
    #	node:$node	array mapping keys to values for the node $node

    # counter is used to give a unique name for unnamed trees
    variable counter 0

    # Only export one command, the one used to instantiate a new tree
    namespace export tree_tcl
}

# ::struct::tree::tree_tcl --
#
#	Create a new tree with a given name; if no name is given, use
#	treeX, where X is a number.
#
# Arguments:
#	name	Optional name of the tree; if null or not given, generate one.
#
# Results:
#	name	Name of the tree created

proc ::struct::tree::tree_tcl {args} {
    variable counter

    set src     {}
    set srctype {}

    switch -exact -- [llength [info level 0]] {
	1 {
	    # Missing name, generate one.
	    incr counter
	    set name "tree${counter}"
	}
	2 {
	    # Standard call. New empty tree.
	    set name [lindex $args 0]
	}
	4 {
	    # Copy construction.
	    foreach {name as src} $args break
	    switch -exact -- $as {
		= - := - as {
		    set srctype tree
		}
		deserialize {
		    set srctype serial
		}
		default {
		    return -code error \
			    "wrong # args: should be \"tree ?name ?=|:=|as|deserialize source??\""
		}
	    }
	}
	default {
	    # Error.
	    return -code error \
		    "wrong # args: should be \"tree ?name ?=|:=|as|deserialize source??\""
	}
    }

    # FIRST, qualify the name.
    if {![string match "::*" $name]} {
        # Get caller's namespace; append :: if not global namespace.
        set ns [uplevel 1 [list namespace current]]
        if {"::" != $ns} {
            append ns "::"
        }

        set name "$ns$name"
    }
    if {[llength [info commands $name]]} {
	return -code error \
		"command \"$name\" already exists, unable to create tree"
    }

    # Set up the namespace for the object,
    # identical to the object command.
    namespace eval $name {
	variable rootname
	set      rootname root

	# Set up root node's child list
	variable children
	set      children(root) [list]

	# Set root node's parent
	variable parent
	set      parent(root) [list]

	# Set up the node attribute mapping
	variable  attribute
	array set attribute {}

	# Set up a counter for use in creating unique node names
	variable nextUnusedNode
	set      nextUnusedNode 1

	# Set up a counter for use in creating node attribute arrays.
	variable nextAttr
	set      nextAttr 0
    }

    # Create the command to manipulate the tree
    interp alias {} $name {} ::struct::tree::TreeProc $name

    # Automatic execution of assignment if a source
    # is present.
    if {$src != {}} {
	switch -exact -- $srctype {
	    tree   {
		set code [catch {_= $name $src} msg]
		if {$code} {
		    namespace delete $name
		    interp alias {} $name {}
		    return -code $code -errorinfo $::errorInfo -errorcode $::errorCode $msg
		}
	    }
	    serial {
		set code [catch {_deserialize $name $src} msg]
		if {$code} {
		    namespace delete $name
		    interp alias {} $name {}
		    return -code $code -errorinfo $::errorInfo -errorcode $::errorCode $msg
		}
	    }
	    default {
		return -code error \
			"Internal error, illegal srctype \"$srctype\""
	    }
	}
    }

    # Give object to caller for use.
    return $name
}

# ::struct::tree::prune_tcl --
#
#	Abort the walk script, and ignore any children of the
#	node we are currently at.
#
# Arguments:
#	None.
#
# Results:
#	None.
#
# Sideeffects:
#
#	Stops the execution of the script and throws a signal to the
#	surrounding walker to go to the next node, and ignore the
#	children of the current node.

proc ::struct::tree::prune_tcl {} {
    return -code 5
}

##########################
# Private functions follow

# ::struct::tree::TreeProc --
#
#	Command that processes all tree object commands.
#
# Arguments:
#	name	Name of the tree object to manipulate.
#	cmd	Subcommand to invoke.
#	args	Arguments for subcommand.
#
# Results:
#	Varies based on command to perform

proc ::struct::tree::TreeProc {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name option ?arg arg ...?\""
    }

    # Split the args into command and args components
    set sub _$cmd
    if { [llength [info commands ::struct::tree::$sub]] == 0 } {
	set optlist [lsort [info commands ::struct::tree::_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    lappend xlist [string range $p 1 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }

    set code [catch {uplevel 1 [linsert $args 0 ::struct::tree::$sub $name]} result]

    if {$code == 1} {
	return -errorinfo [ErrorInfoAsCaller uplevel $sub]  \
		-errorcode $::errorCode -code error $result
    } elseif {$code == 2} {
	return -code $code $result
    }
    return $result
}

# ::struct::tree::_:= --
#
#	Assignment operator. Copies the source tree into the
#       destination, destroying the original information.
#
# Arguments:
#	name	Name of the tree object we are copying into.
#	source	Name of the tree object providing us with the
#		data to copy.
#
# Results:
#	Nothing.

proc ::struct::tree::_= {name source} {
    _deserialize $name [$source serialize]
    return
}

# ::struct::tree::_--> --
#
#	Reverse assignment operator. Copies this tree into the
#       destination, destroying the original information.
#
# Arguments:
#	name	Name of the tree object to copy
#	dest	Name of the tree object we are copying to.
#
# Results:
#	Nothing.

proc ::struct::tree::_--> {name dest} {
    $dest deserialize [_serialize $name]
    return
}

# ::struct::tree::_ancestors --
#
#	Return the list of all parent nodes of a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to look up.
#
# Results:
#	parents	List of parents of node $node.
#		Immediate ancestor (parent) first,
#		Root of tree (ancestor of all) last.

proc ::struct::tree::_ancestors {name node} {
    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::parent
    set a {}
    while {[info exists parent($node)]} {
	set node $parent($node)
	if {$node == {}} break
	lappend a $node
    }
    return $a
}

# ::struct::tree::_attr --
#
#	Return attribute data for one key and multiple nodes, possibly all.
#
# Arguments:
#	name	Name of the tree object.
#	key	Name of the attribute to retrieve.
#
# Results:
#	children	Dictionary mapping nodes to attribute data.

proc ::struct::tree::_attr {name key args} {
    # Syntax:
    #
    # t attr key
    # t attr key -nodes {nodelist}
    # t attr key -glob nodepattern
    # t attr key -regexp nodepattern

    variable ${name}::attribute

    set usage "wrong # args: should be \"[list $name] attr key ?-nodes list|-glob pattern|-regexp pattern?\""
    if {([llength $args] != 0) && ([llength $args] != 2)} {
	return -code error $usage
    } elseif {[llength $args] == 0} {
	# This automatically restricts the list
	# to nodes which can have the attribute
	# in question.

	set nodes [array names attribute]
    } else {
	# Determine a list of nodes to look at
	# based on the chosen restriction.

	foreach {mode value} $args break
	switch -exact -- $mode {
	    -nodes {
		# This is the only branch where we have to
		# perform an explicit restriction to the
		# nodes which have attributes.
		set nodes {}
		foreach n $value {
		    if {![info exists attribute($n)]} continue
		    lappend nodes $n
		}
	    }
	    -glob {
		set nodes [array names attribute $value]
	    }
	    -regexp {
		set nodes {}
		foreach n [array names attribute] {
		    if {![regexp -- $value $n]} continue
		    lappend nodes $n
		}
	    }
	    default {
		return -code error $usage
	    }
	}
    }

    # Without possibly matching nodes
    # the result has to be empty.

    if {![llength $nodes]} {
	return {}
    }

    # Now locate matching keys and their values.

    set result {}
    foreach n $nodes {
	upvar ${name}::$attribute($n) data
	if {[info exists data($key)]} {
	    lappend result $n $data($key)
	}
    }

    return $result
}

# ::struct::tree::_deserialize --
#
#	Assignment operator. Copies a serialization into the
#       destination, destroying the original information.
#
# Arguments:
#	name	Name of the tree object we are copying into.
#	serial	Serialized tree to copy from.
#
# Results:
#	Nothing.

proc ::struct::tree::_deserialize {name serial} {
    # As we destroy the original tree as part of
    # the copying process we don't have to deal
    # with issues like node names from the new tree
    # interfering with the old ...

    # I. Get the serialization of the source tree
    #    and check it for validity.

    CheckSerialization $serial attr p c rn

    # Get all the relevant data into the scope

    variable ${name}::rootname
    variable ${name}::children
    variable ${name}::parent
    variable ${name}::attribute
    variable ${name}::nextAttr

    # Kill the existing parent/children information and insert the new
    # data in their place.

    foreach n [array names parent] {
	unset parent($n) children($n)
    }
    array set parent   [array get p]
    array set children [array get c]
    unset p c

    set nextAttr 0
    foreach a [array names attribute] {
	unset ${name}::$attribute($a)
    }
    foreach n [array names attr] {
	GenAttributeStorage $name $n
	array set ${name}::$attribute($n) $attr($n)
    }

    set rootname $rn

    ## Debug ## Dump internals ...
    if {0} {
	puts "___________________________________ $name"
	puts $rootname
	parray children
	parray parent
	parray attribute
	puts ___________________________________
    }
    return
}

# ::struct::tree::_children --
#
#	Return the list of children for a given node of a tree.
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up.
#
# Results:
#	children	List of children for the node.

proc ::struct::tree::_children {name args} {
    # args := ?-all? node ?filter cmdprefix?

    # '-all' implies that not only the direct children of the
    # node, but all their children, and so on, are returned.
    #
    # 'filter cmd' implies that only those nodes in the result list
    # which pass the test 'cmd' are placed into the final result. 

    set usage "wrong # args: should be \"[list $name] children ?-all? node ?filter cmd?\""

    if {([llength $args] < 1) || ([llength $args] > 4)} {
	return -code error $usage
    }
    if {[string equal [lindex $args 0] -all]} {
	set all 1
	set args [lrange $args 1 end]
    } else {
	set all 0
    }

    # args := node ?filter cmdprefix?

    if {([llength $args] != 1) && ([llength $args] != 3)} {
	return -code error $usage
    }
    if {[llength $args] == 3} {
	foreach {node _const_ cmd} $args break
	if {![string equal $_const_ filter] || ![llength $cmd]} {
	    return -code error $usage
	}
    } else {
	set node [lindex $args 0]
	set cmd {}
    }

    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    if {$all} {
	set result [DescendantsCore $name $node]
    } else {
	variable ${name}::children
	set result $children($node)
    }

    if {[llength $cmd]} {
	lappend cmd $name
	set result [uplevel 1 [list ::struct::list filter $result $cmd]]
    }

    return $result
}

# ::struct::tree::_cut --
#
#	Destroys the specified node of a tree, but not its children.
#	These children are made into children of the parent of the
#	destroyed node at the index of the destroyed node.
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up and cut.
#
# Results:
#	None.

proc ::struct::tree::_cut {name node} {
    variable ${name}::rootname

    if { [string equal $node $rootname] } {
	# Can't delete the special root node
	return -code error "cannot cut root node"
    }

    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::parent
    variable ${name}::children

    # Locate our parent, children and our location in the parent
    set parentNode $parent($node)
    set childNodes $children($node)

    set index [lsearch -exact $children($parentNode) $node]

    # Excise this node from the parent list,
    set newChildren [lreplace $children($parentNode) $index $index]

    # Put each of the children of $node into the parent's children list,
    # in the place of $node, and update the parent pointer of those nodes.
    foreach child $childNodes {
	set newChildren [linsert $newChildren $index $child]
	set parent($child) $parentNode
	incr index
    }
    set children($parentNode) $newChildren

    KillNode $name $node
    return
}

# ::struct::tree::_delete --
#
#	Remove a node from a tree, including all of its values.  Recursively
#	removes the node's children.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to delete.
#
# Results:
#	None.

proc ::struct::tree::_delete {name node} {
    variable ${name}::rootname
    if { [string equal $node $rootname] } {
	# Can't delete the special root node
	return -code error "cannot delete root node"
    }
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::children
    variable ${name}::parent

    # Remove this node from its parent's children list
    set parentNode $parent($node)
    set index [lsearch -exact $children($parentNode) $node]
    ldelete children($parentNode) $index

    # Yes, we could use the stack structure implemented in ::struct::stack,
    # but it's slower than inlining it.  Since we don't need a sophisticated
    # stack, don't bother.
    set st [list]
    foreach child $children($node) {
	lappend st $child
    }

    KillNode $name $node

    while {[llength $st] > 0} {
	set node [lindex $st end]
	ldelete           st end
	foreach child $children($node) {
	    lappend st $child
	}

	KillNode $name $node
    }
    return
}

# ::struct::tree::_depth --
#
#	Return the depth (distance from the root node) of a given node.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to find.
#
# Results:
#	depth	Number of steps from node to the root node.

proc ::struct::tree::_depth {name node} {
    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }
    variable ${name}::parent
    variable ${name}::rootname
    set depth 0
    while { ![string equal $node $rootname] } {
	incr depth
	set node $parent($node)
    }
    return $depth
}

# ::struct::tree::_descendants --
#
#	Return the list containing all descendants of a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to look at.
#
# Results:
#	desc	(filtered) List of nodes descending from 'node'.

proc ::struct::tree::_descendants {name node args} {
    # children -all sucessor, allows filtering.

    set usage "wrong # args: should be \"[list $name] descendants node ?filter cmd?\""

    if {[llength $args] > 2} {
	return -code error $usage
    } elseif {[llength $args] == 2} {
	foreach {_const_ cmd} $args break
	if {![string equal $_const_ filter] || ![llength $cmd]} {
	    return -code error $usage
	}
    } else {
	set cmd {}
    }

    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    set result [DescendantsCore $name $node]

    if {[llength $cmd]} {
	lappend cmd $name
	set result [uplevel 1 [list ::struct::list filter $result $cmd]]
    }

    return $result
}

proc ::struct::tree::DescendantsCore {name node} {
    # CORE for listing of node descendants.
    # No checks ...
    # No filtering ...

    variable ${name}::children

    # New implementation. Instead of keeping a second, and explicit,
    # list of pending nodes to shift through (= copying of array data
    # around), we reuse the result list for that, using a counter and
    # direct access to list elements to keep track of what nodes have
    # not been handled yet. This eliminates a whole lot of array
    # copying within the list implementation in the Tcl core. The
    # result is unchanged, i.e. the nodes are in the same order as
    # before.

    set result  $children($node)
    set at      0

    while {$at < [llength $result]} {
	set n [lindex $result $at]
	incr at
	foreach c $children($n) {
	    lappend result $c
	}
    }

    return $result
}

# ::struct::tree::_destroy --
#
#	Destroy a tree, including its associated command and data storage.
#
# Arguments:
#	name	Name of the tree to destroy.
#
# Results:
#	None.

proc ::struct::tree::_destroy {name} {
    namespace delete $name
    interp alias {} $name {}
}

# ::struct::tree::_exists --
#
#	Test for existence of a given node in a tree.
#
# Arguments:
#	name	Name of the tree to query.
#	node	Node to look for.
#
# Results:
#	1 if the node exists, 0 else.

proc ::struct::tree::_exists {name node} {
    return [info exists ${name}::parent($node)]
}

# ::struct::tree::_get --
#
#	Get a keyed value from a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to query.
#	key	Key to lookup.
#
# Results:
#	value	Value associated with the key given.

proc ::struct::tree::_get {name node key} {
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::attribute
    if {![info exists attribute($node)]} {
	# No attribute data for this node, key has to be invalid.
	return -code error "invalid key \"$key\" for node \"$node\""
    }

    upvar ${name}::$attribute($node) data
    if {![info exists data($key)]} {
	return -code error "invalid key \"$key\" for node \"$node\""
    }
    return $data($key)
}

# ::struct::tree::_getall --
#
#	Get a serialized list of key/value pairs from a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to query.
#
# Results:
#	value	A serialized list of key/value pairs.

proc ::struct::tree::_getall {name node {pattern *}} {
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::attribute
    if {![info exists attribute($node)]} {
	# No attributes ...
	return {}
    }

    upvar ${name}::$attribute($node) data
    return [array get data $pattern]
}

# ::struct::tree::_height --
#
#	Return the height (distance from the given node to its deepest child)
#
# Arguments:
#	name	Name of the tree.
#	node	Node we wish to know the height for..
#
# Results:
#	height	Distance to deepest child of the node.

proc ::struct::tree::_height {name node} {
    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::children
    variable ${name}::parent

    if {[llength $children($node)] == 0} {
	# No children, is a leaf, height is 0.
	return 0
    }

    # New implementation. We iteratively compute the height for each
    # node under the specified one, from the bottom up. The previous
    # implementation, using recursion will fail if the encountered
    # subtree has a height greater than the currently set recursion
    # limit.

    array set h {}

    # NOTE: Check out if a for loop doing direct access, i.e. without
    #       list reversal, is faster.

    foreach n [struct::list reverse [DescendantsCore $name $node]] {
	# Height of leafs
	if {![llength $children($n)]} {set h($n) 0}

	# Height of our parent is max of our and previous height.
	set p $parent($n)
	if {![info exists h($p)] || ($h($n) >= $h($p))} {
	    set h($p) [expr {$h($n) + 1}]
	}
    }

    # NOTE: Check out how much we gain by caching the result.
    #       For all nodes we have this computed. Use cache here
    #       as well to cut the inspection of descendants down.
    #       This may degenerate into a recursive solution again
    #       however.

    return $h($node)
}

# ::struct::tree::_keys --
#
#	Get a list of keys from a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to query.
#
# Results:
#	value	A serialized list of key/value pairs.

proc ::struct::tree::_keys {name node {pattern *}} {
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::attribute
    if {![info exists attribute($node)]} {
	# No attribute data for this node.
	return {}
    }

    upvar ${name}::$attribute($node) data
    return [array names data $pattern]
}

# ::struct::tree::_keyexists --
#
#	Test for existence of a given key for a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to query.
#	key	Key to lookup.
#
# Results:
#	1 if the key exists, 0 else.

proc ::struct::tree::_keyexists {name node key} {
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::attribute
    if {![info exists attribute($node)]} {
	# No attribute data for this node, key cannot exist
	return 0
    }

    upvar ${name}::$attribute($node) data
    return [info exists data($key)]
}

# ::struct::tree::_index --
#
#	Determine the index of node with in its parent's list of children.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to look up.
#
# Results:
#	index	The index of the node in its parent

proc ::struct::tree::_index {name node} {
    variable ${name}::rootname
    if { [string equal $node $rootname] } {
	# The special root node has no parent, thus no index in it either.
	return -code error "cannot determine index of root node"
    }

    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::children
    variable ${name}::parent

    # Locate the parent and ourself in its list of children
    set parentNode $parent($node)

    return [lsearch -exact $children($parentNode) $node]
}

# ::struct::tree::_insert --
#
#	Add a node to a tree; if the node(s) specified already exist, they
#	will be moved to the given location.
#
# Arguments:
#	name		Name of the tree.
#	parentNode	Parent to add the node to.
#	index		Index at which to insert.
#	args		Node(s) to insert.  If none is given, the routine
#			will insert a single node with a unique name.
#
# Results:
#	nodes		List of nodes inserted.

proc ::struct::tree::_insert {name parentNode index args} {
    if { [llength $args] == 0 } {
	# No node name was given; generate a unique one
	set args [list [GenerateUniqueNodeName $name]]
    }
    if { ![_exists $name $parentNode] } {
	return -code error "parent node \"$parentNode\" does not exist in tree \"$name\""
    }

    variable ${name}::parent
    variable ${name}::children
    variable ${name}::rootname

    # Make sure the index is numeric

    if {[string equal $index "end"]} {
	set index [llength $children($parentNode)]
    } elseif {[regexp {^end-([0-9]+)$} $index -> n]} {
	set index [expr {[llength $children($parentNode)] - $n}]
    }

    foreach node $args {
	if {[_exists $name $node] } {
	    # Move the node to its new home
	    if { [string equal $node $rootname] } {
		return -code error "cannot move root node"
	    }
	
	    # Cannot make a node its own descendant (I'm my own grandpa...)
	    set ancestor $parentNode
	    while { ![string equal $ancestor $rootname] } {
		if { [string equal $ancestor $node] } {
		    return -code error "node \"$node\" cannot be its own descendant"
		}
		set ancestor $parent($ancestor)
	    }
	    # Remove this node from its parent's children list
	    set oldParent $parent($node)
	    set ind [lsearch -exact $children($oldParent) $node]
	    ldelete children($oldParent) $ind
	
	    # If the node is moving within its parent, and its old location
	    # was before the new location, decrement the new location, so that
	    # it gets put in the right spot
	    if { [string equal $oldParent $parentNode] && $ind < $index } {
		incr index -1
	    }
	} else {
	    # Set up the new node
	    set children($node) [list]
	}

	# Add this node to its parent's children list
	set children($parentNode) [linsert $children($parentNode) $index $node]

	# Update the parent pointer for this node
	set parent($node) $parentNode
	incr index
    }

    return $args
}

# ::struct::tree::_isleaf --
#
#	Return whether the given node of a tree is a leaf or not.
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up.
#
# Results:
#	isleaf	True if the node is a leaf; false otherwise.

proc ::struct::tree::_isleaf {name node} {
    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::children
    return [expr {[llength $children($node)] == 0}]
}

# ::struct::tree::_move --
#
#	Move a node (and all its subnodes) from where ever it is to a new
#	location in the tree.
#
# Arguments:
#	name		Name of the tree
#	parentNode	Parent to add the node to.
#	index		Index at which to insert.
#	node		Node to move; the node must exist in the tree.
#	args		Additional nodes to move; these nodes must exist
#			in the tree.
#
# Results:
#	None.

proc ::struct::tree::_move {name parentNode index node args} {
    set args [linsert $args 0 $node]

    # Can only move a node to a real location in the tree
    if { ![_exists $name $parentNode] } {
	return -code error "parent node \"$parentNode\" does not exist in tree \"$name\""
    }

    variable ${name}::parent
    variable ${name}::children
    variable ${name}::rootname

    # Make sure the index is numeric

    if {[string equal $index "end"]} {
	set index [llength $children($parentNode)]
    } elseif {[regexp {^end-([0-9]+)$} $index -> n]} {
	set index [expr {[llength $children($parentNode)] - $n}]
    }

    # Validate all nodes to move before trying to move any.
    foreach node $args {
	if { [string equal $node $rootname] } {
	    return -code error "cannot move root node"
	}

	# Can only move real nodes
	if { ![_exists $name $node] } {
	    return -code error "node \"$node\" does not exist in tree \"$name\""
	}

	# Cannot move a node to be a descendant of itself
	set ancestor $parentNode
	while { ![string equal $ancestor $rootname] } {
	    if { [string equal $ancestor $node] } {
		return -code error "node \"$node\" cannot be its own descendant"
	    }
	    set ancestor $parent($ancestor)
	}
    }

    # Remove all nodes from their current parent's children list
    foreach node $args {
	set oldParent $parent($node)
	set ind [lsearch -exact $children($oldParent) $node]

	ldelete children($oldParent) $ind

	# Update the nodes parent value
	set parent($node) $parentNode
    }

    # Add all nodes to their new parent's children list
    set children($parentNode) \
	[eval [list linsert $children($parentNode) $index] $args]

    return
}

# ::struct::tree::_next --
#
#	Return the right sibling for a given node of a tree.
#
# Arguments:
#	name		Name of the tree object.
#	node		Node to retrieve right sibling for.
#
# Results:
#	sibling		The right sibling for the node, or null if node was
#			the rightmost child of its parent.

proc ::struct::tree::_next {name node} {
    # The 'root' has no siblings.
    variable ${name}::rootname
    if { [string equal $node $rootname] } {
	return {}
    }

    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    # Locate the parent and our place in its list of children.
    variable ${name}::parent
    variable ${name}::children

    set parentNode $parent($node)
    set  index [lsearch -exact $children($parentNode) $node]

    # Go to the node to the right and return its name.
    return [lindex $children($parentNode) [incr index]]
}

# ::struct::tree::_numchildren --
#
#	Return the number of immediate children for a given node of a tree.
#
# Arguments:
#	name		Name of the tree object.
#	node		Node to look up.
#
# Results:
#	numchildren	Number of immediate children for the node.

proc ::struct::tree::_numchildren {name node} {
    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::children
    return [llength $children($node)]
}

# ::struct::tree::_nodes --
#
#	Return a list containing all nodes known to the tree.
#
# Arguments:
#	name		Name of the tree object.
#
# Results:
#	nodes	List of nodes in the tree.

proc ::struct::tree::_nodes {name} {
    variable ${name}::children
    return [array names children]
}

# ::struct::tree::_parent --
#
#	Return the name of the parent node of a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to look up.
#
# Results:
#	parent	Parent of node $node

proc ::struct::tree::_parent {name node} {
    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }
    # FRINK: nocheck
    return [set ${name}::parent($node)]
}

# ::struct::tree::_previous --
#
#	Return the left sibling for a given node of a tree.
#
# Arguments:
#	name		Name of the tree object.
#	node		Node to look up.
#
# Results:
#	sibling		The left sibling for the node, or null if node was
#			the leftmost child of its parent.

proc ::struct::tree::_previous {name node} {
    # The 'root' has no siblings.
    variable ${name}::rootname
    if { [string equal $node $rootname] } {
	return {}
    }

    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    # Locate the parent and our place in its list of children.
    variable ${name}::parent
    variable ${name}::children

    set parentNode $parent($node)
    set  index [lsearch -exact $children($parentNode) $node]

    # Go to the node to the right and return its name.
    return [lindex $children($parentNode) [incr index -1]]
}

# ::struct::tree::_rootname --
#
#	Query or change the name of the root node.
#
# Arguments:
#	name	Name of the tree.
#
# Results:
#	The name of the root node

proc ::struct::tree::_rootname {name} {
    variable ${name}::rootname
    return $rootname
}

# ::struct::tree::_rename --
#
#	Change the name of any node.
#
# Arguments:
#	name	Name of the tree.
#	node	Name of node to be renamed
#	newname	New name for the node.
#
# Results:
#	The new name of the node.

proc ::struct::tree::_rename {name node newname} {
    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }
    if {[_exists $name $newname]} {
	return -code error "unable to rename node to \"$newname\",\
		node of that name already present in the tree \"$name\""
    }

    set oldname  $node

    # Perform the rename in the internal
    # data structures.

    variable ${name}::rootname
    variable ${name}::children
    variable ${name}::parent
    variable ${name}::attribute

    set children($newname) $children($oldname)
    unset                   children($oldname)
    set parent($newname)     $parent($oldname)
    unset                     parent($oldname)

    foreach c $children($newname) {
	set parent($c) $newname
    }

    if {[string equal $oldname $rootname]} {
	set rootname $newname
    } else {
	set p $parent($newname)
	set pos  [lsearch -exact $children($p) $oldname]
	lset children($p) $pos $newname
    }

    if {[info exists attribute($oldname)]} {
	set attribute($newname) $attribute($oldname)
	unset                    attribute($oldname)
    }

    return $newname
}

# ::struct::tree::_serialize --
#
#	Serialize a tree object (partially) into a transportable value.
#
# Arguments:
#	name	Name of the tree.
#	node	Root node of the serialized tree.
#
# Results:
#	A list structure describing the part of the tree which was serialized.

proc ::struct::tree::_serialize {name args} {
    if {[llength $args] > 1} {
	return -code error \
		"wrong # args: should be \"[list $name] serialize ?node?\""
    } elseif {[llength $args] == 1} {
	set node [lindex $args 0]

	if {![_exists $name $node]} {
	    return -code error "node \"$node\" does not exist in tree \"$name\""
	}
    } else {
	variable ${name}::rootname
	set node $rootname
    }

    set                   tree [list]
    Serialize $name $node tree
    return               $tree
}

# ::struct::tree::_set --
#
#	Set or get a value for a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to modify or query.
#	args	Optional argument specifying a value.
#
# Results:
#	val	Value associated with the given key of the given node

proc ::struct::tree::_set {name node key args} {
    if {[llength $args] > 1} {
	return -code error "wrong # args: should be \"$name set node key\
		?value?\""
    }
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    # Process the arguments ...

    if {[llength $args] > 0} {
	# Setting the value. This may have to create
	# the attribute array for this particular
	# node

	variable ${name}::attribute
	if {![info exists attribute($node)]} {
	    # No attribute data for this node,
	    # so create it as we need it now.
	    GenAttributeStorage $name $node
	}
	upvar ${name}::$attribute($node) data

	return [set data($key) [lindex $args end]]
    } else {
	# Getting the value

	return [_get $name $node $key]
    }
}

# ::struct::tree::_append --
#
#	Append a value for a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to modify.
#	key	Name of attribute to modify.
#	value	Value to append
#
# Results:
#	val	Value associated with the given key of the given node

proc ::struct::tree::_append {name node key value} {
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::attribute
    if {![info exists attribute($node)]} {
	# No attribute data for this node,
	# so create it as we need it.
	GenAttributeStorage $name $node
    }

    upvar ${name}::$attribute($node) data
    return [append data($key) $value]
}

# ::struct::tree::_lappend --
#
#	lappend a value for a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to modify or query.
#	key	Name of attribute to modify.
#	value	Value to append
#
# Results:
#	val	Value associated with the given key of the given node

proc ::struct::tree::_lappend {name node key value} {
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::attribute
    if {![info exists attribute($node)]} {
	# No attribute data for this node,
	# so create it as we need it.
	GenAttributeStorage $name $node
    }

    upvar ${name}::$attribute($node) data
    return [lappend data($key) $value]
}

# ::struct::tree::_leaves --
#
#	Return a list containing all leaf nodes known to the tree.
#
# Arguments:
#	name		Name of the tree object.
#
# Results:
#	nodes	List of leaf nodes in the tree.

proc ::struct::tree::_leaves {name} {
    variable ${name}::children

    set res {}
    foreach n [array names children] {
	if {[llength $children($n)]} continue
	lappend res $n
    }
    return $res
}

# ::struct::tree::_size --
#
#	Return the number of descendants of a given node.  The default node
#	is the special root node.
#
# Arguments:
#	name	Name of the tree.
#	node	Optional node to start counting from (default is root).
#
# Results:
#	size	Number of descendants of the node.

proc ::struct::tree::_size {name args} {
    variable ${name}::rootname
    if {[llength $args] > 1} {
	return -code error \
		"wrong # args: should be \"[list $name] size ?node?\""
    } elseif {[llength $args] == 1} {
	set node [lindex $args 0]

	if { ![_exists $name $node] } {
	    return -code error "node \"$node\" does not exist in tree \"$name\""
	}
    } else {
	# If the node is the root, we can do the cheap thing and just count the
	# number of nodes (excluding the root node) that we have in the tree with
	# array size.

	return [expr {[array size ${name}::parent] - 1}]
    }

    # If the node is the root, we can do the cheap thing and just count the
    # number of nodes (excluding the root node) that we have in the tree with
    # array size.

    if { [string equal $node $rootname] } {
	return [expr {[array size ${name}::parent] - 1}]
    }

    # Otherwise we have to do it the hard way and do a full tree search
    variable ${name}::children
    set size 0
    set st [list ]
    foreach child $children($node) {
	lappend st $child
    }
    while { [llength $st] > 0 } {
	set node [lindex $st end]
	ldelete st end
	incr size
	foreach child $children($node) {
	    lappend st $child
	}
    }
    return $size
}

# ::struct::tree::_splice --
#
#	Add a node to a tree, making a range of children from the given
#	parent children of the new node.
#
# Arguments:
#	name		Name of the tree.
#	parentNode	Parent to add the node to.
#	from		Index at which to insert.
#	to		Optional end of the range of children to replace.
#			Defaults to 'end'.
#	args		Optional node name; if given, must be unique.  If not
#			given, a unique name will be generated.
#
# Results:
#	node		Name of the node added to the tree.

proc ::struct::tree::_splice {name parentNode from {to end} args} {

    if { ![_exists $name $parentNode] } {
	return -code error "node \"$parentNode\" does not exist in tree \"$name\""
    }

    if { [llength $args] == 0 } {
	# No node name given; generate a unique node name
	set node [GenerateUniqueNodeName $name]
    } else {
	set node [lindex $args 0]
    }

    if { [_exists $name $node] } {
	return -code error "node \"$node\" already exists in tree \"$name\""
    }

    variable ${name}::children
    variable ${name}::parent

    if {[string equal $from "end"]} {
	set from [expr {[llength $children($parentNode)] - 1}]
    } elseif {[regexp {^end-([0-9]+)$} $from -> n]} {
	set from [expr {[llength $children($parentNode)] - 1 - $n}]
    }
    if {[string equal $to "end"]} {
	set to [expr {[llength $children($parentNode)] - 1}]
    } elseif {[regexp {^end-([0-9]+)$} $to -> n]} {
	set to   [expr {[llength $children($parentNode)] - 1 - $n}]
    }

    # Save the list of children that are moving
    set moveChildren [lrange $children($parentNode) $from $to]

    # Remove those children from the parent
    ldelete children($parentNode) $from $to

    # Add the new node
    _insert $name $parentNode $from $node

    # Move the children
    set children($node) $moveChildren
    foreach child $moveChildren {
	set parent($child) $node
    }

    return $node
}

# ::struct::tree::_swap --
#
#	Swap two nodes in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node1	First node to swap.
#	node2	Second node to swap.
#
# Results:
#	None.

proc ::struct::tree::_swap {name node1 node2} {
    # Can't swap the magic root node
    variable ${name}::rootname
    if {[string equal $node1 $rootname] || [string equal $node2 $rootname]} {
	return -code error "cannot swap root node"
    }

    # Can only swap two real nodes
    if {![_exists $name $node1]} {
	return -code error "node \"$node1\" does not exist in tree \"$name\""
    }
    if {![_exists $name $node2]} {
	return -code error "node \"$node2\" does not exist in tree \"$name\""
    }

    # Can't swap a node with itself
    if {[string equal $node1 $node2]} {
	return -code error "cannot swap node \"$node1\" with itself"
    }

    # Swapping nodes means swapping their labels and values
    variable ${name}::children
    variable ${name}::parent

    set parent1 $parent($node1)
    set parent2 $parent($node2)

    # Replace node1 with node2 in node1's parent's children list, and
    # node2 with node1 in node2's parent's children list
    set i1 [lsearch -exact $children($parent1) $node1]
    set i2 [lsearch -exact $children($parent2) $node2]

    lset children($parent1) $i1 $node2
    lset children($parent2) $i2 $node1

    # Make node1 the parent of node2's children, and vis versa
    foreach child $children($node2) {
	set parent($child) $node1
    }
    foreach child $children($node1) {
	set parent($child) $node2
    }

    # Swap the children lists
    set children1 $children($node1)
    set children($node1) $children($node2)
    set children($node2) $children1

    if { [string equal $node1 $parent2] } {
	set parent($node1) $node2
	set parent($node2) $parent1
    } elseif { [string equal $node2 $parent1] } {
	set parent($node1) $parent2
	set parent($node2) $node1
    } else {
	set parent($node1) $parent2
	set parent($node2) $parent1
    }

    return
}

# ::struct::tree::_unset --
#
#	Remove a keyed value from a node.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to modify.
#	key	Name of attribute to unset.
#
# Results:
#	None.

proc ::struct::tree::_unset {name node key} {
    if {![_exists $name $node]} {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    variable ${name}::attribute
    if {![info exists attribute($node)]} {
	# No attribute data for this node,
	# nothing to do.
	return
    }

    upvar ${name}::$attribute($node) data
    catch {unset data($key)}

    if {[array size data] == 0} {
	# No attributes stored for this node, squash the whole array.
	unset attribute($node)
	unset data
    }
    return
}

# ::struct::tree::_walk --
#
#	Walk a tree using a pre-order depth or breadth first
#	search. Pre-order DFS is the default.  At each node that is visited,
#	a command will be called with the name of the tree and the node.
#
# Arguments:
#	name	Name of the tree.
#	node	Node at which to start.
#	args	Optional additional arguments specifying the type and order of
#		the tree walk, and the command to execute at each node.
#		Format is
#		    ?-type {bfs|dfs}? ?-order {pre|post|in|both}? a n script
#
# Results:
#	None.

proc ::struct::tree::_walk {name node args} {
    set usage "$name walk node ?-type {bfs|dfs}? ?-order {pre|post|in|both}? ?--? loopvar script"

    if {[llength $args] > 7 || [llength $args] < 2} {
	return -code error "wrong # args: should be \"$usage\""
    }

    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    set args [WalkOptions $args 2 $usage]
    # Remainder is 'a n script'

    foreach {loopvariables script} $args break

    if {[llength $loopvariables] > 2} {
	return -code error "too many loop variables, at most two allowed"
    } elseif {[llength $loopvariables] == 2} {
	foreach {avar nvar} $loopvariables break
    } else {
	set nvar [lindex $loopvariables 0]
	set avar {}
    }

    # Make sure we have a script to run, otherwise what's the point?
    if { [string equal $script ""] } {
	return -code error "no script specified, or empty"
    }

    # Do the walk
    variable ${name}::children
    set st [list ]
    lappend st $node

    # Compute some flags for the possible places of command evaluation
    set leave [expr {[string equal $order post] || [string equal $order both]}]
    set enter [expr {[string equal $order pre]  || [string equal $order both]}]
    set touch [string equal $order in]

    if {$leave} {
	set lvlabel leave
    } elseif {$touch} {
	# in-order does not provide a sense
	# of nesting for the parent, hence
	# no enter/leave, just 'visit'.
	set lvlabel visit
    }

    set rcode 0
    set rvalue {}

    if {[string equal $type "dfs"]} {
	# Depth-first walk, several orders of visiting nodes
	# (pre, post, both, in)

	array set visited {}

	while { [llength $st] > 0 } {
	    set node [lindex $st end]

	    if {[info exists visited($node)]} {
		# Second time we are looking at this 'node'.
		# Pop it, then evaluate the command (post, both, in).

		ldelete st end

		if {$leave || $touch} {
		    # Evaluate the script at this node
		    WalkCall $avar $nvar $name $node $lvlabel $script
		    # prune stops execution of loop here.
		}
	    } else {
		# First visit of this 'node'.
		# Do *not* pop it from the stack so that we are able
		# to visit again after its children

		# Remember it.
		set visited($node) .

		if {$enter} {
		    # Evaluate the script at this node (pre, both).
		    #
		    # Note: As this is done before the children are
		    # looked at the script may change the children of
		    # this node and thus affect the walk.

		    WalkCall $avar $nvar $name $node "enter" $script
		    # prune stops execution of loop here.
		}

		# Add the children of this node to the stack.
		# The exact behaviour depends on the chosen
		# order. For pre, post, both-order we just
		# have to add them in reverse-order so that
		# they will be popped left-to-right. For in-order
		# we have rearrange the stack so that the parent
		# is revisited immediately after the first child.
		# (but only if there is ore than one child,)

		set clist        $children($node)
		set len [llength $clist]

		if {$touch && ($len > 1)} {
		    # Pop node from stack, insert into list of children
		    ldelete st end
		    set clist [linsert $clist 1 $node]
		    incr len
		}

		for {set i [expr {$len - 1}]} {$i >= 0} {incr i -1} {
		    lappend st [lindex $clist $i]
		}
	    }
	}
    } else {
	# Breadth first walk (pre, post, both)
	# No in-order possible. Already captured.

	if {$leave} {
	    set backward $st
	}

	while { [llength $st] > 0 } {
	    set node [lindex   $st 0]
	    ldelete st 0

	    if {$enter} {
		# Evaluate the script at this node
		WalkCall $avar $nvar $name $node "enter" $script
		# prune stops execution of loop here.
	    }

	    # Add this node's children
	    # And create a mirrored version in case of post/both order.

	    foreach child $children($node) {
		lappend st $child
		if {$leave} {
		    set backward [linsert $backward 0 $child]
		}
	    }
	}

	if {$leave} {
	    foreach node $backward {
		# Evaluate the script at this node
		WalkCall $avar $nvar $name $node "leave" $script
	    }
	}
    }

    if {$rcode != 0} {
	return -code $rcode $rvalue
    }
    return
}

proc ::struct::tree::_walkproc {name node args} {
    set usage "$name walkproc node ?-type {bfs|dfs}? ?-order {pre|post|in|both}? ?--? cmdprefix"

    if {[llength $args] > 6 || [llength $args] < 1} {
	return -code error "wrong # args: should be \"$usage\""
    }

    if { ![_exists $name $node] } {
	return -code error "node \"$node\" does not exist in tree \"$name\""
    }

    set args [WalkOptions $args 1 $usage]
    # Remainder is 'n cmdprefix'

    set script [lindex $args 0]

    # Make sure we have a script to run, otherwise what's the point?
    if { ![llength $script] } {
	return -code error "no script specified, or empty"
    }

    # Do the walk
    variable ${name}::children
    set st [list ]
    lappend st $node

    # Compute some flags for the possible places of command evaluation
    set leave [expr {[string equal $order post] || [string equal $order both]}]
    set enter [expr {[string equal $order pre]  || [string equal $order both]}]
    set touch [string equal $order in]

    if {$leave} {
	set lvlabel leave
    } elseif {$touch} {
	# in-order does not provide a sense
	# of nesting for the parent, hence
	# no enter/leave, just 'visit'.
	set lvlabel visit
    }

    set rcode 0
    set rvalue {}

    if {[string equal $type "dfs"]} {
	# Depth-first walk, several orders of visiting nodes
	# (pre, post, both, in)

	array set visited {}

	while { [llength $st] > 0 } {
	    set node [lindex $st end]

	    if {[info exists visited($node)]} {
		# Second time we are looking at this 'node'.
		# Pop it, then evaluate the command (post, both, in).

		ldelete st end

		if {$leave || $touch} {
		    # Evaluate the script at this node
		    WalkCallProc $name $node $lvlabel $script
		    # prune stops execution of loop here.
		}
	    } else {
		# First visit of this 'node'.
		# Do *not* pop it from the stack so that we are able
		# to visit again after its children

		# Remember it.
		set visited($node) .

		if {$enter} {
		    # Evaluate the script at this node (pre, both).
		    #
		    # Note: As this is done before the children are
		    # looked at the script may change the children of
		    # this node and thus affect the walk.

		    WalkCallProc $name $node "enter" $script
		    # prune stops execution of loop here.
		}

		# Add the children of this node to the stack.
		# The exact behaviour depends on the chosen
		# order. For pre, post, both-order we just
		# have to add them in reverse-order so that
		# they will be popped left-to-right. For in-order
		# we have rearrange the stack so that the parent
		# is revisited immediately after the first child.
		# (but only if there is ore than one child,)

		set clist        $children($node)
		set len [llength $clist]

		if {$touch && ($len > 1)} {
		    # Pop node from stack, insert into list of children
		    ldelete st end
		    set clist [linsert $clist 1 $node]
		    incr len
		}

		for {set i [expr {$len - 1}]} {$i >= 0} {incr i -1} {
		    lappend st [lindex $clist $i]
		}
	    }
	}
    } else {
	# Breadth first walk (pre, post, both)
	# No in-order possible. Already captured.

	if {$leave} {
	    set backward $st
	}

	while { [llength $st] > 0 } {
	    set node [lindex   $st 0]
	    ldelete st 0

	    if {$enter} {
		# Evaluate the script at this node
		WalkCallProc $name $node "enter" $script
		# prune stops execution of loop here.
	    }

	    # Add this node's children
	    # And create a mirrored version in case of post/both order.

	    foreach child $children($node) {
		lappend st $child
		if {$leave} {
		    set backward [linsert $backward 0 $child]
		}
	    }
	}

	if {$leave} {
	    foreach node $backward {
		# Evaluate the script at this node
		WalkCallProc $name $node "leave" $script
	    }
	}
    }

    if {$rcode != 0} {
	return -code $rcode $rvalue
    }
    return
}

proc ::struct::tree::WalkOptions {theargs n usage} {
    upvar 1 type type order order

    # Set defaults
    set type dfs
    set order pre

    while {[llength $theargs]} {
	set flag [lindex $theargs 0]
	switch -exact -- $flag {
	    "-type" {
		if {[llength $theargs] < 2} {
		    return -code error "value for \"$flag\" missing"
		}
		set type [string tolower [lindex $theargs 1]]
		set theargs [lrange $theargs 2 end]
	    }
	    "-order" {
		if {[llength $theargs] < 2} {
		    return -code error "value for \"$flag\" missing"
		}
		set order [string tolower [lindex $theargs 1]]
		set theargs [lrange $theargs 2 end]
	    }
	    "--" {
		set theargs [lrange $theargs 1 end]
		break
	    }
	    default {
		break
	    }
	}
    }

    if {[llength $theargs] == 0} {
	return -code error "wrong # args: should be \"$usage\""
    }
    if {[llength $theargs] != $n} {
	return -code error "unknown option \"$flag\""
    }

    # Validate that the given type is good
    switch -exact -- $type {
	"dfs" - "bfs" {
	    set type $type
	}
	default {
	    return -code error "bad search type \"$type\": must be bfs or dfs"
	}
    }

    # Validate that the given order is good
    switch -exact -- $order {
	"pre" - "post" - "in" - "both" {
	    set order $order
	}
	default {
	    return -code error "bad search order \"$order\":\
		    must be both, in, pre, or post"
	}
    }

    if {[string equal $order "in"] && [string equal $type "bfs"]} {
	return -code error "unable to do a ${order}-order breadth first walk"
    }

    return $theargs
}

# ::struct::tree::WalkCall --
#
#	Helper command to 'walk' handling the evaluation
#	of the user-specified command. Information about
#	the tree, node and current action are substituted
#	into the command before it evaluation.
#
# Arguments:
#	tree	Tree we are walking
#	node	Node we are at.
#	action	The current action.
#	cmd	The command to call, already partially substituted.
#
# Results:
#	None.

proc ::struct::tree::WalkCall {avar nvar tree node action cmd} {

    if {$avar != {}} {
	upvar 2 $avar a ; set a $action
    }
    upvar 2 $nvar n ; set n $node

    set code [catch {uplevel 2 $cmd} result]

    # decide what to do upon the return code:
    #
    #               0 - the body executed successfully
    #               1 - the body raised an error
    #               2 - the body invoked [return]
    #               3 - the body invoked [break]
    #               4 - the body invoked [continue]
    #               5 - the body invoked [struct::tree::prune]
    # everything else - return and pass on the results
    #
    switch -exact -- $code {
	0 {}
	1 {
	    return -errorinfo [ErrorInfoAsCaller uplevel WalkCall]  \
		    -errorcode $::errorCode -code error $result
	}
	3 {
	    # FRINK: nocheck
	    return -code break
	}
	4 {}
	5 {
	    upvar order order
	    if {[string equal $order post] || [string equal $order in]} {
		return -code error "Illegal attempt to prune ${order}-order walking"
	    }
	    return -code continue
	}
	default {
	    upvar 1 rcode rcode rvalue rvalue
	    set rcode $code
	    set rvalue $result
	    return -code break
	    #return -code $code $result
	}
    }
    return {}
}

proc ::struct::tree::WalkCallProc {tree node action cmd} {

    lappend cmd $tree $node $action
    set code [catch {uplevel 2 $cmd} result]

    # decide what to do upon the return code:
    #
    #               0 - the body executed successfully
    #               1 - the body raised an error
    #               2 - the body invoked [return]
    #               3 - the body invoked [break]
    #               4 - the body invoked [continue]
    #               5 - the body invoked [struct::tree::prune]
    # everything else - return and pass on the results
    #
    switch -exact -- $code {
	0 {}
	1 {
	    return -errorinfo [ErrorInfoAsCaller uplevel WalkCallProc]  \
		    -errorcode $::errorCode -code error $result
	}
	3 {
	    # FRINK: nocheck
	    return -code break
	}
	4 {}
	5 {
	    upvar order order
	    if {[string equal $order post] || [string equal $order in]} {
		return -code error "Illegal attempt to prune ${order}-order walking"
	    }
	    return -code continue
	}
	default {
	    upvar 1 rcode rcode rvalue rvalue
	    set rcode $code
	    set rvalue $result
	    return -code break
	}
    }
    return {}
}

proc ::struct::tree::ErrorInfoAsCaller {find replace} {
    set info $::errorInfo
    set i [string last "\n    (\"$find" $info]
    if {$i == -1} {return $info}
    set result [string range $info 0 [incr i 6]]	;# keep "\n    (\""
    append result $replace			;# $find -> $replace
    incr i [string length $find]
    set j [string first ) $info [incr i]]	;# keep rest of parenthetical
    append result [string range $info $i $j]
    return $result
}

# ::struct::tree::GenerateUniqueNodeName --
#
#	Generate a unique node name for the given tree.
#
# Arguments:
#	name	Name of the tree to generate a unique node name for.
#
# Results:
#	node	Name of a node guaranteed to not exist in the tree.

proc ::struct::tree::GenerateUniqueNodeName {name} {
    variable ${name}::nextUnusedNode
    while {[_exists $name "node${nextUnusedNode}"]} {
	incr nextUnusedNode
    }
    return "node${nextUnusedNode}"
}

# ::struct::tree::KillNode --
#
#	Delete all data of a node.
#
# Arguments:
#	name	Name of the tree containing the node
#	node	Name of the node to delete.
#
# Results:
#	none

proc ::struct::tree::KillNode {name node} {
    variable ${name}::parent
    variable ${name}::children
    variable ${name}::attribute

    # Remove all record of $node
    unset parent($node)
    unset children($node)

    if {[info exists attribute($node)]} {
	# FRINK: nocheck
	unset ${name}::$attribute($node)
	unset attribute($node)
    }
    return
}

# ::struct::tree::GenAttributeStorage --
#
#	Create an array to store the attributes of a node in.
#
# Arguments:
#	name	Name of the tree containing the node
#	node	Name of the node which got attributes.
#
# Results:
#	none

proc ::struct::tree::GenAttributeStorage {name node} {
    variable ${name}::nextAttr
    variable ${name}::attribute

    set   attr "a[incr nextAttr]"
    set   attribute($node) $attr
    return
}

# ::struct::tree::Serialize --
#
#	Serialize a tree object (partially) into a transportable value.
#
# Arguments:
#	name	Name of the tree.
#	node	Root node of the serialized tree.
#
# Results:
#	None

proc ::struct::tree::Serialize {name node tvar} {
    upvar 1 $tvar tree

    variable ${name}::attribute
    variable ${name}::parent

    # 'node' is the root of the tree to serialize. The precondition
    # for the call is that this node is already stored in the list
    # 'tvar', at index 'rootidx'.

    # The attribute data for 'node' goes immediately after the 'node'
    # data. the node information is _not_ yet stored, and this command
    # has to do this.


    array set r {}
    set loc($node) 0

    lappend tree $node {}
    if {[info exists attribute($node)]} {
	upvar ${name}::$attribute($node) data
	lappend tree [array get data]
    } else {
	# Encode nodes without attributes.
	lappend tree {}
    }

    foreach n [DescendantsCore $name $node] {
	set loc($n) [llength $tree]
	lappend tree $n $loc($parent($n))

	if {[info exists attribute($n)]} {
	    upvar ${name}::$attribute($n) data
	    lappend tree [array get data]
	} else {
	    # Encode nodes without attributes.
	    lappend tree {}
	}
    }

    return $tree
}


proc ::struct::tree::CheckSerialization {ser avar pvar cvar rnvar} {
    upvar 1 $avar attr $pvar p $cvar ch $rnvar rn

    # Overall length ok ?

    if {[llength $ser] % 3} {
	return -code error \
		"error in serialization: list length not a multiple of 3."
    }

    set rn {}
    array set p    {}
    array set ch   {}
    array set attr {}

    # Basic decoder pass

    foreach {node parent nattr} $ser {

	# Initialize children data, if not already done
	if {![info exists ch($node)]} {
	    set ch($node) {}
	}
	# Attribute length ok ? Dictionary!
	if {[llength $nattr] % 2} {
	    return -code error \
		    "error in serialization: malformed attribute dictionary."
	}
	# Remember attribute data only for non-empty nodes
	if {[llength $nattr]} {
	    set attr($node) $nattr
	}
	# Remember root
	if {$parent == {}} {
	    lappend rn $node
	    set p($node) {}
	    continue
	}
	# Parent reference ok ?
	if {
	    ![string is integer -strict $parent] ||
	    ($parent % 3) ||
	    ($parent < 0) ||
	    ($parent >= [llength $ser])
	} {
	    return -code error \
		    "error in serialization: bad parent reference \"$parent\"."
	}
	# Remember parent, and reconstruct children

	set p($node) [lindex $ser $parent]
	lappend ch($p($node)) $node
    }

    # Root node information ok ?

    if {[llength $rn] < 1} {
	return -code error \
		"error in serialization: no root specified."
    } elseif {[llength $rn] > 1} {
	return -code error \
		"error in serialization: multiple root nodes."
    }
    set rn [lindex $rn 0]

    # Duplicate node names ?

    if {[array size ch] < ([llength $ser] / 3)} {
	return -code error \
		"error in serialization: duplicate node names."
    }

    # Cycles in the parent relationship ?

    array set visited {}
    foreach n [array names p] {
	if {[info exists visited($n)]} {continue}
	array set _ {}
	while {$n != {}} {
	    if {[info exists _($n)]} {
		# Node already converted, cycle.
		return -code error \
			"error in serialization: cycle detected."
	    }
	    set _($n)       .
	    # root ?
	    if {$p($n) == {}} {break}
	    set n $p($n)
	    if {[info exists visited($n)]} {break}
	    set visited($n) .
	}
	unset _
    }
    # Ok. The data is now ready for the caller.

    return
}

##########################
# Private functions follow
#
# Do a compatibility version of [lset] for pre-8.4 versions of Tcl.
# This version does not do multi-arg [lset]!

proc ::struct::tree::K { x y } { set x }

if { [package vcompare [package provide Tcl] 8.4] < 0 } {
    proc ::struct::tree::lset { var index arg } {
	upvar 1 $var list
	set list [::lreplace [K $list [set list {}]] $index $index $arg]
    }
}

proc ::struct::tree::ldelete {var index {end {}}} {
    upvar 1 $var list
    if {$end == {}} {set end $index}
    set list [lreplace [K $list [set list {}]] $index $end]
    return
}

# ### ### ### ######### ######### #########
## Ready

namespace eval ::struct {
    # Put 'tree::tree' into the general structure namespace
    # for pickup by the main management.

    namespace import -force tree::tree_tcl
}

# THATCH /herbie/packages/modules/85/cmdline-1.5.2.tm 
# cmdline.tcl --
#
#	This package provides a utility for parsing command line
#	arguments that are processed by our various applications.
#	It also includes a utility routine to determine the
#	application name for use in command line errors.
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
# Copyright (c) 2001-2015 by Andreas Kupries <andreas_kupries@users.sf.net>.
# Copyright (c) 2003      by David N. Welton  <davidw@dedasys.com>
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

package require Tcl 8.2
package provide cmdline 1.5.2

namespace eval ::cmdline {
    namespace export getArgv0 getopt getKnownOpt getfiles getoptions \
	    getKnownOptions usage
}

# ::cmdline::getopt --
#
#	The cmdline::getopt works in a fashion like the standard
#	C based getopt function.  Given an option string and a
#	pointer to an array or args this command will process the
#	first argument and return info on how to proceed.
#
# Arguments:
#	argvVar		Name of the argv list that you
#			want to process.  If options are found the
#			arg list is modified and the processed arguments
#			are removed from the start of the list.
#	optstring	A list of command options that the application
#			will accept.  If the option ends in ".arg" the
#			getopt routine will use the next argument as
#			an argument to the option.  Otherwise the option
#			is a boolean that is set to 1 if present.
#	optVar		The variable pointed to by optVar
#			contains the option that was found (without the
#			leading '-' and without the .arg extension).
#	valVar		Upon success, the variable pointed to by valVar
#			contains the value for the specified option.
#			This value comes from the command line for .arg
#			options, otherwise the value is 1.
#			If getopt fails, the valVar is filled with an
#			error message.
#
# Results:
# 	The getopt function returns 1 if an option was found, 0 if no more
# 	options were found, and -1 if an error occurred.

proc ::cmdline::getopt {argvVar optstring optVar valVar} {
    upvar 1 $argvVar argsList
    upvar 1 $optVar option
    upvar 1 $valVar value

    set result [getKnownOpt argsList $optstring option value]

    if {$result < 0} {
        # Collapse unknown-option error into any-other-error result.
        set result -1
    }
    return $result
}

# ::cmdline::getKnownOpt --
#
#	The cmdline::getKnownOpt works in a fashion like the standard
#	C based getopt function.  Given an option string and a
#	pointer to an array or args this command will process the
#	first argument and return info on how to proceed.
#
# Arguments:
#	argvVar		Name of the argv list that you
#			want to process.  If options are found the
#			arg list is modified and the processed arguments
#			are removed from the start of the list.  Note that
#			unknown options and the args that follow them are
#			left in this list.
#	optstring	A list of command options that the application
#			will accept.  If the option ends in ".arg" the
#			getopt routine will use the next argument as
#			an argument to the option.  Otherwise the option
#			is a boolean that is set to 1 if present.
#	optVar		The variable pointed to by optVar
#			contains the option that was found (without the
#			leading '-' and without the .arg extension).
#	valVar		Upon success, the variable pointed to by valVar
#			contains the value for the specified option.
#			This value comes from the command line for .arg
#			options, otherwise the value is 1.
#			If getopt fails, the valVar is filled with an
#			error message.
#
# Results:
# 	The getKnownOpt function returns 1 if an option was found,
#	0 if no more options were found, -1 if an unknown option was
#	encountered, and -2 if any other error occurred.

proc ::cmdline::getKnownOpt {argvVar optstring optVar valVar} {
    upvar 1 $argvVar argsList
    upvar 1 $optVar  option
    upvar 1 $valVar  value

    # default settings for a normal return
    set value ""
    set option ""
    set result 0

    # check if we're past the end of the args list
    if {[llength $argsList] != 0} {

	# if we got -- or an option that doesn't begin with -, return (skipping
	# the --).  otherwise process the option arg.
	switch -glob -- [set arg [lindex $argsList 0]] {
	    "--" {
		set argsList [lrange $argsList 1 end]
	    }
	    "--*" -
	    "-*" {
		set option [string range $arg 1 end]
		if {[string equal [string range $option 0 0] "-"]} {
		    set option [string range $arg 2 end]
		}

		# support for format: [-]-option=value
		set idx [string first "=" $option 1]
		if {$idx != -1} {
		    set _val   [string range $option [expr {$idx+1}] end]
		    set option [string range $option 0   [expr {$idx-1}]]
		}

		if {[lsearch -exact $optstring $option] != -1} {
		    # Booleans are set to 1 when present
		    set value 1
		    set result 1
		    set argsList [lrange $argsList 1 end]
		} elseif {[lsearch -exact $optstring "$option.arg"] != -1} {
		    set result 1
		    set argsList [lrange $argsList 1 end]

		    if {[info exists _val]} {
			set value $_val
		    } elseif {[llength $argsList]} {
			set value [lindex $argsList 0]
			set argsList [lrange $argsList 1 end]
		    } else {
			set value "Option \"$option\" requires an argument"
			set result -2
		    }
		} else {
		    # Unknown option.
		    set value "Illegal option \"-$option\""
		    set result -1
		}
	    }
	    default {
		# Skip ahead
	    }
	}
    }

    return $result
}

# ::cmdline::getoptions --
#
#	Process a set of command line options, filling in defaults
#	for those not specified.  This also generates an error message
#	that lists the allowed flags if an incorrect flag is specified.
#
# Arguments:
#	argvVar		The name of the argument list, typically argv.
#			We remove all known options and their args from it.
#                       In other words, after the call to this command the
#                       referenced variable contains only the non-options,
#                       and unknown options.
#	optlist		A list-of-lists where each element specifies an option
#			in the form:
#				(where flag takes no argument)
#					flag comment
#
#				(or where flag takes an argument)
#					flag default comment
#
#			If flag ends in ".arg" then the value is taken from the
#			command line. Otherwise it is a boolean and appears in
#			the result if present on the command line. If flag ends
#			in ".secret", it will not be displayed in the usage.
#	usage		Text to include in the usage display. Defaults to
#			"options:"
#
# Results
#	Name value pairs suitable for using with array set.
#       A modified `argvVar`.

proc ::cmdline::getoptions {argvVar optlist {usage options:}} {
    upvar 1 $argvVar argv

    set opts [GetOptionDefaults $optlist result]

    set argc [llength $argv]
    while {[set err [getopt argv $opts opt arg]]} {
	if {$err < 0} {
            set result(?) ""
            break
	}
	set result($opt) $arg
    }
    if {[info exist result(?)] || [info exists result(help)]} {
	Error [usage $optlist $usage] USAGE
    }
    return [array get result]
}

# ::cmdline::getKnownOptions --
#
#	Process a set of command line options, filling in defaults
#	for those not specified.  This ignores unknown flags, but generates
#	an error message that lists the correct usage if a known option
#	is used incorrectly.
#
# Arguments:
#	argvVar		The name of the argument list, typically argv.  This
#			We remove all known options and their args from it.
#                       In other words, after the call to this command the
#                       referenced variable contains only the non-options,
#                       and unknown options.
#	optlist		A list-of-lists where each element specifies an option
#			in the form:
#				flag default comment
#			If flag ends in ".arg" then the value is taken from the
#			command line. Otherwise it is a boolean and appears in
#			the result if present on the command line. If flag ends
#			in ".secret", it will not be displayed in the usage.
#	usage		Text to include in the usage display. Defaults to
#			"options:"
#
# Results
#	Name value pairs suitable for using with array set.
#       A modified `argvVar`.

proc ::cmdline::getKnownOptions {argvVar optlist {usage options:}} {
    upvar 1 $argvVar argv

    set opts [GetOptionDefaults $optlist result]

    # As we encounter them, keep the unknown options and their
    # arguments in this list.  Before we return from this procedure,
    # we'll prepend these args to the argList so that the application
    # doesn't lose them.

    set unknownOptions [list]

    set argc [llength $argv]
    while {[set err [getKnownOpt argv $opts opt arg]]} {
	if {$err == -1} {
            # Unknown option.

            # Skip over any non-option items that follow it.
            # For now, add them to the list of unknownOptions.
            lappend unknownOptions [lindex $argv 0]
            set argv [lrange $argv 1 end]
            while {([llength $argv] != 0) \
                    && ![string match "-*" [lindex $argv 0]]} {
                lappend unknownOptions [lindex $argv 0]
                set argv [lrange $argv 1 end]
            }
	} elseif {$err == -2} {
            set result(?) ""
            break
        } else {
            set result($opt) $arg
        }
    }

    # Before returning, prepend the any unknown args back onto the
    # argList so that the application doesn't lose them.
    set argv [concat $unknownOptions $argv]

    if {[info exist result(?)] || [info exists result(help)]} {
	Error [usage $optlist $usage] USAGE
    }
    return [array get result]
}

# ::cmdline::GetOptionDefaults --
#
#	This internal procedure processes the option list (that was passed to
#	the getopt or getKnownOpt procedure).  The defaultArray gets an index
#	for each option in the option list, the value of which is the option's
#	default value.
#
# Arguments:
#	optlist		A list-of-lists where each element specifies an option
#			in the form:
#				flag default comment
#			If flag ends in ".arg" then the value is taken from the
#			command line. Otherwise it is a boolean and appears in
#			the result if present on the command line. If flag ends
#			in ".secret", it will not be displayed in the usage.
#	defaultArrayVar	The name of the array in which to put argument defaults.
#
# Results
#	Name value pairs suitable for using with array set.

proc ::cmdline::GetOptionDefaults {optlist defaultArrayVar} {
    upvar 1 $defaultArrayVar result

    set opts {? help}
    foreach opt $optlist {
	set name [lindex $opt 0]
	if {[regsub -- {\.secret$} $name {} name] == 1} {
	    # Need to hide this from the usage display and getopt
	}
	lappend opts $name
	if {[regsub -- {\.arg$} $name {} name] == 1} {

	    # Set defaults for those that take values.

	    set default [lindex $opt 1]
	    set result($name) $default
	} else {
	    # The default for booleans is false
	    set result($name) 0
	}
    }
    return $opts
}

# ::cmdline::usage --
#
#	Generate an error message that lists the allowed flags.
#
# Arguments:
#	optlist		As for cmdline::getoptions
#	usage		Text to include in the usage display. Defaults to
#			"options:"
#
# Results
#	A formatted usage message

proc ::cmdline::usage {optlist {usage {options:}}} {
    set str "[getArgv0] $usage\n"
    set longest 20
    set lines {}
    foreach opt [concat $optlist \
	     {{- "Forcibly stop option processing"} {help "Print this message"} {? "Print this message"}}] {
	set name "-[lindex $opt 0]"
	if {[regsub -- {\.secret$} $name {} name] == 1} {
	    # Hidden option
	    continue
	}
	if {[regsub -- {\.arg$} $name {} name] == 1} {
	    append name " value"
	    set desc "[lindex $opt 2] <[lindex $opt 1]>"
	} else {
	    set desc "[lindex $opt 1]"
	}
	set n [string length $name]
	if {$n > $longest} { set longest $n }
	# max not available before 8.5 - set longest [expr {max($longest, )}]
	lappend lines $name $desc
    }
    foreach {name desc} $lines {
	append str "[string trimright [format " %-*s %s" $longest $name $desc]]\n"
    }

    return $str
}

# ::cmdline::getfiles --
#
#	Given a list of file arguments from the command line, compute
#	the set of valid files.  On windows, file globbing is performed
#	on each argument.  On Unix, only file existence is tested.  If
#	a file argument produces no valid files, a warning is optionally
#	generated.
#
#	This code also uses the full path for each file.  If not
#	given it prepends [pwd] to the filename.  This ensures that
#	these files will never conflict with files in our zip file.
#
# Arguments:
#	patterns	The file patterns specified by the user.
#	quiet		If this flag is set, no warnings will be generated.
#
# Results:
#	Returns the list of files that match the input patterns.

proc ::cmdline::getfiles {patterns quiet} {
    set result {}
    if {$::tcl_platform(platform) == "windows"} {
	foreach pattern $patterns {
	    set pat [file join $pattern]
	    set files [glob -nocomplain -- $pat]
	    if {$files == {}} {
		if {! $quiet} {
		    puts stdout "warning: no files match \"$pattern\""
		}
	    } else {
		foreach file $files {
		    lappend result $file
		}
	    }
	}
    } else {
	set result $patterns
    }
    set files {}
    foreach file $result {
	# Make file an absolute path so that we will never conflict
	# with files that might be contained in our zip file.
	set fullPath [file join [pwd] $file]

	if {[file isfile $fullPath]} {
	    lappend files $fullPath
	} elseif {! $quiet} {
	    puts stdout "warning: no files match \"$file\""
	}
    }
    return $files
}

# ::cmdline::getArgv0 --
#
#	This command returns the "sanitized" version of argv0.  It will strip
#	off the leading path and remove the ".bin" extensions that our apps
#	use because they must be wrapped by a shell script.
#
# Arguments:
#	None.
#
# Results:
#	The application name that can be used in error messages.

proc ::cmdline::getArgv0 {} {
    global argv0

    set name [file tail $argv0]
    return [file rootname $name]
}

##
# ### ### ### ######### ######### #########
##
# Now the typed versions of the above commands.
##
# ### ### ### ######### ######### #########
##

# typedCmdline.tcl --
#
#    This package provides a utility for parsing typed command
#    line arguments that may be processed by various applications.
#
# Copyright (c) 2000 by Ross Palmer Mohn.
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: cmdline.tcl,v 1.28 2011/02/23 17:41:52 andreas_kupries Exp $

namespace eval ::cmdline {
    namespace export typedGetopt typedGetoptions typedUsage

    # variable cmdline::charclasses --
    #
    #    Create regexp list of allowable character classes
    #    from "string is" error message.
    #
    # Results:
    #    String of character class names separated by "|" characters.

    variable charclasses
    #checker exclude badKey
    catch {string is . .} charclasses
    variable dummy
    regexp      -- {must be (.+)$} $charclasses dummy charclasses
    regsub -all -- {, (or )?}      $charclasses {|}   charclasses
    unset dummy
}

# ::cmdline::typedGetopt --
#
#	The cmdline::typedGetopt works in a fashion like the standard
#	C based getopt function.  Given an option string and a
#	pointer to a list of args this command will process the
#	first argument and return info on how to proceed. In addition,
#	you may specify a type for the argument to each option.
#
# Arguments:
#	argvVar		Name of the argv list that you want to process.
#			If options are found, the arg list is modified
#			and the processed arguments are removed from the
#			start of the list.
#
#	optstring	A list of command options that the application
#			will accept.  If the option ends in ".xxx", where
#			xxx is any valid character class to the tcl
#			command "string is", then typedGetopt routine will
#			use the next argument as a typed argument to the
#			option. The argument must match the specified
#			character classes (e.g. integer, double, boolean,
#			xdigit, etc.). Alternatively, you may specify
#			".arg" for an untyped argument.
#
#	optVar		Upon success, the variable pointed to by optVar
#			contains the option that was found (without the
#			leading '-' and without the .xxx extension).  If
#			typedGetopt fails the variable is set to the empty
#			string. SOMETIMES! Different for each -value!
#
#	argVar		Upon success, the variable pointed to by argVar
#			contains the argument for the specified option.
#			If typedGetopt fails, the variable is filled with
#			an error message.
#
# Argument type syntax:
#	Option that takes no argument.
#		foo
#
#	Option that takes a typeless argument.
#		foo.arg
#
#	Option that takes a typed argument. Allowable types are all
#	valid character classes to the tcl command "string is".
#	Currently must be one of alnum, alpha, ascii, control,
#	boolean, digit, double, false, graph, integer, lower, print,
#	punct, space, true, upper, wordchar, or xdigit.
#		foo.double
#
#	Option that takes an argument from a list.
#		foo.(bar|blat)
#
# Argument quantifier syntax:
#	Option that takes an optional argument.
#		foo.arg?
#
#	Option that takes a list of arguments terminated by "--".
#		foo.arg+
#
#	Option that takes an optional list of arguments terminated by "--".
#		foo.arg*
#
#	Argument quantifiers work on all argument types, so, for
#	example, the following is a valid option specification.
#		foo.(bar|blat|blah)?
#
# Argument syntax miscellany:
#	Options may be specified on the command line using a unique,
#	shortened version of the option name. Given that program foo
#	has an option list of {bar.alpha blah.arg blat.double},
#	"foo -b fob" returns an error, but "foo -ba fob"
#	successfully returns {bar fob}
#
# Results:
#	The typedGetopt function returns one of the following:
#	 1	a valid option was found
#	 0	no more options found to process
#	-1	invalid option
#	-2	missing argument to a valid option
#	-3	argument to a valid option does not match type
#
# Known Bugs:
#	When using options which include special glob characters,
#	you must use the exact option. Abbreviating it can cause
#	an error in the "cmdline::prefixSearch" procedure.

proc ::cmdline::typedGetopt {argvVar optstring optVar argVar} {
    variable charclasses

    upvar $argvVar argsList

    upvar $optVar retvar
    upvar $argVar optarg

    # default settings for a normal return
    set optarg ""
    set retvar ""
    set retval 0

    # check if we're past the end of the args list
    if {[llength $argsList] != 0} {

        # if we got -- or an option that doesn't begin with -, return (skipping
        # the --).  otherwise process the option arg.
        switch -glob -- [set arg [lindex $argsList 0]] {
            "--" {
                set argsList [lrange $argsList 1 end]
            }

            "-*" {
                # Create list of options without their argument extensions

                set optstr ""
                foreach str $optstring {
                    lappend optstr [file rootname $str]
                }

                set _opt [string range $arg 1 end]

                set i [prefixSearch $optstr [file rootname $_opt]]
                if {$i != -1} {
                    set opt [lindex $optstring $i]

                    set quantifier "none"
                    if {[regexp -- {\.[^.]+([?+*])$} $opt dummy quantifier]} {
                        set opt [string range $opt 0 end-1]
                    }

                    if {[string first . $opt] == -1} {
                        set retval 1
                        set retvar $opt
                        set argsList [lrange $argsList 1 end]

                    } elseif {[regexp -- "\\.(arg|$charclasses)\$" $opt dummy charclass]
                            || [regexp -- {\.\(([^)]+)\)} $opt dummy charclass]} {
				if {[string equal arg $charclass]} {
                            set type arg
			} elseif {[regexp -- "^($charclasses)\$" $charclass]} {
                            set type class
                        } else {
                            set type oneof
                        }

                        set argsList [lrange $argsList 1 end]
                        set opt [file rootname $opt]

                        while {1} {
                            if {[llength $argsList] == 0
                                    || [string equal "--" [lindex $argsList 0]]} {
                                if {[string equal "--" [lindex $argsList 0]]} {
                                    set argsList [lrange $argsList 1 end]
                                }

                                set oneof ""
                                if {$type == "arg"} {
                                    set charclass an
                                } elseif {$type == "oneof"} {
                                    set oneof ", one of $charclass"
                                    set charclass an
                                }

                                if {$quantifier == "?"} {
                                    set retval 1
                                    set retvar $opt
                                    set optarg ""
                                } elseif {$quantifier == "+"} {
                                    set retvar $opt
                                    if {[llength $optarg] < 1} {
                                        set retval -2
                                        set optarg "Option requires at least one $charclass argument$oneof -- $opt"
                                    } else {
                                        set retval 1
                                    }
                                } elseif {$quantifier == "*"} {
                                    set retval 1
                                    set retvar $opt
                                } else {
                                    set optarg "Option requires $charclass argument$oneof -- $opt"
                                    set retvar $opt
                                    set retval -2
                                }
                                set quantifier ""
                            } elseif {($type == "arg")
                                    || (($type == "oneof")
                                    && [string first "|[lindex $argsList 0]|" "|$charclass|"] != -1)
                                    || (($type == "class")
                                    && [string is $charclass [lindex $argsList 0]])} {
                                set retval 1
                                set retvar $opt
                                lappend optarg [lindex $argsList 0]
                                set argsList [lrange $argsList 1 end]
                            } else {
                                set oneof ""
                                if {$type == "arg"} {
                                    set charclass an
                                } elseif {$type == "oneof"} {
                                    set oneof ", one of $charclass"
                                    set charclass an
                                }
                                set optarg "Option requires $charclass argument$oneof -- $opt"
                                set retvar $opt
                                set retval -3

                                if {$quantifier == "?"} {
                                    set retval 1
                                    set optarg ""
                                }
                                set quantifier ""
                            }
                             if {![regexp -- {[+*]} $quantifier]} {
                                break;
                            }
                        }
                    } else {
                        Error \
			    "Illegal option type specification: must be one of $charclasses" \
			    BAD OPTION TYPE
                    }
                } else {
                    set optarg "Illegal option -- $_opt"
                    set retvar $_opt
                    set retval -1
                }
            }
	    default {
		# Skip ahead
	    }
        }
    }

    return $retval
}

# ::cmdline::typedGetoptions --
#
#	Process a set of command line options, filling in defaults
#	for those not specified. This also generates an error message
#	that lists the allowed options if an incorrect option is
#	specified.
#
# Arguments:
#	argvVar		The name of the argument list, typically argv
#	optlist		A list-of-lists where each element specifies an option
#			in the form:
#
#				option default comment
#
#			Options formatting is as described for the optstring
#			argument of typedGetopt. Default is for optionally
#			specifying a default value. Comment is for optionally
#			specifying a comment for the usage display. The
#			options "--", "-help", and "-?" are automatically included
#			in optlist.
#
# Argument syntax miscellany:
#	Options formatting and syntax is as described in typedGetopt.
#	There are two additional suffixes that may be applied when
#	passing options to typedGetoptions.
#
#	You may add ".multi" as a suffix to any option. For options
#	that take an argument, this means that the option may be used
#	more than once on the command line and that each additional
#	argument will be appended to a list, which is then returned
#	to the application.
#		foo.double.multi
#
#	If a non-argument option is specified as ".multi", it is
#	toggled on and off for each time it is used on the command
#	line.
#		foo.multi
#
#	If an option specification does not contain the ".multi"
#	suffix, it is not an error to use an option more than once.
#	In this case, the behavior for options with arguments is that
#	the last argument is the one that will be returned. For
#	options that do not take arguments, using them more than once
#	has no additional effect.
#
#	Options may also be hidden from the usage display by
#	appending the suffix ".secret" to any option specification.
#	Please note that the ".secret" suffix must be the last suffix,
#	after any argument type specification and ".multi" suffix.
#		foo.xdigit.multi.secret
#
# Results
#	Name value pairs suitable for using with array set.

proc ::cmdline::typedGetoptions {argvVar optlist {usage options:}} {
    variable charclasses

    upvar 1 $argvVar argv

    set opts {? help}
    foreach opt $optlist {
        set name [lindex $opt 0]
        if {[regsub -- {\.secret$} $name {} name] == 1} {
            # Remove this extension before passing to typedGetopt.
        }
        if {[regsub -- {\.multi$} $name {} name] == 1} {
            # Remove this extension before passing to typedGetopt.

            regsub -- {\..*$} $name {} temp
            set multi($temp) 1
        }
        lappend opts $name
        if {[regsub -- "\\.(arg|$charclasses|\\(.+).?\$" $name {} name] == 1} {
            # Set defaults for those that take values.
            # Booleans are set just by being present, or not

            set dflt [lindex $opt 1]
            if {$dflt != {}} {
                set defaults($name) $dflt
            }
        }
    }
    set argc [llength $argv]
    while {[set err [typedGetopt argv $opts opt arg]]} {
        if {$err == 1} {
            if {[info exists result($opt)]
                    && [info exists multi($opt)]} {
                # Toggle boolean options or append new arguments

                if {$arg == ""} {
                    unset result($opt)
                } else {
                    set result($opt) "$result($opt) $arg"
                }
            } else {
                set result($opt) "$arg"
            }
        } elseif {($err == -1) || ($err == -3)} {
            Error [typedUsage $optlist $usage] USAGE
        } elseif {$err == -2 && ![info exists defaults($opt)]} {
            Error [typedUsage $optlist $usage] USAGE
        }
    }
    if {[info exists result(?)] || [info exists result(help)]} {
        Error [typedUsage $optlist $usage] USAGE
    }
    foreach {opt dflt} [array get defaults] {
        if {![info exists result($opt)]} {
            set result($opt) $dflt
        }
    }
    return [array get result]
}

# ::cmdline::typedUsage --
#
#	Generate an error message that lists the allowed flags,
#	type of argument taken (if any), default value (if any),
#	and an optional description.
#
# Arguments:
#	optlist		As for cmdline::typedGetoptions
#
# Results
#	A formatted usage message

proc ::cmdline::typedUsage {optlist {usage {options:}}} {
    variable charclasses

    set str "[getArgv0] $usage\n"
    set longest 20
    set lines {}
    foreach opt [concat $optlist \
            {{help "Print this message"} {? "Print this message"}}] {
        set name "-[lindex $opt 0]"
        if {[regsub -- {\.secret$} $name {} name] == 1} {
            # Hidden option
	    continue
	}

	if {[regsub -- {\.multi$} $name {} name] == 1} {
	    # Display something about multiple options
	}

	if {[regexp -- "\\.(arg|$charclasses)\$" $name dummy charclass] ||
	    [regexp -- {\.\(([^)]+)\)} $opt dummy charclass]
	} {
	    regsub -- "\\..+\$" $name {} name
	    append name " $charclass"
	    set desc [lindex $opt 2]
	    set default [lindex $opt 1]
	    if {$default != ""} {
		append desc " <$default>"
	    }
	} else {
	    set desc [lindex $opt 1]
	}
	lappend accum $name $desc
	set n [string length $name]
	if {$n > $longest} { set longest $n }
	# max not available before 8.5 - set longest [expr {max($longest, [string length $name])}]
    }
    foreach {name desc} $accum {
	append str "[string trimright [format " %-*s %s" $longest $name $desc]]\n"
    }
    return $str
}

# ::cmdline::prefixSearch --
#
#	Search a Tcl list for a pattern; searches first for an exact match,
#	and if that fails, for a unique prefix that matches the pattern
#	(i.e, first "lsearch -exact", then "lsearch -glob $pattern*"
#
# Arguments:
#	list		list of words
#	pattern		word to search for
#
# Results:
#	Index of found word is returned. If no exact match or
#	unique short version is found then -1 is returned.

proc ::cmdline::prefixSearch {list pattern} {
    # Check for an exact match

    if {[set pos [::lsearch -exact $list $pattern]] > -1} {
        return $pos
    }

    # Check for a unique short version

    set slist [lsort $list]
    if {[set pos [::lsearch -glob $slist $pattern*]] > -1} {
        # What if there is nothing for the check variable?

        set check [lindex $slist [expr {$pos + 1}]]
        if {[string first $pattern $check] != 0} {
            return [::lsearch -exact $list [lindex $slist $pos]]
        }
    }
    return -1
}
# ::cmdline::Error --
#
#	Internal helper to throw errors with a proper error-code attached.
#
# Arguments:
#	message		text of the error message to throw.
#	args		additional parts of the error code to use,
#                       with CMDLINE as basic prefix added by this command.
#
# Results:
#	An error is thrown, always.

proc ::cmdline::Error {message args} {
    return -code error -errorcode [linsert $args 0 CMDLINE] $message
}

# THATCH /herbie/packages/modules/85/autoscroll-1.1.tm 
# autoscroll.tcl --
#
#       Package to create scroll bars that automatically appear when
#       a window is too small to display its content.
#
# Copyright (c) 2003    Kevin B Kenny <kennykb@users.sourceforge.net>
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: autoscroll.tcl,v 1.8 2005/06/01 02:37:51 andreas_kupries Exp $

package require Tk
package provide autoscroll 1.1

namespace eval ::autoscroll {
    namespace export autoscroll unautoscroll
    bind Autoscroll <Destroy> [namespace code [list destroyed %W]]
    bind Autoscroll <Map> [namespace code [list map %W]]
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::autoscroll --
 #
 #       Create a scroll bar that disappears when it is not needed, and
 #       reappears when it is.
 #
 # Parameters:
 #       w    -- Path name of the scroll bar, which should already exist
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       The widget command is renamed, so that the 'set' command can
 #       be intercepted and determine whether the widget should appear.
 #       In addition, the 'Autoscroll' bind tag is added to the widget,
 #       so that the <Destroy> event can be intercepted.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::autoscroll { w } {
    if { [info commands ::autoscroll::renamed$w] != "" } { return $w }
    rename $w ::autoscroll::renamed$w
    interp alias {} ::$w {} ::autoscroll::widgetCommand $w
    bindtags $w [linsert [bindtags $w] 1 Autoscroll]
    eval [list ::$w set] [renamed$w get]
    return $w
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::unautoscroll --
 #
 #       Return a scrollbar to its normal static behavior by removing
 #       it from the control of this package.
 #
 # Parameters:
 #       w    -- Path name of the scroll bar, which must have previously
 #               had ::autoscroll::autoscroll called on it.
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       The widget command is renamed to its original name. The widget
 #       is mapped if it was not currently displayed. The widgets
 #       bindtags are returned to their original state. Internal memory
 #       is cleaned up.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::unautoscroll { w } {
    if { [info commands ::autoscroll::renamed$w] != "" } {
        variable grid
        rename ::$w {}
        rename ::autoscroll::renamed$w ::$w
        if { [set i [lsearch -exact [bindtags $w] Autoscroll]] > -1 } {
            bindtags $w [lreplace [bindtags $w] $i $i]
        }
        if { [info exists grid($w)] } {
            eval [join $grid($w) \;]
            unset grid($w)
        }
    }
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::widgetCommand --
 #
 #       Widget command on an 'autoscroll' scrollbar
 #
 # Parameters:
 #       w       -- Path name of the scroll bar
 #       command -- Widget command being executed
 #       args    -- Arguments to the commane
 #
 # Results:
 #       Returns whatever the widget command returns
 #
 # Side effects:
 #       Has whatever side effects the widget command has.  In
 #       addition, the 'set' widget command is handled specially,
 #       by gridding/packing the scroll bar according to whether
 #       it is required.
 #
 #------------------------------------------------------------

proc ::autoscroll::widgetCommand { w command args } {
    variable grid
    if { $command == "set" } {
        foreach { min max } $args {}
        if { $min <= 0 && $max >= 1 } {
            switch -exact -- [winfo manager $w] {
                grid {
                    lappend grid($w) "[list grid $w] [grid info $w]"
                    grid forget $w
                }
                pack {
                    foreach x [pack slaves [winfo parent $w]] {
                        lappend grid($w) "[list pack $x] [pack info $x]"
                    }
                    pack forget $w
                }
            }
        } elseif { [info exists grid($w)] } {
            eval [join $grid($w) \;]
            unset grid($w)
        }
    }
    return [eval [list renamed$w $command] $args]
}


 #----------------------------------------------------------------------
 #
 # ::autoscroll::destroyed --
 #
 #       Callback executed when an automatic scroll bar is destroyed.
 #
 # Parameters:
 #       w -- Path name of the scroll bar
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       Cleans up internal memory.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::destroyed { w } {
    variable grid
    catch { unset grid($w) }
    rename ::$w {}
}


 #----------------------------------------------------------------------
 #
 # ::autoscroll::map --
 #
 #       Callback executed when an automatic scroll bar is mapped.
 #
 # Parameters:
 #       w -- Path name of the scroll bar.
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       Geometry of the scroll bar's top-level window is constrained.
 #
 # This procedure keeps the top-level window associated with an
 # automatic scroll bar from being resized automatically after the
 # scroll bar is mapped.  This effect avoids a potential endless loop
 # in the case where the resize of the top-level window resizes the
 # widget being scrolled, causing the scroll bar no longer to be needed.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::map { w } {
    wm geometry [winfo toplevel $w] [wm geometry [winfo toplevel $w]]
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::wrap --
 #
 #       Arrange for all new scrollbars to be automatically autoscrolled
 #
 # Parameters:
 #       None.
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       ::scrollbar is overloaded to automatically autoscroll any new
 #          scrollbars.
 #
 #----------------------------------------------------------------------

proc ::autoscroll::wrap {} {
    if {[info commands ::autoscroll::_scrollbar] != ""} {return}
    rename ::scrollbar ::autoscroll::_scrollbar
    proc ::scrollbar {w args} {
        eval ::autoscroll::_scrollbar [list $w] $args
        ::autoscroll::autoscroll $w
        return $w
    }
}

 #----------------------------------------------------------------------
 #
 # ::autoscroll::unwrap --
 #
 #       Turns off automatic autoscrolling of new scrollbars. Does not
 #         effect existing scrollbars.
 #
 # Parameters:
 #       None.
 #
 # Results:
 #       None.
 #
 # Side effects:
 #       ::scrollbar is returned to its original state
 #
 #----------------------------------------------------------------------

proc ::autoscroll::unwrap {} {
    if {[info commands ::autoscroll::_scrollbar] == ""} {return}
    rename ::scrollbar {}
    rename ::autoscroll::_scrollbar ::scrollbar
}

# THATCH /herbie/packages/herbie-0.3.tm modbase /herbie/packages
#
# Herbie - The Love Debugger
#
#
# Stuart Cassoff
#
# Fall/Winter/Spring/Summer 2018
# Spring 2019
# Version 0.1
#
# Fall 2022
# Version 0.2
#
# Summer 2023
# Version 0.3
#

package require Tcl 8.6-

namespace eval herbie {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	herbie
		title	Herbie
		descr	{The Love Debugger}
		author	{Stuart Cassoff}
		when	{Summer 2023}
		version	0.3
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2019,2022,2023 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


#
# Init
#

package require tu::tcl

proc ignition {} {
	variable initdone
	if {[info exists initdone] && $initdone} { return }
	package require Tk
	package require tu::combola
	package require tu::epm
	package require tu::file
	package require tu::myfonts
	package require tu::pretty
	package require tu::toplevels::tk
	package require tu::tk
	package require gub
	package require meb
	package require tss
	package require struct::tree
	package require aitcs
	package require autoscroll
	set wm [::gub::gubTkWidgetMap]
#	variable widgetMap [dict merge $wm [dict create C [dict merge [dict get $wm S] [dict create command [namespace current]::autoscrollbar]]]]
	variable widgetMap [dict merge [::gub::gubTkWidgetMap] [::gub::gubAutoscrollbarWidgetMap c]]
	dict set widgetMap Y [dict replace [dict get $widgetMap X] command [list ::tu::combola::combola combobox]]
	::aitcs::aitcs new luv
	#luv {*}[setupFonts]
	luv fonts [::tu::myfonts::MyFonts new Herbie]
	luv fontScale 0
	::aitcs::aitcs new running
	set initdone 1
}


#
# Fonts
#
if 0 {
proc setupFonts {} {
	foreach f {Default Text Fixed Menu Heading} {
		set c [font configure Tk${f}Font]
		font create Herbie${f}Font {*}$c
	}
	font create HerbieBoldFixedFont {*}[font configure HerbieFixedFont] -weight bold
	font create HerbieBigDefaultFont {*}[font configure HerbieDefaultFont] -size -18

	set l [list]
	set d [dict create]
	foreach f {Default Text Fixed Menu Heading BoldFixed BigDefault} {
		set ff Herbie${f}Font
		dict set d $ff [font configure $ff]
		lappend l $ff
	}

	option add *Herbie*Menu*font HerbieMenuFont
#	option add *Herbie*Text*font HerbieFixedFont
	option add *Herbie*TLabel*font HerbieDefaultFont
	option add *Herbie*TEntry*font HerbieTextFont
	option add *Herbie*TCombobox*font HerbieTextFont

	return [dict create fonts $l fontsDefaults $d]
}
}
if 0 {
proc zresetFonts {} {
	dict for {f c} [luv fontsDefaults] {
		font configure $f {*}$c
	}
}

proc zscaleFonts {{s {0}}} {
	dict for {f c} [luv fontsDefaults] {
		set size [dict get $c -size]
		if {$size < 0} {
			set i [expr {-1 * $s}]
		} else {
			set i $s
		}
		incr size $i
		font configure $f -size $size
	}
	luv fontScale $s
}
}

proc resetFonts {} {
	[luv fonts] reset
}

proc scaleFonts {{s {0}}} {
	luv fontScale $s
	[luv fonts] transform-scale [luv fontScale]

	return

	ttk::style configure Treeview -rowheight \
		[expr {int(abs([font configure [ttk::style configure Treeview -font] -size]) * [tk scaling])}]
}


#
# Herbie Gui Bits
#

oo::class create herbieGuiBits {
method textAppend {args} {
	[my wText] insert end {*}$args
	return [self]
}
method textSee {where} {
	[my wText] see $where
	return [self]
}
}; # oo::class create herbieGuiBits


#
# Herbie Misc Bits
#

oo::class create herbieMiscBits {
method tidyTextBlob {blob} {
	return [subst [regsub -line -all {^\t} [string range $blob 1 end-1] {}]]
}
if 0 {
method logColours {} { return [dict create {*}{
	alert     #FF0001
	critical  #FF0002
	debug     #B0C4DE
	emergency #FF0003
	error     #FF0004
	info      #0000FF
	notice    #2E8B57
	warning   #FF4505
}]}
}
}; # oo::class create herbieMiscBits



#
# Workspace
#

oo::class create Workspace {
	mixin herbieGuiBits herbieMiscBits
	variable tss
	variable wordWrap
	variable font
	variable wCombo wLabel wText
	variable dbusses
	variable socks
	variable vblTk vblComm vblDBus vlbSock
	variable wW
	variable wM
	variable filetypes
	variable glovs
constructor {args} {
	::herbie::ignition
	namespace import ::gub::gub
	namespace import ::meb::meb
	namespace import ::tu::toplevels::tk::randWindow ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::onToplevelDelete
	namespace import ::tu::epm::epmW ::tu::epm::epmCreateAnEditMenu ::tu::epm::epmBindMenuTo
	namespace import ::tu::pretty::hexdump
	set wordWrap 1
	set font Fixed
	set filetypes [dict create tcl [list "Tcl Program Files" [list .tcl]] txt [list "Text Files" [list .txt .TXT]]]
	set glovs [list]
	set dbusses [list]
	set socks [list]
	set vblTk   0
	set vblComm 0
	set vblDBus 0
	set vblSoc  0
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
destructor {
	if {[info exists wW]} { destroy $wW }
	if {[info exists wM]} { destroy $wM }
}
method openNew {args} { [info object class [self]] new :: {} {} {*}$args }
method done {} { my destroy }
eval [::tu::tcl::makeROAccessors wW wCombo wLabel wText tss]
method eval {args} { tailcall $tss eval [join $args] }
method openOnIn {where {w {}} {aTss {}} args} {
	set help yes
	set menu yes
	set focus ""
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-help   -
			-menu   -
			-focus  { set state $arg }
			default { break }
			}; # hctiws
		}
		-help  { set state opt; set help  $arg }
		-menu  { set state opt; set menu  $arg }
		-focus { set state opt; set focus $arg }
		}; # hctiws
	}
	set tss [expr {$aTss ne {} ? $aTss : [::tss::Tss new]}]
	$tss cd $where
	$tss env LOPTS "nvapco"
	$tss env FLOPTS "l"
	if {$w eq {}} {
		set w [createToplevelGui "Workspace" -class Herbie]
		eval [gub "-.;|." $w]
		eval [gub "&. Control-q [self] done" $w]
		#bind $w <Control-q> [list [self] done]
	}
	set gob "
		;F t C
		;  C
		;F Y L
		@Y0 combo
		.combo -keyselect listbox -upis down
		.t0 -undo yes -wrap word -tabstyle wordprocessor -width 100 -font Herbie${font}Font
		:F0,t0
		:F1,combo -sticky ew
		#:L0 -sticky w
		-F0,1 ; |F0,1
		&combo Return [self] doCmd
		&combo <<ComboboxSelected>> event generate /combo/ <<LineEnd>>
		&combo Key-Tab break
		&combo Control-Key-Tab continue
		/t0 wText
		/combo wCombo
		/L0 wLabel
		/. wW
	"
	set gui [dict create]
	eval [gub $gob $w -wm $::herbie::widgetMap]
	onToplevelDelete $w [list [self] done]
	foreach v [list wW wCombo wLabel wText] {
		set $v [dict get $gui $v]
	}
	#$wText tag configure found -font HerbieBoldFixedFont
	$wText tag configure found -background red -foreground white
	foreach ww [list $wText $wCombo] {
		#bind $ww <Alt-b> [list [self] browseItKey %W]
		#bind $ww <Alt-B> [list Browser new]
		bind $ww <Alt-c> [list [self] colourTuneItKey %W]
		bind $ww <Alt-d> [list [self] doItKey %W]
		bind $ww <Alt-i> [list [self] inspectItKey %W]
		bind $ww <Alt-I> [list ::herbie::Inspector new $where]
		bind $ww <Alt-m> [list [self] multiWatchItKey %W]
		bind $ww <Alt-p> [list [self] printItKey %W]
		bind $ww <Alt-P> [list [self] printItKey %W 1]
		bind $ww <Alt-u> [list [self] scriptButtonItKey %W]
		bind $ww <Alt-w> [list [self] watchItKey %W]
		bind $ww <Alt-W> [list [self] openNew -focus combo]
	}
	if {$menu} {
		oo::objdefine [self] method wM {} { variable wM; return $wM }
		oo::objdefine [self] mixin -append ::herbie::HerbieMenu
		lassign [my createWorkspaceMenu] wM am
		epmBindMenuTo $wText  $wM
		epmBindMenuTo $wText  $wM -mod Alt- -seq Key- -bn backslash
		epmBindMenuTo $wCombo $wM -mod Alt- -seq Key- -bn backslash
		epmBindMenuTo $wLabel $am -bn 1
	}
	if {$help} {
		oo::objdefine [self] mixin -append ::herbie::HerbieHelp
	}
	oo::objdefine [self] mixin -append ::herbie::HerbieCommands
	my updateLabel
	if 0 {
	dict for {lv colour} [my logColours] {
		$wText tag configure $lv -foreground $colour
	}
	}
	if {$focus eq "combo"} {
		focus $wCombo
	} elseif {$focus eq "text"} {
		focus $wText
	}
}

method scaleFonts {{s {0}}} {
	::herbie::scaleFonts $s
}

method setFont {f} {
	$wText configure -font Herbie${f}Font
}

method setww {} {
	[my wText] configure -wrap [expr {$wordWrap ? "word" : "none"}]
}

method setvblTk {} {
}
method setvblComm {} {
}
method setvblDBus {} {
	if {$vblDBus} {
		dbif connect org.tcllang.debug
		dbif method / Eval {script} {res:s} {uplevel #0 $script}
	}
}
method setvblSock {} {
}

method doItKey           {w} { epmW $w; my doIt }
method inspectItKey      {w} { epmW $w; my inspectIt }
method browseItKey       {w} { epmW $w; my browseIt }
method watchItKey        {w} { epmW $w; my watchIt }
method multiWatchItKey   {w} { epmW $w; my multiWatchIt }
method printItKey        {w {toNew 0}} { epmW $w; my printIt $toNew }
method scriptButtonItKey {w} { epmW $w; my scriptButtonIt }
method colourTuneItKey   {w} { epmW $w; my colourTuneIt }

method doIt {} {
	if {[set blob [my getIt [epmW]]] eq ""} { return "" }
	return [my doBlob $blob]
}
method inspectIt {} {
	if {[set blob [my getIt [epmW]]] eq ""} { return "" }
	return [::herbie::Inspector new $blob]
}
method browseIt {} {
	if {[set blob [my getIt [epmW]]] eq ""} { return "" }
	return [::herbie::Browser new $blob]
}
method watchIt {} {
	if {[set blob [my getIt [epmW]]] eq ""} { return "" }
	return [::herbie::VarWatcher new $blob {} $tss]
}
method initMvw {} {
	if {[::herbie::running? mvw] && [llength [info commands [::herbie::running mvw]]] == 0} { ::herbie::running- mvw }
	if {![::herbie::running? mvw]} {
		::herbie::running mvw [::herbie::MultiVarWatcher new {} {} $tss]
	}
	return [::herbie::running mvw]
}
method multiWatchIt {} {
	if {[set blob [my getIt [epmW]]] eq ""} { return "" }
	raise [[[my initMvw] add $blob] wW]
}
method printIt {{toNew {0}}} {
	set res [my doIt]
	if {$toNew} {
		return [[[[self class] new ::] textAppend $res] textSee end]
	}
	set w [epmW]
	set idx1 [$w index insert]
	if {$w eq $wText} {
		$w tag remove sel 1.0 end
		$w insert insert $res sel
		$w mark set insert $idx1
	} elseif {$w eq $wCombo} {
		set idx2 [expr {$idx1 + [string length $res]}]
		$w insert insert $res
		$w selection range $idx1 $idx2
	}
}
method scriptButtonIt {} {
	set blob [my getIt [epmW]]
	return [::herbie::ScriptButton new Name $blob]
}
method colourTuneIt {} {
	set blob [my getIt [epmW]]
	if {[catch {$blob cget -background}] res} {puts $res; return}
	return [[::herbie::ColourTuner new $blob "$blob configure -background"] openOnIn [$blob cget -background]]
}
method tkOptionEditIt {} {
	if {[set blob [my getIt [epmW]]] eq ""} { return "" }
	return [[[::herbie::tkote::Tkote new] openOnIn] inspect $blob]
}
method doCmd {} {
	set blob [string trim [$wCombo get]]
	$wCombo set $blob
	set res [my doBlob $blob]
	if {$blob ne ""} {
		set values [$wCombo cget -values]
                set vi [lsearch -exact $values $blob]
		if {$vi == -1} {
			set vals [list $blob {*}$values]
		} elseif {$vi > 0} {
			set vals [list $blob {*}[lrange $values 0 $vi-1] {*}[lrange $values $vi+1 end]]
		}
		if {$vi != 0} { $wCombo configure -values $vals }
	}
	$wCombo set ""
	if {$blob eq "clear"} return
	if {[string range $blob 0 1] eq "? "} return
	my textAppend "\n\n"
	if {[string index $blob 0] ne "#"} {
		my textAppend [my poundSpace]
	}
	my textAppend $blob {} \n {} $res {}
	my textSee end
#	focus $wCombo
}
method poundSpace {} {
	return "# "
}
method getIt {w} {
	if {[winfo class $w] eq "Text"} {
		set res [my getItFromText $w]
	} else {
		set res [my getItFromCombo $w]
	}
	return [string trim $res]
}
method getItFromCombo {w} {
	if {![$w selection present]} {
		$w selection range 0 end
	}
	if {![$w selection present]} {
		return ""
	}
	return [string trim [string range [$w get] [$w index sel.first] [$w index sel.last]-1]]
}
method getItFromText {w} {
	if {[llength [set sel [$w tag ranges sel]]] != 2} {
		event generate $w <<LineStart>>
		event generate $w <<SelectLineEnd>>
	}
	if {[llength [set sel [$w tag ranges sel]]] != 2} {
		return ""
	}
	return [$w get {*}$sel]
}
method doBlob {blob} {
	set blob [string trim $blob]
	if {[string index $blob 0] eq "\\"} {
		set blob [string range $blob 1 end]
		set specialsok 0
	} else {
		set specialsok 1
	}
	if {$blob eq ""} { return "" }
	if {$specialsok} {
		set c [string index $blob 0]
		if {$c eq ":" && [string index $blob 1] ne ":"} {
			set res [my {*}[string range $blob 1 end]]
		} elseif {$c eq "~" && [string index $blob 1] eq "~"} {
			catch {$tss {*}[string range $blob 2 end]} res
		} elseif {$c eq "~"} {
			catch {[self] {*}[string range $blob 1 end]} res
		} elseif {$c eq "/"} {
			set blob [string range $blob 1 end]
			if {$blob eq ""} {
				set idxs [list]
			} else {
				set idxs [$wText search -regexp -all $blob 1.0]
			}
			if {$blob eq "" || [llength $idxs] > 0} {
				set ranges [$wText tag ranges found]
				if {[llength $ranges] > 0} {
					$wText tag remove found {*}$ranges
				}
			}
			if {[llength $idxs] > 0} {
				set len [string length $blob]
				foreach idx $idxs {
					$wText tag add found $idx $idx+${len}c
				}
			}
			lassign [$tss env HIGHLIGHT] fg bg
			if {$fg ne {} && $bg ne {}} {
				$wText tag configure found -foreground $fg -background $bg
			}
			set res [expr {$blob eq "" ? "" : [llength $idxs]}]
		} else {
			set i [string first " " $blob]
			if {$i == -1} { set i [string length $blob] }
			set cmd [string range $blob 0 $i-1]
			set rest [string trim [string range $blob $i+1 end]]
			if {$cmd eq "l"} {
				set lopts [$tss env LOPTS]
				if {$lopts eq ""} {
					set res [$tss ls {*}$rest]
				} elseif {[string index $rest 0] eq "-"} {
					set ii [string first " " $rest]
					if {$ii == -1} { set ii [string length $rest] }
					set res [$tss ls -$lopts[string range $rest 1 $ii] {*}[string range $rest $ii+1 end]]
				} else {
$tss env cow "$tss ls -$lopts !$rest!"
					set res [$tss ls -$lopts $rest]
				}
			} elseif {$cmd eq "fl"} {
				set lopts [$tss env FLOPTS]
				if {$lopts eq ""} {
					set res [my fls {*}$rest]
				} elseif {[string index $rest 0] eq "-"} {
					set ii [string first " " $rest]
					if {$ii == -1} { set ii [string length $rest] }
					set res [my fls -$lopts[string range $rest 1 $ii] {*}[string range $rest $ii+1 end]]
				} else {
$tss env fcow "$tss fls -$lopts !$rest!"
					set res [my fls -$lopts $rest]
				}

			} elseif {$cmd eq "cd"} {
				if {$rest eq ""} { set rest [$tss env HOME] }
				set res [$tss $cmd {*}$rest]
			} elseif {[string match {[pvadwn]cat} $cmd]} {
				set rest [list -[string index $cmd 0] {*}$rest]
				set cmd cat
				set res [$tss $cmd {*}$rest]
			} elseif {$cmd in [list "rls" "fls" "fcat" "fput" "fcd" "grep" "glop" "find"]} {
				set res [my $cmd {*}$rest]
			} elseif {$cmd eq "catset"} {
				set cmd cat
				set res "set [lindex $rest end] {[$tss $cmd {*}$rest]}"
			} elseif {$cmd eq "clear"} {
				$wText delete 0.0 end
				set res ""
			} elseif {$cmd eq "?"} {
				set level info
				set ii -1
				if {[string index $rest 0] eq "-"} {
					set ii [string first " " $rest]
					if {$ii > 1} {
						set level [string range $rest 1 $ii-1]
						set rest [string range $rest $ii+1 end]
					}
				}
				my checkGlovs
				if {[llength $glovs] > 0} {
					[lindex $glovs 0] log $level $rest
				}
				set res ""
			} elseif {$cmd eq "help" && $cmd in [set mymethods [info object methods [self] -all]]} {
				if {$rest eq ""} {
					set res [my help]
				} else {
					set hcmd help_$rest
					if {$hcmd in $mymethods} {
						set res [my $hcmd]
					} else {
						set res "No help for \"$rest\""
					}
				}
			} elseif {$cmd in [::tss::tsscmds]} {
				set res [$tss $cmd {*}$rest]
			}
		}
	}
	if {![info exists res]} {
		catch {$tss eval $blob} res
	}
	my updateLabel
	return $res
}
method loadFileIntoText {{fn ""} args} {
	if {[llength $args] == 0} {
		 lappend args -filetypes [list [list [dict get $filetypes tcl] [dict get $filetypes txt]]]
	}
	::tu::file::loadFileIntoText $wText $fn {*}$args
}
method updateLabel {} {
	$wLabel configure -text [format "%s%s %s" [$tss app] [$tss interp] [$tss pwd]]
}
method interpLabel {} {
	return ""
$wLabel configure -text [format "%s%s %s" [expr {"app" in [info object methods $tss -all] ? "[$tss app] " : ""}] [$tss interp] [$tss pwd]]
	$wLabel configure -text [format "%s%s %s" [expr {"app" in [info object methods $tss -all] ? "[$tss app] " : ""}] 
[$tss interp] [$tss pwd]
]
}
method openAGlov {} {
	lappend glovs [::herbie::GlovFull new {}]
	my checkGlovs
}
method checkGlovs {} {
	set glovs [lmap glov $glovs { if {[llength [info commands $glov]] != 1} continue; set glov }]
}
method commOn {{app {}}} {
	package require comm
	if {$app eq {}} return
	set a [split $app ":"]
	if {[llength $a] == 2} {
		set app [lreverse $a]
	}
	return [::comm::comm send $app info tcl]
}
method dbusOn {{app {}}} {
	package require dbif
	if {$app eq {}} return
	if {$app in $dbusses} return
	lappend dbusses $app
	return ""
}
method sockOn {{app {}}} {
	if {$app eq {}} return
	if {$app in $socks} return
	lappend socks $app
	return ""
}
}; # oo::class create Workspace



#
# Herbie Commands
#

oo::class create HerbieCommands {
method rls {args} {
	if {[llength $args] == 0} { return [my doBlob ls] }
	set regexps [lassign $args maybeopts]
	if {[string index $maybeopts 0] eq "-"} {
		set opts $maybeopts
	} else {
		set opts ""
		set regexps [linsert $regexps 0 $maybeopts]
	}
	if {[llength $regexps] == 0} { return [my doBlob "ls $opts"] }
	return [join [lmap line [split [my doBlob "ls $opts *"] \n] {
		if {![::tcl::mathop::+ {*}[lmap regexp $regexps { regexp $regexp [lindex $line 1] }]]} { continue }
		set line
	}] \n]
}
method fls {args} {
	set long no
	set res [list]
	lassign $args opts pat
	if {$opts ne ""} {
		if {[string index $opts 0] eq "-"} {
			foreach c [split [string range $opts 1 end] ""] {
				if {$c eq "l"} {
					set long yes
				}
			}
		} elseif {$pat eq ""} {
			set pat $opts
		}
	}
	if {$pat eq {}} { set pat {*} }
	set fl [my eval glob -nocomplain $pat]
	if {$long} { set res [lmap fn $fl {
		array set stat [my eval "apply {{} {file lstat {$fn} stat; array get stat}}"]
		set mode $stat(mode)
		set smode [format "%s%s"  [expr {[string index $stat(type) 0] eq "d" ? "d" : "-"}]  [join [lmap {mask char} {256 r 128 w 64 x 32 r 16 w 8 x 4 r 2 w 1 x} { expr {$mask & $mode ? $char : "-"} }] ""]  ]
		format {%10s %8d %s %s} $smode $stat(size) [clock format $stat(mtime) -format "%Y/%m/%d %H:%M:%S"] $fn
	}]} else { set res $fl }
	return [join $res \n]
}
method fcat {args} {
	set gunzip  no
	set binary  no
	set utf8    no
	set zutf8   no
	set base64  no
	set hexdump no
	set maxlen   0
	set hexlen  16
	lassign $args opts fn
	if {$opts ne ""} {
		if {[string index $opts 0] eq "-"} { foreach c [split [string range $opts 1 end] ""] {
			      if {$c eq "z"} { set gunzip yes
			} elseif {$c eq "b"} { set binary yes ; set utf8     no
			} elseif {$c eq "u"} { set binary  no ; set utf8    yes
			} elseif {$c eq "a"} { set base64 yes ; set hexdump  no ; set maxlen 68
			} elseif {$c eq "A"} { set base64 yes ; set hexdump  no ; set maxlen  0
			} elseif {$c eq "h"} { set base64  no ; set hexdump yes ; set hexlen 16
			} elseif {$c eq "H"} { set base64  no ; set hexdump yes ; set hexlen 32
			} elseif {$c eq "8"} { set base64  no ; set hexdump yes ; set hexlen  8
		}}} elseif {$fn eq ""} {
			set fn $opts
		}
	}
	if {$fn eq ""} return
	if {$gunzip} {
		set zutf8 $utf8
		set utf8     no
		set binary  yes
	}
	set chanopts [list]
	if {$binary} { lappend chanopts -translation binary }
	if {$utf8}   { lappend chanopts -encoding    utf8   }
	set nnl [expr {$binary ? "" : "-nonewline"}]

	set scr [format {
		apply {{fn nnl chanopts} {
			set f [open $fn r]
			if {[llength $chanopts] > 0} {
				chan configure $f {*}$chanopts
			}
			set blob [read {*}$nnl $f]
			close $f
			set blob
		}} {%s} {%s} {%s}
	} $fn $nnl $chanopts]

	set blob [my eval $scr]

	if {$gunzip} {
		set blob [zlib gunzip $blob]
	}
	if {$zutf8} {
		set blob [encoding convertfrom utf-8 $blob]
	}
	return [if {$base64} {
		binary encode base64 -maxlen $maxlen $blob
	} elseif {$hexdump} {
		join [hexdump $blob $hexlen] \n
	} else {
		set blob
	}]
}
method fcd  {args} { my eval cd {*}$args }
method frm  {args} { if {[llengtth $args] > 0} { my eval file delete {*}$args } }
method fpwd     {} { my eval pwd }
method fput {fn blob} {
	set binary no
	set chanopts [list]
	set putsopts [expr {$binary ? "" : "-nonewline"}]
	set scr [format {
		apply {{fn putsopts chanopts blob} {
			set f [open $fn w]
			if {[llength $chanopts] > 0} {
				chan configure $f {*}$chanopts
			}
			puts {*}$putsopts $f $blob
			close $f
		}} {%s} {%s} {%s} {%s}
	} $fn $putsopts $chanopts $blob]
	my eval $scr
}
method grep {re {tpat {*}}} {
	join [lmap {t p} [[my tss] ls -p $tpat] {
		set blob [[my tss] evalIn [[my tss] pwd] [list info body $p]]
		if {![regexp $re $blob]} continue
		[my tss] cat $p
	}] \n
}
method glop {gpat {tpat {*}}} {
	join [lmap {t p} [[my tss] ls -p $tpat] {
		set blob [[my tss] evalIn [[my tss] pwd] [list info body $p]]
		if {![string match $gpat $blob]} continue
		[my tss] cat -p $p
	}] \n
}
method Pfind {ns pat} {
	set res [list]
	set l [[my tss] ls -p $ns $pat]
#puts "[my tss] ls -p $ns $pat -> !$l!"
	lappend res {*}[lmap p [split $l \n] { string cat $ns :: [lindex $p 1] }]
	foreach nsi [split [[my tss] ls -n $ns] \n] {
		lassign $nsi _ ns
		if {$ns in [list "oo" "tcl"]} continue
		[my tss] cd $ns
		lappend res {*}[my Pfind $ns $pat]
		[my tss] cd ..
	}
	return $res
}
method Bfind {w tag} {
	set res [list]
	if {![my eval winfo exists $w]} {
		if {[lsearch [my eval bind $w] $tag] != -1} {
			lappend res $w
		}
		return $res
	}
	foreach t [my eval bindtags $w] {
		if {[lsearch [my eval bind $t] $tag] != -1} {
			lappend res $t
		}
	}
	foreach c [my eval winfo children $w] { lappend res {*}[my Bfind $c $tag] }
	return $res
}
method find {args} {
	if {[llength $args] == 0} return
	if {[llength $args] > 1 && [lindex $args 0] in [list "-p" "-b"]} {
		set args [lassign $args type]
		set type [string range $type 1 end]
	} else {
		if {[string index [lindex $args 0] 0] eq "<"} {
			set type b
		} else {
			set type p
		}
	}
	return [join [my [string toupper $type]find {*}$args] \n]
}
}; # oo::class create HerbieCommands



#
# Herbie Menu
#

oo::class create HerbieMenu {
	variable filetypes ;# from Workspace
	variable tkInterpMenu
	variable commInterpMenu
	variable dbusInterpMenu
	variable sockInterpMenu
method createWorkspaceMenu {} {
	set wM [randWindow .workspace_]
	set mob "
		.&Do It				-command {[self] doIt}
		.&Print It			-command {[self] printIt}
		.&Inspect It			-command {[self] inspectIt}
#		.&Browse it			-command {[self] browseIt}
		.&Watch It			-command {[self] watchIt}
		.&MultiWatch It			-command {[self] multiWatchIt}
		.ScriptB&utton It		-command {[self] scriptButtonIt}
		.&ColourTune It			-command {[self] colourTuneIt}
		.&TkOptionEdit It		-command {[self] tkOptionEditIt}
#		.P&rint it to new Workspace	-command {[self] printIt 1}
		.&Open				+menuopts {-font HerbieMenuFont}
		. &Workspace			-command {[self] openNew -focus combo}
		. &Inspector			-command {::herbie::Inspector new :: {} {} -focus combo}
#		. Browser			-command {::Browser new}
		. &Errors			-command {::herbie::ErrorCodeVarWatcher new ::errorInfo}
		. &VarWatcher			-command {\[::herbie::VarWatcher new\] openOnIn {}}
		. &MultiVarWatcher		-command {raise \[\[[self] initMvw\] wW\]}
		. ScriptB&utton			-command {::herbie::ScriptButton new}
		. &Styler			-command {\[::herbie::Styler new\] openOnIn}
		. &ColourTuner			-command {\[::herbie::ColourTuner new\] openOnIn}
		. &Glov				-command {[self] openAGlov}
		. &Tk option editor		-command {\[::herbie::tkote::Tkote new\] openOnIn}
		.I&nterp			+menuopts {-font HerbieMenuFont}
		. &Tk				+menuopts {-postcommand {[self] doTkInterpMenu} -font HerbieMenuFont}
		.  tk   interps go here		+save tkinterps +type cascade -menu {{}}
		. &Comm				+menuopts {-postcommand {[self] doCommInterpMenu} -font HerbieMenuFont}
		.  comm interps go here		+save comminterps +type cascade -menu {{}}
		. &DBus				+menuopts {-postcommand {[self] doDBusInterpMenu} -font HerbieMenuFont}
		.  dbus interps go here		+save dbusinterps +type cascade -menu {{}}
		. &Sock				+menuopts {-postcommand {[self] doSockInterpMenu} -font HerbieMenuFont}
		.  sock interps go here		+save sockinterps +type cascade -menu {{}}
		.&Edit				-menu {{}} +type cascade +save edit
		.&Save Selection		-command {::tu::file::saveSelectionToFile \[::tu::epm::epmW\] ? -filetypes [list [list [dict get $filetypes tcl] [dict get $filetypes txt]]]}
		.&Load Something		-command {::tu::file::loadFileIntoText \[::tu::epm::epmW\] ? -filetypes [list [list [dict get $filetypes tcl] [dict get $filetypes txt]]]}
		.&Appearance			+menuopts {-font HerbieMenuFont} +save appearance
		. &Word Wrap			-command {[self] setww} -variable [my varname wordWrap]
		. &Font				+menuopts {-font HerbieMenuFont}
		[join [lmap _ [list Default Fixed] {
			format ".  %s -value %s -variable [my varname font] -command {[self] setFont %s}" $_ $_ $_
		}] \n]
		. Font &Scale
		[subst [join [lmap _ [list -3 -2 -1 0 1 2 3 4 5] {
			format {.  %d -command {[self] scaleFonts %d} -variable [::herbie::luv& fontScale]} $_ $_
		}] \n]]
		.&Visibility
		. &Tk				-command {[self] setvblTk  } -variable [my varname vblTk]
		. &Comm				-command {[self] setvblComm} -variable [my varname vblComm]
		. &DBus				-command {[self] setvblDBus} -variable [my varname vblDBus]
		. &Sock				-command {[self] setvblSock} -variable [my varname vblSock]
		.Help				-command {[self] textAppend \[[self] poundSpace\]help\\n\[[self] doBlob help\]}
		.Close				-command {::event generate \[::tu::epm::epmW\] <Control-q>}
	"

	set moob [meb -menuname $wM -menuopts {-postcommand ::tu::epm::epmAdjustMenu -font HerbieMenuFont} -varname plop $mob]
	eval [dict get $moob tcl src]

	set   tkInterpMenu [dict get $plop   tkinterps in]
	set commInterpMenu [dict get $plop comminterps in]
	set dbusInterpMenu [dict get $plop dbusinterps in]
	set sockInterpMenu [dict get $plop sockinterps in]

	set editCascade [dict get $plop edit]
	lassign [epmCreateAnEditMenu $wM -font HerbieMenuFont] em emmap
	[dict get $editCascade in] entryconfigure [dict get $editCascade index] -menu $em

	return [list $wM [dict get $plop appearance cascade]]
}
method doTkInterpMenu {} {
	my clearAnInterpMenu $tkInterpMenu
	my makeAnInterpMenu $tkInterpMenu Tk [winfo interps]
}
method doCommInterpMenu {} {
	my clearAnInterpMenu $commInterpMenu
	if {[package provide comm] eq {}} return
	my makeAnInterpMenu $commInterpMenu Comm [::comm::comm interps]
}
method doDBusInterpMenu {} {
	my clearAnInterpMenu $dbusInterpMenu
	if {[package provide dbif] eq {}} return
	if {[llength $dbusses] == 0} return
	my makeAnInterpMenu $dbusInterpMenu DBus $dbusses
}
method doSockInterpMenu {} {
	my clearAnInterpMenu $sockInterpMenu
	if {[llength $socks] == 0} return
	my makeAnInterpMenu $sockInterpMenu Socket $socks
}
method clearAnInterpMenu {menu} {
	while {[set i [$menu index end]] ne "none"} {
		set m [$menu entrycget $i -menu]
		$menu delete $i
		destroy $m
	}
}
method makeAnInterpMenu {menu type ids} {
	set i -1
	set mob ""
	foreach id $ids {
		append mob "
			.&[incr i] $id  +menuopts {-font HerbieMenuFont}
			. &Workspace    -command {::herbie::Workspace new :: {} \[::tss::Tss$type new {$id}\] -focus combo}
			. &Inspector    -command {::herbie::Inspector new :: {} \[::tss::Tss$type new {$id}\] -focus tree}
		"
	}
	set moob [meb -nomenu -menuname $menu [string map {# ``} $mob]]
	eval [string map {`` #} [dict get $moob tcl src]]
}
}; # oo::class create HerbieMenu



#
# Herbie Help
#

oo::class create HerbieHelp {
method help {} { return [my tidyTextBlob {
	Commands:
	  [::tss::tsscmds]
	  catset rls fls fcat fput fcd fpwd l fl
	  grep glop find clear / ~ ~~ help

	Use \\ to escape commands.

	Mouse:
	  Button 3 in text area for menu.
	  Button 1 on label for appearance menu.

	Keyboard:
	  Alt-d, Alt-D	Do It
	  Alt-p		Print It
	  Alt-P		Print It to new Workspace
	  Alt-W		Open new Workspace
	  Alt-I		Open new Inspector
	  Alt-u		ScriptButton It
	  Alt-c		ColourTune It
	  Alt-w		Watch It
	  Alt-m		Multiwatch It
	  Control-Tab	Switch between text area and combo
	  Control-q	Close window
	  Alt-q		Close initial window
	  Alt-\\		Menu

	Do, Print, ScriptButton, ColourTuner and Watch
	operate on the currently selected text.
	The current line is used if no text is selected.

	Other:
	  /text		Search for text
	  /		Clear search
	  ~		Access Herbie
	  ~~		Access Tss
	  help		This help
	  help <topic>	Help on topic
	  help topsum	Summary of available topics
	  help topics	Available topics
}]}
method help_help {} { return [my tidyTextBlob {
	Main help or help for a topic.

	Usage: help ?topic?
}]}
method topicslist {} {
	set o [list]
	set l [lmap h [lsearch -inline -all [info object methods [self] -all] help_*] { string range $h 5 end }]
	set l [lmap e [lsort $l] { if {$e in [list "/"]} { lappend o $e; continue }; set e }]
	lappend l {*}$o
}
method help_topics {} {
	set s [join [my topicslist]]
	set p 0
	while {[set i [string first " " $s [incr p 40]]] != -1} {
		set s [string cat [string range $s 0 $i] \n [string range $s $i+1 end]]
	}
	return $s
}
method help_topsum {} {
	set topics [my topicslist]
	set max [::tcl::mathfunc::max {*}[lmap t $topics { string length $t }]]
	join [lmap topic $topics {
		if {$topic in [list "topsum" "topics"]} continue
		format "%${max}s  %s" $topic [lindex [split [my help_$topic] \n] 0]
	}] \n
}
method help_about {} {
	package require central
	central::central discover
	set a [list]
	lappend a "About various things." {}
	set pn ::herbie
	set pi [central::central package $pn]
	lappend a [string cat [dict get $pi title] " - " [dict get $pi descr]]
	lappend a [string cat "Version " [dict get $pi version] " - " [dict get $pi when]]
	lappend a "By [dict get $pi author]" {}
	lappend a "Things:"
	set l [list]
	foreach pn [central::central packages] {
		if {$pn eq "::herbie"} continue
		set pi [central::central package $pn]
		set name    [expr {[dict exists $pi title  ] ? [dict get $pi title  ] : [string trimleft $pn ":"]}]
		set version [expr {[dict exists $pi version] ? [dict get $pi version] : {}}]
		lappend l [list $name $version]
	}
	lappend a [::herbie::tblfmt $l]
	return [join $a \n]
}
method help_clear {} { return [my tidyTextBlob {
	Delete the contents of the workspace.
}]}
method help_pwd {} { return [my tidyTextBlob {
	Print current namespace.
}]}
method help_cd {} { return [my tidyTextBlob {
	Change current namespace.

	With no arguments, changes the current namespace
	to the HOME env entry if it exists, otherwise
	changes the current namespace to ::.

	Namespaces may be specicied as namespaces or
	like file paths, using / instead of ::.

	The characters .. can be used to specify
	a parent namespace.

	Prints the final namespace.
}]}
method help_cat {} { return [my tidyTextBlob {
	Display the contents of a proc, var, object, method, window, binding or namespace ensemble.

	Usage: cat ?-type? thing
	Or:    <type>cat thing
	Also: cat Obj::>::Method

	If more than one type of thing exists
	(ex. a proc and a var) then all will be displayed.

	The type of thing is auto-detected except for dicts.

	Use one of the following type chars to
	filter for one type of thing to display.
	  p Proc
	  v Variable
	  a Array
	  d Dict
	  w Window
	  b Binding
	  n Namespace ensemble

	To cat a method of an object,
	join the object and method names together
	with the string "::>::".
}]}
method help_catsort {} { return [my tidyTextBlob {
	Sorted cat.
}]}
method help_catset {} { return [my tidyTextBlob {
	Wrap the output of cat in a set statement.
}]}
method help_ls {} { return [my tidyTextBlob {
	List the contents of one or more namespaces.

	Usage: ls ?-filterkeys? ?pattern ...?

	If the first argument to ls begins with a "-"
	then the remainder of that argument will be
	interpreted as a string of filter keys.

	Namespace is current namespace unless pat
	contains namespace separators.

	Contents are listed in two columns: type, name.
	Globals are capitalized.

	The filterkeys string is composed of characters
	found in the first column of "help types".

	Example: List all the things that begin with "t",
	but only global arrays and procedures in the
	current namespace.

	ls -Ap t*
}]}
method help_l {} { return [my tidyTextBlob {
	Alias of ls.

	Run the ls command using the LOPTS env entry for the filterkeys.
}]}
method help_rls {} { return [my tidyTextBlob {
	Regexp ls.

	Like ls but with a regular expression instead of a glob pattern.
}]}
method help_grep {} { return [my tidyTextBlob {
	Match proc bodies with regular expressions.

	Usage: grep re ?tpat?

	Search all procs whose names match ?tpat? (default: *)
	and cat any whose bodies contain a match
	for the regular expression re.
}]}
method help_glop {} { return [my tidyTextBlob {
	Match proc bodies with glob patterns.

	Usage: glop gpat ?tpat?

	Like grep but with a glob pattern instead of a regular expression.
}]}
method help_find {} { return [my tidyTextBlob {
	Find things.

	Usage: find ?pat?
}]}
method help_fls {} { return [my tidyTextBlob {
	Display a list of files.

	Usage: fls ?-options? ?pattern?

	Options:
		l Long listing
}]}
method help_fl {} { return [my tidyTextBlob {
	Alias of fls.

	Run the fls command using the FLOPTS env entry for options.
}]}
method help_fpwd {} { return [my tidyTextBlob {
	Return working directory.

	Usage: fpwd
}]}
method help_fcd {} { return [my tidyTextBlob {
	Change working directory.

	Usage: fcd ?dirname?
}]}
method help_fcat {} { return [my tidyTextBlob {
	Display the contents of a file.

	Usage: fcat ?-options? filename

	Options:
		b Read the file in binary mode (disables utf-8)
		u Read the file in utf-8 mode (disables binary)
		z File is gzipped
		a Display as base64 wrapped at 68 chars
		A Display as base64 not wrapped
		h Display as 16-wide hexdump
		H Display as 32-wide hexdump
		8 Display as  8-wide hexdump

	If z then the file will be read in binary mode.
	If z and u then the gunzipped data will be utf-8 decoded.

	If none of b, u, or z then the file is
	read with default configuration.
}]}
method help_fput {} { return [my tidyTextBlob {
	Write data to a file.

	Usage: fput filename data

	The file is written with default configuration.
}]}
method help_env {} { return [my tidyTextBlob {
	Manipulate env.

	With no arguments, prints env.
	With one argument, prints value of entry.
	With two arguments, sets entry and prints value.
}]}
method help_env- {} { return [my tidyTextBlob {
	Remove entry from env.
}]}
method help_/ {} { return [my tidyTextBlob {
	Searches for text in the text area.

	Usage: /?text?

	The number of matches is printed and
	all matching text is highlighted.

	If text isn't supplied then the search is cleared.

	Whitespace is allowed between the "/" and the search text.

	The default highlight is {fg bg}: {white red},
	and can be changed via the HIGHLIGHT env entry.
	Example: env HIGHLIGHT {blue yellow}
}]}
method help_types {} {
	return Tss\ Types.\n\n[::tss::twocols [dict keys [::tss::tsstypes]] [dict values [::tss::tsstypes]]]
}
method help_comm {} { return [my tidyTextBlob {
	Connecting to an interp via Comm.

	Comm is not loaded by default.

	To connect to an interpreter running comm, type
	    ~commOn <comm interp>

	and if successful, the Tcl version of the comm interp will be printed
	and the comm interp will appear in the Interp->Comm menu.

	The <comm interp> may be specified as a comm interp id,
	or as host:port.

	To simply load the comm package, type
	    ~commOn
}]}
method help_dbus {} { return [my tidyTextBlob {
	Connecting to an interp via the DBus.

	DBus is not loaded by default.

	To connect to an interpreter listening on the DBus, type
	    ~dbusOn <app>

	To simply load the DBus package, type
	    ~dbusOn

	Schelte sez:
	To be able to discover debuggable applications,	they should emit
	an Appname signal on startup and upon a Discover signal.
}]}
method help_sock {} { return [my tidyTextBlob {
	Connecting to an interp via a socket.

	To connect to an interpreter running on a socket, type
	    ~sockOn host:port
	and the sock interp will appear in the Interp->Sock menu.
}]}
method help_workspace {} { return [my tidyTextBlob {
	Debugging environment.

	A workspace is a class that provides a text widget
	(aka text area), a combobox widget (aka combo),
	and a label widget (aka label).

	Selected text in the text area can be evaluated.

	The combo provides a REPL-like interface.

	The label to the right of the combo shows the
	current interpreter and namespace.

	A workspace provides a combination of:
	  * A Unix-y, terminal-style environment,
	    with scrolling text and command history.
	  * A Smalltalk-inspired freeform environment
	    where any text can be evaluated in
	    the context of a running program.

	The goal is to allow the user to live debug their programs
	in the fashion of working with common Unix filesystem commands
	in a Unix terminal, and to augment this experience with tools
	and capabilities that a GUI can provide.
}]}
method help_inspector {} { return [my tidyTextBlob {
	Debugging environment with Tree.

	An inspector is a class that provides a
	workspace with a tree stuck to it.
	Button 3 in the tree area for menu.
}]}
method help_varwatcher {} { return [my tidyTextBlob {
	Watch a variable.
}]}
method help_multivarwatcher {} { return [my tidyTextBlob {
	Watch one or more variables.
}]}
method help_scriptbutton {} { return [my tidyTextBlob {
	Eval a script on a click.

	A ScriptButton is a button that runs a script.

	How to use:

	* Select some text and hit Alt-u (ScriptButton It).
	  A ScriptButton edit window will appear.

	* Make any changes to the text for the button
	  and the script that will be evaluated.
	  Click on Discard to cancel.
	  Click on Keep and the ScriptButton appears.

	* Click on the button to evaluate the script.

	* Control-q to destroy the ScriptButton.

	* Control-e to edit the ScriptButton.
}]}
method help_tss {} { return [my tidyTextBlob {
	Tcl Shell-Shell.

	Tss is an extension that provides a class
	that can encapsulate a process-local, Tk,
	socket or comm interpreter.

	Tss also provides the ability to view
	the namespace hierarchy like a filesystem.
	Methods ls, cd, pwd and cat function like
	the familiar filesystem commands.
	Methods ls, cd and pwd operate on namespaces.
	The cat method can return the contents of things.

	For the cd method, namespace paths may be
	written like filesystem paths, with /
	replacing :: and .. meaning parent.

}]}
method help_herbie {} { return [my tidyTextBlob {
	The Love Debugger.

	Herbie is an extension that provides a set of classes
	for use in a debugging environment.

	Some classes are: workspace, inspector, and scriptbutton.

	Herbie was designed to be embedded in programs,
	giving them built-in debugging facilities.

	Herbie uses Tss objects to access interpreters.
}]}
method help_tennessee {} { return [my tidyTextBlob {
	Stand-alone Debugger.

	Tennessee is a stand-alone program that uses Herbie
	objects to provide a debugging environment.

	Running Tennessee from the command-line
	brings up a workspace.

	Tennessee augments the initial workspace
	with command history loading and saving.
}]}
method help_using {} { return [my tidyTextBlob {
	Using the debugger.

	* Examine environment with commands like
	  "cat tcl_platform", or actions like typing
	  "info tcl" on a line by itself in the text
	  area and hitting Alt-p.

	* Use pwd, cd and ls to get to a namespace
	  containing a proc of interest.

	* Use cat <procname> to print the proc.

	* Make changes in the text area.

	* Select all the text of the proc.

	* Hit Alt-p or select Print It from the menu.
	  Alt-d (Do It) will also evaluate the
	  selected text but it won't print any errors.

	In this way one can browse around and modify a live program.

	Don't forget that these are only changes to a running program
	so any polished procs should probably be copied and pasted
	back into their original source files!
}]}
method help_notes {} { return [my tidyTextBlob {
	Release Notes.

	2023/06/01

	Socket interps don't work.
	Var tracing is iffy.
}]}
method Banner {} { return [my tidyTextBlob {
	    #     # ####### ######  ######  ### #######
	    #     # #       #     # #     #  #  #
	    #     # #       #     # #     #  #  #
	    ####### #####   ######  ######   #  #####
	    #     # #       #   #   #     #  #  #
	    #     # #       #    #  #     #  #  #
	    #     # ####### #     # ######  ### #######

	                The Love Debugger


Version 0.3.
Type "help" for help.
}]}
method banner {{rep {#}}} { return [string map [list {#} [string index $rep 0]] [my Banner]] }
}; # oo::class create HerbieHelp



#
# Inspector
#

oo::class create Inspector {
	variable ts
	variable tss
	variable tssts
	variable M
	variable whar
	variable workspace
	variable wordWrap
	variable wTree
	variable wWorkspace
	variable wW
constructor {args} {
	::herbie::ignition
	namespace import ::tu::toplevels::tk::randWindow ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::onToplevelDelete
	namespace import ::tu::epm::epmBindMenuTo ::tu::epm::epmW
	namespace import ::gub::gub
	namespace import ::meb::meb
	set wordWrap 1
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
destructor {
	if {[info exists wW]} { destroy $wW }
	if {[info exists ts]} { $ts destroy }
}
method eval {args} { tailcall $tss eval [join $args] }
method go {where} {
	lassign [$tssts go $where] path type
	set tsn root
	$ts set $tsn path $path
	$ts set $tsn type $type
	$ts set $tsn text $path
	$ts set $tsn values [list $tsn $type]
	$tssts fillAtNode root
}
method updateNode {node} {
	my ts->tv $node {}
	$workspace updateLabel
}
method openOnIn {where {w {}} {aTss {}} args} {
	set orient horizontal
	set focus ""
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-orient -
			-focus  { set state $arg }
			default { break }
			}; # hctiws
		}
		-orient { set state opt; set orient $arg }
		-focus  { set state opt; set focus  $arg }
		}; # hctiws
	}
	set tss [expr {$aTss ne {} ? $aTss : [::tss::Tss new]}]
	if {$w eq {}} {
		set w [createToplevelGui "Inspector" -class Herbie]
		eval [gub "-.;|." $w]
		bind $w <Control-q> [list [self] destroy]
	}
	set ts [::struct::tree]
	set tssts [::herbie::TssTs new $tss $ts]
	my go $where
	set whar $where
	set gob "
		;P F V C
		;    C
		;  F
		.P0 -orient $orient
		:P0,V0
		-F0,1 ; |F0,1
		/V0 wTree
		/F1 wWorkspace
		/. wW
	"
	set gui [dict create]
	eval [gub $gob $w -wm $::herbie::widgetMap]
	onToplevelDelete $w [list [self] destroy]
	foreach v [list wW wTree wWorkspace] {
		set $v [dict get $gui $v]
	}
	set workspace [::herbie::Workspace new $where $wWorkspace $tss]
	my configureTree

	epmBindMenuTo $wTree [my createInspectorMenu]
	my updateNode root
	if {$focus eq "combo"} {
		focus [$workspace wCombo]
	} elseif {$focus eq "text"} {
		focus [$workspace wText]
	} elseif {$focus eq "tree"} {
		focus $wTree
	}
}
method createInspectorMenu {} {
	set M [randWindow .inspector_]
	set mob "
		&Inspect it	-command {[self] inspectIt}
		&Script Buttton	-command {[self] scriptbuttonIt}
		&Colour Tuner	-command {[self] colourtuneIt}
		&Watch it	-command {[self] watchIt}
		&MultiWatch it	-command {[self] multiwatchIt}
		Close		-command {::event generate \[::tu::epm::epmW\] <Control-q>}
	"
	set moob [meb -menuname $M $mob]
 	eval [dict get $moob tcl src]
	return $M
}
method configureTree {} {
	set cols {treenode type}
	$wTree configure -columns $cols
	$wTree configure -displaycolumns {type}
	$wTree configure -selectmode browse
	$wTree heading \#0 -text !
	$wTree heading treenode -text tsn
	$wTree heading type -text Type
	$wTree column type -width 40
#	$wTree configure -show tree
	bind $wTree <<TreeviewSelect>> [list [self] tvselect]
	bind $wTree <<TreeviewOpen>> [list [self] tvopen]
	foreach col $cols {
		$wTree heading $col -text [string totitle $col] -command [list ::tu::tk::tvSortBy $wTree $col 0]
	}
	$wTree heading #0 -text $whar
}
method tvopen {} {
	set tsn [$wTree set [lindex [$wTree selection] 0] treenode]
	$tssts fillAtNode $tsn
	my ts->tv $tsn [$wTree selection]
}
method tvselect {} {
	set tsn [$wTree set [lindex [$wTree selection] 0] treenode]
	set type [$ts get $tsn type]
	set path [$ts get $tsn path]
	set w [$workspace wText]
	$w delete 1.0 end
	if {$type eq "win"} {
		$w insert end [join [my eval $path configure] \n]
	} elseif {$type eq "prop"} {
		lassign $path wpath cpath
		$w insert end [my eval $wpath cget $cpath]
	} elseif {$type in {arr Arr}} {
		set l [list]
		foreach {n v} [my eval array get $path] {
			lappend l [list $n $v]
		}
		$w insert end [::herbie::tblfmt [lsort -dictionary -index 0 $l]]
	} elseif {$type in {var method obj Var method Obj proc Proc mtd Mtd}} {
		$w insert end [$tss cat $path]
	} else {
		$w insert end $path
	}
	if {0 && $type eq "ns"} {
		$tssts fillAtNode $tsn
		my ts->tv $tsn [$wTree selection]
	}
}
method ts->tv {tsnode tvnode} {
	set tv $wTree
	$tv delete [$tv children $tvnode]
	foreach tsn [$ts children $tsnode] {
		set nid [$tv insert $tvnode end -text [$ts get $tsn text] -values [$ts get $tsn values]]
		if {[$ts get $tsn type] in {ns arr Arr win obj Obj}} {
			$tv insert $nid end
		}
		#my ts->tv $tsn $nid
	}
	if {[$ts numchildren $tsnode] == 0 && [$ts get $tsnode type] in {ns arr Arr win obj Obj}} {
		$tv insert $tvnode end
	}
}
method scriptbuttonIt {} {
	set tv [epmW]
	if {[llength [set sell [$tv selection]]] == 0} { return "" }
	set sel [lindex $sell 0]
	set treenode [$tv set $sel treenode]
	set path [$ts get $treenode path]
	if {$path eq ""} { return "" }
	return [::herbie::ScriptButton new $path $path]
}
method colourtuneIt {} {
	set tv [epmW]
	if {[llength [set sell [$tv selection]]] == 0} { return "" }
	set sel [lindex $sell 0]
	set treenode [$tv set $sel treenode]
	set path [$ts get $treenode path]
	if {$path eq ""} { return "" }
	return [[::herbie::ColourTuner new $path "$path configure -background"] openOnIn [$path cget -background]]
}
method inspectIt {} {
	set tv [epmW]
	if {[llength [set sell [$tv selection]]] == 0} { return "" }
	set sel [lindex $sell 0]
	set treenode [$tv set $sel treenode]
	set path [$ts get $treenode path]
	if {$path eq ""} { return "" }
	[self class] new $path
}
method watchIt {} {
	set tv [epmW]
	if {[llength [set sell [$tv selection]]] == 0} { return "" }
	set sel [lindex $sell 0]
	set treenode [$tv set $sel treenode]
	set path [$ts get $treenode path]
	if {$path eq ""} { return "" }
	::herbie::VarWatcher new $path
}
method multiwatchIt {} {
	set tv [epmW]
	if {[llength [set sell [$tv selection]]] == 0} { return "" }
	set sel [lindex $sell 0]
	set treenode [$tv set $sel treenode]
	set path [$ts get $treenode path]
	if {$path eq ""} { return "" }
	if {![::herbie::running? mvw]} {
		::herbie::running mvw [::herbie::MultiVarWatcher new $path {} $tss]
	}
#	[::herbie::running mvw] add $path
}
}; # oo::class create Inspector



#
# Tcl Shell Shell <-> Tree Struct
#

oo::class create TssTs {
    variable tss
    variable ts
constructor {aTss aTs {where {}}} {
	set tss $aTss
	set ts  $aTs
}
method eval {args} { tailcall $tss eval [join $args] }
method go {where} {
	if {$where eq "::"} {
		$tss cns $where
		set path [$tss cns]
		set type ns
	} else {
		set pat [namespace tail $where]
		set wr  [namespace qualifiers $where]
		set g [$tss glob $pat $wr]
		lassign [lindex $g 0] type name
		#set pwd $name
		set name ${wr}::$name
		$tss cd $name
		if {$type eq "win"} {
			set path $name
		} else {
			set path [$tss cns]
		}
	}
	return [list $path $type]
}
method fillAtNode {tsnode} {
	set nodetype [$ts get $tsnode type]
	if {$nodetype ni {ns arr Arr win obj Obj}} { return }
	set path [$ts get $tsnode path]
	if {$path eq "::"} { set path "" }
#$tv delete [$tv children $node]
	set thangs [list]
	set thongs [list]
	if {$nodetype eq "win"} {
		set newnodetype win
		set l [my eval winfo children [namespace tail $path]]
		foreach n [lsort $l] {
			lappend thongs [list $newnodetype $n]
			lappend thangs $n
		}
		set newnodetype prop
		set l [list]
		foreach o [my eval $path configure] {
			lappend l [lindex $o 0]
		}
		foreach n [lsort $l] {
			lappend thongs [list $newnodetype [list $path $n]]
			lappend thangs $n
		}
		set path [namespace qualifiers $path]
	} elseif {$nodetype in {obj Obj}} {
		set objtype [expr {[my eval info object isa class $path] ? "class" : "object"}]
		set newnodetype [string map {obj mtd Obj mtd} $nodetype]
		foreach n [lsort -dictionary [my eval info $objtype methods $path]] {
			lappend thongs [list $newnodetype $n]
			lappend thangs $n
		}
#		set path [namespace qualifiers $path]
	} elseif {$nodetype in {arr Arr}} {
		set newnodetype [string map {arr var Arr Var} $nodetype]
		foreach n [lsort -dictionary [my eval array names $path]] {
			lappend thongs [list $newnodetype [namespace tail $path]($n)]
			lappend thangs $n
		}
		set path [namespace qualifiers $path]
	} else {
		set thongs [$tss glob * $path]
	}
	foreach thong $thongs thang $thangs {
		lassign $thong type thing
		if {[string is upper [string index $type 0]] && $path ne ""} {continue}
		if {$type eq {"win"}} continue
		if {$thang eq ""} {
			set thang $thing
		}
		set nn [lindex [$ts insert $tsnode end] 0]
		$ts set $nn path ${path}::[expr {$type eq "mtd" ? ">::" : ""}]$thing
		$ts set $nn type $type
		$ts set $nn text $thang
		$ts set $nn values [list $nn $type]
	}
}
}; # oo::class create TssTs



#
# BasicWin
#

oo::class create BasicWin {
	variable wW myW
constructor {args} {
	::herbie::ignition
	namespace import ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::onToplevelDelete
	namespace import ::gub::gub
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
destructor {
	if {[info exists myW]} { bind $myW <Destroy> {}; destroy $myW }
}
method openOnIn {{_ {}} {w {}}} {
	if {$w eq {}} {
		set w [createToplevelGui {} -class Herbie]
		eval [gub "-.;|." $w]
		eval [gub "&. Control+q [self] destroy" $w]
		set myW $w
	} else {
		eval [gub "F;:F0;-F0;|F0;/F0 myW" $w -wm $::herbie::widgetMap -vn {}]
	}
	set wW $w
	onToplevelDelete $myW [list [self] destroy]
}
}; # oo::class create BasicWin



#
# Dict Editor
#
# ::herbie::DictEditor new {apply {{res dct} { puts $res\n$dct}}} "Edit Something" [dict create cheese Brie veg Tomato bread Baguette vin Rouge lang Japanese]

oo::class create DictEditor {
	variable wW myW
	variable gui
	variable dict
	variable cx
	variable res
	variable rdict
	variable wtitle
constructor {commandPrefix title args} {
	::herbie::ignition
	namespace import ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::onToplevelDelete
	namespace import ::gub::gub
	set cx $commandPrefix
	if {$title ne ""} {
		set wtitle $title
	} else {
		set wtitle "Dict Editor"
	}
	set res 0
	set rdict [dict create]
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
destructor {
	if {[info exists myW]} { bind $myW <Destroy> {};destroy $myW }
	if {[info exists cx]} {
		uplevel #0 $cx $res [list $rdict]
		#uplevel #0 {*}$cx $res [list $rdict]
	}
}
eval [::tu::tcl::makeROAccessors wW]
method openOnIn {d {w {}}} {
	if {$w eq {}} {
		set w [createToplevelGui $wtitle -class Herbie]
		eval [gub "-.;|." $w]
		bind $w <Control-q> [list [self] done 0]
		set myW $w
	} else {
		eval [gub "F;/F0 myW" $w -vn {}]
	}
	set wW $w
	set dict $d
	set gob ""
	append gob "F;"
	set i -1
	dict for {k v} $dict {
		incr i
		append gob "  L E;"
		append gob ".L${i} -text {$k};"
		append gob "/E${i} $k;"
	}
	append gob ":L -sticky w;"
	append gob "F B B;"
	append gob ".B0 -text Keep;"
	append gob ".B1 -text Discard;"
	append gob ".B0 -command {[self] done 1};"
	append gob ".B1 -command {[self] done 0};"
	set gui [dict create]
	eval [gub $gob $myW -wm $::herbie::widgetMap]
	onToplevelDelete $myW [list [self] done 0]
	dict for {k v} $dict {
		[dict get $gui $k] insert end $v
	}
	focus [dict get $gui [lindex [dict keys $dict] 0]]
}
method done {{result {0}}} {
	set res $result
	if {$res} { dict for {k v} $dict {
		dict set rdict $k [[dict get $gui $k] get]
	}}
	my destroy
}
}; # oo::class create dictEditor



#
# ScriptButton
#

proc classOf {o} {
	return [namespace tail [info object class $o]]
}
oo::class create ScriptButton {
	variable wW
	variable wButton
	variable script
	variable tss
constructor {{text ""} {script ""} {aTss {}}} {
	::herbie::ignition
	namespace import ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::getToplevelTitle
	namespace import ::gub::gub
	namespace import ::herbie::DictEditor
	set tss $aTss
	DictEditor new [list [self] createDone] [::herbie::classOf [self object]] [dict create text $text script $script]
}
destructor {
	if {[info exists wW]} { destroy $wW }
	if {[info exists tss] && $tss ne {}} {$tss destroy }
}
method go {} { tailcall $tss eval $script }
method createDone {res dict} {
	if {$res} {
		if {$tss eq {}} {
			set tss [::tss::Tss new]
		}
		set script [dict get $dict script]
		my createIn [dict get $dict text]
	} else {
		my destroy
	}
}
method createIn {text {w {}}} {
	if {$w eq {}} {
		set w [createToplevelGui [::herbie::classOf [self class]] -class Herbie]
	}
	eval [gub "/B0 wButton;/. wW;B;.B -text {$text} -command {[self] go}" $w -wm $::herbie::widgetMap]
	dict with gui {}
	bind $wW <Control-q> [list [self] createDone 0 {}]
	bind $wW <Control-e> [list [self] edit]
	focus $wButton
}
method edit {} {
	DictEditor new [list [self] editDone] "Edit [::herbie::classOf [self object]]" [dict create text [$wButton cget -text] script $script]
}
method editDone {res dict} {
	if {!$res} { return }
	set script [dict get $dict script]
	$wButton configure -text [dict get $dict text]
}
}; # oo::class create ScriptButton



#
# ColourTuner
#

oo::class create ColourTuner {
	superclass ::herbie::BasicWin
	variable wW myW
	variable wLabel
	variable script
	variable text
	variable tss
	variable colourBox
	variable wEdit
constructor {{aText ""} {aScript ""} {aTss {}} args} {
	next
	package require colourbox
	namespace import ::colourbox::ColourBox
	namespace import ::herbie::DictEditor
	set text $aText
	set script $aScript
	set tss $aTss
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
destructor {
	if {[info exists colourBox]} { $colourBox destroy }
	if {$tss ne {}} { $tss destroy }
	next
}
eval [::tu::tcl::makeRWAccessors wW wEdit]
method openOnIn {{colour {}} {w {}}} {
	if {$tss eq {}} { package require tss; set tss [::tss::Tss new] }
	next {} $w
	set colourBox [::colourbox::ColourBox new $wW -command [list [self] setColour] -defaultColour $colour]
	$colourBox open
	grid [$colourBox wColourBox] -sticky w
	grid [set wLabel [ttk::label $wW.lName -text $text]]
	grid [set wEdit [ttk::frame $wW.fEdit]] -sticky ew
	bind $wW <Control-q> [list [self] destroy]
	bind $wW <Control-e> [list [self] edit]
	return [self]
}
method edit {} {
	if {[llength [winfo children $wEdit]] > 0} return
	DictEditor new [list [self] editDone] \
		"Edit [::herbie::classOf [self object]]" \
		[dict create text [$wLabel cget -text] script $script] \
		$wEdit
	grid $wEdit
}
method editDone {res dict} {
	grid remove $wEdit
	if {!$res} { return }
	set script [dict get $dict script]
	$wLabel configure -text [dict get $dict text]
}
method setColour {colourBox} {
	set colour [$colourBox colour]
	if {![regsub -all {%c} $script $colour scr]} {
		set scr [string cat $script " " $colour]
	}
	catch { $tss eval $scr }
}
}; # oo::class create ColourTuner



#
# Var Watcher
#

oo::class create VarWatcher {
	mixin herbieGuiBits
	variable tss
	variable var
	variable vtype
	variable watch
	variable cont
	variable wordWrap
	variable wW
	variable wText
	variable wVar
constructor {args} {
	::herbie::ignition
	namespace import ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::onToplevelDelete
	namespace import ::gub::gub
	set watch 0
	set cont 1
	set wordWrap 1
	if {[llength $args] > 0} { set watch 1; my openOnIn {*}$args; my watch }
}
destructor {
	if {[info exists wW]} { destroy $wW  }
	if {[info exists watch]} { set watch 0; my watch }
	$tss destroy
}
method done {} { my destroy }
eval [::tu::tcl::makeRWAccessors wW wCombo wText wVar]
method eval {args} { tailcall $tss eval [join $args] }
method normalize {variable} { string cat "::" [string trimleft $variable ":"] }
method openOnIn {variable {w {}} {aTss {}}} {
	set tss [expr {$aTss ne {} ? $aTss : [::tss::Tss new]}]
	if {$w eq {}} {
		set w [createToplevelGui "Var Watcher" -class Herbie]
		eval [gub "-.;|." $w]
		bind $w <Control-q> [list [self] done]
	}
	set var [my normalize $variable]
	set vtype [expr {[my eval array exists $var] ? "arr" : "var"}]
	set gob "
		;F t C
		;  C
		;F L E KwKcKaB B
		.L0 -text Variable
		.E0 -textvariable {[my varname var]}
		.Kw -text Watch -variable [my varname watch] -command {[self] watch}
		.Kc -text Continuous -variable [my varname cont] -command {[self] activity $var {} write}
		.Ka -text WW -variable [my varname wordWrap] -command {[self] setww}
		.B0 -text Clear -command {/t0/ delete 1.0 end}
		.B1 -text Close -command {[self] destroy}
		.t0 -wrap word -tabstyle wordprocessor -width 40 -height 18
		:t,F0
		:F1 -sticky ew
		:B0 -sticky e -padx {12 0}
		:E0 -padx {6 0} -sticky ew
		:Kw -padx {24 0}
		:Kc -padx 12
		-F0 ; |F0
		|F1 1
		/t0 wText
		/E0 wVar
		/. wW
	"
	set gui [dict create]
	eval [gub $gob $w -wm $::herbie::widgetMap]
	foreach v [list wW wText wVar] {
		set $v [dict get $gui $v]
	}
	onToplevelDelete $w [list [self] destroy]
	if {$watch} {
		my activity $var {} write
	}
	focus [dict get $gui wVar]
}
method activity {name1 name2 op} {
	set variable $name1
	if {$vtype eq "var" && $name2 ne ""} {
		append variable "(" $name2 ")"
	}
	if {!$cont} { $wText delete 1.0 end }
	catch {$tss cat $variable} res
	my dispVal $res
}
method watch {} {
	$tss [expr {$watch ? "" : "un"}]watch $var [list [self] activity]
}
method dispVal {val} {
	my textAppend $val
	my textAppend \n
	if {$cont} {
		my textAppend [string repeat "-" 40]
		my textAppend \n
	}
	my textSee end
}
method setww {} {
	[my wText] configure -wrap [expr {$wordWrap ? "word" : "none"}]
}
}; # oo::class create VarWatcher


#
# ErrorCode Var Watcher
#

oo::class create ErrorCodeVarWatcher {
	superclass VarWatcher
method dispVal {val} {
	variable cont
	my textAppend $::errorCode
	my textAppend \n\n
	my textAppend $val
	my textAppend \n
	if {$cont} {
		my textAppend [string repeat "-" 40]
		my textAppend \n
	}
	my textSee end
}
}; # oo::class create ErrorWatcher


#
# Multi Var Watcher
#

oo::class create MultiVarWatcher {
	mixin herbieGuiBits
	variable tss
	variable var
	variable watch
	variable wW
	variable wTv
	variable wVar
	variable wWatch
	variable vars
	variable items
	variable watchcmd
constructor {args} {
	::herbie::ignition
	namespace import ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::onToplevelDelete
	namespace import ::gub::gub
	set watch no
	set var {}
	set vars [dict create]
	set items [dict create]
	set watchcmd [list [self] activity]
	if {[llength $args] > 0} { return [my openOnIn {*}$args] }
}
destructor {
	if {[info exists wW]} { destroy $wW  }
	foreach variable $vars {
		$tss unwatch $variable $watchcmd
	}
}
method done {} { my destroy }
eval [::tu::tcl::makeRWAccessors wW wTv wVar wWatch]
method eval {args} { tailcall $tss eval [join $args] }
method normalize {variable} { string cat "::" [string trimleft $variable ":"] }
method openOnIn {{variable {}} {w {}} {aTss {}}} {
	set tss [expr {$aTss ne {} ? $aTss : [::tss::Tss new]}]
	if {$w eq {}} {
		set w [createToplevelGui "Multi Var Watcher" -class Herbie]
		eval [gub "-.;|." $w]
		bind $w <Control-q> [list [self] done]
	}
	onToplevelDelete $w [list [self] destroy]
	set cols [dict create watch 50 variable 100 value 214]
	set gui [dict create]
	eval [gub "
		;F V C
		;  C
		;F L E K B B
		.V0 -columns {[dict keys $cols]} -show headings -selectmode browse
		.L0 -text Variable
		.E0 -textvariable {[my varname var]}
		.K0 -text Watch -variable [my varname watch] -command {[self] watch} -onvalue yes -offvalue no -state disabled
		.B0 -text Add -command {[self] addFromGui}
		.B1 -text Remove -command {[self] remove}
		:F0,V0
		:F1 -sticky ew
		:E0 -padx {6 0} -sticky ew
		:K0 -padx {12 12}
		-F0 ; |F0
		|F1 1
		&V0 <<TreeviewSelect>> [self] selected
		&V0 Double-Button-1 /K0/ invoke
		/V0 wTv
		/E0 wVar
		/K0 wWatch
		/. wW
	" $w -wm $::herbie::widgetMap]
	foreach v [list wW wTv wVar wWatch] {
		set $v [dict get $gui $v]
	}
	dict for {col width} $cols {
		$wTv heading $col -text [string totitle $col]
		$wTv column $col -width $width
	}
	if {$variable ne {}} {
		my add $variable
	}
	focus $wVar
}
method selected {} {
	if {[llength [set sel [$wTv selection]]] == 0} { return }
	set idx [lindex $sel 0]
	set values [$wTv item $idx -values]
	set watch [lindex $values 0]
	set var [lindex $values 1]
}
method addFromGui {} {
	if {[set variable [string trim [$wVar get]]] eq ""} { return }
	my add $variable
}
method add {variable} {
	set variable [my normalize $variable]
	if {[my eval array exists $variable]} {
		my addArray $variable
	} else {
		my addVar $variable
	}
	return [self]
}
method addArray {variable} {
	set variable [my normalize $variable]
	foreach e [my eval array names $variable] {
		my addVar ${variable}($e)
	}
}
method addVar {variable} {
	set variable [my normalize $variable]
	if {[dict exists $vars $variable]} { return }
	set item [$wTv insert {} end -value [list yes $variable {}]]
	dict set vars $variable $item
	dict set items $item $variable
	my activity $variable {} write
	set var $variable; set watch yes; my watch
	$wTv selection set [list $item]
	$wTv focus $item
	$wWatch state {!disabled}
}
method remove {} {
	set variable $var
	if {[set item [my getItemForVariable $variable]] eq {}} { return }
	set watch no
	my watch
	$wTv delete [list $item]
	set items [$wTv children {}]
	if {[llength $items] == 0} {
		$wWatch state {disabled}
	} else {
		set item [lindex $items 0]
		$wTv selection set [list $item]
		$wTv focus $item
	}
}
method getItemForVariable {variable} {
	return [dict get $vars $variable]
}
method getValuesForVariable {variable} {
	return [$wTv item [dict get $vars $variable] -values]
}
method activity {name1 name2 op} {
	set variable $name1
	if {$name2 ne ""} {
		append variable "(" $name2 ")"
	}
	if {[llength [set values [my getValuesForVariable $variable]]] == 0} { return }
	if {[set item [my getItemForVariable $variable]] eq {}} { return }
	catch {$tss cat $variable} res
	lset values 2 $res
	$wTv item $item -values $values
}
method watch {} {
	set variable $var
	$tss [expr {$watch ? "" : "un"}]watch $variable $watchcmd
	if {$watch} {
		my activity $variable {} write
	}
	if {[llength [set values [my getValuesForVariable $variable]]] == 0} { return }
	if {[set item [my getItemForVariable $variable]] eq {}} { return }
	lset values 0 [expr {$watch ? "yes" : "no"}]
	$wTv item $item -values $values
}
}; # oo::class create MultiVarWatcher


#
# Styler
#

namespace eval styler {
	set savedWs [list wW wEtv wEtext wStv wStext wLtv wLtext wMtv wMtext]
	set things [lsort [list {*}{
		.

		TButton
		TCheckbutton
		TRadiobutton
		TMenubutton
		TEntry
		TCombobox
		TSpinbox
		TLabelframe
		TLabelframe.Label
		TScrollbar
		TScale
		TProgressbar
		TNotebook
		TNotebook.Tab
		Heading
		Treeview
		ComboboxPopdownFrame
		Toolbutton
		TFrame

		Item
		Cell
		Row

		Tab

		TLabel
		Label
		TPanedwindow
		Horizontal.Sash
		Vertical.Sash
		Horizontal.TScale
		Vertical.TScale
		Horizontal.TScrollbar
		Vertical.TScrollbar
		TSeparator
		TSizegrip
		TSquare
	}]]
oo::class create Styler {
	mixin ::herbie::herbieGuiBits
	foreach v $::herbie::styler::savedWs { variable $v }
	variable theme
	variable styles layouts maps
constructor {args} {
	::herbie::ignition
	namespace import ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::onToplevelDelete
	namespace import ::gub::gub
	set theme default
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
destructor {
	if {[info exists wW]} { destroy $wW  }
}
method done {} { my destroy }
eval [::tu::tcl::makeROAccessors {*}$::herbie::styler::savedWs]
method openOnIn {{aTheme {}} {w {}}} {
	if {$w eq {}} {
		set w [createToplevelGui "Styler" -class Herbie]
		eval [gub "-.;|." $w]
		bind $w <Control-q> [list [self] done]
	}
	if {$aTheme ne {}} {
		set theme $aTheme
	}

	set ecols [list element]
	set scols [list style]
	set lcols [list layout]
	set mcols [list map]

	set themes [::ttk::style theme names]

	set gob "
		;P PaM P F V C
		;          C
		;        F t C
		;          C
		;    M P F V C
		;          C
		;        F t C
		;          C
		;  PbM P F V C
		;          C
		;        F t C
		;          C
		;    M P F V C
		;          C
		;        F t C
		;          C
		;FtLtYt
		#
		.{M} -text Elements Styles Layouts Maps
		.{V} -columns {$ecols} {$scols} {$lcols} {$mcols}
		.V -show {} -selectmode browse
		.Lt -text Theme
		.Yt -values {$themes} -textvariable [my varname theme] -state readonly
		.P0 -orient vertical
		.P -orient horizontal
		.t -height 12 -width 30
		+P<ab>,M -weight 1
#		+M0,1 -weight 1
		:P,V,t
#		:M2 -sticky ew
		:Ft -sticky w
		:Lt -padx {0 4}
		-.,M,F ; |.,M,F
		&V0 <<TreeviewSelect>> [self] showElementInfo
		&V1 <<TreeviewSelect>> [self] showStyleInfo
		&V2 <<TreeviewSelect>> [self] showLayoutInfo
		&V3 <<TreeviewSelect>> [self] showMapInfo
		&Y0 <<ComboboxSelected>> [self] useTheme
		/{V} {} wEtv   wStv   wLtv   wMtv
		/{t} {} wEtext wStext wLtext wMtext
		/. wW
	"
	set gui [dict create]
	eval [gub $gob $w -wm $::herbie::widgetMap]
	foreach v $::herbie::styler::savedWs {
		set $v [dict get $gui $v]
	}
	foreach {tv cols} [list $wEtv $ecols $wStv $scols $wLtv $lcols $wMtv $mcols] {
		foreach col $cols {
			$tv heading $col -text [string totitle $col]
		}
	}
	foreach tv [list $wStv $wLtv $wMtv] {
		my fillTree $tv $::herbie::styler::things
	}
	my fillTree $wEtv [::ttk::style theme settings $theme {::ttk::style element names}]
}

method showElementInfo {} {
	set tv $wEtv
	set t  $wEtext
	if {[llength [set sel [$tv selection]]] == 0} { return }
	set options [lsort [::ttk::style element options [$tv set [lindex $sel 0] element]]]
	$t delete 1.0 end
	$t insert end [join $options \n]
}

method showStyleInfo {} {
	set tv $wStv
	set t  $wStext
	if {[llength [set sel [$tv selection]]] == 0} { return }
	set options [::ttk::style configure [$tv set [lindex $sel 0] style]]
	set l [list]
	foreach o [lsort [dict keys $options]] {
		lappend l [list $o [dict get $options $o]]
	}
	$t delete 1.0 end
	$t insert end [join $l \n]
}

method showLayoutInfo {} {
	set tv $wLtv
	set t  $wLtext
	if {[llength [set sel [$tv selection]]] == 0} { return }
	catch {::ttk::style layout [$tv set [lindex $sel 0] layout]} options
	$t delete 1.0 end
	$t insert end $options
}

method showMapInfo {} {
	set tv $wMtv
	set t  $wMtext
	if {[llength [set sel [$tv selection]]] == 0} { return }
	set map [::ttk::style map [$tv set [lindex $sel 0] map]]
	$t delete 1.0 end
	$t insert end [join [lmap {p v} $map { string cat $p "\n  " [join [lmap {a b} $v {list $a $b}] "\n  "]}] \n]
}

method fillElements {} {
$wEtv tag configure present -background orchid
	set elems [::ttk::style theme settings $theme {::ttk::style element names}]
	$wEtv tag remove present
	foreach item [$wEtv children {}] {
		$wEtv tag [expr {[$wEtv set $item element] in $elems ? "add" : "remove"}] present $item
	}
}

method fillTree {tv items} {
	$tv delete [$tv children {}]
	foreach item $items { $tv insert {} end -values $item }
}

method useTheme {} {
	my fillElements
#	::ttk::style theme use
}

}; # oo::class create Styler

namespace export Styler

}; #namespace eval styler

namespace import styler::Styler


#
# Tw
#

namespace eval tw {

set savedWs [list Tw_wW Tw_wText]

oo::class create Tw {
	mixin ::herbie::herbieGuiBits
	foreach v $::herbie::tw::savedWs { variable $v }
constructor {args} {
	::herbie::ignition
	namespace import ::tu::toplevels::tk::createToplevelGui ::tu::toplevels::tk::onToplevelDelete
	namespace import ::gub::gub
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
destructor {
	if {[info exists Tw_wW]} { destroy $Tw_wW  }
}
eval [::tu::tcl::makeROAccessors {*}$::herbie::tw::savedWs]
method openOnIn {{w {}}} {
	if {$w eq {}} {
		set w [createToplevelGui "Tw" -class Herbie]
		eval [gub "-.;|." $w]
		bind $w <Control-q> [list [self] destroy]
	}
	set gob "
		;F t C
		;  C
		.t -wrap word
#		.t -tabstyle wordprocessor
		:F0,t0
		-F0 ; |F0
		/t0 Tw_wText
		/. Tw_wW
	"
	set gui [dict create]
	eval [gub $gob $w -wm $::herbie::widgetMap]
	foreach v $::herbie::tw::savedWs {
		set $v [dict get $gui $v]
	}
}
}; #oo::class create Tw

namespace export Tw

}; #namespace eval tw

namespace import tw::Tw


#
# Glov
#

namespace eval glov {

set savedWs [list]

oo::class create Glov {
	superclass ::herbie::Tw
	foreach v $::herbie::tw::savedWs { variable $v }
	foreach v $::herbie::glov::savedWs { variable $v }
	variable showTime
	variable showLevel
	variable lvColours
constructor {args} {
	set showTime 1
	set showLevel 1
	set lvColours [dict create {*}{
		alert     #FF0001
		critical  #FF0002
		debug     #B0C4DE
		emergency #FF0003
		error     #FF0004
		info      #0000FF
		notice    #2E8B57
		warning   #FF4505
	}]
	next {*}$args
}
eval [::tu::tcl::makeROAccessors {*}$::herbie::glov::savedWs]
method openOnIn {{w {}}} {
	next $w
	$Tw_wText tag configure time  -elide [expr {!$showTime}]
	$Tw_wText tag configure level -elide [expr {!$showLevel}]
	$Tw_wText tag configure text -lmargin2 [expr {[font measure [$Tw_wText cget -font] \t] * 10}]
	my configureColours
}
method configureColours {} {
	dict for {lv colour} $lvColours {
		$Tw_wText tag configure $lv -foreground $colour
	}
}
method formatLogMsg {level text} {
	return [list [clock format [clock seconds] -format "%Y-%m-%dT%H:%M:%S"]\t time $level\t level $text [list text $level] \n]
}
method logPuts {level text} {
	set msg [my formatLogMsg $level $text]
	set w $Tw_wText
	if {[set state [$w cget -state]] ne "normal"} { $w configure -state normal }
	$w insert end {*}$msg
	$w see end
	if {$state ne "normal"} { $w configure -state $state }
}
method log  {level text} { my logPuts $level $text }
#method log  {level text} { ::log::log $level $text }
method ilog {text} { my log info    $text }
method elog {text} { my log error   $text }
method wlog {text} { my log warning $text }
}; # oo::class create Glov


set gfSavedWs [list Gf_wW]

oo::class create GlovFull {
	superclass Glov
	variable Tw_wText
	variable showTime
	variable showLevel
	variable lvColours
	variable vLevel
	variable vText
	foreach v $::herbie::glov::gfSavedWs { variable $v }
constructor {args} {
	package require colourbox
#	set vLevel [lindex [dict keys $lvColours] 0]
	set vLevel info
	set vText ""
	next {*}$args
}
destructor {
	if {[info exists Gf_wW]} { destroy $Gf_wW  }
}
method openOnIn {{w {}}} {
	if {$w eq {}} {
		set w [createToplevelGui "Tw" -class Herbie]
		eval [gub "-.;|." $w]
		bind $w <Control-q> [list [self] destroy]
	}
	set nColours [dict size $lvColours]
	set gob "
		;P Fl
		;  FCFc
		;      [string repeat ";      L;      F" [expr {$nColours + 2}]]
		;FoKtKlXlLmXtBnB B B
		.P0 -orient horizontal
		.{L} -text [dict keys $lvColours] Background Foreground Message
		.Kt -text Time -variable [my varname showTime] -command {[self] toggleTag time}
		.Kl -text Level -variable [my varname showLevel] -command {[self] toggleTag level}
		.Xl -value {[dict keys $lvColours]} -textvariable [my varname vLevel] -state readonly -width 10
		.Xt -textvariable [my varname vText]
		.B -width 4
		.Bn -width 6
		.{B} -text Newline Clear Test Log
		.{B} -command {[self] newline} {[self] clear} {[self] test} {[self] logFromGui}
		:P
		:L -sticky w
		:Fo -sticky ew
		:Fc -sticky nsew -padx {0 4} -pady {0 4}
		:Xt -sticky ew -padx 2
		:K,Xl -padx 2
		+{F<lC>} -weight 1 0
		-F<lC> ; |F<lC>
		|Fo 4
		!:-L1-[expr {$nColours+1}]*2 -2
		!:|L1-[expr {$nColours+1}]*2  1
		!:|F4-[expr {$nColours+4}]*2  1
		:L1-[expr {$nColours+4}]*2 -padx {14 0} -sticky w
		:F4-[expr {$nColours+4}]*2 -padx {14 0}
		&Xt Return [self] logFromGui
		/Fl log
		/Fc colour
		/F3-[expr {$nColours+4}]
		/Xt
		/. Gf_wW
	"
	set gui [dict create]
	eval [gub $gob $w -wm $::herbie::widgetMap]
	foreach v $::herbie::glov::gfSavedWs {
		set $v [dict get $gui $v]
	}
	next [dict get $gui log]
	for {set i 0} {$i < $nColours} {incr i} {
		set lv [lindex [dict keys $lvColours] $i]
		grid [[[::colourbox::ColourBox new \
			[dict get $gui F[expr {$i+3}]] \
			-layout horizontal \
			-defaultColour [dict get $lvColours $lv] \
			-command [list [self] setColour $lv] \
		] open] wColourBox]
	}
	grid [[[::colourbox::ColourBox new \
		[dict get $gui F[expr {$i+3}]] \
		-layout horizontal \
		-defaultColour [$Tw_wText cget -background] \
		-command [list [self] setTextColour background] \
	] open] wColourBox]
	incr i
	grid [[[::colourbox::ColourBox new \
		[dict get $gui F[expr {$i+3}]] \
		-layout horizontal \
		-defaultColour [$Tw_wText cget -foreground] \
		-command [list [self] setTextColour foreground] \
	] open] wColourBox]
	my ilog "Glov provides a way to evaluate colours for"
	my ilog "the different logging levels found in"
	my ilog "the Tcllib log extension."
	my newline
	my log notice "Glov uses ColourBox objects."
	my log notice "Button-3 on a colour swatch to copy/paste."
	my log notice "Double-Button-1 on a colour swatch to reset."
	my newline
	my log alert "Control-q to close."
	set vText "Type a message and hit Return."
	focus [dict get $gui Xt]
	event generate [dict get $gui Xt] <<TraverseIn>>
}
method setColour {level colourBox} {
	dict set lvColours $level [$colourBox colour]
	$Tw_wText tag configure $level -foreground [$colourBox colour]
}
method setTextColour {which colourBox} {
	$Tw_wText configure -$which [$colourBox colour]
}
method logFromGui {} {
	set vText [string trim $vText]
	if {$vText eq ""} { return }
	my log $vLevel $vText
	set vText ""
}
method toggleTag {tag} {
	$Tw_wText tag configure $tag -elide [expr {![$Tw_wText tag cget $tag -elide]}]
}
method test {} {
	foreach level [dict keys $lvColours] {
		my log $level [string totitle $level]
	}
}
method clear {} {
	$Tw_wText delete 1.0 end
	set vText ""
}
method newline {} {
	$Tw_wText insert end \n
}
}; #oo::class create GlovFull

namespace export Glov GlovFull

}; #namespace eval glov

namespace import glov::Glov glov::GlovFull


#
# Tkot
#

namespace eval tkot {

set savedWs [list wTv]

oo::class create Tkot {
	superclass ::herbie::BasicWin
	variable wW myW
	mixin ::herbie::herbieGuiBits
	foreach v $::herbie::tkot::savedWs { variable $v }
eval [::tu::tcl::makeROAccessors {*}$::herbie::tkot::savedWs]
method openOnIn {{_ {}} {w {}}} {
	next $_ $w
	set cols {opt db class def cur}
	eval [gub "
		;F V C
		;  C
		.V0 -columns {$cols} -selectmode browse -show headings
		:F0,V0
		-F0 ; |F0
		/V0 wTv
	" $myW -wm $::herbie::widgetMap -vn {}]
	foreach col $cols {
		$wTv heading $col -text [string totitle $col]
		$wTv column  $col -width 120
	}
}
method put {conf} {
	$wTv delete [$wTv children {}]
	foreach c $conf {
		if {[llength $c] == 2} { continue }
		$wTv insert {} end -values $c
	}
}
foreach dir [list up down] {
	method $dir {} [format {set f [focus]; focus $wTv; event generate $wTv <Key-%s>; focus $f} [string totitle $dir]]
}
}; #oo::class create Tkot

namespace export Tkot

}; #namespace eval tkot

namespace import tkot::Tkot


#
# Tkote
#

namespace eval tkote {

oo::class create Tkote {
	superclass ::herbie::BasicWin
	variable wW myW
	mixin ::herbie::herbieGuiBits
	variable curw curo curv
	variable wTv
	variable curi
constructor {args} {
	set curw ""
	set curo ""
	set curv ""
	set curi {}
	next {*}$args
}
eval [::tu::tcl::makeRWAccessors curw]
method openOnIn {{aWin {}} {w {}}} {
	next $aWin $w
	eval [gub "
		;F Fl
		;  FeLwEwBwLeEeBe
		.Lw -text Window
		.Le -textvariable [my varname curo]
		.Le -width [string length "-highlightbackground"]
		.Le -anchor e
		.{B} -text Inspect Keep
		.{B} -command {[self] inspect} {[self] keep}
		.{E} -textvariable [my varname curw] [my varname curv]
		:F0,Fl
		:Fe -sticky ew
		:Le -sticky e
		-F0,Fl ; |F0,Fl
		|Fe 3
		&Ew Return /Bw/ invoke
		&Ee Return /Be/ invoke
		&Ee Key-Up [namespace current]::tkot up
		&Ee Key-Down [namespace current]::tkot up
		/Fl wList
		/Ee wEdit
		/Ew wWindow
	" $myW -wm $::herbie::widgetMap -vn {}]
	::herbie::tkot::Tkot create tkot _ $wList
	set wTv [tkot wTv]
	bind $wTv <<TreeviewSelect>> [list [self] select]
	::oo::objdefine [self] forward put tkot put
#	::tu::epm::epmBindEpmMenuTo $wEdit
	focus $wWindow
	return [self]
}
method select {} {
	set sel [$wTv selection]
	if {[llength $sel] == 0} { return }
	set curi [lindex $sel 0]
	lassign [$wTv item $curi -values] curo {} {} {} curv
}
method inspect {{w {}}} {
	if {$w ne {}} { set curw $w }
	set curw [string trim $curw]
	if {$curw eq {}} return
	if {[llength $curw] == 2} {
		set w [lindex $curw 0]
		if {![winfo exists $w]} return
		tkot put [$w itemconfigure [lindex $curw 1]]
	} else {
		set w $curw
		if {![winfo exists $w]} return
		tkot put [$w configure]
	}
	return [self]
}
method keep {} {
	if {[llength $curw] == 2} {
		[lindex $curw 0] itemconfigure [lindex $curw 1] $curo $curv
	} else {
		$curw configure $curo $curv
	}
	$wTv set $curi cur $curv

}
}; #oo::class create Tkote

namespace export Tkote

}; #namespace eval tkote

namespace import tkote::Tkote


#
# Data Editor
#

namespace eval dataeditor {

	variable blank [dict create name {} type text value {} default {}]

oo::class create DataEditor {
	superclass ::herbie::BasicWin
	variable wW myW
	variable gui
	variable data
	variable cx
	variable notified
	variable wtitle
constructor {commandPrefix title args} {
	namespace import ::gub::guhb
	set cx $commandPrefix
	if {$title ne ""} {
		set wtitle $title
	} else {
		set wtitle "Data Editor"
	}
	set notified no
	if {[llength $args] > 11} {
		next [lindex $args 1]
	} else {
		next
	}
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
destructor {
	my notify 0
	next
}
method blank {} {
	return $::herbie::dataeditor::blank
}
method notify {ok} {
	if {$notified} { return 1 }
	set notified yes
	if {![info exists cx]} { return 1 }
	return [uplevel #0 $cx $ok [list $data]]
#	return [uplevel #0 {*}$cx $ok [list $data]]
}
method setdata {newdata} {
	set data [list]
	foreach d $newdata {
		lappend data [dict merge [my blank] $d]
	}
}
method openOnIn {newdata {w {}}} {
	next {} $w
	my setdata $newdata
	set g [list]
	lappend g "F;@F0 q"
	foreach d $data {
		dict with d {}
		lappend g ">q"
		if {$type eq "text"} {
			lappend g "L E"
			lappend g "/E0 $name"
		}
		lappend g ".L0 -text {$name}"
		lappend g ":L -sticky w"
	}
	lappend g ">."
	lappend g "F B B"
	lappend g ".B0 -text Keep"
	lappend g ".B1 -text Discard"
	lappend g ".B0 -command {[self] done 1}"
	lappend g ".B1 -command {[self] done 0}"
	set gui [dict create]
	eval [guhb [join $g \n] $myW -wm $::herbie::widgetMap]
	my refresh
	focus [dict get $gui [dict get [lindex $data 0] name]]
}
method done {{ok {0}}} {
	set rdata [list]
	foreach d $data {
		dict set d value [[dict get $gui [dict get $d name]] get]
		lappend rdata $d
	}
	set data $rdata
	if {[my notify $ok]} {
		set notified no
		return
	}
	my destroy
}
method refresh {} {
	foreach d $data {
		set w [dict get $gui [dict get $d name]]
		$w delete 0 end
		$w insert end [dict get $d value]
	}
}
}; # oo::class create DataEditor

namespace export DataEditor

}; # namespace eval dataeditor

#
# Basic Array Editor
#
# herbie::BasicArrayEditor new varName

namespace eval basicarrayeditor {
oo::class create BasicArrayEditor {
	superclass ::herbie::BasicWin
	variable wW myW
	variable gui
	variable data
	variable wtitle
constructor {args} {
	namespace import ::tu::tcl::dgwd
	namespace import ::gub::guhb
	if {0 && $title ne ""} {
		set wtitle $title
	} else {
		set wtitle "Basic Array Editor"
	}
	next
	if {[llength $args] > 0} {
		my openOnIn {*}$args
	}
}
method openOnIn {array {flddesc {}} {w {}}} {
	next {} $w
	upvar #0 $array arr
	if {[dict size $flddesc] == 0} {
		foreach k [array names arr] {
			dict set flddesc $k [dict create]
		}
	}
	set g [list]
	lappend g "F;@F0 q"
	dict for {name desc} $flddesc {
		set type [dgwd string $desc type]
		lappend g ">q"
		if {$type eq "string"} {
			lappend g "L E"
			lappend g ".E0 -textvariable ${array}($name)"
			lappend g "/E0 $name"
		}
		lappend g ".L0 -text {$name}"
		lappend g ":L -sticky w"
	}
	eval [guhb [join $g \n] $myW -wm $::herbie::widgetMap]
}
}; # oo::class create BasicArrayEditor

namespace export BasicArrayEditor

}; # namespace eval basicarrayeditor

#
# DruMa Editor
#

if {![catch {package require druma}]} {
namespace eval drumaeditor {
oo::class create DruMaEditor {
	superclass ::herbie::BasicWin ::druma::DruMa
	# From BasicWin
	variable wW myW
	# From DruMa
	variable aitcs
	# This class
	variable wtitle
	variable gui
constructor {aname {w {}} args} {
	namespace import ::tu::tcl::dgwd
	namespace import ::gub::guhb
	if {0 && $title ne ""} {
		set wtitle $title
	} else {
		set wtitle "DruMa Editor"
	}
	nextto ::herbie::BasicWin
	nextto ::druma::DruMa $aname {*}$args
	my openOnIn _ $w
}
method openOnIn {{_ {}} {w {}}} {
	next $w
	set flddesc [my flddesc]
	set groups [my groups]
	set fc [expr {[dict size $groups] > 1 ? "M" : "F"}]; # frame char
	set g [list]
	lappend g "F;@F0 cow;:F0;|F0"
	dict for {group fields} $groups {
		lappend g "> cow"
		lappend g "$fc;@${fc}0 g$group"
		lappend g ".M0 -text {$group}"
		lappend g ":${fc}0 -sticky ew"
		lappend g "|${fc}0"
		foreach field $fields {
			set type [my type $field]
			lappend g ">g$group"
			if {$type eq "string"} {
				lappend g "L E"
				lappend g ".E0 -textvariable [$aitcs& $field]"
				lappend g ":E0 -sticky e"
				lappend g "/E0 $field"
			}
			lappend g ".L0 -text {$field}"
			lappend g ":L -sticky w"
		}
	}
	eval [guhb [join $g \n] $myW -wm $::herbie::widgetMap]
}
method gui {e} {
	return [dict get $gui $e]
}
}; # oo::class create DruMaEditor

namespace export DruMaEditor

}; # namespace eval drumaeditor

namespace import drumaeditor::DruMaEditor
namespace export DruMaEditor

}; # catch package require druma

#
# Misc
#

proc sed {ssre blob} {
	lassign [split $ssre /] cmd match sub opts
	catch {regsub -all -line $match $blob $sub} res
	return $res
}

proc tblfmt {wut} {
	lassign [gettblfmt $wut] ncols fmt
	set l [list]
	foreach r $wut {
		lappend l [format $fmt {*}$r]
	}
	return [join $l \n]
}

proc gettblfmt {wut} {
	set ncols [llength [lindex $wut 0]]
	array set lens {}
	for {set col 0} {$col < $ncols} {incr col} {
		set lens($col) 0
	}
	foreach r $wut {
		set col -1
		foreach c $r {
			incr col
			if {[set len [string length $c]] > $lens($col)} {
				set lens($col) $len
			}
		}
	}
	set fmt ""
	for {set col 0} {$col < $ncols} {incr col} {
		append fmt % - [expr $lens($col) + 2] s
	}
	return [list $ncols $fmt]
}

proc getGubOffsets {{w .}} {
	return [gubOffsets [winfo children $w]]
}

proc autoscrollbar {w args} {
	set w [ttk::scrollbar $w {*}$args]
	::autoscroll::autoscroll $w
	return $w
}

namespace export Workspace Inspector VarWatcher MultiVarWatcher
namespace export DictEditor DataEditor ScriptButton
namespace export ErrorCodeVarWatcher
namespace export Styler Tw Glov GlovFull
namespace export Tkot
namespace import dataeditor::DataEditor
namespace export DataEditor
namespace import basicarrayeditor::BasicArrayEditor
namespace export BasicArrayEditor

namespace export herbie

package provide herbie 0.3

}; # namespace eval herbie

# EOF

# THATCH /herbie/packages/aitcs-2.0.1.tm modbase /herbie/packages
#
# AitCS - Array in the Command's Shadow
#
# Stuart Cassoff
#
#
# Spring 2018
# Version 1.1
#
# Winter 2020
# Version 1.2
#
# Spring 2020
# Version 1.3, 1.4, 1.5, 1.6, 1.7, 1.8
#
# Fall 2020
# Version 1.9
# Version 2.0b1
#
# Fall 2022
# Version 2.0
#
# Summer 2023
# Version 2.0.1
#


if {$::tcl_platform(engine) eq "Tcl"} { package require Tcl 8.6- }

namespace eval aitcs {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	aitcs
		title	AitCS
		descr	{Array in the Command's Shadow}
		author	{Stuart Cassoff}
		when	{Summer 2023}
		version	2.0.1
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2020,2022,2023 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}


## Internal

if {$::tcl_platform(engine) eq "Jim"} {
	proc Alias {this that args} {        alias    $this    $that {*}$args }
} else {
	proc Alias {this that args} { interp alias {} $this {} $that {*}$args }
}

proc Randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

proc Randvar {name variable_namespace} {
	return [string trimright $variable_namespace :]::__${name}__[Randstr]_
}

proc MakeCommands {type name varname cns} {
	set jim [expr {$::tcl_platform(engine) eq "Jim"}]

	set cns [string trimright $cns :]
	set smae [expr {[namespace qualifiers $varname] eq $cns}]
	set cns $cns:

	set ops   [list ""   ?       -      ^    &    ->    ]
	set procs [list Vset Vexists Vunset Vgwd Vref Vaitcs]

	if {$type eq "eagle"} {
		lappend ops   =      +        +=
		lappend procs Vsetto Vlappend Vincr
	}

	if {$jim} {
		set hows [lrepeat [llength $ops] full]
	} else {
		if {$smae} {
			set hows [lrepeat [llength $ops] var]
		} else {
			set hows [lrepeat [llength $ops] full]
			set pl [list Vexists Vgwd]
			if {$type eq "eagle"} {
				lappend pl Vsetto
			}
			foreach p $pl {
				lset hows [lsearch $procs $p] partial
			}
		}
		lset hows [lsearch $procs Vref] full
	}

	foreach op $ops proc $procs how $hows {
		set pargs [info args $proc]
		if {!$jim} {
			set pargs [lmap arg $pargs {
				expr {[info default $proc $arg def] ? [list $arg $def] : $arg}
			}]
		}
		set body [info body $proc]
		if {[lindex $pargs 0] eq "varname"} {
			set pargs [lrange $pargs 1 end]
			# For itiming tests, consult Makefile.aitcs.in when modifying this line: set how ?
			set body [switch -exact -- $how {
				var { # Change "upvar ..." to "variable ..." and hardcode [namespace tail $varname] everywhere
					set vn [namespace tail $varname]
					string map [list {upvar 1 $varname _v_} "variable $vn" _v_ $vn {$varname} $vn {${varname}} $vn] $body
				}
				full { # Remove "upvar ..." and hardcode $varname everywhere
					string map [list {upvar 1 $varname _v_} {} _v_ $varname {$varname} $varname {${varname}} $varname] $body
				}
				partial { # Hardcode $varname where used
					string map [list {$varname} $varname {${varname}} $varname] $body
				}
				setvar { # Add "set varname $varname" to top of proc
					format {set varname {%s}%s%s} $varname \n $body
				}
				alias { # Use aliases
					if {$proc eq "Vaitcs"} {
						if {[llength [info commands Vaitcs_alias]] == 0} {
							proc Vaitcs_alias {varname elem args} [string map [list \
								{tailcall $_v_($elem)$op {*}$args} \
								{uplevel 1 $_v_($elem)$op $args}] $body]
						}
						#set proc [string cat $proc _alias]
						append proc _alias
					}
					Alias $cns:$name$op [namespace current]::$proc $varname
					continue
				}
			}]
		}

		# Debug
		#puts "$name$op {$pargs} {$body\n}"

		proc $cns:$name$op $pargs $body

		# Debug
		#puts $name$op\n[::tcl::unsupported::disassemble proc $cns:$name$op]
	}

	# Debug
	#exit

	Alias $cns:$name! upvar #0 $varname $name

	# For itiming tests, consult Makefile.aitcs.in when modifying these 3 lines
	# Alias $cns:$name&&    return -level 0 $cns:$name
	proc    $cns:$name&& {} [list return $cns:$name]

	set --- [join [list {} \
		"unset $varname" \
		{*}[lmap op [list {*}$ops ! && ---] {
			format {rename %s%-3s {}} $cns:$name $op
		}] \
	] \n\t]\n

	# For itiming tests, consult Makefile.aitcs.in when modifying these 3 lines
	# Alias $cns:$name--- eval ${---}
	proc    $cns:$name--- {}   ${---}
}

proc New {type args} {
	set name ""
	set cns ""
	set vns ""
	set argies [list]
	set nameset no
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-cns - -command_namespace - -vns - -variable_namespace { set state $arg }
			-- { set state -- }
			default { if {$nameset} { lappend argies $arg } else { set name $arg; set nameset yes } }
			}; # hctiws
		}
		-cns - -command_namespace  { set state opt; set cns $arg }
		-vns - -variable_namespace { set state opt; set vns $arg }
		-- { lappend argies $arg }
		default { error "Huh? <[expr {int(rand()*100000)}]>" }
		}; # hctiws
	}
	if {$state ni [list opt --]} { return -code error "no argument given for option \"$state\"" }
	if {[expr {[llength $argies] % 2}] != 0} { return -code error "initial values must be \"key value ...\"" }
	if {$vns eq ""} { set vns [uplevel 1 {namespace current}] }
	if {$cns eq ""} { set cns $vns }
	set vns [string trimright $vns :]
	set cns [string trimright $cns :]
	if {$name eq ""} { for {*}{
		{ set name [Randstr]; set i 0 }
		{ $i < 1000 && [llength [info commands ${cns}::$name]] > 0 }
		{ set name [Randstr]; incr i }
		{}
	}}
	if {[llength [info commands ${cns}::$name]] > 0} { return -code error "${cns}::$name already exists" }
	set varname [Randvar $name $vns]
	MakeCommands $type $name $varname $cns
	array set $varname $argies
	return ${cns}::$name
}


## Operators

# AitCS / AitCS+

#
proc Vset {varname elem args} {
	upvar 1 $varname _v_
	if {[llength $args] == 0} {
		set _v_($elem)
	} elseif {[llength $args] == 1} {
		set _v_($elem) [lindex $args 0]
	} else {
		array set _v_ [list $elem {*}$args]
		set _v_($elem)
	}
}

# -
proc Vunset {varname elem args} {
	upvar 1 $varname _v_
	foreach e $args { array unset _v_ $e }
	array unset _v_ $elem
}

# &
proc Vref {varname {elem ""}} {
	if {$elem eq ""} { return $varname }
	return ${varname}($elem)
}

# ?
proc Vexists {varname elem {myVar ""}} {
	upvar 1 $varname _v_
	if {![info exists _v_($elem)]} { return 0 }
	if {$myVar eq {}} { return 1 }
	upvar 1 $myVar var
	set var $_v_($elem)
	return 1
}

# ^
proc Vgwd {varname elem {default ""} {myVar ""}} {
	upvar 1 $varname _v_
	if {$myVar ne ""} { upvar 1 $myVar var }
	set var [expr {[info exists _v_($elem)] ? $_v_($elem) : $default}]
}

# ->
proc Vaitcs {varname elem args} {
	upvar 1 $varname _v_
	# Avoid mismatches by checking double-char operators before single-char operators
	if {[set op [string range $elem end-1 end]] in [list += && ->]} {
		set elem [string range $elem 0 end-2]
	} elseif {[set op [string index $elem end]] in [list ? - ^ & = + ! :]} {
		set elem [string range $elem 0 end-1]
	} elseif {[set op [string range $elem end-2 end]] eq "---"} {
		set elem [string range $elem 0 end-3]
	} else {
		set op ""
	}
	tailcall $_v_($elem)$op {*}$args
}

# AitCS+

# =
proc Vsetto {varname elem elemOrVal args} {
	upvar 1 $varname _v_
	set a [list $elem $elemOrVal {*}$args]
	set val [lindex $a end]
	foreach e [lrange $a 0 end-1] { set _v_($e) $val }
	return $val
}

# +
proc Vlappend {varname elem args} {
	upvar 1 $varname _v_
	lappend _v_($elem) {*}$args
}

# +=
proc Vincr {varname elem {incr {1}} args} {
	upvar 1 $varname _v_
	foreach i $args { incr _v_($elem) $i }
	incr _v_($elem) $incr
}


## Setup

if {$::tcl_platform(engine) eq "Jim"} {
	foreach plus [list "" +] type [list snake eagle] {
		proc aitcs$plus {cmd args} [string map [list @plus@ $plus @type@ $type] {
			if {$cmd eq "operators"} {
				tailcall ::aitcs::$cmd@plus@ {*}$args
			} elseif {$cmd eq "new"} {
				tailcall ::aitcs::New @type@ {*}$args
			} else {
				return -code error "unknown subcommand \"$cmd\": must be new, or operators"
			}
		}]
	}
} else {
	foreach plus [list "" +] type [list snake eagle] {
		namespace ensemble create -prefixes no \
			-command aitcs$plus \
			-map [dict create new [list New $type] operators operators$plus]
	}
}


# These things
proc operators  {} { return [list "" ? - ^ & -> ! && ---] }
proc operators+ {} { return [list {*}[operators] = + +=] }


namespace export aitcs aitcs+

package provide aitcs 2.0.1


}; # namespace eval aitcs


# EOF

# THATCH /herbie/packages/gub-1.0a5.tm modbase /herbie/packages
#
# Gub - GUiBuilder
#
# "The world's fastest gui builder."
# "An nroff for Tk."
#
# Stuart Cassoff
#
# Fall/Winter 2008
# Version 0.3
#
# Summer 2009
# Version 0.4
#
# Fall 2011, Winter 2012, Spring 2012
# Version 0.5
#
# Summer 2012, Winter 2013, Spring 2015
# Version 0.6
#
# Spring 2015
# Version 0.7
#
# Spring 2015
# Version 0.8
#
# Fall/Winter/Spring 2018
# Version 0.9
#
# Summer 2022
# Version 1.0a1
#
# Winter 2023
# Version 1.0a2
# Version 1.0a3
# Version 1.0a4
# Version 1.0a5
#


[namespace eval gub {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	gub
		title	Gub
		descr	{Gui Builder}
		author	{Stuart Cassoff}
		when	{Winter 2023}
		version	1.0a5
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2008,2011,2012,2015,2018,2022,2023 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


#---------------------------------------------------------------------
# Package declaration, exports, variables
#---------------------------------------------------------------------

proc provide {} {
	package provide gub 1.0a5
	return on
}

proc require {} {
	package require Tcl 8.5-
	return true
}

proc export {} {
	namespace export {*}{
		gub gubParse gubAssemble
		guhb guhbParse guhbAssemble guhbGubParse
		gubWidgetMap gubTkWidgetMap gubAutoscrollbarWidgetMap
		gubSaveGob gubButtonGob gubOpGob gubPropGob
		gubBindGob gubAddGob gubWrapToggleGob
		gubOpListGob gubMultiGob
		gubGoobKeys gubGoobWidgetKeys
		gubWidgetMapElementKeys
		gubAutoUnderlines
		gubLangs gubTLangs
		gubUA gubParentPath
		gubOffsets
		gubExpandEventKeys
		gubParseArgs
		gubOperators gubOperatorNames
		guhbOperators guhbOperatorNames
		gubSubOperators gubSubOperatorNames
		gubBoilerplate gubBoilerplateHead gubBoilerplateTail gubWrapSource
	}
	return yes
# gubWidgetMapKeys
}

proc initvars {} {
	variable WidgetMap             [gubTkWidgetMap]
	variable GubOperators          [gubOperators]
	variable GubSubOperators       [gubSubOperators]
	variable GubScrollbarSelectors [dict create - H | V]
	return true
}


#---------------------------------------------------------------------
# Info
#---------------------------------------------------------------------

proc guhbOperators {} {
	list >
}

proc guhbOperatorNames {} {
	dict create > in
}

proc gubOperators {} {
	list @ / ^ & - | . : + % ~ !
}

proc gubOperatorNames {} {
	dict create {*}{
		@  label
		/  save
		^  sibling
		&  bind
		-  rowconfigure
		|  columnconfigure
		.  property
		:  grid
		+  add
		%  scrollbarlink
		~  bindtags
		!  alternatefunctionality
	}
}

proc gubSubOperators {} {
	list / : + ! % ~
}

proc gubSubOperatorNames {} {
	dict create {*}{
		/  altsave
		:  grid
		+  add
		!  name
		%  disableautoscrollbarlinking
		~  widgetmap
	}
}

proc gubGoobKeys {} {
	list in widgets adds rccs binds bindtags labels N F E offsets
}

proc gubGoobWidgetKeys {} {
	list widgetchar n path options row col savename gridoptions links
}

if 0 {
proc gubWidgetMapKeys {} {
	list widgetchar widgetmapelement
}
}

proc gubWidgetMapElementKeys {} {
	list command properties
}

proc gubAutoUnderlines {} {
	list -text
}

proc gubTLangs {} { list &Tcl R&uby &Lua &Perl P&ython }
proc  gubLangs {} { list  tcl  ruby  lua  perl  python }


#---------------------------------------------------------------------
# Widget map
#---------------------------------------------------------------------

proc gubTkWidgetMap {} {
	dict create {*}{
		b  {command  tk::button}
		c  {command  tk::canvas		properties scrollable}
		k  {command  tk::checkbutton}
		e  {command  tk::entry}
		f  {command  tk::frame		properties container}
		l  {command  tk::label}
		m  {command  tk::labelframe	properties container}
		o  {command  tk::listbox	properties scrollable}
		u  {command  tk::menubutton}
		p  {command  tk::panedwindow	properties adder}
		r  {command  tk::radiobutton}
		a  {command  tk::scale}
		s  {command  tk::scrollbar	properties scroller}
		i  {command  tk::spinbox}
		t  {command  tk::text		properties scrollable}
		z  {command  tk::toplevel	properties toplevel}
		B  {command ttk::button}
		K  {command ttk::checkbutton}
		X  {command ttk::combobox}
		E  {command ttk::entry}
		F  {command ttk::frame		properties container}
		L  {command ttk::label}
		M  {command ttk::labelframe	properties container}
		U  {command ttk::menubutton}
		N  {command ttk::notebook	properties adder}
		P  {command ttk::panedwindow	properties adder}
		O  {command ttk::progressbar}
		R  {command ttk::radiobutton}
		A  {command ttk::scale}
		S  {command ttk::scrollbar	properties scroller}
		T  {command ttk::separator}
		G  {command ttk::sizegrip}
		I  {command ttk::spinbox}
		V  {command ttk::treeview	properties scrollable}
	}
}

proc gubAutoscrollbarWidgetMap {{wc {y}}} {
	set wc [string tolower $wc]
	set WC [string toupper $wc]
	return [dict create \
		$wc {command ::gub::tk::autoscrollbar  properties scroller} \
		$WC {command ::gub::ttk::autoscrollbar properties scroller} \
	]
}

proc gubWidgetMap {{newWidgetMap {}}} {
	variable WidgetMap
	set oldWidgetMap $WidgetMap
	if {[dict size $newWidgetMap] > 0} { set WidgetMap $newWidgetMap }
	return $oldWidgetMap
}


#---------------------------------------------------------------------
# Gub
#---------------------------------------------------------------------

proc gubGuh {} { dict replace [dict create {*}{
	in		.
	lang		tcl
	varname		gui
	vartype		dict
	setcmd		set
	widgetmap	{}
	offsets		{}
	offsetrow	0
	offsetcol	0
	widgetchar_in	""
	autoul		{!gub!}
	compact		no
	eolChar		\n
	guhb		auto
	gob		""
	rgob		{}
	inline_put	yes
	inline_link	yes
	inline_save	yes
	sections	{create put add link rcc bindtags bind save}
	wrap		no
}] langs [gubLangs]}

proc gubParseArgs {args} {
	set opts    [list -i        -gb        -gh        -lg        -wr \
			  -a        -sc        -vn        -co        -au \
			  -ec       -os        -or        -oc        -iw \
			  -rg       -wm                              -sp \
			                                             -ss \
			                                             -sl]

	set options [list -in       -gob       -guhb      -lang      -wrap           \
			  -array    -setcmd    -varname   -compact   -autounderlines \
			  -eolchar  -offsets   -offsetrow -offsetcol -inwidgetchar   \
			  -readygob -widgetmap                       -separateput    \
			                                             -separatesave   \
			                                             -separatelink]

	set binopts    [list -a     -co      -wr   -sp          -ss           -sl          ]
	set binoptions [list -array -compact -wrap -separateput -separatesave -separatelink]

	set binvals [dict create -a array -co yes -wr yes -sp no -ss no -sl no]

	set elems [dict create -a vartype -co compact -wr wrap -i in -lg lang -vn varname \
		-sc setcmd -wm widgetmap -sp inline_put -ss inline_save -sl inline_link \
		-os offsets -or offsetrow -oc offsetcol \
		-iw widgetchar_in -au autoul -ec eolChar -gh guhb -gb gob -rg rgob]

	set optIs gob
	set xtra -3
	set state opt
	set guh [gubGuh]

	foreach arg $args {
		if {$state in $opts} {
			dict set guh [dict get $elems $state] $arg
			set state opt
		} elseif {[set i [lsearch -exact $binopts $arg]] != -1 || [set i [lsearch -exact $binoptions $arg]] != -1} {
			set arg [lindex $binopts $i]
			dict set guh [dict get $elems $arg] [dict get $binvals $arg]
		} elseif {[set i [lsearch -exact $opts $arg]] != -1 || [set i [lsearch -exact $options $arg]] != -1} {
			set state [lindex $opts $i]
		} else {
			dict set guh $optIs $arg; set optIs [string map [list gob in in varname varname 0 $xtra [incr xtra]] $optIs]
		}
	}
	return $guh
}

proc gub {args} {
	set guh [gubParseArgs {*}$args]
	checkArg $guh guhb
	if {[gobIsGohb [dict get $guh gob]] && [dict get $guh guhb] in [list auto yes]} { tailcall guhbIt $guh }
	dict set guh goob [gubParse $guh]
	set src [dict get [gubAssemble $guh] [dict get $guh lang]]
	return [if {[dict get $guh wrap]} {
		gubWrapSource $src [dict get $guh lang]
	} else {
		set src
	}]
}


#---------------------------------------------------------------------
# Guhb
#---------------------------------------------------------------------

proc getPathForSavename {wigs savename} {
	foreach widget $wigs {
		if {[dgwd {} $widget savename] eq $savename} {
			return [dict get $widget path]
		}
	}
	return {}
}

proc getWigPropertyForPath {wigs path {property {widgetchar}}} {
	foreach widget $wigs {
		if {[dgwd {} $widget path] eq $path} {
			return [dgwd {} $widget $property]
		}
	}
	return {}
}

proc guhbResolveReferences {guh src} {
	set wigs [dict get $guh wigs]
	set labs [dict get $guh labs]
	set refmap [dict create]
	foreach {ref name} [regexp -inline -all {/>(.+?)/} $src] {
		if {[dict exists $refmap $ref]} continue
		if {[set path [getPathForSavename $wigs $name]] eq "" && [set path [dgwd "" $labs $name]] eq ""} continue
		dict set refmap $ref $path
	}
	if {[dict size $refmap] > 0} {
		return [string map $refmap $src]
	} else {
		return $src
	}
}

proc guhbParse {guh} {
	set g [list]
	set in "."
	set rgob [list]
	set args [list]
	# Split up the gobs
	foreach l [prepGob [dict get $guh gohb]] {
		if {[string index $l 0] eq ">"} {
			lappend g $in $rgob $args
			set args [lassign [string trim [string range $l 1 end]] in]
			set rgob [list]
		} else {
			lappend rgob $l
		}
	}
	lappend g $in $rgob $args
	# Process the gobs
	set goohb [list]
	foreach {in rgob args} $g {
		if {[llength $args] == 0} {
			lappend goohb $in [guhbSubI $rgob 0]
			continue
		}
		optsLenMustBeEven guhb $args
		set keys [list]
		set vals [list]
		set state opt
		foreach arg $args {
			if {$state eq "-keys"} {
				if {[llength $arg] > 0} { # Empty list not allowed
					lappend keys $arg
				}
				set state opt
			} elseif {$state eq "-values"} {
				lappend vals $arg
				set state opt
			} elseif {$arg in [list "-keys" "-values"]} {
				set state $arg
			}
		}
		# Ensure there are the same number of -keys and -values
		set d [expr {[llength $keys] - [llength $vals]}]
		if {$d < 0} {
			set vals [lrange $vals 0 end$d]
		} elseif {$d > 0} {
			set keys [lrange $keys 0 end-$d]
		}
		set nPairs [llength $keys]
		if {$nPairs == 0} {
			lappend goohb $in [guhbSubI $rgob 0]
			continue
		}
		set valCounts [dict create]
		set nTodo 0
		for {set i 0} {$i < $nPairs} {incr i} {
			dict set valCounts $i 0
			set nTodo [expr {max($nTodo, int(ceil(1.0*[llength [lindex $vals $i]]/[llength [lindex $keys $i]])))}]
		}
		for {set ghi 0} {$ghi < $nTodo} {incr ghi} {
			set map [list]
			set ki 0
			foreach ks $keys {
				foreach key $ks {
					lappend map $key [lindex $vals $ki [dict get $valCounts $ki]]
					dict incr valCounts $ki
				}
				incr ki
			}
			lappend goohb $in [guhbSubI [string map $map $rgob] $ghi]
		}
	}
	return $goohb
}

proc guhbGubParse {guh} {
	set gooohb [list]
	set in [dict get $guh in]
	set offsets [dict get $guh offsets]
	set rofs [dict create $in [dict get $guh offsetrow]]
	set wigs [list [dict create savename . path $in]]
	set labs [list]
	set gubargs [list \
		-in $in \
		-wm [dict get $guh widgetmap] \
		-vn [dict get $guh varname] \
		-sc [dict get $guh setcmd] \
		-au [dict get $guh autoul] \
	]
	if {[dict get $guh vartype] eq "array"} { lappend gubargs -a  }
	if {[dict get $guh compact]}            { lappend gubargs -co }
	if {![dict get $guh inline_put]}        { lappend gubargs -sp }
	if {![dict get $guh inline_link]}       { lappend gubargs -sl }
	if {![dict get $guh inline_save]}       { lappend gubargs -ss }
	set lang [dict get $guh lang]
	set goohb [dict get $guh goohb]
	set done [list]
	set nTodo [expr {[llength $goohb]/2}]
	set nDone -1
	while {$nDone != 0 && $nDone != $nTodo} { set nDone 0; set i -1; foreach {ghin rgob} $goohb {
		incr i
		if {$i in $done} continue
		if {[set path [getPathForSavename $wigs $ghin]] eq "" && [set path [dgwd "" $labs $ghin]] eq ""} continue
		if {![dict exists $rofs $path]} { dict set rofs $path 0 }
		set iwc [expr {$path eq $in ? [dict get $guh widgetchar_in] : [getWigPropertyForPath $wigs $path widgetchar]}]
		set gguh [gubParseArgs {*}$gubargs -in $path -iw $iwc -offsets $offsets -offsetrow [dict get $rofs $path] -rg $rgob -lg $lang]
		dict set gguh goob [gubParse $gguh]
		lappend gooohb $gguh
		set offsets [dict get $gguh goob offsets]
		lappend wigs {*}[dgwd {} $gguh goob widgets]
		lappend labs {*}[dgwd {} $gguh goob labels]
		dict incr rofs $path
		foreach {path offset} [dict get $gguh goob in_row] {
			dict set rofs $path [expr {$offset + 1}]
		}
		lappend done $i
		incr nDone
	}}
	return [list $gooohb $wigs [dict create {*}$labs]]
}

proc guhbAssemble {guh {langs {}}} {
	if {[llength $langs] == 0} { set langs [dict get $guh langs] }
	set nl [expr {[dict get $guh compact] ? "" : "\n"}]
	set res [dict create tcl "" ruby "" lua "" perl "" python ""]
	foreach gguh [dict get $guh gooohb] {
		set assembled [gubAssemble $gguh]
		foreach lang $langs {
			dict append res $lang [dict get $assembled $lang] \n $nl
		}
	}
	dict for {lang src} $res { dict set res $lang [guhbResolveReferences $guh $src] }
	return $res
}

proc guhbIt {guh} {
	set in [string trimright [dict get $guh in]]
	dict set guh in [expr {$in eq "" ? "." : $in}]
	dict set guh gohb [dict get $guh gob]
	dict unset guh gob
	dict set guh goohb [guhbParse $guh]
	lassign [guhbGubParse $guh] gooohb wigs labs
	dict set guh gooohb $gooohb
	dict set guh wigs $wigs
	dict set guh labs $labs
	return [dict get [guhbAssemble $guh [list [dict get $guh lang]]] [dict get $guh lang]]
}

proc guhb {args} {
	return [guhbIt [gubParseArgs {*}$args]]
}


#---------------------------------------------------------------------
# Parse
#---------------------------------------------------------------------

proc gubParse {guh} {
	variable WidgetMap
	variable GubOperators
	variable GubSubOperators
	variable GubScrollbarSelectors

	# Create as local vars some elements from the guh and adjust as needed
	foreach k [list in widgetmap offsetrow offsetcol widgetchar_in autoul compact gob rgob] {
		set $k [dict get $guh $k]
	}
	if {$in eq ""} { set in [dict get [gubParseArgs] in] }
	if {[llength $autoul] == 1 && [lindex $autoul 0] eq "!gub!" } { set autoul [gubAutoUnderlines] }
	set wm [expr {[dict size $widgetmap] == 0 ? $WidgetMap : $widgetmap}]
	incr offsetrow -1

	# Create dicts
	foreach d {N O P G Q M S L K D E F R C B I H V U Z A X} { set $d [dict create] }

	# N widget counters
	# O widget counter offsets
	# P widget opts
	# G grid opts
	# Q add opts
	# M moveups
	# S saves
	# L labels
	# K reverse-lookup labels
	# D label definitions (unexpanded widgetSpecs)
	# E row offsets
	# F column offsets
	# R rowconfigures
	# C columnconfigures
	# B binds
	# I bindtags
	# H horiscrolls
	# V vertiscrolls
	# U putcmds
	# Z altsaves
	# A names
	# X alternate functionality

	# Initialize widget counters
	foreach wc [dict keys $wm] { dict set N [string tolower $wc] -1 }

	# Bring in offsets
	dict for {wc v} [dict get $guh offsets] {
		if {![dict exists $N $wc]} { return -code error "bad offset \"$wc\"" }
		if {$v < 1} continue
		dict set O $wc $v
	}

	# Setup "in"
	set orig_in $in
	set in [string trimright $in "."]
	#set goob [dict create]
	#dict set goob in [expr {$in eq "" ? "." : $in}]
	set goob_in [expr {$in eq "" ? "." : $in}]

	# Widget map changes
	set wmchgs [dict create]

	# The names of "all"
	set alls [list "" "all" "{}" {""}]

	# Parse: separate layout and process operators

	# Layout lines accumulated here
	set layout [list]

	# Bindings & Bindtags
	set tobind    [list]
	set tobindtag [list]

	# Brace expansion
	set braceMax -1
	set braceNum  0
	set braceParm ""

	# Scrollbar auto-linking
	# "aslH", "aslV": autoScrollLink (H)orizontal, (V)ertical
	set aslH yes
	set aslV yes

	# Use gob if supplied regardless if rgob exists
	if {$gob ne ""} {
		set rgob [prepGob $gob]
	}

	# Run through the lines
	foreach line $rgob {
		# Clean and separate
		set ln [string trimleft $line]
		if {$ln eq ""} explode
		set op [string index $ln 0]
		if {$op eq "#"} explode
		if {$op ni $GubOperators} {
			lappend layout $line
			continue
		}

		# Tabs count as spaces. Tricky.
		set ln [string map [list \t " "] $ln[set ln {}]]
		# name, value
		if {[set i [string first " " $ln]] == -1} {
			set n [string range $ln 1 end]
			set v ""
		} else {
			set n [string range $ln 1 $i-1]
			set v [string range $ln $i+1 end]
		}
		set vv [string trim $v]

		# Bind and break & Not-gub-widget bindings
		if {$op eq "&"} {
			set c [string index $n 0]
			if {$c eq "!"} { # Bind and break
				set bab [expr {[string first " " $vv] == -1 ? "" : ";break"}]
				set n [string range $n 1 end]
			} else {
				set bab ""
			}
			if {[string index $n 0] eq "`"} { # Not-gub-widget bindings
				lappend tobind [string range $n 1 end] [list $vv$bab]
				continue
			}
		}

		# Scrollbar linking
		if {$op eq "%"} {
			if {[string length $n] < 2} continue
			set selector [string index $n 0]
			checkOpt selector $selector selector
			set whichLink [dict get $GubScrollbarSelectors $selector]
			set n [string range $n 1 end]
		}

		# Alternate functionality
		if {$op eq "!"} {
			if {$n ne "" && [set subop [string index $n 0]] in $GubSubOperators} {
				set n [string range $n 1 end]
			} else {
				set subop ""
			}
		}
		if {$op eq "!" && $subop ne ""} {
			if {$subop eq ":"} { #grid
				if {[string index $n 0] in [list - |]} { #row/column offset
					set offsetsVar [string map [list | F - E] [string index $n 0]]
					set n [string range $n 1 end]
					if {$n eq ""} continue
				} else { #putcmd
					set offsetsVar ""
				}
			} elseif {$subop eq "%"} { #disable autoscrollbarlinking
				set selector [string index $n 0]
				if {$selector eq ""} {
					set aslH no
					set aslV no
				} else {
					checkOpt selector $selector selector
					set n [string range $n 1 end]
					set linkDir [dict get $GubScrollbarSelectors $selector]
					set asl$linkDir no
				}
				continue
			} elseif {$subop eq "~"} { #widget map
				if {[set i [string first " " $v]] == -1} {
					set wc $v
					set wp ""
				} else {
					set wc [string range $v 0 $i-1]
					set wp [string range $v $i+1 end]
				}
				if {$n ne "" && $wc ne ""} {
					set nme [dict create command $wc]
					if {[llength $wp] > 0} {
						dict set nme properties $wp
					}
					set wmchgs [dict replace $wmchgs $n $nme]
					set wm [dict merge $wm $wmchgs]
					if {[string tolower $n] ni $N} {
						dict set N [string tolower $n] -1
					}
				}
				continue
			}
		}; # alternatefunctionality if

		if {$n in $alls} { set n "all" } ;# Nothing is everything

		# Label definitions (unexpanded widgetSpecs), needed for brace expansion
		if {$op eq "@"} { foreach q $vv {
			if {$q in $alls} continue ; # "all" is reserved for Gub
			dict set D $q $n
		}}

		# Setup brace expansion
		if {[string index $n 0] eq "\{" && [string index $n end] eq "\}"} {
			set n [string range $n 1 end-1]
			set n [join [lmap q [split $n ","] { if {[dict exists $D $q]} { dict get $D $q } else { set q } }] ","]
			set vv [lassign $vv[set vv {}] braceParm]
			set braceMax [llength $vv]
			set braceCount 0
			set braceNum 0
			# Drive the brace loop with a ranged widgetSpec
			if {[string length $n] == 1} { # Single widgetchar; counter set to # of items
				set n ${n}0-[expr {$braceMax - 1}]
			} elseif {[regexp {^(.)(\d+)-$} $n -> c s]} { # Widgetchar and num with trailing "-"; start counting at num, counter set to # of items
				set n ${c}${s}-[expr {$braceMax - 1 + $s}]
			}
		} else {
			# Reset or else a prior brace expansion will resume if the 
			# number of widgets was less than the brace expansion args length
			set braceMax -1
		}

		# Skip for operators that don't make use of an empty $vv
		if {$vv eq "" && $op in [list . ~]} continue

		# Operators
		foreach n [expand $n[set n {}]] {
			# Brace expansion
			if {$braceNum < $braceMax} {
				set v [gubSubI [list {*}$braceParm [lindex $vv [expr {$braceCount % $braceMax}]]] $braceCount]
				incr braceCount
				if {[incr braceNum] == $braceMax} { set braceNum 0 } ;# Wrap around
			} else {
				set v $vv
			}
			# Operators
			switch -exact -- $op {
			& { dict lappend B $n $v$bab; #bind }
			~ { dict lappend I $n $v; #bindtags }
			. { dict lappend P $n {*}[continueIfLenOdd $v]; #property }
			- { if {$n eq "zall"} continue; dict lappend R $n $v; #rowconfigure }
			| { if {$n eq "zall"} continue; dict lappend C $n $v; #columnconfigure }
			@ { if {$n eq "all" && [llength $v] > 0} continue elseif {[llength $v] == 0} { set v [list {}] }; foreach q $v { if {$q in $alls} { set q "" }; dict lappend L $n $q; dict lappend K $q $n }; #label }
			/ { dict set S $n [expr {[llength $v] == 1 && [llength [lindex $v 0]] > 1 ? [join $v] : $v}]; #save }
			^ { dict set M $n [expr {$v eq "" ? "1" : $v}]; #sibling }
			: { dict set G $n [expr {$v eq "" ? "-sticky nesw" : [continueIfLenOdd $v]}]; #grid }
			+ { dict set Q $n [continueIfLenOdd $v]; #add }
			% { dict set $whichLink $n $v; #scrollbarlinking }
			! { #alternatefunctionality
			          if {$subop eq ":" && $offsetsVar ne ""} { dict set $offsetsVar $n $v; #row/column offset
			    } elseif {$subop eq ":"} { dict set U $n $v; #grid
			    } elseif {$subop eq "+"} { dict set U $n $v; #add
			    } elseif {$subop eq "!"} { dict set A $n $v; #name
			    } elseif {$subop eq "/"} { if {$n eq "!" && $v ne ""} { dict set guh varname $v } else { dict set Z $n $v }; #save
			    } else                   { dict set X $n $v; #alternatefunctionality }
			}
			default {}
			}; # operators switch
		}; # operators loop
	}; # lines loop


	# Process layout; create widget hierarchy, resolve widget options

	set tops     [list "."  ".."    "\\."]
	set topNames [list "in" "parent" "root"]

	set paths [dict create]
	#dict set paths "."   [dict get $goob in]
	dict set paths "."   $goob_in
	dict set paths ".."  [gubParentPath [dict get $paths "."]]
	dict set paths "\\." "."

	set widgets     [dict create]
	set adders      [dict create]
	set auto_adds   [dict create]
	set scrollers   [dict create]
	set scrollables [list]

	set in_x   [dict create]
	set in_row [dict create]
	set in_col [dict create]
	set is_first_child [dict create]

	set initial_in $in

	dict set in_x    -1 $in
	dict set in_row $in $offsetrow
	dict set in_col $in $offsetcol
	dict set is_first_child $in yes
	set y -1
	set container no

	if {$widgetchar_in ne "" && [Is adder [dict get $wm $widgetchar_in]]} {
		dict set adders $initial_in "."
	}

	foreach l $layout {
		incr y
		set x -2
		foreach p [dict values $in_x] {
			dict set in_col $p 0
		}
		dict set in_col $initial_in $offsetcol
		foreach child [dict keys $is_first_child] {
			dict set is_first_child $child yes
		}
		if {[string length $l] % 2 != 0} { append l " " }
		foreach {c m} [split $l ""] {
			incr x 2;
			foreach xp [lsort -integer -decreasing [dict keys $in_x]] { #lreverse can be used?
				if {$x > $xp} {
					set in [dict get $in_x $xp]
					break
				}
			}
			if {[dict exists $wm $c]} {
				# lc widgetchar and num for path et al
				set lc [string tolower $c]
				dict incr N $lc
				set n [dict get $N $lc]
				set ofs_n $n
				if {[dict exists $O $lc]} {
					incr ofs_n [dict get $O $lc]
				}

				# thouse (most-to-least specific):
				#   wnum ?wnumlabel? wchar all
				# those (most-to-least specific):
				#   wnum ?wnumlabel? ?wnam ?wnamlabel?? wchar ?wcharlabel? all
				# these (least-to-most specific):
				#   all ?wcharlabel? wchar ??wnamlabel? wnam? ?wnumlabel? wnum
				set thouse [list $c$n]
				if {$m ne " "} { lappend thouse $c$m }
				lappend thouse $c all
				set those [list]
				foreach what $thouse {
					lappend those $what
					if {[dict exists $L $what]} {
						lappend those {*}[lreverse [dict get $L $what]]
					}
				}
				set these [lreverse $those]

				set ths [list all $c]
				if {$m ne " "} { lappend ths $c$m }
				lappend ths $c$n
				foreach what $ths[set ths {}] {
					if {[dict exists $L $what]} {
						lappend ths {*}[dict get $L $what]
					}
					lappend ths $what
				}
				if {$ths ne $these} explode

				# Flying Teapot
				foreach what $these {
					if {![dict exists $M $what]} continue
					set path_or_level [dict get $M $what]
					if {[dict exists $paths $path_or_level]} {
						set in [dict get $paths $path_or_level]
						set in [join [lrange [split $in "."] 0 end-1] "."]
						set zin $in
						set in [gubParentPath [dict get $paths $path_or_level]]
						if {$in eq "."} { set in "" }
						if {$zin ne $in} {puts !$zin,$in!;set in $zin}
						dict incr in_col $in
					} elseif {[scan $path_or_level %d i] == 1} {
						set pp [gubParentPath $in $i]
						if {$pp eq "."} { set pp "" }
						for {} {$i > 0} {incr i -1} {
							set zin [join [lrange [split $in "."] 0 end-1] "."]
							set in [gubParentPath $in]
							if {$in eq "."} { set in "" }
							if {$zin ne $in} {puts ^$zin,$in^;set in $zin}
						}
						if {$pp ne $in} {puts @$pp,$in@}
						dict incr in_col $in
					}
				}

				# Widget Pathfinder (Fourth Sector ;-)
				set wut ""
				foreach what $those {
					if {[dict exists $A $what]} {
						set wut $what
						break
					}
				}
				if {$wut ne ""} { # A widget by any other path would smell just as sweet
					set wpath [dict get $A $wut]
					if {$wpath eq ""} {
						if {[dict exists $L $what]} {
							set wpath [lindex [dict get $L $what] end]
						}
					}
					if {$wpath eq ""} {
						set wpath {%gwc%%gwi%}
					}
					# Substitutions:
					# %gwi%   Gub Widget Integer (widget counter)
					# %gwii%  Gub Widget Integer Integer (gwi + 1)
					# %gwc%   Gub Widget Command
					# %gwC%   Gub Widget Command (titled)
					# %gwn%   Gub Widget Name ("wNamepaceCommand0")
					# %grs%   Gub Random String (6 char random lowercase string)
					# The "!!" (name) operator can set this up
					if {[string first {%g} $wpath] != -1} {
						set command [dict get $wm $c command]
						set cmd [namespace tail $command]
						set wpath [string map [list \
							%gwi%  $ofs_n \
							%gwii% [expr {$ofs_n + 1}] \
							%gwc%  $cmd \
							%gwC%  [string totitle $cmd] \
							%gwn%  w[makeName $command]$ofs_n \
							%gwa%  $lc \
							%gwA%  $c \
						] $wpath]
					}
					while {[string first {%grs%} $wpath] != -1} {
						set wpath [regsub {%grs%} $wpath [randstr]]
					}
					if {$wpath eq ""} {
						explode
						set wpath "$lc$ofs_n"
					}
					if {[string index $wpath 0] eq "."} {
						set path $wpath
						set in [gubParentPath $wpath]
						if {![dict exists $is_first_child $in]} { dict set is_first_child $in yes }
						if {![dict exists $in_row $in]} { dict set in_row $in 0 }
						if {![dict exists $in_col $in]} { dict set in_col $in 0 }
					} else {
						set path "$in.$wpath"
					}
				} else {
					set path "$in.$lc$ofs_n"
				}
			}; # c in wm if

			if {$c eq " " && $m eq " "} {
				dict incr in_col $in
				continue
			}

			if {[dict get $is_first_child $in]} {
				dict incr in_row $in
				dict set is_first_child $in no
			}

			if {![dict exists $wm $c]} continue

			# Labels
			# Adjust for widgetchar/all and/or empty name.
			foreach what $thouse {
				if {![dict exists $L $what]} continue
				set adjust no
				set names [dict get $L $what]
				foreach name $names[set names {}] {
					if {$name eq ""} {
						if {$what in [list $c "all"]} {
							set name $c$n
						} else {
							set name $what
						}
						set adjust yes
					}
					lappend names $name
				}
				set nms $names
				set names [lmap name [dict get $L $what] {
					if {$name eq ""} {
						if {$what in [list $c "all"]} {
							set name $c$n
						} else {
							set name $what
						}
						set adjust yes
					}
					set name
				}]
				if {$nms ne $names} explode
				if {$what in [list $c "all"]} {
					set what $c$n
					foreach name $names {
						if {[dict exists $L $name] && $what in [dict get $L $name]} continue
						dict lappend L $what $name
						set adjust yes
					}
				} elseif {$adjust} {
					dict set L $what $names
				}
				if {$adjust} { foreach name $names {
					if {[dict exists $K $name] && $what in [dict get $K $name]} continue
					dict lappend K $name $what
				}}
			}

			# Start making widget
			set widget [dict create widgetchar $c n $ofs_n path $path]

			# Paths
			dict set paths $c$n $path
			if {$m ne " "} {
				dict set paths $c$m $path
			}

			# Options, gridoptions, addoptions
			foreach optdct {P G Q} key {options gridoptions addoptions} {
				set opts [dict create]
				set dct [set $optdct]
				if {$optdct eq "P"} { foreach what $these {
					if {[dict exists $dct $what]} {
						foreach {k v} [dict get $dct $what] { dict set opts $k $v }
					}
				}} else { foreach what $those {
					if {[dict exists $dct $what]} {
						foreach {k v} [dict get $dct $what] { dict set opts $k $v }
						break
					}
				}}
				if {[dict size $opts] == 0} continue
				foreach toul $autoul {
					if {![dict exists $opts $toul] || [dict exists $opts -underline]} continue
					set uopt [gubUA [dict get $opts $toul]]
					dict set opts $toul [lindex $uopt 0]
					if {[llength $uopt] > 1} {
						dict set opts "-underline" [lindex $uopt 2]
					}
				}
				dict set widget $key $opts
			}

			# Save
			foreach what $those {
				if {![dict exists $S $what]} continue
				set name [dict get $S $what]
				if {$name eq ""} {
					if {$what in [list $c all]} {
						set name $c$n
					} else {
						set name $what
					}
				}
				dict set widget savename $name
				if {[dict exists $Z $what]} {
					dict set widget altsave 1
				}
				break
			}

			# Putcmd
			foreach what $those {
				if {![dict exists $U $what]} continue
				dict set widget putcmd [dict get $U $what]
				break
			}
if 0 {
			# Row force
			if {[dict exists $widget gridoptions -row]} {
				set row [dict get $widget gridoptions -row]
				if {$row eq {}} {
					set hasrow no
				} elseif {[regexp {^\d+$} $row r]} {
					dict set in_row $in $r
				}
				dict unset widget gridoptions -row
			}

			# Column force
			set what col
			set opt -column
			if {[dict exists $widget gridoptions $opt]} {
				dict set in_$what $in [dict get $widget gridoptions $opt]
				dict unset widget gridoptions $opt
			}
} else {
			# Row and column force
			foreach what {row col} opt {-row -column} {
				if {![dict exists $widget gridoptions $opt]} continue
				dict set in_$what $in [dict get $widget gridoptions $opt]
				dict unset widget gridoptions $opt
			}
}
			# Row offset
			foreach what $these {
				if {![dict exists $E $what]} continue
				set offset [dict get $E $what]
				if {$offset eq {}} { #snap-back
					set l [string length $in]
					set max -1
					foreach k [dict keys $widgets $in.*] {
						if {[string last . $k] != $l} continue
						if {[dict get $widgets $k col] != [dict get $in_col $in]} continue
						if {[dict get $widgets $k row] <= $max} continue
						set max [dict get $widgets $k row]
					}
					dict set in_row $in [incr max]
				} elseif {[string index $offset 0] in [list - +]} { #relative
					dict incr in_row $in $offset
				} else { #absolute or script
					if {[string is integer $offset]} {
						dict set in_row $in $offset
					} else {
						set offset [string map [list "%gr%" [dict get $in_row $in] "%gc%" [dict get $in_col $in]] $offset]
						if {![catch {uplevel #0 $offset} offset]} {
							dict set in_row $in $offset
						}
					}
				}
			}

			# Column offset
			foreach what $these {
				if {![dict exists $F $what]} continue
				set offset [dict get $F $what]
				if {$offset eq {}} { #snap-back
					set l [string length $in]
					set max -1
					foreach k [dict keys $widgets $in.*] {
						if {[string last . $k] != $l} continue
						if {[dict get $widgets $k row] != [dict get $in_row $in]} continue
						if {[dict get $widgets $k col] <= $max} continue
						set max [dict get $widgets $k col]
					}
					dict set in_col $in [incr max]
				} elseif {[string index $offset 0] in [list - +]} { #relative
					dict incr in_col $in $offset
				} else { #absolute or script
					if {[string is integer $offset]} {
						dict set in_col $in $offset
					} else {
						set offset [string map [list "%gr%" [dict get $in_row $in] "%gc%" [dict get $in_col $in]] $offset]
						if {![catch {uplevel #0 $offset} offset]} {
							dict set in_col $in $offset
						}
					}
				}
			}

			# Finalized row and column
			dict set widget row [dict get $in_row $in]
			dict set widget col [dict get $in_col $in]

			# All done, add to the widgets
			dict set widgets [dict get $widget path] $widget

			# Add scrollables and scrollers
			if {[Is scrollable [dict get $wm $c]]} {
				lappend scrollables $c$n $x $y
			}
			if {[Is scroller [dict get $wm $c]]} {
				dict set scrollers $x $y $c$n
			}

			# Add to auto_adds if parent is an adder and widget don't have no putcmd ("" counts as a putcmd).
			if {[dict exists $adders $in] && (![dict exists $widget putcmd] || [dict get $widget putcmd] ne "")} {
#puts "				dict lappend auto_adds [dict get $paths [dict get $adders $in]] [dict get $paths $c$n]"
				dict lappend auto_adds [dict get $paths [dict get $adders $in]] [dict get $paths $c$n]
			}

			# Add Adders
			if {[Is adder [dict get $wm $c]]} {
				dict set adders $path $c$n
			}

			# Binds
			set ztobind $tobind
			set l [list]
			foreach what $these {
				if {[dict exists $X $what]} continue
				if {![dict exists $B $what]} continue
				lappend l [dict get $B $what] $path
			}
			lappend ztobind {*}[lreverse $l]
			foreach what $those {
				if {[dict exists $X $what]} continue
				if {![dict exists $B $what]} continue
				lappend tobind $path [dict get $B $what]
			}
			if {$ztobind ne $tobind} explode

			# Bindtags
			set ztobindtag $tobindtag
			set l [list]
			foreach what $those {
				if {[dict exists $X $what]} break
				if {![dict exists $I $what]} continue
				lappend l [dict get $I $what] $path
				break
			}
			lappend ztobindtag {*}[lreverse $l]
			foreach what $those {
				if {[dict exists $X $what]} break
				if {![dict exists $I $what]} continue
				lappend tobindtag $path [dict get $I $what]
				break
			}
			if {$ztobindtag ne $tobindtag} explode

			# Container ops
			if {[dict exists $in_x $x]} {
				set w [dict get $in_x $x]
				set al $in_x
				foreach i [lsearch -glob -all $al $w*] {
					set ii [lindex $al [incr i -1]]
					dict unset in_row [dict get $in_x $ii]
					dict unset in_col [dict get $in_x $ii]
					dict unset in_x $ii
				}
			}

			# Container override
			if {[set container [Is container [dict get $wm $c]]]} {
				foreach what $those {
					if {[dict exists $X $what]} {
						set container no
						break
					}
				}
			}

			if {$container} {
				set in $path
				dict set in_x $x $in
				dict set in_row $in 0
				dict set in_col $in 0
				dict set is_first_child $in no
			} else {
				dict incr in_col $in
			}
		}; # cm loop
	}; # layout loop

	if {$container} {
		dict unset in_row $in
		dict unset in_col $in
	}

	# Adds
	set adds [dict create]
	dict for {k vv} $auto_adds { foreach v $vv {
		if {[dict exists $adds $k] && [lsearch -index 0 [dict get $adds $k] $v] != -1} continue
		dict lappend adds $k [list $v [dgwd {} $widgets $v addoptions]]
	}}

	# Scrollbars, row/col configures, binds, bindtags

	# Horizontal scrollbar auto-links
	if {$aslH} { foreach {w x y} $scrollables {
		set i [expr {$y + 1}]
		if {![dict exists $scrollers $x $i]} continue
		set s [dict get $scrollers $x $i]
		if {[dict exists $H $s]} continue
		if {![dict exists $paths $s]} continue
		set p [dict get $paths $s]
		set widget [dict get $widgets $p]
		dict set H $s $w
		dict set widgets $p     options [dict replace [dgwd {} $widget     options] -orient horizontal]
		dict set widgets $p gridoptions [dict replace [dgwd {} $widget gridoptions] -sticky ew]
	}}

	# Vertical scrollbar auto-links
	if {$aslV} { foreach {w x y} $scrollables {
		set i [expr {$x + 2}]
		if {![dict exists $scrollers $i $y]} continue
		set s [dict get $scrollers $i $y]
		if {[dict exists $V $s]} continue
		if {![dict exists $paths $s]} continue
		set p [dict get $paths $s]
		set widget [dict get $widgets $p]
		dict set V $s $w
		dict set widgets $p     options [dict replace [dgwd {} $widget     options] -orient vertical]
		dict set widgets $p gridoptions [dict replace [dgwd {} $widget gridoptions] -sticky ns]
	}}

	# Link scrollers (k) and scrollables (v)
	foreach d {H V} a {x y} { dict for {k v} [set $d] {
		foreach q {k v} { set qq [set $q]; if {![dict exists $paths $qq] && [dict exists $K $qq]} { set $q [lindex [dict get $K $qq] end] } }
		if {![dict exists $paths $k] || ![dict exists $paths $v]} continue
		set kp      [dict get $paths   $k ] ; set vp      [dict get $paths   $v ]
		set kwidget [dict get $widgets $kp] ; set vwidget [dict get $widgets $vp]
		dict set widgets $kp links [list {*}[dgwd {} $kwidget links]           -command [dict get $paths $v] ${a}view]
		dict set widgets $vp links [list {*}[dgwd {} $vwidget links] -${a}scrollcommand [dict get $paths $k] set]
	}}

	# Row and column configures
	set rccs [list]
	foreach d {R C} which {row column} { dict for {k v} [set $d] { set ks [expr {![dict exists $paths $k] && [dict exists $K $k] ? [dict get $K $k] : [list $k]}]; foreach o $v { foreach k $ks {
		if {[dict exists $paths $k]} {
			set wl [list [dict get $paths $k]]
		} else {
			set wl [lsort -dictionary [dict values [dict filter $paths key $k*]]]
		}
		if {[llength $wl] == 0} continue
		if {$o eq ""} { set o "0" }
		if {[set i [string first " " $o]] == -1} {
			set oo " -weight 1"
			set i end
		} else {
			set oo [string range $o $i end]
			incr i -1
		}
		set er [expandRange [string range $o 0 $i]]
		if {$compact} {
			foreach w $wl {
				lappend rccs [list $which $w "[list $er]$oo"]
			}
		} else {
			foreach w $wl { foreach o $er {
				lappend rccs [list $which $w "$o$oo"]
			}}
		}
	}}}}

	# Binds & Bindtags

	# Tops and their labels
	set these [list]
	foreach what $tops {
		if {[dict exists $L $what]} {
			lappend these {*}[dict get $L $what]
		}
		lappend these $what
	}

	# Binds & Bindtags for Tops
	foreach d {B I} l {tobind tobindtag} { foreach what $these {
		if {![dict exists [set $d] $what]} continue
		foreach k [expr {![dict exists $paths $what] && [dict exists $K $what] ? [dict get $K $what] : [list $what]}] {
			if {$k ni $tops} continue
			lappend $l [dict get $paths $k] [dict get [set $d] $what]
		}
	}}

	# Binds
	set binds [list]
	foreach {path vv} $tobind {
		foreach v $vv {
			if {[set i [string first " " $v]] == -1} {
				set seq $v
				#set tag [dict get $goob in]
				set tag $goob_in
				if {$seq in $tops} { set seq [dict get $paths $seq] }
			} else {
				set tag [string range $v 0 $i-1]
				set seq [string trim [string range $v $i+1 end]]
				if {$tag in $tops} { set tag [dict get $paths $tag] }
				if {$seq eq ""} {
					set seq $tag
					#set tag [dict get $goob in]
					set tag $goob_in
				}
			}
			if {[string index $tag 0] in [list . /]} {
				set script "$path invoke;break"
				if {[string index $seq 0] in [list . /]} {
					set tag $seq
					set seq ""
				}
			} else {
				set script $seq
				if {$script eq "{}"} { set script {} }
				set seq $tag
				set tag $path
			}
			set auto_ok [expr {[string index $seq end] in [list - +]}]
			if {$auto_ok} {
				if {[string index $seq end-1] eq "\\"} {
					set seq [string cat [string range $seq 0 end-2] [string range $seq end end]]
					set auto_ok no
				}
			}
			set options [dgwd {} $widgets $path options]
			if {    ($seq eq "" || $auto_ok)
			     && [dict exists $options -text]
			     && [dict exists $options -underline]
			} {
				if {$seq eq ""} { set seq "+" } elseif {$seq eq "-"} { set seq "" }
				append seq [string index [dict get $options -text] [dict get $options -underline]]
			}
			if {$seq eq ""} {
				continue
			}
			foreach sequence [gubExpandEventKeys $seq] {
		                lappend binds [list $tag $sequence $script]
			}
		}
	}

	# Bindtags
	set bindtags [list]
	foreach {path vv} $tobindtag { foreach v $vv {
		set tags [lassign $v idx]
		if {[llength $tags] == 0} {
			set tags $idx
			set idx 1
		}
                lappend bindtags [list $path $tags $idx]
	}}

	# Save Tops
	set save_tops [dict create]
	set i -1
	foreach what $tops {
		incr i
		set owhat $what
		set label no
		if {![dict exists $S $what] && [dict exists $L $what]} {
			foreach wut [dict get $L $what] {
				if {![dict exists $S $wut]} continue
				set what $wut
				set label yes
				break
			}
		}
		if {![dict exists $S $what]} continue
		set s [dict get $S $what]
		if {$s eq ""} { set s [expr {$label ? $what : [lindex $topNames $i]}] }
		dict lappend save_tops [dict get $paths $owhat] $s
	}

	# References, labels, refmap
	set refmap [list]
	set labels [dict create]
	dict for {k v} $L {
		if {![dict exists $paths $k]} continue
		foreach vv $v {
			dict lappend labels $vv [dict get $paths $k]
			lappend refmap /$vv/ [dict get $paths $k]
		}
	}
	dict for {k v} $paths {
		lappend refmap /$k/ $v
	}

	# Put it all together
	set goob [dict create in $goob_in]
	if {[dict size $widgets] > 0} {
		dict set goob widgets [dict values $widgets]
	}
	foreach what {rccs binds bindtags} {
		if {[llength [set $what]] > 0} { dict set goob $what [set $what] }
	}
	foreach what {adds labels wmchgs save_tops} {
		if {[dict size [set $what]] > 0} { dict set goob $what [set $what] }
	}
	foreach what {N E F} {
		dict set goob $what [set $what]
	}
	dict set goob offsets [dict create]
	dict for {k v} [dict get $goob N] {
		if {[dict exists $O $k]} { incr v [dict get $O $k] }
		if {$v == -1} continue
		dict set goob offsets $k [incr v]
	}
	dict set goob in_row $in_row
	dict set goob in_col $in_col
	if {$initial_in ne $orig_in} { ; # should only happen with root window?
		# Move any saved values to orig and remove initial
		if {[dict exists $goob in_row $initial_in]} {
			dict   set goob in_row $orig_in [dict get $goob in_row $initial_in]
			dict unset goob in_row $initial_in
		}
		if {[dict exists $goob in_col $initial_in]} {
			dict   set goob in_col $orig_in [dict get $goob in_col $initial_in]
			dict unset goob in_col $initial_in
		}
	}

	return [string map $refmap $goob]
}


#---------------------------------------------------------------------
# Assemble
#---------------------------------------------------------------------

proc gubAssemble {guh} {
	variable WidgetMap

	set wm [expr {[dict size [dict get $guh widgetmap]] > 0 ? [dict get $guh widgetmap] : $WidgetMap}]

	set goob [dict get $guh goob]

	if {[dict exists $goob wmchgs]} {
		set wm [dict merge $wm[set wm {}] [dict get $goob wmchgs]]
	}

	set in [dict get $goob in]

	set inline_put  [dict get $guh inline_put]
	set inline_link [dict get $guh inline_link]

	set langs [dict get $guh langs]
	foreach lang $langs {
		set gen_$lang [dict create]
		foreach e [dict get $guh sections] {
			dict set gen_$lang $e [list]
		}
	}

	foreach e [dict get $guh sections] {
		set g_$e [expr {1 ? $e : "create"}]
	}

	set adds [dict create]
	if {[dict exists $goob adds]} { dict for {w toadds} [dict get $goob adds] {
		foreach toadd $toadds {
			dict set adds [lindex $toadd 0] [list $w [lindex $toadd 1]]
		}
	}}

	set slinks [dict create]
	foreach lang $langs {
		dict set slinks $lang [list]
	}

	set put [dict create]
	set add [dict create]

if 0 {
	set weegs [dict create]
	set chils [dict create]
	if {[dict exists $goob widgets]} { foreach widget [dict get $goob widgets] {
		set path [dict get $widget path]
		dict set weegs $path $widget
		dict lappend chils [gubParentPath $path] $path
	}}
	set order "natural"
#	set order "tree"
	if {$order eq "natural"} {
		set todo [dict keys $weegs]
	} elseif {$order eq "tree"} {
		set stk [list [dict get $guh in]]
		#set stk [dict get $chils [dict get $guh in]]
		set out [list]
		while {[llength $stk] > 0} {
			set w [lindex $stk 0]
			set stk [lrange $stk 1 end]
			lappend out $w
			if {[dict exists $chils $w]} {
				lappend stk {*}[dict get $chils $w]
			}
		}
		set todo $out
	} else {
		return -code error "unknown order: $order"
	}
	foreach weeg $todo {}
}
	if {[dict exists $goob widgets]} { foreach widget [dict get $goob widgets] {
		#if {![dict exists $weegs $weeg]} { continue }
		#set widget  [dict get $weegs $weeg]
		set wc      [dict get $widget widgetchar]
		set path    [dict get $widget path]
		set upath   [d2u $path]
		set ppath   [gubParentPath $path]
		set uppath  [d2u $ppath]
		set command [dict get $wm [dict get $widget widgetchar] command]
		set options [dgwd [list] $widget options]

		set cmds [dict create]
		dict set cmds tcl    $command
		dict set cmds ruby   [tc2rc $command]
		dict set cmds lua    [tc2lc $command]
		dict set cmds perl   [tc2pc $command]
		dict set cmds python [tc2yc $command]

		set ppaths [dict create]
		dict set ppaths tcl    $ppath
		dict set ppaths ruby   [rd2u $ppath]
		dict set ppaths lua    [pine [ed2u $ppath]]
		dict set ppaths perl   [rd2u $ppath]
		dict set ppaths python [rd2u $ppath]

		set paths [dict create]
		dict set paths tcl    $path
		dict set paths ruby   [rd2u $path]
		dict set paths lua    $upath
		dict set paths perl   [rd2u $path]
		dict set paths python [rd2u $path]

		set links [dict create]
		foreach lang $langs {
			dict set links $lang [list]
		}
		set lnks $links
		if {[dict exists $widget links]} { foreach {opt win cmd} [dict get $widget links] {
			set uwin [d2u $win]
			dict lappend links tcl    $opt $win\ $cmd
			dict lappend links ruby   $opt proc\{|*args|$uwin.$cmd\(*args)\}
			dict lappend links lua    $opt $uwin..\"\ $cmd\"
			dict lappend links perl   $opt \[\$$uwin,'$cmd'\]
			dict lappend links python $opt $uwin.$cmd
		}}
		if {$inline_link} {
			set lnks $links
		} else {
			foreach lang $langs {
				if {[llength [set l [dict get $links $lang]]] == 0} { continue }
				dict lappend slinks $lang [dict get $paths $lang] $l
			}
		}

		set opts [dict create]
		dict set opts tcl    [to2to [concat $options [dict get $lnks    tcl]]]
		dict set opts ruby   [to2ro [concat $options [dict get $lnks   ruby]]]
		dict set opts lua    [to2lo [concat $options [dict get $lnks    lua] -path $path] ""]
		dict set opts perl   [to2po [concat $options [dict get $lnks   perl]] {}]
		dict set opts python [to2yo [concat $options [dict get $lnks python]]]

		set wigs [dict create]
		dict set wigs tcl    [dict get $cmds tcl]\ [dict get $paths tcl][dict get $opts tcl]
		dict set wigs ruby   [dict get $cmds ruby]([dict get $ppaths ruby][dict get $opts ruby])
		#dict set wigs lua    ltk.[dict get $cmds lua][dict get $ppaths lua]\{[dict get $opts lua]\}
		dict set wigs lua    ltk.[dict get $cmds lua]\{[dict get $opts lua]\}
		dict set wigs perl   \$[dict get $ppaths perl]->[dict get $cmds perl]([dict get $opts perl])
		dict set wigs python [dict get $cmds python]([dict get $ppaths python][dict get $opts python])

		dict for {lang sv} [gubAssembleSetVar $widget $wigs $guh] {
			set w_$lang [dict get $sv create]
			set    save [dict get $sv save]
			if {$save ne ""} {
				dict lappend gen_$lang ${g_save} $save
			}
		}

		set putcmd [expr {[dict exists $adds $path] ? "add" : "grid"}]
		set putcmd [dgwd $putcmd $widget putcmd]
		if {$inline_put && $putcmd ne ""} {
			set w_tcl "\[$w_tcl\]"
		}

		set putpaths [dict create]
		if {$inline_put} {
			foreach lang $langs {
				dict set putpaths $lang [set w_$lang]
			}
		} else {
			dict set putpaths tcl    $path
			dict set putpaths ruby   [rd2u $path]
			dict set putpaths lua    $upath
			dict set putpaths perl   [rd2u $path]
			dict set putpaths python [rd2u $path]
		}

		foreach lang $langs {
			dict set put $lang ""
			dict set add $lang ""
		}

		if {[dict exists $adds $path]} {
			lassign [dict get $adds $path] add_w add_opts
			set uadd_w [d2u $add_w]
			dict set add tcl    "$add_w $putcmd [dict get $putpaths tcl]"
			dict set add ruby   "[rd2u $add_w].add(($upath=[dict get $putpaths ruby])"
			dict set add lua    "$uadd_w.add\{[dict get $putpaths lua]"
			dict set add perl   "\$[rd2u $add_w]->add\("
			dict set add python "[rd2u $add_w].add("
			if {[Is container [dict get $wm $wc]]} {
				dict append add perl   "my \$$upath=[dict get $putpaths perl]"
				dict append add python "$upath:=[dict get $putpaths python]"
			} else {
				dict append add perl   [dict get $putpaths perl]
				dict append add python [dict get $putpaths python]
			}
			if {$add_opts ne ""} {
				dict append add tcl    [to2to $add_opts]
				dict append add ruby   [to2ro $add_opts]
				dict append add lua    [to2lo $add_opts]
				dict append add perl   [to2po $add_opts]
				dict append add python [to2yo $add_opts]
			}
			dict append add ruby   ")"
			dict append add lua   "\}"
			dict append add perl   ")"
			dict append add python ")"
		} elseif {$putcmd eq ""} {
			if 1 {
			foreach lang $langs {
				dict lappend gen_$lang create [set w_$lang]
				#dict lappend gen_$lang create [set putpath_$lang]
			}
			} else {
			dict lappend gen_tcl    create $w_tcl
			dict lappend gen_ruby   create $w_ruby
			dict lappend gen_lua    create $w_lua
			dict lappend gen_perl   create $w_perl
			dict lappend gen_python create $w_python
			}
		} else {
			set row [dict get $widget row]
			set col [dict get $widget col]
			if {[dict exists $widget gridoptions]} {
				set gridoptions [dict get $widget gridoptions]
			} else {
				set gridoptions ""
			}
			dict set put tcl  "$putcmd [dict get $putpaths tcl][to2to $gridoptions] -row $row -column $col"
			dict set put ruby "($upath=[dict get $putpaths ruby]).$putcmd\(:row=>$row,:column=>$col[to2ro $gridoptions])"
			dict set put lua  "$upath=[dict get $putpaths lua] ltk.$putcmd\{$upath,row=$row,column=$col[to2lo $gridoptions]\}"
			if {1 ||[Is adder [dict get $wm $wc]]} {
				dict set put perl   "(my \$$upath=[dict get $putpaths perl])"
				dict set put python "($upath:=[dict get $putpaths python])"
			} else {
				dict set put perl   [dict get $putpaths perl]
				dict set put python [dict get $putpaths python]
			}
			dict append put perl   "->g_$putcmd\(-row=>$row,-column=>$col[to2po $gridoptions])"
			dict append put python ".$putcmd\(row=$row,column=$col[to2yo $gridoptions])"
		}

		if {$inline_put} {
			set cmds [list ${g_create} ${g_create}]
		} else {
			foreach lang $langs {
				dict lappend gen_$lang ${g_create} [set w_$lang]
			}
			set cmds [list ${g_put} ${g_add}]
		}
		set hows [list put add]
		foreach lang $langs { foreach how $hows cmd $cmds {
			set howd [set $how]
			if {[dict get $howd $lang] ne ""} {
				dict lappend gen_$lang $cmd [dict get $howd $lang]
			}
		}}
	}}; # widgets loop

	if {!$inline_link} {
		foreach {wig opts} [dict get $slinks tcl] {
			#dict lappend gen_tcl ${g_link} "$wig configure [join $opts]"
			dict lappend gen_tcl ${g_link} "$wig configure[to2to $opts]"
		}
		foreach {wig opts} [dict get $slinks ruby] {
			foreach {n v} $opts {
				dict lappend gen_ruby ${g_link} "$wig\['[string range $n 1 end]'\]=$v"
			}
		}
		foreach {wig opts} [dict get $slinks lua] {
			dict lappend gen_lua ${g_link} "ltk.wcmd($wig)\{'configure'[to2lo $opts]\}"
		}
		foreach {wig opts} [dict get $slinks perl] {
			dict lappend gen_perl ${g_link} "\$$wig->configure([to2po $opts {}])"
		}
		foreach {wig opts} [dict get $slinks python] {
			dict lappend gen_python ${g_link} "$wig.configure([to2yo $opts {}])"
		}
	}

	if {[dict exists $goob rccs]} { foreach rcc [dict get $goob rccs] {
		lassign $rcc which window options
		if {$window eq ""} explode
		set w [expr {$window eq "" ? $in : ""}]
		#set w [expr {$window eq "" && $w eq "" ? "." : $w}]
		set wwindow ${w}$window
		set uwwindow [d2u $wwindow]
		#set lua_uwwindow [expr {$uwwindow eq "_" ? "'.'" : $uwwindow}]
		dict lappend gen_tcl    ${g_rcc} "grid ${which}configure ${w}${window} $options"
		dict lappend gen_ruby   ${g_rcc} "TkGrid.${which}configure([rd2u $wwindow],[lindex $options 0][to2ro [lrange $options 1 end]])"
		dict lappend gen_lua    ${g_rcc} "ltk.grid{'${which}configure',${w}${window},'[lindex $options 0]'[to2lo [lrange $options 1 end]]}"
		dict lappend gen_perl   ${g_rcc} "\$[rd2u $wwindow]->g_grid_${which}configure([lindex $options 0][to2po [lrange $options 1 end]])"
		dict lappend gen_python ${g_rcc} "[rd2u $wwindow].${which}configure([lindex $options 0][to2yo [lrange $options 1 end]])"
	}}; # rccs

	if {[dict exists $goob binds]} { foreach binding [dict get $goob binds] {
		lassign $binding tag seq script
		dict lappend gen_tcl    ${g_bind} "bind $tag $seq [braceIfNeeded $script]"
		dict lappend gen_ruby   ${g_bind} "[rd2u $tag].bind('[string range $seq 1 end-1]', $script)"
		dict lappend gen_lua    ${g_bind} "ltk.bind('$tag','$seq',{$script})"
		dict lappend gen_perl   ${g_bind} "\$[rd2u $tag]->g_bind('$seq',sub{[d2u $script]})"
		dict lappend gen_python ${g_bind} "[rd2u $tag].bind('$seq',$script)"
	}}; # binds

	if {[dict exists $goob bindtags]} { foreach bindtag [dict get $goob bindtags] {
		lassign $bindtag window tags idx
		if {[regexp {^-?[0-9]+$} $idx]} {
			if {$idx == -1} { set idx "end" }
			dict lappend gen_tcl    ${g_bindtags} "bindtags $window \[linsert \[bindtags $window\] $idx $tags\]"
#		dict lappend gen_ruby   ${g_bind} "[rd2u $window].bind('[string range $seq 1 end-1]', $script)"
#		dict lappend gen_lua    ${g_bind} "ltk.bind('$window','$seq',{$script})"
#		dict lappend gen_perl   ${g_bind} "\$[rd2u $window]->g_bind('$seq',sub{[d2u $script]})"
			dict lappend gen_python ${g_bindtags} "[rd2u $window].bindtags(bt\[:$idx\]+tags+bt\[$idx:\]))([rd2u $window].bindtags(),('[join $tags "','"]',)))"
		} elseif {$idx == "="} {
			dict lappend gen_tcl    ${g_bindtags} "bindtags $window {$tags}"
			dict lappend gen_python ${g_bindtags} "[rd2u $window].bindtags(('[join $tags "','"]',))"
		} elseif {[string index $idx 0] == "="} {
			set idx [string range $idx 1 end]
			dict lappend gen_tcl    ${g_bindtags} [format {bindtags %s [apply {{bt tags idx} {expr {[set i [lsearch $bt $idx]] == -1 ? $bt : [lreplace $bt $i $i {*}$tags]}}} [bindtags %s] {%s} {%s}]} $window $window $tags $idx]
			dict lappend gen_python ${g_bindtags} "[rd2u $window].bindtags((lambda bt,tags,key:(bt if key not in bt else bt\[:(i:=bt.index(key))-1\]+tags+bt\[i:\]))([rd2u $window].bindtags(),('[join $tags "','"]',),'$idx'))"
		} else {
			set incr 0
			if       {[string index $idx 0] == "-"} {
				set idx [string range $idx 1 end]
			} elseif {[string index $idx 0] == "+"} {
				set idx [string range $idx 1 end]
				incr incr
			}
			dict lappend gen_tcl    ${g_bindtags} [format {bindtags %s [apply {{bt tags idx} {expr {[set i [lsearch $bt $idx]] == -1 ? $bt : [linsert $bt [incr i %d] {*}$tags]}}} [bindtags %s] {%s} {%s}]} $window $incr $window $tags $idx]
			dict lappend gen_python ${g_bindtags} "[rd2u $window].bindtags((lambda bt,tags,key:(bt if key not in bt else bt\[:(i:=bt.index(key)+$incr)\]+tags+bt\[i:\]))([rd2u $window].bindtags(),('[join $tags "','"]',),'$idx'))"
		}
	}}; # bindtags

	# Save Tops
	if {[dict exists $goob save_tops]} {
		set emptyWigs [dict create]
		foreach lang $langs { dict set emptyWigs $lang {} }
		dict for {path names} [dict get $goob save_tops] { foreach name $names {
			dict for {lang sv} [gubAssembleSetVar [dict create path $path savename $name altsave 1] $emptyWigs $guh] {
				dict lappend gen_$lang save [dict get $sv save]
			}
		}}
	}

	# Two ways to achieve the same result
	# Second way kept around for possible future usefulness

	set gen [dict create]
	foreach lang $langs { dict set gen $lang [set gen_$lang] }
	return [gubAssembleScripts $gen $guh]

	#set src [dict create]
	#foreach lang $langs { dict set src $lang [gubAssembleScript [set gen_$lang] $guh $lang] }
	#return $src
}

proc gubAssembleScript {gen guh {lang {}}} {
	if {$lang eq {}} {
		set lang [dict get $guh lang]
	}
	set c  [dict get $guh eolChar]
	set lc ""
	if {$lang eq "perl"} {
		set c  ";$c"
		set lc ";"
	}
	return [join [lmap e [dict get $guh sections] { set l [dict get $gen $e]; if {[llength $l] == 0} continue; join $l $c }] $c]$lc
}

proc gubAssembleScripts {gen guh} {
	set src [dict create]
	set sections [dict get $guh sections]
	foreach lang [dict get $guh langs] {
		set c  [dict get $guh eolChar]
		set lc ""
		if {$lang eq "perl"} {
			set c  ";$c"
			set lc ";"
		}
		dict set src $lang [join [lmap s $sections { set l [dict get $gen $lang $s]; if {[llength $l] == 0} continue; join $l $c }] $c]$lc
	}
	return $src
}

proc gubAssembleSetVar {widget wigs guh} {
	set langs [dict get $guh langs]
	set create [dict create]
	set   save [dict create]
	foreach lang $langs {
		dict set create $lang [dict get $wigs $lang]
		dict set   save $lang ""
	}
	if {![dict exists $widget savename]} {
		set sv [dict create]
		foreach lang $langs {
			dict set sv $lang [dict create create [dict get $create $lang] save [dict get $save $lang]]
		}
		return $sv
	}
	set varname     [dict get $guh varname]
	set vartype     [dict get $guh vartype]
	set setcmd      [dict get $guh setcmd] 
	set inline_save [dict get $guh inline_save]
	set savename    [dict get $widget savename]
	set altsave     [dgwd 0 $widget altsave]
	set path        [dict get $widget path]
	if {$inline_save && !$altsave} {
		dict set create tcl "\[[dict get $create tcl]\]"
	}
	if {$varname eq ""} {
		dict set save tcl    "$setcmd $savename "
		dict set save ruby   "($savename="
		dict set save python "($savename:="
		if {$inline_save && !$altsave} {
			dict append save tcl    [dict get $create tcl]
			dict set  create tcl    [dict get $save tcl]
			dict set    save tcl    ""
			dict append save ruby   [dict get $create ruby])
			dict set  create ruby   [dict get $save ruby]
			dict set    save ruby   ""
			dict append save python [dict get $create python])
			dict set  create python [dict get $save python]
			dict set    save python ""
		} else {
			dict append save tcl    $path
			dict append save ruby   "'$path')"
			dict append save python "'$path')"
		}
	} else {
		if {$vartype eq "dict"} {
			set setcmd_tcl "dict set"
			if {$setcmd ne "set"} {
				set setcmd_tcl $setcmd
			}
			dict set save tcl "$setcmd_tcl $varname $savename"
			if {$inline_save && !$altsave} {
				dict append save tcl " " [dict get $create tcl]
			} else {
				dict append save tcl " " $path
			}
			if {!$altsave} {
				if {(![dict exists $widget putcmd] || [dict get $widget putcmd] ne "") && $inline_save && [dict get $guh inline_put]} {
					dict set save tcl [join [list "dict get" "\[[dict get $save tcl]\]" {*}[lrange [split $varname] 1 end] $savename]]
				}
			}
			if {$inline_save && !$altsave} {
				dict set create tcl [dict get $save tcl]
				dict set save tcl ""
			}
		} elseif {$vartype eq "array"} {
			if {[string first " " $savename] == -1} {
				set lb ""   ; set rb ""
			} else {
				set lb "\{" ; set rb "\}"
			}
			dict set save tcl "$setcmd $lb$varname\($savename)$rb"
			if {$inline_save && !$altsave} {
				dict set create tcl "[dict get $save tcl] [dict get $create tcl]"
				dict set save tcl ""
			} else {
				dict append save tcl " " $path
			}
		} else {
			return -code error "unknown vartype: $vartype"
		}
		dict set save ruby   "($varname\[\"$savename\"\]="
		dict set save lua    "$varname\[\"$savename\"\]="
		dict set save perl   "\$$varname\{'$savename'\}="
		dict set save python "($varname\[\"$savename\"\]:="
		if {$inline_save && !$altsave} {
			dict set create ruby   [dict get $save   ruby][dict get $create   ruby])
			dict set create lua    [dict get $save    lua][dict get $create    lua]
			dict set create perl   [dict get $save   perl][dict get $create   perl]
			dict set create python [dict get $save python][dict get $create python])
			dict set   save ruby   ""
			dict set   save lua    ""
			dict set   save perl   ""
			dict set   save python ""
		} else {
			dict append save ruby   "'$path'"
			dict append save lua    "'$path'"
			dict append save perl   "'$path'"
			dict append save python "'$path'"
		}
	}

	set sv [dict create]
	foreach lang $langs {
		dict set sv $lang [dict create create [dict get $create $lang] save [dict get $save $lang]]
	}

	return $sv
}


#---------------------------------------------------------------------
# Expansion
#---------------------------------------------------------------------

proc looksLikeARange {s} {
	return [expr {[string first "-" $s] > 1 || [string first "*" $s] != -1 || [string first "+" $s] != -1}]
}

proc expandRange {range} {
	set expanded [list 0]
	foreach chunk [split $range ","] {
		if {[set inc [lassign [split $chunk "*"] chunk]] eq ""} { set inc 1 }
		if {[string first "+" $chunk] != -1} {
			lassign [split $chunk "-"] c1 c2
			if {[string index $c1 0] eq "+"} { set c1 [expr {[lindex $expanded end] + [string range $c1 1 end]}] }
			if {[string index $c2 0] eq "+"} { set c2 [expr {$c1+ [string range $c2 1 end]}] }
			set chunk [join [list $c1 {*}$c2] "-"]
		}
		foreach hi [lassign [split $chunk "-"] chunk] {
			for {set lo $chunk; set chunk {}} {$lo <= $hi} {incr lo $inc} { lappend chunk $lo }
		}
		lappend expanded {*}$chunk
	}
	return [lrange $expanded 1 end]
}

proc expandImplied {n} {
	set cen {}
	set wc "" ;# malformed input may cause error below
	if {[string index $n 0] eq "+"} { return $n }
	foreach s [split $n ","] {
		set ch [string index $s 0]
		if {$ch eq "+"} {
			set comma [expr {[llength $cen] == 0 ? "" : ","}]
			lset cen end [lindex $cen end]$comma$s
			continue
		}
		if {[string is digit $ch] || $ch eq "<"} {
			set s $wc$s
		} else {
			set wc [string index $s 0]
		}
		lappend cen $s
	}
	set en {}
	foreach s $cen {
		if {![looksLikeARange $s]} { lappend en $s; continue }
		set wc [string index $s 0]
		foreach i [expandRange [string range $s 1 end]] { lappend en $wc$i }
	}
	return [join $en ","]
}

proc expandLtGt {s} {
	set s ",$s,"
	while {[regexp -indices {,([^,]+?)<(.+?)>,} $s alli nami vali]} {
		lassign $nami l r; set nam [string range $s $l $r]
		lassign $vali l r; set val [string range $s $l $r]
		set n {}
		foreach b [split $val ""] {
			lappend n $nam$b
		}
		lassign $alli l r
		set s [string replace $s [incr l] [incr r -1] [join $n ","]]
	}
	return [string range $s 1 end-1]
}

proc expand {n} {
	return [split [expandLtGt [expandImplied $n]] ","]
}

proc gubExpandEventKeys {event} {
	set n [expr {[string match "<<?*>>" $event] ? 2 : 1}]
	lassign [list [string repeat "<" $n] [string repeat ">" $n]] lb rb
	set event [string trim $event "<>"]
	if {[llength [set l [split $event "+"]]] > 1 && [string length [lindex $l end]] == 1} {
		lassign $l mod key
		if {$mod ne ""} { append mod "-" }
		set f "$lb$mod%s$rb"
		set k [string tolower $key]
		set e [list [format $f $k]]
		if {[set K [string toupper $key]] ne $k} {
			lappend e [format $f $K]
		}
		return $e
	}
	return [list $lb$event$rb]
}


#---------------------------------------------------------------------
# Prep & Check
#---------------------------------------------------------------------

#
# Backslash-Escaped Split
# Split on one char, join backslash-escaped lines with another char
#
proc beSplit {blob splitChar joinChar} {
	set res [list]
	set hold [list]
	foreach line [split $blob $splitChar] {
		if {[regexp {\\+$} $line eolbs] && [string length $eolbs] % 2 != 0} {
			lappend hold [string range $line 0 end-1]
			continue
		}
		if {[llength $hold] > 0} {
			lappend hold $line
			lappend res [join $hold $joinChar]
			set hold [list]
		} else {
			lappend res $line
		}
	}
	if {[llength $hold] > 0} {
		lappend res [join $hold $joinChar]
	}
	return $res
}

proc removeCommentsAndBlanks {lines} {
	return [lmap line $lines {
		set ln [string trim $line]
		if {$ln eq "" || [string index $ln 0] eq "#"} continue
		set line
	}]
}

proc prepGob {gob} {
	set lines [list]
	foreach line [removeCommentsAndBlanks [beSplit $gob "\n" ""]] {
		lappend lines {*}[beSplit $line ";" ";"]
	}
	return [lmap line [removeCommentsAndBlanks $lines] { string trimleft [string trimright $line] "\t" }]
}

proc optsLenMustBeEven {name opts} {
	if {[llength $opts] % 2 == 0} { return $opts }
	return -code error "$name options ($opts) must have an even number of elements"
}

proc continueIfLenOdd {opts} {
	if {[llength $opts] % 2 == 0} { return $opts }
	return -code continue
}

proc checkOpt {name opt {name {option}}} {
	variable GubScrollbarSelectors
	if {	($name eq "guhb"     && $opt ni [set o [list auto yes no]])
	     || ($name eq "selector" && $opt ni [set o [dict keys $GubScrollbarSelectors]])
	} {
		return -code error "unknown $name \"$opt\": must be [oneOf $o]"
	}
}

proc checkArg {guh name} {
	set opt [dict get $guh $name]
	if {$name eq "guhb"} { checkOpt $name $opt }
}

#
# "one of" message
#
proc oneOf {l} {
	if {[llength $l] < 2} { return [join $l] }
	return [strcat "one of " [join [lrange $l 0 end-1] ", "] [expr {[llength $l] > 2 ? "," : ""}] " or " [lindex $l end]]
}


#---------------------------------------------------------------------
# Gub Support
#---------------------------------------------------------------------

proc Is {what widgetmapelement} {
	if {![dict exists $widgetmapelement properties]} { return 0 }
	set is [expr {$what in [dict get $widgetmapelement properties]}]
	# Special case: adders are also containers
	set is [expr {$is || ($what eq "container" && [Is adder $widgetmapelement])}]
	# Special case: toplevels are also containers
	set is [expr {$is || ($what eq "container" && [Is toplevel $widgetmapelement])}]
	return $is
}

proc gobIsGohb {gob} { regexp {(?:^|\n|;)[[:space:]]*>[[:space:]]*?.} $gob }

proc gubSubI {s {i 0} {x ""}} {
	if {[string first "%g" $s] == -1} { return $s }
	# Substitutions:
	# %g?wi%   Gub Widget Integer
	# %g?wii%  Gub Widget Integer Integer (gwi + 1)
	return [string map [list \
		%g${x}i%  $i \
		%g${x}ii% [expr {$i + 1}] \
	] $s]
}

proc guhbSubI {s {i 0}} { gubSubI $s $i h }

proc makeName {cmd} {
	return [join [lmap e [split $cmd ":"] {
		if {$e eq {}} continue
		string totitle $e
	}] ""]
}

proc braceIfNeeded {script} {
	if {[string index $script 0] ne "\{" && [string index $script end] ne "\}"} {
		return "{$script}"
	} else {
		return $script
	}
}

#
# Underline Ampersand
# Convert ampersand in text to -underline option
#
proc gubUA {text} {
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}

if 0 {
#
# Underlined char or "" if none
#
proc gubUlChar {s} {
	set u [gubUA $s]
	if {[llength $u] == 1} {
		return ""
	} else {
		return [string index [lindex $u 0] [lindex $u end]]
	}
}
}

proc gubParentPath {win {level 1}} {
	return [strcat "." [string trimleft [join [lrange [split $win "."] 0 end-$level] "."] "."]]
}

proc gubOffsets {{wpaths {}}} {
	set offsets [dict create]
	foreach w $wpaths {
 		if {![regexp {^(.)(\d+)$} [lindex [split $w "."] end] -> wc wn]} continue
		if {$wn >= [dgwd -1 $offsets $wc]} { dict set offsets $wc [incr wn] }
	}
	return $offsets
}


#---------------------------------------------------------------------
# Misc
#---------------------------------------------------------------------

#
# Dict Get With Default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

#
# Random [a-z] string
#
proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

proc K {x args} { return $x }

#
# 8.5 doesn't have [string cat]
#
proc strcat {args} {
	return [join $args ""]
}

#
# 8.5 doesn't have [lmap]
#
if {[llength [info commands lmap]] == 0} { proc lmap args {
	# 8.5 doesn't have lmap
	# https://wiki.tcl-lang.org/page/lmap+forward+compatibility
	set body [lindex $args end]
	set args [lrange $args 0 end-1]
	set n 0
	set pairs [list]
	foreach {varnames listval} $args {
		set varlist [list]
		foreach varname $varnames {
			upvar 1 $varname var$n
			lappend varlist var$n
			incr n
		}
		lappend pairs $varlist $listval
	}
	set temp [list]
	foreach {*}$pairs {
		lappend temp [uplevel 1 $body]
	}
	return $temp
}}; # lmap

#
# Return the widget after configure.
#
proc gubConfigureWidget {w args} {
	if {[llength $args] > 0} { $w configure {*}$args }
	return $w
}


#---------------------------------------------------------------------
# Lang
#---------------------------------------------------------------------

#
# Dot To Underscore
#
proc d2u {s} {
	return [string map [list . _] $s]
}

#
# 'Root'ed Dot To Underscore
#
proc rd2u {s} {
	return [expr {$s eq "." ? "root" : [d2u $s]}]
}
proc ed2u {s} {
	return [expr {$s eq "." ? "" : [d2u $s]}]
}
proc pine {s} {
	if {$s eq ""} {
		return ""
	} else {
		return "($s)"
	}
}

#
# Tcl Option To Tcl Option
#
proc to2to {options {firstChar { }}} {
	if {[llength $options] == 0} { return "" }
	return $firstChar$options
}

#
# Tcl Option To Ruby Option
#
proc to2ro {options {firstChar {,}}} {
	if {[llength $options] == 0} { return "" }
	return ${firstChar}[join [lmap {n v} $options {
		set q [expr {$n in [list -weight -command -xscrollcommand -yscrollcommand -validatecommand -textvariable -variable] ? "" : "'"}]
		K ":[string range $n 1 end]=>$q$v$q"
	}] ","]
}

#
# Tcl Option To Lua Option
#
proc to2lo {options {firstChar {,}}} {
	if {[llength $options] == 0} { return "" }
	return ${firstChar}[join [lmap {n v} $options {
		set q [expr {$n in [list -weight -command -xscrollcommand -yscrollcommand -validatecommand] ? "" : "'"}]
		K "[string range $n 1 end]=$q$v$q"
	}] ","]
}

#
# Tcl Option To Perl Option
#
proc to2po {options {firstChar {,}}} {
	if {[llength $options] == 0} { return "" }
	return ${firstChar}[join [lmap {n v} $options {
		set q [expr {$n in [list -weight -command -xscrollcommand -yscrollcommand -validatecommand -textvariable -variable] ? "" : "'"}]
		K "$n=>$q$v$q"
	}] ","]
}

#
# Tcl Option To Python Option
#
proc to2yo {options {firstChar {,}}} {
	if {[llength $options] == 0} { return "" }
	return ${firstChar}[join [lmap {n v} $options {
		set q [expr {$n in [list -weight -orient -command -xscrollcommand -yscrollcommand -validatecommand -textvariable -variable] ? "" : "'"}]
		if {$n in [list -orient]} { set v tkinter.[string toupper $v[set v {}]] }
		K "[string range $n 1 end]=$q$v$q"
	}] ","]
}

#
# Tcl Command To Ruby Call
#
proc tc2rc {c} {
	set c [string map {:: .} [string trim $c[set c {}] ":"]]
	set l [split $c "."]
	lset l end [string totitle [lindex $l end]]
	set ns [lindex $l 0]
	if {$ns eq "tk"} {
		set l [lrange $l[set l {}] 1 end]
	} elseif {$ns eq "ttk"} {
		lset l 0 Tile
		if {[lindex $l end] in [list Button]} {
			lset l end T[lindex $l end]
		}
	}
	return Tk::[join $l "::"].new
}

#
# Tcl Command To Lua Call
#
proc tc2lc {c} {
	set c [string map {:: .} [string trim $c[set c {}] ":"]]
	set l [split $c "."]
	set ns [lindex $l 0]
	if {$ns eq "tk"} {
		set l [lrange $l[set l {}] 1 end]
	} elseif {$ns eq "ttk"} {
		set l [lrange $l[set l {}] 1 end]
	}
	return [join $l "moo"]
}

#
# Tcl Command To Perl Call
#
proc tc2pc {c} {
	return new_[string map {: _} $c]
}

#
# Tcl Command To Python Call
#
proc tc2yc {c} {
	set c [string map {:: .} [string trim $c ":"]]
	set l [split $c "."]
	lset l end [string totitle [lindex $l end]]
	set ns [lindex $l 0]
	if {$ns eq "tk"} {
		if {[llength $l] == 2} {
			lset l 1 [string map [list "Labelframe" "LabelFrame" "Panedwindow" "PanedWindow"] [lindex $l 1]]
		}
		lset l 0 tkinter
	} elseif {$ns eq "ttk"} {
		lset l 0 tkinter.ttk
	}
	return [join $l "."]
}

#
# Tcl Namespace To Python Namespace
#
#proc tns2yns {ns} { string map {:: .} [string trim $ns ":"] }

proc gubBoilerplateHead {lang} {
	return [switch -exact -- $lang {
	tcl    { strcat {package require Tk} }
	ruby   { strcat {require 'tk'} "\n" {require 'tkextlib/tile'} "\n" {gui={}} "\n" {root = TkRoot.new()} }
	lua    { strcat {ltk = require('ltk')} \n {root = nil} }
	python { strcat {import tkinter} "\n" {import tkinter.ttk} "\n" {gui={}} "\n" {root = tkinter.Tk()} }
	perl   { strcat {use strict;} "\n" {use Tkx;} "\n" {my @gui;} "\n" {my $root = Tkx::widget->new(".");} }
	default explode
	}]
}

proc gubBoilerplateTail {lang} {
	return [switch -exact -- $lang {
	tcl    { strcat {#vwait forever} }
	ruby   { strcat {Tk.mainloop()} }
	lua    { strcat {ltk.mainloop()} }
	python { strcat {root.mainloop()} }
	perl   { strcat {Tkx::MainLoop();} }
	default explode
	}]
}

proc gubBoilerplate {lang} {
	strcat [gubBoilerplateHead $lang] "\n\n" {#@source@} "\n\n" [gubBoilerplateTail $lang]
}

proc gubWrapSource {src lang} {
	string map [list #@source@ $src] [gubBoilerplate $lang]
}


#---------------------------------------------------------------------
# Gob Procs
#---------------------------------------------------------------------

proc gubSaveGob {widgetChar args} {
	set s ""; set i -1
	foreach a $args { append s / $widgetChar [incr i] " " $a \n }
	return $s
}

proc gubButtonGob {text args} {
	set args [list $text {*}$args]
	set index 0
	set window ""
	set modifiers [list ]
	set buttonchar b
	set gob ""
	set ai -1
	set state opt
	foreach arg $args {
		incr ai
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-first      -
			-window     -
			-modifiers  -
			-buttonchar { set state $arg }
			--          { incr ai; break }
			default     { break }
			}; # hctiws
		}
		-first      { set state opt; set index      $arg }
		-window     { set state opt; set window     $arg }
		-modifiers  { set state opt; set modifiers  $arg }
		-buttonchar { set state opt; set buttonchar $arg }
		}; # hctiws
	}
	set window [string trim $window]
	if {$window eq ""} {
		set window " "
	} else {
		set window " $window "
	}
	incr index -1
	foreach text [lrange $args $ai end] {
		incr index
		set buttonOpts [gubUA $text]
		append gob . ${buttonchar} ${index} " -text " $buttonOpts ";"
		if {[llength $buttonOpts] > 1 && [llength $modifiers] > 0} {
			set key [string index [lindex $buttonOpts 0] [lindex $buttonOpts end]]
			foreach mod $modifiers {
				append gob & ${buttonchar} ${index} ${window} ${mod}+${key} ";"
			}
		}
	}
	return $gob
}

proc gubOpListGob {op todo property properties} {
	set ldiff [expr {[llength $todo] - [llength $properties]}]
	if {$ldiff > 0} { lappend properties {*}[lrepeat $ldiff [lindex $properties end]] }
	set l {}
	set gi 0
	foreach td $todo p $properties {
		lappend l [strcat $op $td " " $property " " [gubSubI $p $gi]]
		incr gi
	}
	return [join $l "\n"]
}

proc gubPropGob {widgetchar property args} {
	set l {}
	set gi 0
	foreach a $args {
		lappend l [strcat "." $widgetchar $gi " " $property " " [gubSubI $a $gi]]
		incr gi
	}
	return [join $l "\n"]
}

proc gubBindGob {w ev z script} {
	set l {}
	set gi 0
	foreach q $z {
		lappend l [strcat "&" $w " " $ev$q " " [gubSubI $script $gi]]
		incr gi
	}
	return [join $l "\n"]
}

proc gubAddGob {to toadd property properties} {
	set l {}
	set gi 0
	foreach ta $toadd p $properties {
		lappend l [strcat "+" $to " " $ta " " $property " " [gubSubI $p $gi]]
		incr gi
	}
	return [join $l "\n"]
}

proc gubWrapToggleGob {w t args} {
	set l {}
	foreach {w t} [list $w $t {*}$args] {
		lappend l [strcat "." $w " " "-command {/$t/ configure -wrap \[expr {\[/$t/ cget -wrap\] eq \"none\" ? \"word\" : \"none\"}\]}"]
	}
	return [join $l "\n"]
}

proc gubMultiGob {n g {pre ""}} {
	set gob $pre
	regexp {^ *} $pre spaces
	if {$pre ne ""} {
		append gob " "
		append spaces "  "
	}
	append gob $g
	append gob "\n"
	for {set i 1} {$i < $n} {incr i} {
		append gob ";$spaces$g" "\n"
	}
	return $gob
}


#---------------------------------------------------------------------
# Autoscrollbar procs
#---------------------------------------------------------------------

proc initas {} { foreach ns {tk ttk} wc {s S} {
	namespace eval $ns [string map [list @wc@ $wc] { if {[llength [info commands autoscrollbar]] == 0} {
		proc autoscrollbar {path args} {
			set cmd [dict get [dict get [[namespace parent]::gubWidgetMap] @wc@] command]
			if {$cmd eq [lindex [info level 0] 0]} {
				set cmd [dict get [dict get [[namespace parent]::gubTkWidgetMap] @wc@] command]
			}
			set w [$cmd $path {*}$args]
			::autoscroll::autoscroll $path
			return $path
		}
	}}]}
	return yes
}


#---------------------------------------------------------------------
# Setup
#---------------------------------------------------------------------

proc setup {} { if {!([require] && [initas] && [initvars] && [export] && [provide])} { error "Something went wrong." } }

namespace current}]::setup


# EOF

# THATCH /herbie/packages/meb-0.5.tm modbase /herbie/packages
#
# Meb - MEnuBuilder
#
# Stuart Cassoff
#
# Spring 2012
# Version 0.1
#
# Summer 2012
# Spring 2015
# Version 0.2
#
# Spring 2015
# Version 0.3
#
# Winter 2018
# Spring 2018
# Version 0.4
#
# Summer 2022
# Version 0.5
#

[namespace eval meb {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	meb
		title	Meb
		descr	{Menu Builder}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.5
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2012,2015,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


#---------------------------------------------------------------------
# Package declaration, exported procs
#---------------------------------------------------------------------

proc provide {} {
	package provide meb 0.5
	return on
}

proc require {} {
	package require Tcl 8.6-
	return true
}

proc export {} {
	namespace export {*}{
		meb
		mebMoobKeys mebItemKeys
		mebLangs mebTLangs
		mebUA
	}
	return yes
}

#---------------------------------------------------------------------
# Meb
#---------------------------------------------------------------------

proc mebMoobKeys {} {
	return [list items script saves]
}

proc mebItemKeys {} {
	return [list autocmd index inName inNumber level options save spaces type]
}

proc mebTLangs {} { list &Tcl R&uby &Lua &Perl P&ython }
proc mebLangs  {} { list  tcl  ruby  lua  perl  python }

proc meb {args} {
	if {[llength $args] == 0} { return "" }
	set ns ::
	set menuname .m
	set names 0
	set autocmd 1
	set menucmd menu
	set menu 1
	set menuopts {}
	set varname gui
	set vartype dict
	set state opt
	set mob [lindex $args end]
	foreach arg [lrange $args 0 end-1] {
		switch -- $state {
		opt {
			switch -- $arg {
			-a  - -array     { set vartype array }
			-n  - -names     { set names   1     }
			-na - -noautocmd { set autocmd 0     }
			-nm - -nomenu    { set menu    0     }
			-mc - -menucmd   { set state -mc     }
			-mn - -menuname  { set state -mn     }
			-mo - -menuopts  { set state -mo     }
			-ns - -namespace { set state -ns     }
			-vn - -varname   { set state -vn     }
			default { error "unknown opt \"$arg\"" }
			}
		}
		-mc { set state opt; set menucmd  $arg }
		-mn { set state opt; set menuname $arg }
		-mo { set state opt; set menuopts $arg }
		-ns { set state opt; set ns       $arg }
		-vn { set state opt; set varname  $arg }
		}
	}
	set ns ::[string trimleft $ns :]
	set ns [string trimright $ns :]::
	set items {}
	array set vars {}
	set spaces {}

	# Parse
	foreach l [split $mob \n] {
		set leadin ""; set rest ""; set space ""; set text ""; set comments ""
		regexp {^([[:space:]]*[;.])?(.*)$} $l -> leadin rest
		regexp {^([[:space:]]*)(.*)$} $rest -> space text
		set text [regsub -all -- {(-[^\s]+?)\s+#([0-9a-fA-F]{6})} $text {\1 `\2}]
		regexp {^(.*?)(#.*)$} $text -> text comments
		set text [regsub -all -- {(-[^\s]+?)\s+`([0-9a-fA-F]{6})} $text {\1 #\2}]
		if {$text eq ""} { continue }
		set space [string trimleft $space \t]
		set text [string trimright $text]
		set type command
		set usertype ""
		set options [dict create]
		set opts {}
		set ol [dict create spaces [string length $space]]

		if {[regexp {^[-=]+$} $text]} {
			set type separator
		} else {
			regexp {^(.+?)[[:space:]]+([-+].*)?$} $text -> text opts
			foreach {n v} [linsert [mebUA $text] 0 -label] {
				dict set options $n $v
			}
		}
		foreach {n v} $opts {
			if {[string index $n 0] eq "+"} {
				switch -exact -- [string range $n 1 end] {
				mo - menuopts - menuoptions {dict set ol menuoptions $v}
				save { dict set ol save $v }
				type { set usertype $v }
				}
				continue
			}
			switch -exact -- $n {
			-value {set type radiobutton}
			-onvalue - -offvalue {set type checkbutton}
			-menu {set type cascade}
			-variable {
				set type checkbutton
				incr vars($v)
			}
			}
			dict set options $n $v
		}
		if {$usertype ne ""} {
			set type $usertype
			dict set ol usertype $usertype
		}
		dict set ol type $type
		dict set ol options $options
		lappend items $ol
	}

	# Auto Radiobuttons
	foreach z $items[set items {}] {
		if {![dict exists $z usertype] && [dict get $z type] eq "checkbutton" \
				&& [dict exists $z options -variable] \
				&& $vars([dict get $z options -variable]) > 1} {
			dict set z type radiobutton
		}
		lappend items $z
	}

	# Levels
	set s {}
	foreach z $items {
		lappend s [dict get $z spaces]
	}
	set i -1
	set l [lsort -integer -unique $s]
	foreach n $l {
		set levs($n) [incr i]
	}
	foreach z $items[set items {}] {
		dict set z level $levs([dict get $z spaces])
		lappend items $z
	}

	# Cascades
	set ritems {}
	set l 0
	foreach z [lreverse $items] {
		if {[dict get $z level] < $l} {
			dict set z type cascade
		}
		set l [dict get $z level]
		lappend ritems $z
	}
	set items [lreverse $ritems]

	set olev 0

	# Paths, numbers, names, indexes, autocommands
	foreach z $items[set items {}] {
		set lev [dict get $z level]
		if {[dict get $z type] eq "cascade"} {
			if {![info exists cnt($lev)]} {
				set cnt($lev) 0
			} else {
				incr cnt($lev)
			}
			foreach e [array names cnt] {
				if {$e > $lev} {
					unset cnt($e)
				}
			}
		}

		while {$olev > $lev} {
			unset idx($olev)
			incr olev -1
		}
		set olev $lev

		if {![info exists idx($lev)]} {
			set idx($lev) 0
		} else {
			incr idx($lev)
		}
		set index $idx($lev)
		set L([dict get $z level]) [string totitle [regsub -all {[[:space:]]} [dgwd {} $z options -label] {}]]
		set inNumber $menuname
		set inName   $menuname
		set ac ""
		for {set i 0} {$i < [dict get $z level]} {incr i} {
			append inNumber .m$cnt($i)
			append inName .m$L($i)
			if {$i == 0} {
				append ac [string tolower $L($i)]
			} else {
				append ac $L($i)
			}
		}
		if {[info exists L($i)]} {
			if {$i == 0} {
				append ac [string tolower $L($i)]
			} else {
				append ac $L($i)
			}
		}
		dict set z inNumber $inNumber
		dict set z inName   $inName
		dict set z autocmd  $ac
		dict set z index    $index

		# Also determine whether or not to create a menu for a cascade
		if {[dict get $z type] eq "cascade"} {
			dict set z addmnu [expr {![dict exists $z options -menu]}]
		}

		lappend items $z
	}

	# Attach cascades
	set ritems {}
	set l 0
	foreach z [lreverse $items] {
		if {[dict get $z type] eq "cascade"} {
			dict set z menuNumber $menuNumber
			dict set z menuName $menuName
		}
		set menuNumber [dict get $z inNumber]
		set menuName [dict get $z inName]
		lappend ritems $z
	}
	set items [lreverse $ritems]

	# Menu, opts
	set menu0 "$menucmd $menuname"
	set addtro 1
	foreach {n v} $menuopts[set menuopts {}] {
		if {$n eq "-tearoff"} {
			set addtro 0
		}
		lappend menuopts $n $v
	}
	if {$addtro} {
		lappend menuopts -tearoff 0
	}
	append menu0 " " $menuopts

	set ruby_menucmd Tk[string totitle $menucmd].new
	set ruby_menu0 [d2u $menuname]=$ruby_menucmd\([to2ro $menuopts]\)

	set lua_menucmd ltk.$menucmd
	set lua_menu0 [d2u $menuname]=$lua_menucmd\{[to2lo $menuopts]\}

	set perl_menucmd new_$menucmd
	set perl_menu0 my\ \$[d2u $menuname]=\$root->$perl_menucmd\([to2po $menuopts])

	set python_menucmd [string totitle $menucmd]
	set python_menu0 [d2u $menuname]=$python_menucmd\(root,[to2yo $menuopts]\)

	# Assemble
	set        o {}
	set   ruby_o {}
	set    lua_o {}
	set   perl_o {}
	set python_o {}
	set int [lindex {inNumber inName} $names]
	set mnt [lindex {menuNumber menuName} $names]
	if {$menu} {
		if {[string length $menu0] > 0} {
			lappend o $menu0
		}
		if {[string length $ruby_menu0] > 0} {
			lappend ruby_o $ruby_menu0
		}
		if {[string length $lua_menu0] > 0} {
			lappend lua_o $lua_menu0
			lappend lua_o [d2u $menuname]_wc=ltk.wcmd([d2u $menuname])
		}
		if {[string length $perl_menu0] > 0} {
			lappend perl_o $perl_menu0
		}
		if {[string length $python_menu0] > 0} {
			lappend python_o $python_menu0
		}
	}
	foreach z $items {
		set type [dict get $z type]
		set zint [dict get $z $int]
		set uzint [d2u $zint]
		set        l   "$zint add $type"
		set    lua_l   "$uzint\_wc\{'add','$type'"
		set   ruby_l   "$uzint.add_$type\("
		set   perl_l "\$$uzint->add_$type\("
		set python_l   "$uzint.add_$type\("
		set addcmd [expr {$autocmd && $type eq "command"}]
		set addtro 1
		set   ruby_options {}
		set    lua_options {}
		set   perl_options {}
		set python_options {}
		dict for {n v} [dict get $z options] {
			if {[regexp {[[:space:]]} $v]} {
				if {$n eq "-command"} {
					set v "\{$v\}"
				}
				if {$n eq "-label"} {
					set v "\{$v\}"
				}
			}
			if {$n eq "-command"} {
				set addcmd 0
				if {[string range $v 0 1] ne "::" && [string index $v 0] ne "\{"} {
					set v "$ns$v"
				}
			}
			if {$n eq "-variable"} {
				if {[string range $v 0 1] ne "::"} {
					set v "$ns$v"
				}
			}

			append l " " $n " " $v

			set cv $v
			if {[regexp {[[:space:]]} $cv]} {
				if {[string index $cv 0] eq "\{" && [string index $cv end] eq "\}"} {
					set cv [string range $cv 1 end-1]
				}
			}
			if {[string index $cv 0] ne "\"" && [string index $cv end] ne "\""} {
				if {$n in [list "-command"]} {
					set cv "\"$cv\""
				}
			}

			if {$n ni [list "-variable" "-command"]} {
				lappend lua_options $n $cv
			}

			if {$n ni [list "-variasble"]} {
				if {$n in [list "-command"]} {
					lappend python_options $n [tns2yns $cv]
				} else {
					#lappend python_options $n $cv
					lappend python_options $n [tns2yns $cv]
				}
			}

			lappend ruby_options $n $cv
			lappend perl_options $n $cv

		}
		if {$addcmd} {
			set ac [dict get $z autocmd]
			append l " " -command " " $ns$ac
			lappend    lua_options -command [tns2yns $ns$ac]
			lappend python_options -command "\"[tns2yns $ns$ac]\""
		}
		if {[dict get $z type] eq "cascade"} {
			set menuopts [dgwd {} $z menuoptions]
			set nmenuopts {}
			set addtro 1
			foreach {n v} $menuopts {
				if {$n eq "-tearoff"} {
					set addtro 0
				}
				if {$n eq "-postcommand"} {
					set addcmd 0
					if {[string range $v 0 1] ne "::" && [string index $v 0] ne "\{"} {
						set v $ns$v
					}
				}
				lappend nmenuopts $n $v
			}
			set menuopts $nmenuopts
			if {$addtro} {
				lappend menuopts -tearoff 0
			}
			if {[dict get $z addmnu]} {
				append l " " -menu " "
				append l "\["
				append l $menucmd " " [dict get $z $mnt]
				append l " " $menuopts
				append l "\]"
				set umnt [d2u [dict get $z $mnt]]
				lappend   ruby_options -menu   $umnt
				lappend    lua_options -menu   $umnt
				lappend   perl_options -menu \$$umnt
				lappend python_options -menu   $umnt
			}
		}
		lappend o $l
		append ruby_l [to2ro $ruby_options]
		append ruby_l "\)"
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend ruby_o [d2u [dict get $z $mnt]]=$ruby_menucmd\([d2u [dict get $z $int]],[to2ro $menuopts]\)
		}
		lappend ruby_o $ruby_l
		if {[llength $lua_options] > 0} {append lua_l "," }
		append lua_l [to2lo $lua_options]
		append lua_l "\}"
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend lua_o [d2u [dict get $z $mnt]]=$lua_menucmd\{[to2lo $menuopts]\}
		}
		lappend lua_o $lua_l
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend lua_o [d2u [dict get $z $mnt]]_wc=ltk.wcmd([d2u [dict get $z $mnt]])
		}
		append perl_l [to2po $perl_options]
		append perl_l "\)"
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend perl_o my\ \$[d2u [dict get $z $mnt]]=\$[d2u [dict get $z $int]]->$perl_menucmd\([to2po $menuopts]\)
		}
		lappend perl_o $perl_l
		append python_l [to2yo $python_options]
		append python_l "\)"
		if {[dict get $z type] eq "cascade" && [dict get $z addmnu]} {
			lappend python_o [d2u [dict get $z $mnt]]=$python_menucmd\([d2u [dict get $z $int]],[to2yo $menuopts]\)
		}
		lappend python_o $python_l
	}

	# Saves
	foreach z $items {
		if {![dict exists $z save]} { continue }
		set save [dict get $z save]
		if {$save eq ""} {
			set save [dict get $z autocmd]
		}
		if {$vartype eq "array"} {
			set s "set ${varname}($save)"
		} else {
			set s "dict set $varname $save"
		}
		append s " "
		append s "\["
		append s "dict create in [set _in [dict get $z $int]] index [set _index [dict get $z index]] configure {$_in entryconfigure $_index}"
		dict set saves [dict get $z $int] [dict create index [dict get $z index]]
		if {[dict get $z type] eq "cascade"} {
			append s " " cascade " " [dict get $z $mnt]
			dict set saves [dict get $z $int] cascade [dict get $z $mnt]
		}
		append s "\]"
		lappend o $s
	}

	set d [dict create items $items]
	dict set d tcl    [dict create src [join $o        \n]]
	dict set d ruby   [dict create src [join $ruby_o   \n]]
	dict set d lua    [dict create src [join $lua_o    \n]]
	dict set d perl   [dict create src [join $perl_o \;\n]\;]
	dict set d python [dict create src [join $python_o \n]]
	if {[info exists saves]} {
		dict set d saves $saves
	}
	return $d
}

#---------------------------------------------------------------------
# Misc
#---------------------------------------------------------------------

#
# Dict Get With Default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

#
# Underline Ampersand
# Convert ampersand in text to -underline option
#
proc mebUA {text} {
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}

#---------------------------------------------------------------------
# Lang
#---------------------------------------------------------------------

#
# Dot To Underscore
#
proc d2u {s} {
	return [string map [list . _] $s]
}

#
# Tcl Option To Ruby Option
#
proc to2ro {options} {
	set z {}
	foreach {n v} $options {
		switch -- $n {
		-command - -tearoff - -underline - -menu { set l ""; set r "" }
		default { set l "'"; set r "'" }
		}
		lappend z ":[string range $n 1 end]=>$l$v$r"
	}
	return [join $z ,]
}

#
# Tcl Option To Lua Option
#
proc to2lo {options} {
	set z {}
	foreach {n v} $options {
		switch -- $n {
		-command - -tearoff - -underline - -menu { set l ""; set r "" }
		default { set l "'"; set r "'" }
		}
		lappend z [string range $n 1 end]=$l$v$r
	}
	return [join $z ,]
}

#
# Tcl Option To Perl Option
#
proc to2po {options} {
	set z {}
	foreach {n v} $options {
		switch -- $n {
		-command - -tearoff - -underline - -menu { set l ""; set r "" }
		default { set l "\""; set r "\"" }
		}
		lappend z $n=>$l$v$r
	}
	return [join $z ,]
}

#
# Tcl Option To Python Option
#
proc to2yo {options} {
	set z {}
	foreach {n v} $options {
		switch -- $n {
		-command - -tearoff - -underline - -menu { set l ""; set r "" }
		default { set l "\""; set r "\"" }
		}
		lappend z [string range $n 1 end]=$l$v$r
	}
	return [join $z ,]
}

#
# Tcl Namespace To Python Namespace
#
proc tns2yns {ns} {
	set q 0
	if {[string index $ns 0] eq "\"" && [string index $ns end] eq "\""} {
		set ns [string range $ns 1 end-1]
		set q 1
	}
	set ns [string map {:: _} [string trim $ns :]]
	if {$q} {
		set ns "\"$ns\""
	}
	return $ns
}

#---------------------------------------------------------------------
# Setup
#---------------------------------------------------------------------

proc setup {} {
	if {![require] || ![export] || ![provide]} { error "Something went wrong." }
}

namespace current}]::setup

# EOF

# THATCH /herbie/packages/tss-0.4.tm modbase /herbie/packages
#
# TSS - Tcl Shell Shell
#
#
# Stuart Cassoff
#
# Fall 2017
# Version 0.1
#
# Winter 2020
# Version 0.2
#
# Fall 2022
# Version 0.3
#
# Spring 2023
# Version 0.4
#


package require Tcl 8.6-


namespace eval tss {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tss
		title	Tss
		descr	{Tcl Shell Shell}
		author	{Stuart Cassoff}
		when	{Spring 2023}
		version	0.4
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig initializer Central_Init
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2020,2022,2023 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


## Tss

oo::class create Tss {

variable Refcount

variable interp
variable cns
variable watching
variable env
variable app

constructor {} {
	namespace import ::tss::twocols
	set Refcount 1
	set interp {}
	set cns "::"
	set watching [dict create]
	set env [dict create]
	set app ""
}
method yourself {} { self }
method destroy args { if {[incr Refcount -1] == 0} { next } }
method makeReference {} { incr Refcount; return [self] }

method interp {} {
	return $interp
}

method app {} {
	return $app
}

method watching {} {
	return $watching
}

method eval {script} {
	return [my evalIn [my cns] $script]
}

method evalIn {where args} {
	return [interp eval [my interp] namespace eval $where "{[join $args]}"]
}
if 0 {
method ArgsPrep {argies} {
	lmap arg $argies {
		if {[string first " " $arg] != -1} {
			list $arg
		} else {
			set arg
		}
	}
}
}

method cns {{ns {}}} {
	if {$ns ne {}} { set cns $ns }
	return $cns
}
method pwd {} { return $cns }

method getns {where} {
	set nns $cns
	set where [string trim $where]
	if {$where eq "."} { return "" }
	if {$where eq "::"} {
		set nns $where
		return $nns
	}
	if {$where eq ".." || $where eq "../"} {
		set nns [namespace qualifiers $nns]
		if {$nns eq ""} {
			set nns "::"
		}
		return $nns
	}
	if {[string range $where 0 1] eq ".."} {
		set where [namespace qualifiers $nns]/[string range $where 2 end]
	}
	set where [string map {:: /} $where]
	set l [list]
	foreach w [split $where /] {
		if {$w eq ".."} {
			set l [lrange $l 0 end-1]
		} else {
			lappend l $w
		}
	}
	set where [join $l ::]
	if {[string index $where 0] ne ":"} {
		set ns $nns\::$where
	} else {
		set ns $where
	}

	return $ns
}

method cd {{where "::"}} {
	set ns [my getns $where]
	if {[my evalIn :: namespace exists $ns]} {
		set cns [my evalIn $ns namespace current]
	}
	return [my pwd]
}

method glob {{pat "*"} {where ""}} {
	set nspat $pat
	if {$nspat eq "*"} { set nspat "" }
	if {$where eq ""} { set where $cns }
	set vs [my evalIn $where info vars     $pat]
	set cs [my evalIn $where info commands $pat]
	set ps [my evalIn $where info procs    $pat]
	set vg [my evalIn ::     info vars     $pat]
	set cg [my evalIn ::     info commands $pat]
	set ns [my evalIn $where namespace children $where {*}$nspat]

	set l [list]
	foreach v $vs {
		if {[my evalIn $where array exists $v]} {
			set t arr
		} else {
			set t var
		}
		if {$v in $vg} {
			set t [string totitle $t]
		}
		lappend l [list $t $v]
	}
	foreach c $cs {
		if {$c in $ps} {
			if {[my evalIn $where info object isa object [list $c]]} {
				set t obj
			} else {
				set t proc
			}
		} elseif {[string index $c 0] eq "."} {
			set t win
		} else {
			if {[my evalIn $where info object isa object [list $c]]} {
				set t obj
			} else {
				set t cmd
			}
			#set s 0
		}
		if {$c in $cg} {
			set t [string totitle $t]
		}

		set t [string map {Win win Obj obj} $t]

		lappend l [list $t $c]
	}

	foreach n $ns {
		set n [namespace tail $n]
		set t ns
		#set s 0
		lappend l [list $t $n]
	}

	set l [lsort -dictionary -index 1 $l]
	set l [lsort -dictionary -index 0 $l]

	return $l
}
method ls {args} {
	set rtypes [dict create {*}[lreverse [::tss::tsstypes]]]
	set filter [dict values $rtypes]
	set where [list]
	if {[llength $args] > 0} {
		set tfi 0
		set arg [lindex $args 0]
		if {[string index $arg 0] eq "-"} {
			set filter [lrange [split $arg ""] 1 end]
			set tfi 1
		}

		foreach arg [lrange $args $tfi end] {
			#lappend where [list * [my getns $arg]]
			lappend where [list $arg ""]
		}
	}
puts $where
	if {[llength $where] == 0} {
		lappend where [list "*" $cns]
	}
	set l1 [list]
	set l2 [list]
	foreach w $where { foreach l [my glob {*}$w] { if {[dict get $rtypes [lindex $l 0]] in $filter} {
		lappend l1 [lindex $l 0]
		lappend l2 [lindex $l 1]
	}}}
	return [twocols $l1 $l2]
}
method cat {args} {
	if {[llength $args] < 1 ||[llength $args] > 2} { return "usage: cat ?-type? thing" }
	if {[llength $args] == 1} {
		set what [lindex $args 0]
		set type ""
	} else {
		set what [lindex $args 1]
		set type [lindex $args 0]
	}

	set where $cns

#	set what [my getns $what]

	set tmap [list isvar v isprc p isobj o ismtd m iswin w isnse n isdct d isbdg b]

	set vars [list isvar iscmd isobj ismtd isprc iswin isnse isdct isbdg]

	foreach v $vars { set $v no; set ck_$v yes }

	if {[regexp {^(.+)::>::(.+)$} $what -> obj mtd]} {
		foreach v $vars { set ck_$v no }
		set what [list $obj $mtd]
		set ismtd yes
	}

	set todo [list]
	if {$type ne ""} {
		foreach v $vars { set ck_$v no }
		set char [string range $type 1 end]
		if {$char eq "a"} { set char v }
		set rtmap [lreverse $tmap]
		if {[dict exists $rtmap $char]} {
			set ck_[dict get $rtmap $char] yes
		}
	}

	if {$ck_isnse} { set isnse [my evalIn $where namespace ensemble exists $what] }
	if {$ck_isvar} { set isvar [my evalIn $where info exists $what] }
	if {$ck_isbdg} { set isbdg [expr {[my evalIn $where bind $what] ne {}}] }
	if {$ck_iscmd || $ck_isprc} { set iscmd [expr {[llength [my evalIn $where info commands $what]] == 1}] }
	if {$iscmd} {
		set isobj [my evalIn $where info object isa object $what]
		set iscmd [expr {!$isobj}]
	}
	if {$iscmd} {
		set isprc [expr {[llength [my evalIn $where info procs $what]] == 1}]
		set iscmd [expr {!$isprc}]
	} else {
		set isprc no
	}
	if {$iscmd || $ck_iswin} {
		set hastk [expr {[llength [my evalIn $where info commands winfo]] == 1}]
		if {$hastk} {
			set iswin [my evalIn $where winfo exists $what]
			set iscmd [expr {!$iswin}]
		}
	}

	set todo [lmap {name char} $tmap { if {![set $name]} continue; set char }]

	if {[llength $todo] == 0} {
		set ret ""
	} else {
		set ret [join [lmap do $todo { my ${do}cat $where $what }] \n]
	}

	return $ret
}
method catsort {args} {
	return [join [lsort [split [my cat {*}$args] \n]] \n]
}
method pcat {where what} {
	set body [my evalIn $where info body $what]
	set args [my evalIn $where info args $what]
	set argies [list]
	foreach arg $args {
		#set myvar [namespace current]::plop
		set myvar ::_default
		if {[my evalIn $where info default $what $arg $myvar]} {
			lappend argies "{$arg {[my evalIn $where set $myvar]}}"
		} else {
			lappend argies $arg
		}
		unset -nocomplain $myvar
	}
	return [string cat "proc " $what " {" [join $argies] "} {" $body "}"]
}
method mcat {where what} { tailcall my omcat $where $what }
method omcat {where what} {
#	lassign $what oobj omtd
	lassign $what  obj  mtd
if {$mtd eq "destroy"} return ""
	if {![my evalIn $where info object isa object $obj]} { return "" }
	foreach op [list call definition] {
		set class [my evalIn $where info object isa class $obj]
		set thing [expr {$class ? "class" : "object"}]
		set thang [expr {$class ? ""      : "obj"   }]
		set res [my evalIn $where info $thing $op $obj $mtd]
		if {$op eq "call"} {
			lassign [lindex $res 0] impl name obj type
		} elseif {$op eq "definition"} {
			lassign $res argies body
		} else { explode }
	}
#	lassign [lindex [my evalIn $where info $thing call $obj $mtd] 0] _ mtd obj
#	return [list "::oo::${thang}define" $obj method $mtd [join $argies] $body]
	return [string cat "::oo::${thang}define " $obj " method " $mtd " {" [join $argies] "} {" $body "}"]
#	return [format {::oo::%sdefine %s method %s {%s} {%s}} $thang $obj $mtd [join $argies] $body]
}
method ovcat {where what} {
	lassign $what obj var
	if {![my evalIn $where info object isa object $obj]} { return "" }
	set class [my evalIn $where info object isa class $obj]
	set thing [expr {$class ? "class" : "object"}]
	set thang [expr {$class ? ""      : "obj"   }]
#	set res [my evalIn $where info $thing variable $obj $var]
#	return $res
	return [string cat "::oo::${thang}define " $obj " variable " $var]
}
method oncat {where what} {
	return [string cat namespace " " $what]
}
method ofcat {where what} {
	return [string cat filter " " $what]
}
method occat {where what} {
	return [string cat class " " $what]
}
method ocat {where what} {
	set type [expr {[my evalIn $where info object isa class $what] ? "class" : "object"}]
	set toshow [dict create]
	if {$type eq "class"} {
		set things {filters variables methods}
	} else {
		set things {namespace class filters vars methods}
	}
	set opts [lrepeat [llength $things] {}]
	lset opts end [linsert [list [lindex $opts end]] end -all]
	foreach thing $things opt $opts {
		dict set toshow $thing [my evalIn $where info $type $thing $what {*}$opt]
	}
	set l [lmap {thing blob} $toshow {
		set m o[string index $thing 0]cat
		if {$m ni [info object methods [self] -all]} continue
		join [lmap b [split $blob] {
			my $m $where [list $what $b]
		}] \n
	}]
	return [join $l \n]
}
method wcat {where what} {
	set  names [list]
	set values [list]
	set  abbrs [dict create]
	foreach opts [my evalIn $where $what configure] {
		if {[llength $opts] == 2} {
			dict set abbrs {*}[lreverse $opts]
		} else {
			lappend  names [lindex $opts 0]
			lappend values [lindex $opts 4]
		}
	}
	dict for {name abbr} $abbrs {
		set i [lsearch -exact $names $name]
		append name , [dict get $abbrs $name]
		lset names $i $name
	}
	return [twocols $names $values]
}
method vcat {where what} {
	if {[my evalIn $where array exists $what]} {
		set d [dict create {*}[my evalIn $where array get $what]]
		set res [twocols [dict keys $d] [dict values $d]]
	} else {
		set res [my evalIn $where set $what]
	}
	return $res
}
method ncat {where what} {
	set d [dict create {*}[my evalIn $where namespace ensemble configure $what]]
	return [twocols [dict keys $d] [dict values $d]]
}
method dcat {where what} {
	set d [my evalIn $where set $what]
	return [twocols [dict keys $d] [dict values $d]]
}
method bcat {where what} {
	set q ""
	dict for {k v} [my bindings $what] {
		append q $k \n
		dict for {b s} $v {
			append q "  " [format "%-30s" $b] " " [string trim $s] \n
		}
	}
	return $q
}
method env {args} {
	if {[llength $args] == 0} {
		return [twocols [dict keys $env] [dict values $env]]
	}
	if {[llength $args] == 1} {
		return [::tss::dgwd "" $env [lindex $args 0]]
	}
	if {[llength $args] == 2} {
		set k [lindex $args 0]
		dict set env $k [lindex $args 1]
		return [my env $k]
	}
	return {}
}
method env- {args} {
	foreach arg $args {
		dict unset env $arg
	}
}
method setwatch {what} {
	my eval [list interp alias [my interp] _watcher {} [self] watcher]
	my eval [list trace add variable $what [list write unset] _watcher]
}
method unsetwatch {what} {
	my eval [list trace remove variable $what [list write unset] _watcher]
#	my eval [list interp alias [my interp] _watcher {}]
}
method watch {what who} {
	if {![dict exists $watching $what]} {
		my setwatch $what
		dict set watching $what [list]
	}
	if {$who ni [dict get $watching $what]} {
		dict lappend watching $what $who
	}
}
method unwatch {what who} {
	if {![dict exists $watching $what]} { return }
	set wl [dict get $watching $what]
	set wi [lsearch -exact $wl $who]
	set wl [lreplace $wl $wi $wi]
	if {[llength $wl] > 0} {
		dict set watching $what $wl
		return
	}
	my unsetwatch $what
	dict unset watching $what
}
method watcher {name1 name2 op} {
	set name1 [my eval "namespace which -variable {$name1}"]
	set variable $name1
	if {$name2 ne ""} {
		append variable "(" $name2 ")"
	}
	set whom [list]
	if {[dict exists $watching $variable]} {
		lappend whom {*}[dict get $watching $variable]
	}
	if {$name1 ne $variable && [dict exists $watching $name1]} {
		lappend whom {*}[dict get $watching $name1]
	}
#	if {[llength $whom] == 0} { return }
	foreach who $whom {
		uplevel #0 $who [list $name1 $name2 $op]
	}
}
method interpLabel {} {
	return ""
}
method bindings {tag} {
	set d [dict create]
	set l [expr {[string index $tag 0] eq "." ? [bindtags $tag] : [list $tag]}]
	foreach t $l { foreach b [bind $t] {
		dict set d $t $b [bind $t $b]
	}}
	return $d
}
}; # oo::class create Tss



oo::class create TssTk {
	superclass Tss
	variable app
constructor {anApp} {
	next
	set app $anApp
}
method evalIn {where args} {
	return [send $app eval [list namespace eval $where $args]]
}
method makewatchblob {what traceop} {
return [string map [list @what@ $what @myname@ [tk appname] @self@ [self] @traceop@ $traceop] {
trace @traceop@ variable @what@ {write} {
::apply {{name1 name2 op} {
catch [list send {@myname@} @self@ watcher [namespace which -variable $name1] [list $name2] $op]
}
}}}]
}
method setwatch {what} {
	my eval [my makewatchblob $what add]
}
method unsetwatch {what} {
	my eval [my makewatchblob $what remove]
#	my eval [list catch trace remove variable $what {write} _watcher]
#	my eval [list interp alias [my interp] _watcher {}]
}
}; # oo::class create TssTk


oo::class create TssComm {
	superclass Tss
	variable app
constructor {anApp} {
	next
	set app $anApp
}
method evalIn {where args} {
	return [::comm::comm send $app eval [list namespace eval $where $args]]
#	return [::comm::comm send $app [list namespace eval $where $args]]
	return [send $app eval [list namespace eval $where [my ArgPrep $args]]]

}
method makewatchblob {what traceop} {
return [string map [list @what@ $what @myid@ [::comm::comm self] @self@ [self] @traceop@ $traceop] {
trace @traceop@ variable @what@ {write} {
::apply {{name1 name2 op} {
catch [list ::comm::send {@myname@} @self@ watcher [namespace which -variable $name1] [list $name2] $op]
}
}}}]
}
method setwatch {what} {
	my eval [my makewatchblob $what add]
}
method unsetwatch {what} {
	my eval [my makewatchblob $what remove]
#	my eval [list catch trace remove variable $what {write} _watcher]
#	my eval [list interp alias [my interp] _watcher {}]
}
}; # oo::class create TssComm


oo::class create TssDBus {
	superclass Tss
	variable app
	variable bus
constructor {anApp} {
	next
	set app $anApp
	dbus connect
}
method evalIn {where args} {
	set tout 1000
	set bus session
	set obj org.tcllang.debug
	set path /
	return [dbus call $bus -dest $app $path $obj Eval [list namespace eval $where $args]]
}
}; # oo::class create TssDBus


oo::class create TssSocket {
	superclass Tss
	variable app
	variable host
	variable port
constructor {anApp} {
	next
	set app $anApp
	lassign [split $anApp :] host port
}
method evalIn {where args} {
	set s [socket $host $port]
	puts $s [list namespace eval $where {*}$args]
	flush $s
	close $s w
	set res [read $s]
	close $s
	return $res
	return [send $app eval [list namespace eval $where [my ArgPrep $args]]]

}
method makewatchblob {what traceop} {
return [string map [list @what@ $what @myid@ [::comm::comm self] @self@ [self] @traceop@ $traceop] {
trace @traceop@ variable @what@ {write} {
::apply {{name1 name2 op} {
catch [list puts {@myname@} @self@ watcher [namespace which -variable $name1] [list $name2] $op]
}
}}}]
}
method setwatch {what} {
	my eval [my makewatchblob $what add]
}
method unsetwatch {what} {
	my eval [my makewatchblob $what remove]
#	my eval [list catch trace remove variable $what {write} _watcher]
#	my eval [list interp alias [my interp] _watcher {}]
}
}; # oo::class create TssSocket



proc tsscmds {} { return [list {*}{
	cd
	pwd
	ls
	cat
	catsort
	env
	env-
}]}

proc tsstypes {} { return [dict create {*}{
	n ns
	v var
	a arr
	p proc
	c cmd
	o obj
	w win
	b bdg
	V Var
	A Arr
	P Proc
	C Cmd
	O Obj
	W Win
}]}

proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

proc twocols {l1 l2} {
	set max 0
	foreach e $l1 { set max [::tcl::mathfunc::max $max [string length $e]] }
	set l [list]
	return [join [lmap e1 $l1 e2 $l2 { format "%-${max}s %s" $e1 $e2 }] \n]
}


## Setup

namespace export Tss TssTk TssComm TssSocket TssDBus twocols

package provide tss 0.4


}; # namespace eval tss


# EOF

# THATCH /herbie/packages/tu/combola-0.4.tm modbase /herbie/packages
#
# Tu Combola - Tcl Usefulness Combobox Enhancements
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#
# Fall 2022
# Version 0.2
#
# Winter 2023
# Version 0.3
# Version 0.4
#

package require Tcl 8.5-

namespace eval tu { namespace eval combola {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::combola
		title	{Tu Combola}
		descr	{Combobox Enhancements}
		author	{Stuart Cassoff}
		when	{Winter 2023}
		version	0.4
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022,2023 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


### Procedures copied from other places

namespace eval tu {
namespace eval tcl {
proc getNextOrPrevMatch {aList aString {curIdx 0} {backwards no}} {
	set ixs [lsearch -all -nocase $aList $aString*]
	if {[llength $ixs] == 0} { return -1 }
	if {[llength $ixs] == 1 && [lindex $ixs 0] == $curIdx} { return -1 }
	set p -1
	foreach ix $ixs {
		if {$ix < $curIdx} { set p $ix }
		if {$ix > $curIdx} break
	}
	if {$backwards} {
		if {$p == -1} {
			set idx [lindex $ixs end]
		} else {
			set idx $p
		}	
	} else {
		if {$ix == $curIdx} {
			set idx [lindex $ixs 0]
		} else {
			set idx $ix
		}
	}
	return $idx
}

proc whereBe {where curIdx nitems {height {10}}} {
	set idx $curIdx
	switch -exact -- $where {
	home     {  set idx  0       }
	end      {  set idx  $nitems }
	next     { incr idx  1       }
	prev     { incr idx -1       }
	nextpage { incr idx  $height ; if {$idx > $nitems} { set idx $nitems } }
	prevpage { incr idx -$height ; if {$idx < 0      } { set idx 0       } }
	default  {  set idx  $where  }
	}
	if {$idx < 0 || $idx > $nitems || $idx == $curIdx} { set idx -1 }
	return $idx
}

namespace export getNextOrPrevMatch whereBe
}; # namespace eval tcl

namespace eval tk {
proc bindAndBreak {w args} {
	dict for {seq script} $args {
		if {$script eq ""} {
			bind $w $seq break
		} elseif {$script in [list break continue]} {
			bind $w $seq $script
		} else {
			bind $w $seq $script
			bind $w $seq +break
		}
	}
}

namespace export bindAndBreak
}; # namespace eval tk
}; # namespace eval tu


### Combola

## Combo

proc comboGoto {combo idx} {
	$combo current $idx
	event generate $combo <<SelectAll>>
	event generate $combo <<ComboboxSelected>> -when tail
}

proc combolistGoto {combolist idx} {
	$combolist activate $idx
	$combolist see $idx
	event generate $combolist <<Invoke>> -when tail
}

proc comboGo {combo where} {
	set curIdx [$combo current]
	set nitems [expr {[llength [$combo cget -values]] - 1}]
	set height 10
	if {[set idx [whereBe $where $curIdx $nitems $height]] >= 0} {
		comboGoto $combo $idx
	}
}

proc combolistGo {combolist where} {
	set curIdx [CombolistCurrent $combolist]
	set nitems [expr {[$combolist index end] - 1}]
	set height [$combolist cget -height]
	if {[set idx [whereBe $where $curIdx $nitems $height]] >= 0} {
		combolistGoto $combolist $idx
	}
}

proc comboKey {combo key} {
	if {[string length $key] != 1} return
	set values [$combo cget -values]
	set curIdx [$combo current]
	set backwards [string is upper $key]
	if {[set idx [getNextOrPrevMatch $values $key $curIdx $backwards]] >= 0} {
		comboGoto $combo $idx
	}
}

proc combolistKey {combolist key} {
	if {[string length $key] != 1} return
	set values [[::ttk::combobox::LBMaster $combolist] cget -values]
	set curIdx [CombolistCurrent $combolist]
	set backwards [string is upper $key]
	if {[set idx [getNextOrPrevMatch $values $key $curIdx $backwards]] >= 0} {
		combolistGoto $combolist $idx
	}
}

proc combolaBeginEdit {combo} {
	set curIdx [$combo current]
	set ns [namespace current]:
	set bt [bindtags $combo]
	set tag [MkTag $combo EditStop]
	set ix [lsearch -exact $bt CombolaKeySelectTCombobox]
	if {$ix == -1} { set ix [lsearch -exact $bt CombolaTCombobox] }
	set comboi [dict create lastIdx $curIdx combo $combo bindtags $bt tag $tag]
	bind $tag <<CombolaAcceptEdit>>  [list $ns:combolaEndEdit [dict replace $comboi eres accept]]
	bind $tag <<CombolaDiscardEdit>> [list $ns:combolaEndEdit [dict replace $comboi eres discard]]
	bind $tag <<CombolaCancelEdit>>  [list $ns:combolaEndEdit [dict replace $comboi eres cancel]]
	set bt [lreplace $bt $ix $ix $tag CombolaEditTCombobox]
	foreach seq [list . all] {
		set ix [lsearch -exact $bt $seq]
		set bt [lreplace $bt $ix $ix]
	}
	bindtags $combo $bt
	$combo state !readonly
	event generate $combo <<CombolaEditBegin>> -data [dict create value [$combo get]] -when tail
}

proc combolaEndEdit {comboi} {
	set combo [dict get $comboi combo]
	set value [$combo get]
	bindtags $combo [dict get $comboi bindtags]
	$combo state readonly
	set lastIdx [dict get $comboi lastIdx]
	if {$lastIdx >= 0} {
		comboGoto $combo $lastIdx
	} else {
		$combo set ""
	}
	RmTag [dict get $comboi tag]
	event generate $combo <<CombolaEditEnd>> -data [dict create value $value eres [dict get $comboi eres]] -when tail
}


## Commands

proc enhance {combo args} {
	Init
	set tags [list]
	set upis [expr {[dict exists $args -upis] ? [dict get $args -upis] : "up"}]
	if {$upis ni [list down up]} {
		return -code error "bad value \"$upis\": must be one of down, or up"
	}
	set keyselect [expr {[dict exists $args -keyselect] ? [dict get $args -keyselect] : "both"}]
	if {$keyselect ni [list both combobox listbox none]} {
		return -code error "bad value \"$keyselect\": must be one of both, combobox, listbox, or none"
	}
	if {$keyselect in [list combobox both]} {
		lappend tags CombolaKeySelectTCombobox
	}
	if {[dict exists $args -edit]} {
		set edit [dict get $args -edit]
		set tag [MkTag $combo EditReady]
		bind $tag <<CombolaEditBegin>> [list uplevel #0 [list $edit begin %W %d]]
		bind $tag <<CombolaEditEnd>>   [list uplevel #0 [list $edit end   %W %d]]
		bind $tag         <Destroy>    [list [namespace current]::RmTag $tag]
		lappend tags $tag CombolaEditReadyTCombobox
	}
	if {$keyselect in [list listbox both]} {
		lappend tags CombolaOpenKeySelectTCombobox
		lappend tags CombolaKeySelectButtons
	} else {
		lappend tags CombolaOpenTCombobox
		lappend tags CombolaButtons
	}
	lappend tags CombolaScrollableUpIs[string totitle $upis]
	lappend tags CombolaTCombobox
	set idx [lsearch -exact [set bt [bindtags $combo]] [winfo class $combo]]
	set bt [linsert $bt $idx {*}$tags]
	bindtags $combo $bt
	if {$keyselect in [list combobox both]} {
		$combo state readonly
	}
	return $combo
}

proc combobox {w args} {
	Init
	lassign [Args $args] targs cargs
	return [enhance [ttk::combobox $w {*}$targs] {*}$cargs]
}


## Util

proc Args {argies} {
	set targs [dict create]
	set cargs [dict create]
	dict for {arg val} $argies {
		dict set [expr {$arg in [list -edit -keyselect -upis] ? "c" : "t"}]args $arg $val
	}
	return [list $targs $cargs]
}

proc CombolistCurrent {combolist} {
	set sel [$combolist curselection]
	return [expr {$sel ne {} ? [lindex $sel 0] : -1}]
}

proc MkEgt {event args} { list event generate %W $event {*}$args -when tail }
proc MkTag {combo sufx} { format {Combola`%s`%s} $combo $sufx }
proc RmTag {tag}        { foreach seq [bind $tag] { bind $tag $seq {} } }


## Init

proc Init {} {
	variable Initted
	if {$Initted} return

	set ns [namespace current]:

	# Create bindings

	# Combobox
	bind CombolaTCombobox <<CombolaKey>> [list $ns:comboKey %W %d]
	bind CombolaTCombobox <<CombolaGo>>  [list $ns:comboGo  %W %d]

	# Listbox
	bind CombolaTComboboxListbox <<CombolaListKey>> [list $ns:combolistKey %W %d]
	bind CombolaTComboboxListbox <<CombolaListGo>>  [list $ns:combolistGo  %W %d]

	# Edit Ready
	bind CombolaEditReadyTCombobox <<CombolaBeginEdit>> [list $ns:combolaBeginEdit %W]

	# Open
	bind CombolaOpenTCombobox          <<CombolaOpenList>> [list ::ttk::combobox::CombolaPost %W]
	bind CombolaOpenKeySelectTCombobox <<CombolaOpenList>> [list ::ttk::combobox::CombolaKeySelectPost %W]

	# Combobox
	bind CombolaTCombobox <<CombolaGoNext>>     [MkEgt <<CombolaGo>> -data next    ]
	bind CombolaTCombobox <<CombolaGoPrev>>     [MkEgt <<CombolaGo>> -data prev    ]
	bind CombolaTCombobox <<CombolaGoHome>>     [MkEgt <<CombolaGo>> -data home    ]
	bind CombolaTCombobox <<CombolaGoEnd>>      [MkEgt <<CombolaGo>> -data end     ]
	bind CombolaTCombobox <<CombolaGoNextPage>> [MkEgt <<CombolaGo>> -data nextpage]
	bind CombolaTCombobox <<CombolaGoPrevPage>> [MkEgt <<CombolaGo>> -data prevpage]

	# Listbox
	bind CombolaTComboboxListbox <<CombolaListGoHome>>     [MkEgt <<CombolaListGo>> -data home    ]
	bind CombolaTComboboxListbox <<CombolaListGoEnd>>      [MkEgt <<CombolaListGo>> -data end     ]
	bind CombolaTComboboxListbox <<CombolaListGoNextPage>> [MkEgt <<CombolaListGo>> -data nextpage]
	bind CombolaTComboboxListbox <<CombolaListGoPrevPage>> [MkEgt <<CombolaListGo>> -data prevpage]

	# Scrollable
	bindAndBreak CombolaScrollableUpIsUp \
		<Alt-Key-Down>  continue                      \
		    <Key-Down>  [MkEgt <<CombolaGoNext>>    ] \
		    <Key-Up>    [MkEgt <<CombolaGoPrev>>    ] \
		    <Key-Next>  [MkEgt <<CombolaGoNextPage>>] \
		    <Key-Prior> [MkEgt <<CombolaGoPrevPage>>] \
	    <Control-Key-Home>  [MkEgt <<CombolaGoHome>>    ] \
	    <Control-Key-End>   [MkEgt <<CombolaGoEnd>>     ]

	bindAndBreak CombolaScrollableUpIsDown \
		<Alt-Key-Down>  continue                      \
		    <Key-Down>  [MkEgt <<CombolaGoPrev>>    ] \
		    <Key-Up>    [MkEgt <<CombolaGoNext>>    ] \
		    <Key-Next>  [MkEgt <<CombolaGoPrevPage>>] \
		    <Key-Prior> [MkEgt <<CombolaGoNextPage>>] \
	    <Control-Key-Home>  [MkEgt <<CombolaGoEnd>>     ] \
	    <Control-Key-End>   [MkEgt <<CombolaGoHome>>    ]

	# Combobox
	bindAndBreak CombolaTCombobox \
		<Alt-Key-Down>  [MkEgt <<CombolaOpenList>>  ] \
		<Alt-Key-space> [MkEgt <<CombolaOpenList>>  ]

#		<<TraverseIn>>  break

	bindAndBreak CombolaButtons \
		 <Button-1>     [list ::ttk::combobox::CombolaPress ""  %W %x %y] \
	   <Shift-Button-1>     [list ::ttk::combobox::CombolaPress "s" %W %x %y] \
	  <Double-Button-1>     [list ::ttk::combobox::CombolaPress "2" %W %x %y] \
	  <Triple-Button-1>     [list ::ttk::combobox::CombolaPress "3" %W %x %y]

	# Listbox
	bind CombolaTComboboxListbox     <Key-Home>  [MkEgt <<CombolaListGoHome>>] 
	bind CombolaTComboboxListbox     <Key-End>   [MkEgt <<CombolaListGoEnd>>] 
	bind CombolaTComboboxListbox <Alt-Key-space> [MkEgt <Key-Escape>]
	bind CombolaTComboboxListbox     <Key-space> [MkEgt <Key-Escape>]

	bindAndBreak CombolaTComboboxListbox \
		    <Key-Next>  [MkEgt <<CombolaListGoNextPage>>] \
		    <Key-Prior> [MkEgt <<CombolaListGoPrevPage>>] \
		<Alt-Key-Up>	[MkEgt <Key-Escape>]

	# Key Select
	bind CombolaKeySelectTCombobox <Alt-Key>      continue
	bind CombolaKeySelectTCombobox <Control-Key>  continue
	bind CombolaKeySelectTCombobox <Key>          [MkEgt <<CombolaKey>> -data %K]
	bind CombolaKeySelectTCombobox <<TraverseIn>> [MkEgt <<SelectAll>>]
	bind CombolaKeySelectTCombobox <FocusIn>      [MkEgt <<SelectAll>>]

	bindAndBreak CombolaKeySelectTCombobox \
		    <Key-Home>  [MkEgt <<CombolaGoHome>>  ] \
		    <Key-End>   [MkEgt <<CombolaGoEnd>>   ] \
		    <Key-Right> {} \
		    <Key-Left>  {} \
		    <Key-space> [MkEgt <<CombolaOpenList>>]

	bindAndBreak CombolaKeySelectButtons \
		 <Button-1>     [list ::ttk::combobox::CombolaKeySelectPress ""  %W %x %y] \
	   <Shift-Button-1>     [list ::ttk::combobox::CombolaKeySelectPress "s" %W %x %y] \
	  <Double-Button-1>     [list ::ttk::combobox::CombolaKeySelectPress "2" %W %x %y] \
	  <Triple-Button-1>     [list ::ttk::combobox::CombolaKeySelectPress "3" %W %x %y]

	# Key Select Listbox
	bind CombolaKeySelectTComboboxListbox <Key> [MkEgt <<CombolaListKey>> -data %K]

	# Edit Ready
	bindAndBreak CombolaEditReadyTCombobox \
	    <Control-Key-Return> [MkEgt <<CombolaBeginEdit>>]

	# Edit
	bindAndBreak CombolaEditTCombobox \
	    <Control-Key-Return> [MkEgt <<CombolaAcceptEdit>> ] \
		    <Key-Escape> [MkEgt <<CombolaDiscardEdit>>]

if 0 {
	bindAndBreak CombolaEditTCombobox \
		<<NextWindow>>       {} \
		<<PrevWindow>>       {}
}


	# Create procs

	set tns ::ttk::combobox:

	foreach {
		srcproc search
		dstproc replace
	} {
		PopdownWindow			{ComboboxListbox}
		CombolaPopdownWindow		{CombolaTComboboxListbox ComboboxListbox}

		Post				{set popdown \[PopdownWindow \$cb\]}
		CombolaPost			{set popdown [CombolaPopdownWindow $cb]}

		PopdownWindow			{ComboboxListbox}
		CombolaKeySelectPopdownWindow	{CombolaKeySelectTComboboxListbox CombolaTComboboxListbox ComboboxListbox}

		Post				{set popdown \[PopdownWindow \$cb\]}
		CombolaKeySelectPost		{set popdown [CombolaKeySelectPopdownWindow $cb]}
	} {
		if {[regsub $search [set body [info body $tns:$srcproc]] $replace body] != 1} {
			puts stderr "Couldn't substitute for $tns:$dstproc"
		}
		proc $tns:$dstproc {cb} $body
	}

	foreach {
		srcproc search
		dstproc replace
	} {
		Press				{Post \$w}
		CombolaPress			{CombolaPost $w}

		Press				{Post \$w}
		CombolaKeySelectPress		{CombolaKeySelectPost $w}
	} {
		if {[regsub $search [set body [info body $tns:$srcproc]] $replace body] != 1} {
			puts stderr "Couldn't substitute for $tns:$dstproc"
		}
		proc $tns:$dstproc {mode w x y} $body
	}

	set Initted yes
}


## Setup

variable Initted no

namespace import tu::tcl::getNextOrPrevMatch tu::tcl::whereBe
namespace import tu::tk::bindAndBreak

namespace ensemble create -command combola -prefixes no \
	-subcommands [list combobox enhance go] \
	-map [list go comboGo]

namespace export combola

package provide tu::combola 0.4


}}; # namespace tu::combola


# EOF

# THATCH /herbie/packages/tu/epm-0.3.tm modbase /herbie/packages
#
# Tu Epm - Tcl Usefulness Edit Program Menu
#
# Keeps track of two menus that can be used by a program:
#   An epm menu with items for choosing files etc., and edit items.
#   An epm edit menu with only edit items.
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#
# Fall 2022
# Version 0.2
# Version 0.3
#

package require Tcl 8.5-

namespace eval tu { namespace eval epm {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::epm
		title	{Tu Epm}
		descr	{Edit Program Menu}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.3
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


### Procedures copied from other places

namespace eval tu {
namespace eval tcl {
proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

# dict get with default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

namespace export randstr dgwd
}; # namespace eval tcl

namespace eval tk {
#
# Underline Ampersand
# Convert ampersand in text to -underline option
# From Gub
#
proc ua {text} {	
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}

namespace export ua
}; # namespace eval tk
}; # namespace eval tu


### Util

proc loadFileDialog {args} {
	if {[package provide tu::file] ne {}} {
		tailcall ::tu::file::loadFileDialog {*}$args
	} else {
		tailcall tk_getOpenFile {*}$args

	}
}
proc chooseDirectoryDialog {args} {
	if {[package provide tu::file] ne {}} {
		tailcall ::tu::file::chooseDirectoryDialog {*}$args
	} else {
		tailcall tk_chooseDirectory {*}$args
	}

}
proc saveSelectionToFile {args} {
	if {[package provide tu::file] ne {}} {
		tailcall ::tu::file::saveSelectionToFile {*}$args
	} else {
	}
}


### Epm

variable M {} ;# The epm menu
variable Mmap [dict create] ;# Named items and their indexes
variable E {} ;# The epm edit menu
variable Emap [dict create] ;# Named items and their indexes
# The 'mu' for E
variable epmmu [dict create -win {} -mmap {} -defs {} -ftmap {} -inserter {} -getwrap {} -setwrap {}]
variable epmww ;# Current word wrap state


# Create an epm menu
proc epmCreateMenu {{w ""} args} {
	variable M
	variable Mmap
	set M [menu $w.epm -tearoff 0 -postcommand [namespace current]::epmAdjustMenu {*}$args]
	foreach item [list "Choose &File" "Choose &Directory" "Choose C&olour" "&Set To Default" Cu&t &Copy &Paste C&lear "Select &All"] {
		$M add command -command [namespace current]::epm[string map {{ } {}} [string map {"&" ""} $item]] -label {*}[ua $item]
	}
	$M add checkbutton -command [namespace current]::epmAdjustWordWrap -variable [namespace current]::epmww -onvalue word -offvalue none -label {*}[ua "&Word Wrap"]
	$M insert 4 separator
	set Mmap [dict create cut 5 copy 6 paste 7 clear 8 wrap 10 setdef 3]
	return $M
}

# Return the epm menu, create if doesn't exist
proc epmMenu {} {
	variable M
	if {$M eq {}} { epmCreateMenu }
	return $M
}

# Create an edit menu
proc epmCreateAnEditMenu {{w ""} args} {
	set e [menu $w.eepm_[randstr] -tearoff 0 {*}$args]
	foreach item [list Cu&t &Copy &Paste C&lear "Select &All" "Sa&ve As"] {
		$e add command -command [namespace current]::epm[string map {{ } {}} [string map {"&" ""} $item]] -label {*}[ua $item]
	}
	$e add checkbutton -command [namespace current]::epmAdjustWordWrap -variable [namespace current]::epmww -onvalue word -offvalue none -label {*}[ua "&Word Wrap"]
	set emap [dict create cut 0 copy 1 paste 2 clear 3 saveas 5 wrap 6]
	epmSetEpmAdjustMenu $e $emap
	return [list $e $emap]
}

# Setup epm adjustment for a menu
proc epmSetEpmAdjustMenu {m mmap} {
	$m configure -postcommand [list [namespace current]::epmAdjustMenu $m $mmap]
}

# Create the epm edit menu
proc epmCreateEditMenu {{w ""}} {
	variable E
	variable Emap
	lassign [epmCreateAnEditMenu $w] E Emap
	return $E
}

# Return the epm edit menu, create if doesn't exist
proc epmEditMenu {} {
	variable E
	if {$E eq {}} { epmCreateEditMenu }
	return $E
}

# Have the epm menu popup for a widget
proc epmBindEpmMenuTo {w args} {
	variable Mmap
	# Call epmMenu to ensure M's menu exists
	epmBindMenuTo $w [epmMenu] {*}$args -mmap $Mmap
}

# Have the epm edit menu popup for a widget
proc epmBindEditMenuTo {w args} {
	variable Emap
	# Call epmEditMenu to ensure E's menu exists
	epmBindMenuTo $w [epmEditMenu] {*}$args -mmap $Emap
}

# Make a 'mu'
proc epmMakeMu {m args} {
	set ns [namespace current]:
	set mu [dict merge $args [dict create -menu $m \
		-mmap     [dgwd [dict create]      $args -mmap]     \
		-defs     [dgwd [dict create]      $args -defs]     \
		-ftmap    [dgwd [dict create]      $args -ftmap]    \
		-inserter [dgwd $ns:epmInserter    $args -inserter] \
		-getwrap  [dgwd $ns:epmGetWordWrap $args -getwrap]  \
		-setwrap  [dgwd $ns:epmSetWordWrap $args -setwrap]  \
	]]
	return $mu
}

# Have a menu popup for a widget
proc epmBindMenuTo {w m args} {
	set mu [epmMakeMu $m {*}$args]
	set mod   [dgwd ""        $args -mod]
	set seq   [dgwd "Button-" $args -seq]
	set bn    [dgwd 3         $args -bn]
	bind $w <$mod$seq$bn> [list [namespace current]::epmMenuGo %X %Y %W $mu]
}

# Popup activator
proc epmMenuGo {x y w mu} {
	epmMu [dict merge $mu [dict create -win $w -x $x -y $y]]
	tk_popup [dict get [epmMu] -menu] $x $y
}

# Get/set the 'mu'
proc epmMu {args} {
	variable epmmu
	if {[llength $args] == 1} {
		set epmmu [lindex $args 0]
	}
	return $epmmu
}

# Get/set current window for a popup
proc epmW {args} {
	variable epmmu
	if {[llength $args] == 1} {
		dict set epmmu -win [lindex $args 0]
	}
	return [dict get $epmmu -win]
}

# Get the word wrap state for a window. Used by the default 'mu'.
proc epmGetWordWrap {w} {
	return [$w cget -wrap]
}

# Get the word wrap state for a window. Used by the default 'mu'.
proc epmSetWordWrap {w wrap} {
	$w configure -wrap $wrap
}

# Adjust a window's word wrap state according to epm state
proc epmAdjustWordWrap {} {
	variable epmmu
	variable epmww
	if {[dict get $epmmu -setwrap] eq {}} { return }
	[dict get $epmmu -setwrap] [dict get $epmmu -win] $epmww
}

# Ajdust a menu according to a window's state
proc epmAdjustMenu {{m {}} {mmap {}}} {
	variable epmmu
	variable epmww
	set w [dict get $epmmu -win]
	if {$w eq "" || ![winfo exists $w]} {
		if {![catch {info level -1} res]} {
			puts "epmAdjustMenu !$res!"
		}
		return

	}

if {[dict get $epmmu -getwrap] eq {}} { return }
#if {[dict get $epmmu -setwrap] eq {}} { return }

	set defs [dict get $epmmu -defs]
	if {$m eq {}} {
		set m [dict get $epmmu -menu]
	}
	if {$mmap eq {}} {
		set mmap [dict get $epmmu -mmap]
	}
	set items [list]
	if {[dict size $mmap] > 0} { foreach what [list cut copy paste clear wrap setdef saveas] {
		if {![dict exists $mmap $what]} { continue }
		lappend items $what
	}}
	if {[llength $items] == 0} { return }
	set states [dict create]
	dict set states setdef [expr {[dict exists $defs $w] ? "normal" : "disabled"}]
	set wclass [winfo class $w]
	if {$wclass eq "Text"} {
		set epmww [[dict get $epmmu -getwrap] $w]
		if {$epmww eq "char"} { set epmww word }
		dict set states wrap normal
		set sel [expr {[llength [$w tag ranges sel]] > 0 ? "normal" : "disabled"}]
	} elseif {$wclass eq "TLabel"} {
		set epmww none
		dict set states wrap disabled
		set sel normal
	} elseif {$wclass eq "TButton"} {
		set epmww none
		dict set states wrap disabled
		set sel disabled
	} else {
		set epmww none
		dict set states wrap disabled
		set sel [expr {[$w selection present] ? "normal" : "disabled"}]
	}
	dict set states copy $sel
	dict set states saveas $sel
	if {[$w cget -state] eq "normal"} {
		foreach item [list cut clear] {
			dict set states $item $sel
		}
		dict set states paste normal
	} else {
		foreach item [list cut clear paste setdef saeveas] {
			dict set states $item disabled
		}
	}
	if {[dict get $states paste] eq "normal" && "paste" in $items} {
		if {[catch {selection get -displayof $w -selection CLIPBOARD} s]} { set s "" }
		dict set states paste [expr {[string length $s] > 0 ? "normal" : "disabled"}]
	}
	foreach item $items {
		$m entryconfigure [dict get $mmap $item] -state [dict get $states $item]
	}
}

proc epmInserter {where what} {
	epmSelectAll ; epmClear
	$where insert end $what
}

proc epmChooseFile {} {
	variable epmmu
	set w [dict get $epmmu -win]
	set ftmap [dict get $epmmu -ftmap]
	set filetypes [dgwd {} $ftmap $w]
	if {[lsearch -exact -index {0 1} $filetypes [list *]] == -1} {
		lappend filetypes [list "All Files" [list *]]
	}
	if {[set v [loadFileDialog -initialdir [file dirname [$w get]] -initialfile [$w get] -filetypes $filetypes]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmChooseDirectory {} {
	variable epmmu
	set w [dict get $epmmu -win]
	if {[set v [chooseDirectoryDialog -initialdir [$w get]]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmChooseColour {} {
	variable epmmu
	set w [dict get $epmmu -win]
	if {[winfo class $w] eq "Text"} {
		set v ""
	} else {
		set v [string trim [$w get]]
	}
	if {$v eq ""} { set v #000000 }
	if {[catch {winfo rgb . $v}]} { set v olive }
	if {[set v [tk_chooseColor -initialcolor $v]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmSaveAs {} {
	variable epmmu
	saveSelectionToFile [dict get $epmmu -win]
}

proc epmSetToDefault {} {
	variable epmmu
	set w [dict get $epmmu -win]
	set defs [dict get $epmmu -defs]
	if {![dict exists $defs $w]} { return }
	set v [dict get $defs $w]
	[dict get $epmmu -inserter] $w $v
}

foreach e [list Cut Copy Paste Clear SelectAll] {
	proc epm$e {} [string map [list @e@ $e] {
		variable epmmu
		set w [dict get $epmmu -win]
		event generate $w <<@e@>>
		focus $w
	}]
}; unset e


## Setup

namespace import tu::tcl::randstr tu::tcl::dgwd
namespace import tu::tk::ua

namespace export epmCreateMenu epmMenu
namespace export epmCreateEditMenu epmCreateAnEditMenu epmEditMenu
namespace export epmBindMenuTo epmBindEpmMenuTo epmBindEditMenuTo
namespace export epmMakeMu epmSetEpmAdjustMenu
namespace export epmMu epmW

package provide tu::epm 0.3

}}; # namespace tu::epm


# EOF

# THATCH /herbie/packages/tu/file-0.2.tm modbase /herbie/packages
#
# Tu File - Tcl Usefulness File Utils
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#
# Fall 2022
# Version 0.2
#

package require Tcl 8.5-

namespace eval tu { namespace eval file {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::file
		title	{Tu File}
		descr	{File Utils}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.2
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


### File

proc getFileFor {what {fn ""} args} {
	if {$fn ni {"" "?"}} { return $fn }
	if {[package provide Tk] eq ""} { return "" }
	return [${what}FileDialog {*}$args]
}

proc getLoadFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor load $fn {*}$args]
}

proc getSaveFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor save $fn {*}$args]
}

proc openFile {fn {mode r}} {
	if {$fn eq "-"} { return [expr {$mode eq "r" ? "stdin" : "stdout"}] }
	if {$fn in {stdin stdout stderr}} { return $fn }
	if {[catch {open $fn $mode} f]} { return "" }
	return $f
}

proc closeFile {f} {
	if {$f in {stdin stdout stderr}} { return }
	close $f
}

proc loadFile {{fn ""} args} {
	set notok [list "" "" 0]
	set fn [getLoadFile $fn {*}$args]
	if {$fn eq ""} { return $notok] }
	if {[set f [openFile $fn]] eq ""} { return $notok }
	set data [read -nonewline $f]
	closeFile $f
	return [list $fn $data 1]
}

proc saveFile {fn data args} {
	set notok [list "" 0]
	set fn [getSaveFile $fn {*}$args]
	if {$fn eq ""} { return $notok }
	if {[set f [openFile $fn w]] eq ""} { return $notok }
	puts -nonewline $f $data
	closeFile $f
	return [list $fn 1]
}

proc loadListFile {{fn ""} args} {
	if {![lassign [loadFile $fn {*}$args] fn data]} { return [list $fn "" 0] }
	return [list $fn [processListFileBlob $data] 1]
}

proc processListFileBlob {blob} {
	set res [list]
	foreach l [split $blob \n] {
		set l [string trim $l]
		if {$l eq {} || [string index $l 0] eq {#}} { continue }
		if {[catch {set len [llength $l]}] || $len < 1} { continue }
		lappend res $l
	}
	return $res
}

proc filemode->text {mode} {
    foreach { mask pairs } {
	00400 { 00400 r }
	00200 { 00200 w }
	04100 { 04100 s 04000 S 00100 x }
	00040 { 00040 r }
	00020 { 00020 w }
	02010 { 02010 s 02000 S 00010 x }
	00004 { 00004 r }
	00002 { 00002 w }
	01001 { 01001 t 01000 T 00001 x }
    } {
	set value [expr $mode & $mask]
	set bit -
	foreach { x b } $pairs {
	    if { $value == $x } {
		set bit $b
	    }
	}
	append retval $bit
    }
    return $retval
}

proc loadFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getOpenFile {*}$args
	} else {
		tailcall ttk::getOpenFile {*}$args
	}
}

proc saveFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getSaveFile {*}$args
	} else {
		tailcall ttk::getSaveFile {*}$args
	}
}

proc chooseDirectoryDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_chooseDirectory {*}$args
	} else {
		tailcall ttk::chooseDirectory {*}$args
	}
}

proc loadFileIntoText {w {fn ""} args} {
	lassign [loadFile $fn {*}$args] fn data
	if {$fn eq ""} { return "" }
	$w delete 1.0 end
	$w insert end $data
	return $fn
}

proc saveSelectionToFile {w {fn ""} args} {
	if {[set blob [getSelectionFromWidget $w]] eq ""} { return }
	saveFile $fn $blob {*}$args
	return $fn
}

#these should be in tu::tk
proc getSelectionFromWidget {w} {
	if {[winfo class $w] eq "Text"} {
		set res [getSelectionFromText $w]
	} else {
		set res [getSelectionFromOther $w]
	}
	return [string trim $res]
}
proc getSelectionFromText {w} {
	if {[llength [set sel [$w tag ranges sel]]] != 2} {
		return ""
	}
	return [$w get {*}$sel]
}
proc getSelectionFromOther {w} {
	if {[catch {$w selection present}]} { return "" }
	if {![$w selection present]} {
		$w selection range 0 end
	}
	if {![$w selection present]} {
		return ""
	}
	return [string range [$w get] [$w index sel.first] [$w index sel.last]-1]
}


## Setup

namespace export openFile closeFile getLoadFile loadFile getSaveFile saveFile
namespace export loadListFile processListFileBlob
namespace export loadFileDialog saveFileDialog
namespace export loadFileIntoText saveSelectionToFile
namespace export filemode->text

package provide tu::file 0.2

}}; # namespace tu::file


# EOF

# THATCH /herbie/packages/tu/myfonts-0.2.tm modbase /herbie/packages
#
# Tu MyFonts - Tcl Usefulness My Fonts
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#
# Fall
# Version 0.2
#

package require Tcl 8.5-
package require TclOO

namespace eval tu { namespace eval myfonts {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::myfonts
		title	{Tu MyFonts}
		descr	{My Fonts}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.2
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


### Procedures copied from other places

namespace eval tu {
namespace eval tcl {
proc makeROAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " "{}" " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}
}; # namespace eval tcl
}; # namespace eval tu


### MyFonts

::oo::class create MyFonts {
	variable TkFonts
	variable class
	variable fonts
	variable defaults
constructor {myClass args} {
	set class $myClass
	set TkFonts [list Default Text Fixed Menu Heading]
	set fonts [dict create]
	my setup {*}$args
}
eval [::tu::myfonts::tu::tcl::makeROAccessors class TkFonts fonts]
method setup {args} {
	foreach f $TkFonts {
		set c [font configure Tk${f}Font]
		dict set fonts [font create ${class}${f}Font {*}$c {*}$args] {}
	}
	foreach {attr which props} [list {*}{
		Bigger Text {-size 11}
		Log    Text {}
	}] {
		dict set fonts [font create ${class}${attr}${which}Font {*}[font configure ${class}${which}Font] {*}$props {*}$args] {}
	}
#		Bold   Text {-weight bold}
#		Bold Fixed   {-weight bold}
#		Big  Default {-size -18}

	foreach f [dict keys $fonts] { dict set fonts $f [font configure $f] }
	option add *${class}*Menu*font      ${class}MenuFont
	option add *${class}*Text*font      ${class}FixedFont
	option add *${class}*TLabel*font    ${class}DefaultFont
	option add *${class}*TEntry*font    ${class}TextFont
	option add *${class}*TCombobox*font ${class}TextFont
if 0 {
	ttk::style configure Heading           -font ${class}HeadingFont
	ttk::style configure Treeview          -font ${class}DefaultFont
	ttk::style configure TButton           -font ${class}DefaultFont
	ttk::style configure TCheckbutton      -font ${class}DefaultFont
	ttk::style configure TRadiobutton      -font ${class}DefaultFont
	ttk::style configure TNotebook.Tab     -font ${class}DefaultFont
	ttk::style configure TLabelframe.Label -font ${class}DefaultFont
}
}
method defaults {font} {
	return [dict get $fonts $font]
}
method reset {} {
	dict for {f c} $fonts { font configure $f {*}$c }
}
method transform-scale {{s 0} args} {
	if {[llength $args] > 0} {
		set fnts [dict create]
		foreach arg $args { dict set fnts $arg [dict get $fonts $arg] }
	} else {
		set fnts $fonts
	}
	dict for {f c} $fnts {
		set size [dict get $c -size]
		font configure $f -size [expr {$size + (($size < 0 ? -1 : 1) * $s)}]
	}
}
}; # ::oo::class create MyFonts


## Setup

#namespace export MyFonts

package provide tu::myfonts 0.2

}}; # namespace tu::myfonts


# EOF

# THATCH /herbie/packages/tu/pretty-0.2.tm modbase /herbie/packages
#
# Tu Pretty - Tcl Usefulness Pretty Printing
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#
# Fall 2022
# Version 0.2
#

package require Tcl 8.5-

namespace eval tu { namespace eval pretty {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::pretty
		title	{Tu Pretty}
		descr	{Pretty Printing}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.2
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


## Tcl

proc hexdump {blob {bits 8}} {
	set l [string length $blob]
	set dump {}
	for {set i 0; set ii [expr {$bits - 1}]; set addr 0} {$i < $l} {incr i $bits; incr ii $bits; incr addr $bits} {
		binary scan [string range $blob $i $ii] @0H*@0a* hex ascii
		regsub -all -- {[^[:graph:] ]} $ascii {.} ascii
		regsub -all -- {..} $hex {& } hex
		lappend dump [format "%04X  %-[expr {3 * $bits}]s %-${bits}s" $addr $hex $ascii]
	}
	return $dump
}

proc prettyBindings {bindings} {
	set q ""
	dict for {k v} $bindings {
		append q $k \n
		dict for {b s} $v {
			append q "  " [format "%-30s" $b] " " [string trim $s] \n
		}
	}
	return $q
}

proc mpd {dict {pat {*}}} {
	set f %-[::tcl::mathfunc::max 0 {*}[lmap k [set keys [lsort -dictionary [dict keys $dict $pat]]] { string length $k }]]s\ \ %s
	return [lmap k $keys { format $f $k [dict get $dict $k] }]
}
interp alias {} [namespace current]::makePrintDict {} [namespace current]::mpd

proc pd {dict {pat {*}}} {
	return [join [mpd $dict $pat] \n]
}
interp alias {} [namespace current]::printDict {} [namespace current]::pd

proc pa {arr {pat {*}}} {
	return [pd [array get $arr] $pat]
}
interp alias {} [namespace current]::printArray {} [namespace current]::pa

proc pv {var {pat {*}}} {
	upvar 1 $var v
	if {[array exists v]} { return [pa $var $pat] }
	if {![catch {dict size $v}]} { return [pd $v $pat] }
	return $v
}
interp alias {} [namespace current]::printVar {} [namespace current]::pv


## Tk
proc pb {tag} { package require tu::tk; return [prettyBindings [::tu::tk::bindings $tag]] }
interp alias {} [namespace current]::prettyBindingsOfTag {} [namespace current]::pb


## Setup

namespace export hexdump prettyBindings
namespace export mpd pv pa pd pb
namespace export makePrintDict printVar printArray printDict
namespace export prettyBindingsOfTag

package provide tu::pretty 0.2

}}; # namespace tu::pretty


# EOF

# THATCH /herbie/packages/tu/tcl-0.1.tm modbase /herbie/packages
#
# Tu Epm - Tcl Usefulness Tcl Utils
#
#
# Stuart Cassoff
#
# Fall 2022
# Version 0.1
#
# ?
# Version 0.2
#

package require Tcl 8.5-

namespace eval tu { namespace eval tcl {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::tcl
		title	{Tu Tcl}
		descr	{Tcl Utils}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

#
# dict get with default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

#
# In My Namespace
#
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

#
# Default value of arg of caller
#
proc myArgDef {arg} {
	set proc [lindex [info level -1] 0]
	if {$arg ni [info args $proc]} { return "" }
	if {[info default $proc $arg val]} { return $val }
	return ""
}

#
# String Cat
#
proc strcat {args} { join $args "" }

#
# [lmap] replacement if not available
#
if {[llength [info commands lmap]] == 0} { proc lmap args {
	# 8.5 doesn't have lmap
	# https://wiki.tcl-lang.org/page/lmap+forward+compatibility
	set body [lindex $args end]
	set args [lrange $args 0 end-1]
	set n 0
	set pairs [list]
	foreach {varnames listval} $args {
		set varlist [list]
		foreach varname $varnames {
			upvar 1 $varname var$n
			lappend varlist var$n
			incr n
		}
		lappend pairs $varlist $listval
	}
	set temp [list]
	foreach {*}$pairs {
		lappend temp [uplevel 1 $body]
	}
	set temp
}}; # lmap

proc getNextOrPrevMatch {aList aString {curIdx 0} {backwards no}} {
	set idxs [lsearch -all -nocase $aList $aString*]
	if {[llength $idxs] == 0} { return -1 }
	if {[llength $idxs] == 1 && [lindex $idxs 0] == $curIdx} { return -1 }
	set p -1
	foreach idx $idxs {
		if {$idx < $curIdx} { set p $idx }
		if {$idx > $curIdx} { break }
	}
	if {$backwards} {
		if {$p == -1} {
			set newIdx [lindex $idxs end]
		} else {
			set newIdx $p
		}	
	} else {
		if {$idx == $curIdx} {
			set newIdx [lindex $idxs 0]
		} else {
			set newIdx $idx
		}
	}
	return $newIdx
}

proc whereBe {where curIdx nitems {height {10}}} {
	set idx $curIdx
	switch -exact -- $where {
	home     {  set idx  0       }
	end      {  set idx  $nitems }
	next     { incr idx  1       }
	prev     { incr idx -1       }
	nextpage { incr idx  $height ; if {$idx > $nitems} { set idx $nitems } }
	prevpage { incr idx -$height ; if {$idx < 0      } { set idx 0       } }
	default  {  set idx  $where  }
	}
	if {$idx < 0 || $idx > $nitems || $idx == $curIdx} { set idx -1 }
	return $idx
}

proc laddAndSort {list value {sort {-dictionary}}} {
	set i -1
	set values [lsort {*}$sort -index 0 [lmap e [linsert $list end $value] { list $e [incr i] }]]
	return [list [lsearch -index 1 $values $i] [lmap e $values { lindex $e 0 }]]
}

proc makeRWAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " args " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			if {[llength $args] == 1} { set @m@ [lindex $args 0] }
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}
proc makeROAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " "{}" " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}

#
# Remove blank lines and comment lines, add space-backslash to all lines but the last.
# If a line contains ";#" (spaces ok too) then chares from those to the end of the line will be removed
# Useful for making big lists in code without all the unpleasant backslashes.
#
proc backslashify {blob} { return [join [lmap l [split $blob \n] { if {[regexp {^[[:space:]]*(;?#.*)?$} $l]} { continue } else { lindex [regexp -inline -- {^(.+?)(;\s*#.*)?$} $l] 1 } }] " \\\n"] }

proc maxstrlen {args} { return [::tcl::mathfunc::max 0 {*}[lmap s $args { string length $s }]] }

proc oneOf {l} {
	if {[llength $l] < 2} { return [join $l] }
	return [strcat "one of " [join [lrange $l 0 end-1] ", "] [expr {[llength $l] > 2 ? "," : ""}] " or " [lindex $l end]]
}



## Setup

namespace export makeRWAccessors makeRWOccessors
namespace export randstr dgwd imn backslashify maxstrlen oneOf
namespace export getNextOrPrevMatch whereBe laddAndSort

package provide tu::tcl 0.1

}}; # namespace tu::tcl


# EOF

# THATCH /herbie/packages/tu/tk-0.1.tm modbase /herbie/packages
#
# Tu Epm - Tcl Usefulness Tk Utils
#
#
# Stuart Cassoff
#
# Fall 2022
# Version 0.1
#

package require Tcl 8.5-

namespace eval tu { namespace eval tk {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::tk
		title	{Tu Tk}
		descr	{Tk Utils}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


proc setEntryText {w text} {
	$w delete 0 end
	if {$text ne ""} { $w insert end $text }
}

proc setTextText {w text} {
	if {[set state [$w cget -state]] ne "normal"} { $w configure -state normal }
	$w delete 1.0 end
	if {$text ne ""} { $w insert end $text }
	#if {$text ne ""} { $w insert end $text[expr {[string index $text end] eq "\n" ? "" : "\n"}] }
	if {$state ne "normal"} { $w configure -state $state }
}

proc bindings {tag} {
	set d [dict create]
	set l [expr {[string index $tag 0] eq "." ? [bindtags $tag] : [list $tag]}]
	foreach t $l { foreach b [bind $t] {
		dict set d $t $b [bind $t $b]
	}}
	return $d
}

proc fb {window seq} {
	set found [list]
	foreach t [bindtags $window] {
		if {[set s [bind $t $seq]] eq ""} continue
		lappend found [list $t $s]
	}
	return $found
}
interp alias {} [namespace current]::findBinding {} [namespace current]::fb

proc cloneBinding {window fromseq toseq {map {}}} {
	set found [findBinding $window $fromseq]
	foreach tb $found {
		if {$map ne {}} {
			set tb [lmap e $tb { string map $map $e }]
		}
		bind {*}[linsert $tb 1 $toseq]
	}
	return $found
}

proc copyBindings {fromTag toTag} {
	foreach seq [bind $fromTag] {
		bind $toTag $seq [bind $fromTag $seq]
	}
}

# From the Tk widget demo "A multi-column list of countries"
# Code to do the sorting of the tree contents when clicked on
#
proc tvSortBy {tree col direction} {
	# Build something we can sort
	set data {}
	foreach row [$tree children {}] {
		lappend data [list [$tree set $row $col] $row]
	}

	set dir [expr {$direction ? "-decreasing" : "-increasing"}]
	set r -1

	# Now reshuffle the rows into the sorted order
	foreach info [lsort -dictionary -index 0 $dir $data] {
		$tree move [lindex $info 1] {} [incr r]
	}

	# Switch the heading so that it will sort in the opposite direction
	$tree heading $col -command [list [namespace current]::tvSortBy $tree $col [expr {!$direction}]]
}

#
# Underline Ampersand
# Convert ampersand in text to -underline option
# From Gub
#
proc ua {text} {	
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}

proc bindAndBreak1 {w args} {
	foreach {seq script} $args {
		if {$script ni [list break continue]} {
			if {$script ne ""} { append script ";" }
			append script break
		}
		bind $w $seq $script
	}
}
proc bindAndBreak {w args} {
	dict for {seq script} $args {
		if {$script eq ""} {
			bind $w $seq break
		} elseif {$script in [list break continue]} {
			bind $w $seq $script
		} else {
			bind $w $seq $script
			bind $w $seq +break
		}
	}
}


## Setup

namespace export ua
namespace export bindings
namespace export fb
namespace export findBinding cloneBinding copyBindings
namespace export tvSortBy
namespace export setTextText setEntryText
namespace export bindAndBreak

package provide tu::tk 0.1

}}; # namespace tu::tk


# EOF

# THATCH /herbie/packages/tu/toplevels/tk-0.2.tm modbase /herbie/packages
#
# Tu Toplevels Tk - Tcl Usefulness Toplevels Tk
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#
# Fall 2022
# Version 0.2
#

package require Tcl 8.5-

namespace eval tu { namespace eval toplevels { namespace eval tk {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::toplevels::tk
		title	{Tu Toplevels Tk}
		descr	{Tk Toplevels}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.2
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


### Procedures copied from other places

namespace eval tu {
namespace eval tcl {
proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}
namespace export randstr
}; # namespace eval tcl
}; # namespace eval tu


### Tu Toplevels Tk

variable defaultWindow .

proc randWindow {{w {.}}} {
	while {[winfo exists [set rw $w[randstr]]]} {}
	return $rw
}

proc windowIsToplevel {w} {
#	return [expr {![string match .?*.?* $w]}]
#	return [expr {[winfo toplevel $w] eq $w}]
	return [expr {[winfo toplevel $w] eq $w || [winfo class $w] eq "Toplevel"}]
}

proc newToplevel {{w ""} args} {
	if {$w eq ""} { set w [randWindow] }
	toplevel $w {*}$args
	return $w
}

proc groupToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	wm group $w $W
}

proc destroyToplevel {w} {
	destroy $w
}

proc transientizeToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	wm transient $w $W
}

proc withdrawToplevel {w} {
	wm withdraw $w
}

proc forgetToplevel {w} {
	wm forget $w
}

proc iconifyToplevel {w} {
	wm iconify $w
}

proc deiconifyToplevel {w} {
	wm deiconify $w
}

proc toplevelIsZoomable {w} {
	return [expr {"-zoomed" in [dict keys [wm attributes $w]]}]
}

proc toplevelZoomed {w {zoomed {}}} {
	if {$zoomed ne {}} { wm attributes $w -zoomed $zoomed }
	return [wm attributes $w -zoomed]
}

proc toplevelFullscreen {w {fullscreen {}}} {
	if {$fullscreen ne {}} { wm attributes $w -fullscreen $fullscreen }
	return [wm attributes $w -fullscreen]
}

proc toplevelGeometry {w args} {
	if {[llength $args] == 1} { wm geometry $w [lindex $args 0] }
	return [wm geometry $w]
}

proc manageToplevel {w} {
	wm manage $w
}

proc toplevelResizable {w args} {
	if {[llength $args] == 2} { wm resizable $w {*}$args }
	return [wm resizable $w]
}

proc setToplevelTitle {w t} {
	wm title $w $t
}

proc getToplevelTitle {w} {
	return [wm title $w]
}

proc centreToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	::tk::PlaceWindow $w widget $W
}
interp alias {} [namespace current]::centerToplvel {} [namespace current]::centreToplvel

proc onToplevelDelete {w {s ""}} {
	return [if {[winfo toplevel $w] eq $w} {
		if {$s ne ""} { wm protocol $w WM_DELETE_WINDOW $s }
		wm protocol $w WM_DELETE_WINDOW
	} else {
		if {$s ne ""} { bind $w <Destroy> $s }
		bind $w <Destroy>
	}]
}

proc createToplevelGui {{title ""} args} {
	set w [newToplevel {} {*}$args]
	groupToplevel $w
	if {$title ne ""} { setToplevelTitle $w $title }
	return $w
}


## Setup

namespace import tu::tcl::randstr

namespace export \
	randWindow windowIsToplevel newToplevel groupToplevel destroyToplevel \
	transientizeToplevel centreToplevel centerToplevel \
	withdrawToplevel forgetToplevel iconifyToplevel deiconifyToplevel \
	onToplevelDelete setToplevelTitle getToplevelTitle \
	toplevelIsZoomable toplevelZoomed toplevelFullscreen \
	toplevelGeometry manageToplevel toplevelResizable \
	createToplevelGui

package provide tu::toplevels::tk 0.2

}}}; # namespace tu::toplevels::tk


# EOF

# THATCH /herbie/packages/colourbox-1.0.tm modbase /herbie/packages
#
# ColourBox - Colour selector
#
# Stuart Cassoff
#
# Summer 2018
# Version 0.1
#
# Summer 2023
# Version 1.0
#

package require Tcl 8.6-

namespace eval colourbox {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	colourbox
		title	ColourBox
		descr	{Colour Selector}
		author	{Stuart Cassoff}
		when	{Summer 2023}
		version	1.0
		example	yes
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2023 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


oo::class create ColourBox {
	variable colour
	variable defaultColour
	variable command
	variable R G B
	variable layout
	variable lazout
	variable compact
	variable create
	variable horizontal
	variable vertical
	variable horizontalCompact
	variable verticalCompact
	variable wColourBox
	variable wSwatch
	variable wW
	variable toplevel
	variable makeToplevel
	variable xR xG xB
	variable row
	variable column
	variable open
constructor {{w {}} args} {
	package require tu::tcl;
	package require tu::epm
	package require tu::toplevels::tk; namespace import ::tu::toplevels::tk::newToplevel ::tu::toplevels::tk::onToplevelDelete
	package require gub
	namespace import ::gub::gub ::gub::gubWidgetMap

	set R 0
	set G 0
	set B 0
	set layout horizontal
	set lazout 0
	set compact ""
	set command {}
	set row 0
	set column 0
	set open 0

	set toplevel [expr {$w in [list "" "."]}]
	set makeToplevel [expr {$w eq {}}]
	if {$w eq {}} { set w .[::tu::tcl::randstr] }
	set wW $w

	dict for {a v} $args {
		set [string range $a 1 end] $v
	}
	if {![info exists colour]} {
		if {[info exists color]} {
			set colour $color
		} else {
			set colour #000000
		}
	}
	if {![info exists defaultColour]} {
		if {[info exists defaultColor]} {
			set defaultColour $defaultColor
		} else {
			set defaultColour $colour
		}
	}

	my prepare
	if {$open} { my open }
}
method zconstructor {args} {
	package require tu::tcl;
	package require tu::epm
	package require tu::toplevels::tk; namespace import ::tu::toplevels::tk::newToplevel ::tu::toplevels::tk::onToplevelDelete
	package require gub
	namespace import ::gub::gub ::gub::gubWidgetMap
	set R 0
	set G 0
	set B 0
	set layout horizontal
	set lazout 0
	set compact ""
	set command {}
	set row 0
	set column 0
	set open 0
	set colour #000000
	set defaultColour $colour
	if {[llength $args] > 0} { my openOnIn {*}$args }
}
method openOnIn {{aColour {}} {w {}} args} {
	if {$aColour ne {}} {
		set colour $aColour
	}
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-colour - color
				{ set state -colour }
			-defaultcolour - defaultcolor
				{ set state -defaultcolour }
			}; # hctiws
		}
		-colour        { set state opt; set colour        $arg }
		-defaultcolour { set state opt; set defaultColour $arg }
		}; # hctiws
	}
	set toplevel [expr {$w in [list "" "."]}]
	set makeToplevel [expr {$w eq {}}]
	if {$w eq {}} { set w .[::tu::tcl::randstr] }
	set wW $w
	my prepare
}
method done {} { my destroy }
destructor {
	set w [expr {$toplevel ? $wW : $wColourBox}]
	if {[llength [info commands winfo]] == 1 && [winfo exists $w]} { destroy $w }
}
method dup {} {
	[info object class [self]] new {} -defaultColour $colour -layout $layout -open 1
}
method prepare {} {
	set w $wW

	# Create a WidgetMap where the widget commands
	# only configure and not create their widgets
	set wm [dict create]
	dict for {c p} [gubWidgetMap] {
		dict set wm $c [dict set p command ::gub::gubConfigureWidget]
	}

	set colourBoxFrameGob "
		;Fc
#		!!Fc colourBox_[::tu::tcl::randstr]
	"

	set createFrame [gub "
		$colourBoxFrameGob
		[expr {$toplevel ? "" : "!:"}]
	" $w]

	eval [gub "
		$colourBoxFrameGob
		/Fc
		!:
	" $w weegs -wm $wm]

	set wColourBox [dict get $weegs Fc]

	set colourBoxGobVerticalBase [gub "
		$colourBoxFrameGob
		:Fc -sticky ns
		-Fc
		|Fc 0 -weight 0
		!:
	" $w -wm $wm -oc $column]

	set colourBoxGobHorizontalBase [gub "
		$colourBoxFrameGob
		:Fc -sticky ew
		|Fc
		-Fc 0 -weight 0
		!:
	" $w -wm $wm -or $row]

	set colourBoxGutsGobVertical "
		;ArAgAb
		;LrLgLb
		;LRLGLB
		;Lc
		:A  -sticky ns   -rowspan 1 -columnspan 1
		:Lc -sticky nsew -rowspan 1 -columnspan 3
	"
	set colourBoxGutsGobVerticalCompact "
		;ArAgAbLrLgLb
		;      LRLGLB
		;      Lc
		:A -sticky  ns   -rowspan 3 -columnspan 1
		:Lc -sticky nsew -rowspan 1 -columnspan 3
	"

	set colourBoxGutsGobHorizontal "
		;ArLr
		;AgLg
		;AbLb
		;    LR
		;    LG
		;    LB
		;      Lc
		!:-LR -3
		!:-Lc -3
		:A -sticky {} -columnspan 1
		:Lc -sticky nsew -rowspan 3 -columnspan 1
	"
	set colourBoxGutsGobHorizontalCompact "
		;Ar
		;Ag
		;Ab
		;    Lr
		;    Lg
		;    Lb
		;  LR
		;  LG
		;  LB
		;Lc
		!:-LR -3
		!:-Lc -3
		:A  -sticky   ew -rowspan 1 -columnspan 3
		:Lc -sticky nsew -rowspan 3 -columnspan 1
	"

	set colourBoxGutsGobCreate [set colourBoxGutsGob[string totitle $layout]]

	set colourBoxGutsGobCreateOptions "
		/Lc swatch
	"

	set colourBoxGutsGobConfigureOptions "
		.{L<rgb>} -text R G B
		.{L<RGB>} -textvariable [my varname xR] [my varname xG] [my varname xB]
		.{A} -variable [my varname R] [my varname G] [my varname B]
		.Lc -width 6 ;#-relief ridge
		.A -from 0 -to 255 -command {[self] updateColour}
		.Lc -text \\n\\n
	"

	set colourBoxGutsGobBindings "
		&A<rgb>       Button-4 event generate %W <<PrevWord>>
		&A<rgb> Shift-Button-4 event generate %W <<PrevChar>>
		&A<rgb>       Button-5 event generate %W <<NextWord>>
		&A<rgb> Shift-Button-5 event generate %W <<NextChar>>

		&L<rR>       Button-4 event generate /Ar/ <<PrevWord>>
		&L<gG>       Button-4 event generate /Ag/ <<PrevWord>>
		&L<bB>       Button-4 event generate /Ab/ <<PrevWord>>
		&L<rR> Shift-Button-4 event generate /Ar/ <<PrevChar>>
		&L<gG> Shift-Button-4 event generate /Ag/ <<PrevChar>>
		&L<bB> Shift-Button-4 event generate /Ab/ <<PrevChar>>
		&L<rR>       Button-5 event generate /Ar/ <<NextWord>>
		&L<gG>       Button-5 event generate /Ag/ <<NextWord>>
		&L<bB>       Button-5 event generate /Ab/ <<NextWord>>
		&L<rR> Shift-Button-5 event generate /Ar/ <<NextChar>>
		&L<gG> Shift-Button-5 event generate /Ag/ <<NextChar>>
		&L<bB> Shift-Button-5 event generate /Ab/ <<NextChar>>

		&.,L<rgbRGBc>,A<rgb>       Control-Button-4 +event generate /Ar/ <<PrevWord>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-4 +event generate /Ag/ <<PrevWord>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-4 +event generate /Ab/ <<PrevWord>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-4 +event generate /Ar/ <<PrevChar>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-4 +event generate /Ag/ <<PrevChar>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-4 +event generate /Ab/ <<PrevChar>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-5 +event generate /Ar/ <<NextWord>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-5 +event generate /Ag/ <<NextWord>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-5 +event generate /Ab/ <<NextWord>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-5 +event generate /Ar/ <<NextChar>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-5 +event generate /Ag/ <<NextChar>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-5 +event generate /Ab/ <<NextChar>>
#		&L<rR>,A<r> Control-Button-1 +/Ag/ set \[/Ar/ get\]
#		&L<rR>,A<r> Control-Button-1 +/Ab/ set \[/Ar/ get\]
#		&L<gG>,A<g> Control-Button-1 +/Ar/ set \[/Ag/ get\]
#		&L<gG>,A<g> Control-Button-1 +/Ab/ set \[/Ag/ get\]
#		&L<bB>,A<b> Control-Button-1 +/Ar/ set \[/Ab/ get\]
#		&L<bB>,A<b> Control-Button-1 +/Ag/ set \[/Ab/ get\]
		&Lc Double-Button-1 [self] updateColour \[[self] colour \[[self] defaultColour\]\]
		&Lc Shift-Double-Button-1 [self] defaultColour \[[self] colour\]
		&Lc <<Paste>> [self] paste
		&Lc <<Copy>>  [self] copy
	"

	set colourBoxGutsGobCommonOptions "
		!!Lc swatch
	"

	set createColourBox [gub "
		$colourBoxGutsGobCreate
		$colourBoxGutsGobCommonOptions
		$colourBoxGutsGobBindings
		$colourBoxGutsGobConfigureOptions
		!:
	" $wColourBox]

	eval [gub "
		$colourBoxGutsGobCreate
		$colourBoxGutsGobCreateOptions
		$colourBoxGutsGobCommonOptions
		!:
	" $wColourBox weegs -wm $wm]

	set vertical [string cat $colourBoxGobVerticalBase \n [gub "
		$colourBoxGutsGobVertical
		.A -orient vertical
		.Lc -borderwidth 3
		.L<RGB> -wraplength 1 -width 1
		$colourBoxGutsGobCommonOptions
	" $wColourBox -wm $wm]]

	set verticalCompact [string cat $colourBoxGobVerticalBase \n [gub "
		$colourBoxGutsGobVerticalCompact
		.A -orient vertical
		.Lc -borderwidth 3
		.L<RGB> -wraplength 1 -width 1
		$colourBoxGutsGobCommonOptions
	" $wColourBox -wm $wm]]

	set horizontal [string cat $colourBoxGobHorizontalBase \n [gub "
		$colourBoxGutsGobHorizontal
		.A -orient horizontal
		.Lc -borderwidth 0
		.L<RGB> -wraplength 0 -width 3 -anchor n
		$colourBoxGutsGobCommonOptions
	" $wColourBox -wm $wm]]

	set horizontalCompact [string cat $colourBoxGobHorizontalBase \n [gub "
		$colourBoxGutsGobHorizontalCompact
		.A -orient horizontal
		.Lc -borderwidth 0
		.L<RGB> -wraplength 0 -width 3 -anchor n
		$colourBoxGutsGobCommonOptions
	" $wColourBox -wm $wm]]

	set create [string cat $createFrame \n $createColourBox]
	set wSwatch [dict get $weegs swatch]
}
method zprepare {} {
	set w $wW

	# Create a WidgetMap where the widget commands
	# only configure and not create their widgets
	set wm [dict create]
	dict for {c p} [gubWidgetMap] {
		dict set wm $c [dict set p command ::gub::gubConfigureWidget]
	}

	set colourBoxFrameGob "
		;Fc
#		!!Fc colourBox_[::tu::tcl::randstr]
	"

	set createFrame [gub "
		$colourBoxFrameGob
		[expr {$toplevel ? "" : "!:"}]
	" $w]

	eval [gub "
		$colourBoxFrameGob
		/Fc
		!:
	" $w weegs -wm $wm]

	set wColourBox [dict get $weegs Fc]

	set colourBoxGobVerticalBase [gub "
		$colourBoxFrameGob
		:Fc -sticky ns
		-Fc
		|Fc 0 -weight 0
		!:
	" $w -wm $wm -oc $column]

	set colourBoxGobHorizontalBase [gub "
		$colourBoxFrameGob
		:Fc -sticky ew
		|Fc
		-Fc 0 -weight 0
		!:
	" $w -wm $wm -or $row]

	set colourBoxGutsGobVertical "
		;ArAgAb
		;LrLgLb
		;LRLGLB
		;Lc
		:A  -sticky ns   -rowspan 1 -columnspan 1
		:Lc -sticky nsew -rowspan 1 -columnspan 3
	"
	set colourBoxGutsGobVerticalCompact "
		;ArAgAbLrLgLb
		;      LRLGLB
		;      Lc
		:A -sticky  ns   -rowspan 3 -columnspan 1
		:Lc -sticky nsew -rowspan 1 -columnspan 3
	"

	set colourBoxGutsGobHorizontal "
		;ArLr
		;AgLg
		;AbLb
		;    LR
		;    LG
		;    LB
		;      Lc
		!:-LR -3
		!:-Lc -3
		:A -sticky {} -columnspan 1
		:Lc -sticky nsew -rowspan 3 -columnspan 1
	"
	set colourBoxGutsGobHorizontalCompact "
		;Ar
		;Ag
		;Ab
		;    Lr
		;    Lg
		;    Lb
		;  LR
		;  LG
		;  LB
		;Lc
		!:-LR -3
		!:-Lc -3
		:A  -sticky   ew -rowspan 1 -columnspan 3
		:Lc -sticky nsew -rowspan 3 -columnspan 1
	"

	set colourBoxGutsGobCreate [set colourBoxGutsGob[string totitle $layout]]

	set colourBoxGutsGobCreateOptions "
		/Lc swatch
	"

	set colourBoxGutsGobConfigureOptions "
		.{L<rgb>} -text R G B
		.{L<RGB>} -textvariable [my varname xR] [my varname xG] [my varname xB]
		.{A} -variable [my varname R] [my varname G] [my varname B]
		.Lc -width 6 ;#-relief ridge
		.A -from 0 -to 255 -command {[self] updateColour}
		.Lc -text \\n\\n
	"

	set colourBoxGutsGobBindings "
		&A<rgb>       Button-4 event generate %W <<PrevWord>>
		&A<rgb> Shift-Button-4 event generate %W <<PrevChar>>
		&A<rgb>       Button-5 event generate %W <<NextWord>>
		&A<rgb> Shift-Button-5 event generate %W <<NextChar>>

		&L<rR>       Button-4 event generate /Ar/ <<PrevWord>>
		&L<gG>       Button-4 event generate /Ag/ <<PrevWord>>
		&L<bB>       Button-4 event generate /Ab/ <<PrevWord>>
		&L<rR> Shift-Button-4 event generate /Ar/ <<PrevChar>>
		&L<gG> Shift-Button-4 event generate /Ag/ <<PrevChar>>
		&L<bB> Shift-Button-4 event generate /Ab/ <<PrevChar>>
		&L<rR>       Button-5 event generate /Ar/ <<NextWord>>
		&L<gG>       Button-5 event generate /Ag/ <<NextWord>>
		&L<bB>       Button-5 event generate /Ab/ <<NextWord>>
		&L<rR> Shift-Button-5 event generate /Ar/ <<NextChar>>
		&L<gG> Shift-Button-5 event generate /Ag/ <<NextChar>>
		&L<bB> Shift-Button-5 event generate /Ab/ <<NextChar>>

		&.,L<rgbRGBc>,A<rgb>       Control-Button-4 +event generate /Ar/ <<PrevWord>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-4 +event generate /Ag/ <<PrevWord>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-4 +event generate /Ab/ <<PrevWord>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-4 +event generate /Ar/ <<PrevChar>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-4 +event generate /Ag/ <<PrevChar>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-4 +event generate /Ab/ <<PrevChar>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-5 +event generate /Ar/ <<NextWord>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-5 +event generate /Ag/ <<NextWord>>
		&.,L<rgbRGBc>,A<rgb>       Control-Button-5 +event generate /Ab/ <<NextWord>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-5 +event generate /Ar/ <<NextChar>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-5 +event generate /Ag/ <<NextChar>>
		&.,L<rgbRGBc>,A<rgb> Shift-Control-Button-5 +event generate /Ab/ <<NextChar>>
#		&L<rR>,A<r> Control-Button-1 +/Ag/ set \[/Ar/ get\]
#		&L<rR>,A<r> Control-Button-1 +/Ab/ set \[/Ar/ get\]
#		&L<gG>,A<g> Control-Button-1 +/Ar/ set \[/Ag/ get\]
#		&L<gG>,A<g> Control-Button-1 +/Ab/ set \[/Ag/ get\]
#		&L<bB>,A<b> Control-Button-1 +/Ar/ set \[/Ab/ get\]
#		&L<bB>,A<b> Control-Button-1 +/Ag/ set \[/Ab/ get\]
		&Lc Double-Button-1 [self] updateColour \[[self] colour \[[self] defaultColour\]\]
		&Lc Shift-Double-Button-1 [self] defaultColour \[[self] colour\]
		&Lc <<Paste>> [self] paste
		&Lc <<Copy>>  [self] copy
	"

	set colourBoxGutsGobCommonOptions "
		!!Lc swatch
	"

	set createColourBox [gub "
		$colourBoxGutsGobCreate
		$colourBoxGutsGobCommonOptions
		$colourBoxGutsGobBindings
		$colourBoxGutsGobConfigureOptions
		!:
	" $wColourBox]

	eval [gub "
		$colourBoxGutsGobCreate
		$colourBoxGutsGobCreateOptions
		$colourBoxGutsGobCommonOptions
		!:
	" $wColourBox weegs -wm $wm]

	set vertical [string cat $colourBoxGobVerticalBase \n [gub "
		$colourBoxGutsGobVertical
		.A -orient vertical
		.Lc -borderwidth 3
		.L<RGB> -wraplength 1 -width 1
		$colourBoxGutsGobCommonOptions
	" $wColourBox -wm $wm]]

	set verticalCompact [string cat $colourBoxGobVerticalBase \n [gub "
		$colourBoxGutsGobVerticalCompact
		.A -orient vertical
		.Lc -borderwidth 3
		.L<RGB> -wraplength 1 -width 1
		$colourBoxGutsGobCommonOptions
	" $wColourBox -wm $wm]]

	set horizontal [string cat $colourBoxGobHorizontalBase \n [gub "
		$colourBoxGutsGobHorizontal
		.A -orient horizontal
		.Lc -borderwidth 0
		.L<RGB> -wraplength 0 -width 3 -anchor n
		$colourBoxGutsGobCommonOptions
	" $wColourBox -wm $wm]]

	set horizontalCompact [string cat $colourBoxGobHorizontalBase \n [gub "
		$colourBoxGutsGobHorizontalCompact
		.A -orient horizontal
		.Lc -borderwidth 0
		.L<RGB> -wraplength 0 -width 3 -anchor n
		$colourBoxGutsGobCommonOptions
	" $wColourBox -wm $wm]]

	set create [string cat $createFrame \n $createColourBox]
	set wSwatch [dict get $weegs swatch]
}
method open {} {
	package require Tk
	if {$toplevel} {
		set w $wW
		if {$makeToplevel} { newToplevel $w }
		eval [gub "-.;|." $w]
		eval [gub "&. Control-q [self] done" $w]
		onToplevelDelete $w [list [self] destroy]
	}
	eval $create
	my layout
	set m [menu $wSwatch.mSwatch -tearoff 0]
	$m add command -label Copy  -command ::tu::epm::epmCopy
	$m add command -label Paste -command ::tu::epm::epmPaste
	if {$toplevel} {
		$m add radiobutton -variable [my varname layout]  -command [list [self] layout] -value horizontal -label Horizontal
		$m add radiobutton -variable [my varname layout]  -command [list [self] layout] -value vertical   -label Vertical
		$m add checkbutton -variable [my varname compact] -command [list [self] layout] -label Compact -offvalue "" -onvalue Compact
		$m add command -label Dup   -command [list [self] dup]
		$m add command -label Close -command [list [self] done]
	}
	::tu::epm::epmBindMenuTo $wSwatch $m
	::tu::epm::epmSetEpmAdjustMenu $m {paste 1}
	my colour $defaultColour
	my updateColour
	return [self]
}
method layout {{newLayout {}}} {
	if {$newLayout ne {}} {
		set layout $newLayout
	}
	eval [set $layout]
}
method colour {{newColour {}}} {
	if {$newColour eq {}} { return $colour }
	my scanColour_RGB $newColour r g b
	set colour [my formatColour_RGB $r $g $b]
	my scanColour_RGB $colour R G B
	$wSwatch configure -background $colour
	lassign [format {%02X %02X %02X} $R $G $B] xR xG xB
	return $colour
}
method updateColour {{v {unused!}}} {
	foreach e [list R G B] { set $e [::tcl::mathfunc::int [set $e]] }
	my colour [my formatColour_RGB $R $G $B]
	if {$command ne {}} { uplevel #0 $command [self] }
}
method paste {} {
	if {![catch {selection get -displayof $wSwatch -selection CLIPBOARD} s]} { my colour $s; my updateColour }
}
method copy {} {
	clipboard clear  -displayof $wSwatch
	clipboard append -displayof $wSwatch -- [$wSwatch cget -background]
}
method formatColour_RGB {r g b}      { return [format {#%02X%02X%02X} $r $g $b] }
method scanColour_RGB   {c rv gv bv} { return [uplevel 1 scan $c {#%02X%02X%02X} $rv $gv $bv] }
method command          {args}       { if {[llength $args] == 1} { set command       [lindex $args 0] }; return $command       }
method defaultColour    {args}       { if {[llength $args] == 1} { set defaultColour [lindex $args 0] }; return $defaultColour }
method wColourBox       {}           { return $wColourBox }

method color           {args} { tailcall my colour           {*}$args }
method formatColor_RGB {args} { tailcall my formatColour_RGB {*}$args }
method scanColor_RGB   {args} { tailcall my scanColour_RGB   {*}$args }
method defaultColor    {args} { tailcall my defaultColour    {*}$args }
method wColorBox       {args} { tailcall my wColourBox       {*}$args }
method updateColor     {args} { tailcall my updateColour     {*}$args }

}; #oo::class create ColourBox

namespace export ColourBox

package provide colourbox 1.0

}; #namespace eval ::colourbox

# EOF

# THATCH /herbie/packages/druma-0.3.tm modbase /herbie/packages
#
# DruMa - Teach an AitCS new moves.
#
# Stuart Cassoff
#
# Summer 2018
# Version 0.1
#
# Summer 2019
# Spring 2020
# Version 0.2
#
# Fall 2020
# Version 0.3
#

package require Tcl 8.6-

package require aitcs


namespace eval druma {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	druma
		title	DruMa
		descr	{Teach an AitCS new moves}
		author	{Stuart Cassoff}
		when	{Fall 2020}
		version	0.3
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2019,2020 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


### DruMa

oo::class create DruMa {
	variable name
	variable aitcs
constructor {aname {createAitCS {yes}}} {
	if {[set name [string trim $aname]] eq ""} { return -code error "name cannot be an empty string" }
	if {$createAitCS eq "+"} { set createAitCS yes; set plus + } else { set plus "" }
	if {$createAitCS} { ::aitcs::aitcs$plus new $name -cns [uplevel 1 {namespace current}] -vns [namespace current] }
	set aitcs [uplevel 1 $name&&]
	::oo::objdefine [self] forward Aitcs $aitcs

	#::oo::objdefine [self] method DestroyAitCS {} [interp alias {} $aitcs---]
	::oo::objdefine [self] method DestroyAitCS {} [info body $aitcs---]
	interp alias {} $aitcs--- {} [self] destroy
	interp alias {} $aitcs:   {} [self]
}
destructor {
	if {![info exists aitcs]} { return }
	interp alias {} $aitcs: {}
	my DestroyAitCS
}
method unknown {{method {}} args} { return [expr {$method eq {} ? [self] : [next $method {*}$args]}] }
method name  {} { return $name  }
method aitcs {} { return $aitcs }
}; #oo::class create DruMa

namespace export DruMa

package provide druma 0.3

}; # namespace eval druma


# EOF
